/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/


#ifndef ALGEBRA_HEADER_FILE
#define ALGEBRA_HEADER_FILE

#ifndef GENERAL_HEADER_FILE
#include "general.h"
#endif

#define M_DIM 6


#define BOUND 2.0
#define MIN_TIME_STEP 0.00001
#define MIN_ROTATION 0.001
#define mval(m, i, j) m[i][j]
#define div(n, d) (d == 0 ? (div_zero = TRUE) : (n)/(d))
#define vec_cross(v1, v2, v3) v1[0]=v2[1]*v3[2]-v2[2]*v3[1]; v1[1]=v2[2]*v3[0]-v2[0]*v3[2]; v1[2]=v2[0]*v3[1]-v2[1]*v3[0]
#define vec_cross_pl(v1, v2, v3) v1[0]+=v2[1]*v3[2]-v2[2]*v3[1]; v1[1]+=v2[2]*v3[0]-v2[0]*v3[2]; v1[2]+=v2[0]*v3[1]-v2[1]*v3[0]
#define vec_plus(v1, v2, v3) v1[0]=v2[0]+v3[0]; v1[1]=v2[1]+v3[1];v1[2]=v2[2]+v3[2]
#define vec_neg(v1) v1[0]=-v1[0]; v1[1]=-v1[1];v1[2]=-v1[2]
#define vec_subt(v1, v2, v3) v1[0]=v2[0]-v3[0]; v1[1]=v2[1]-v3[1];v1[2]=v2[2]-v3[2]
#define vec_dot(v1, v2) ((v1[0]*v2[0])+(v1[1]*v2[1])+(v1[2]*v2[2]))
#define vs_mult(v, s) v[0]=v[0]*s; v[1]=v[1]*s; v[2]=v[2]*s
#define vs_div(v, s) v[0]=v[0]/s; v[1]=v[1]/s; v[2]=v[2]/s
#define vs4_mult(v, s) v[0]=v[0]*s; v[1]=v[1]*s; v[2]=v[2]*s; v[3]=v[3]*s
#define vsv_mult(dv, sc, sv) dv[0]=sc*sv[0];dv[1]=sc*sv[1];dv[2]=sc*sv[2]
#define vsv_mult_pl(dv, sc, sv) dv[0]+=sc*sv[0];dv[1]+=sc*sv[1];dv[2]+=sc*sv[2]
#define vec_copy(d, s) d[0]=s[0]; d[1]=s[1]; d[2]=s[2]
#define vec_sp_copy(d, s) d[0]=s[0]; d[1]=s[1]; d[2]=s[2]; d[3]=s[3]; d[4]=s[4]; d[5]=s[5]
#define vv3_plus(d, s) d[0]+=s[0];d[1]+=s[1];d[2]+=s[2]
#define vvv3_plus(d, v1, v2) d[0]=v1[0]+v2[0];d[1]=v1[1]+v2[1];d[2]=v1[2]+v2[2]

#ifndef NULL
#define NULL 0
#endif

#define absv(val) (((val) >= 0.0) ? (val) : -(val))
#define sqr(val) (val)*(val)
#define maxv(val1, val2) (((val1) > (val2)) ? (val1) : (val2))
#define minv(val1, val2) (((val1) < (val2)) ? (val1) : (val2))
#define vs6_mult(v, s) v[0]=s*v[0];v[1]=s*v[1];v[2]=s*v[2];v[3]=s*v[3];v[4]=s*v[4];v[5]=s*v[5]
#define vv6_plus(d, s) d[0]+=s[0];d[1]+=s[1];d[2]+=s[2];d[3]+=s[3];d[4]+=s[4];d[5]+=s[5]
#define vvv6_plus(d, v1, v2) d[0]=v1[0]+v2[0];d[1]=v1[1]+v2[1];d[2]=v1[2]+v2[2];d[3]=v1[3]+v2[3];d[4]=v1[4]+v2[4];d[5]=v1[5]+v2[5]
#define vvv6_subt(d, v1, v2) d[0]=v1[0]-v2[0];d[1]=v1[1]-v2[1];d[2]=v1[2]-v2[2];d[3]=v1[3]-v2[3];d[4]=v1[4]-v2[4];d[5]=v1[5]-v2[5]
#define vsv6_mult(d, s, v) d[0]=s*v[0];d[1]=s*v[1];d[2]=s*v[2];d[3]=s*v[3];d[4]=s*v[4];d[5]=s*v[5]
#define vsv6_pmult(d, s, v) d[0]+=s*v[0];d[1]+=s*v[1];d[2]+=s*v[2];d[3]+=s*v[3];d[4]+=s*v[4];d[5]+=s*v[5]

Real distance2(Real p1[2], Real p2[2]);


Real simplify_angle(Real angle);
Real simplify_angle_180(Real angle);
Real angle_btw_pts(Real cx, Real cy, Real p1x, Real p1y, Real p2x, Real p2y);
Real angle_btw_vec3(Real vec1[3], Real vec2[3]);

void m3_print(int m[3][3], char *string);
void m3_print(Real m[3][3], char *string);
void m4_print(Real m[4][4], char *string);
void sp_m_print(Real m[6][6], char *string);
void sp_v_print(Real v[6], char *string);
void v_print(Real v[3], char *string);
void v4_print(Real v[4], char *string);

int vec_norm(Real v[3]);

void mm3_transpose(Real dst[3][3], Real src[3][3]);

Real svTv6_mult(Real v1[6], Real v2[6]);
Real svv6_mult(Real v1[6], Real v2[6]);
void vvm4_mult(Real dst[3], Real src[3], Real mat[4][4]);
void vmv_mult(Real dst[3], Real mat[3][3], Real src[3]);
void vmv4_mult(Real dst[3], Real mat[4][4], Real src[3]);
void vmv443_mult(Real dst[4], Real mat[4][3], Real src[3]);
void mm3_copy(Real dst[3][3], Real msrc[3][3]);
void mm3_copy(int dst[3][3], int msrc[3][3]);
void mmm3_mult(int dst[3][3], int mat1[3][3], int mat2[3][3]);
void mmm3_mult(Real dst[3][3], Real mat1[3][3], Real mat2[3][3]);
void mmm_mult(Real dst[4][4], Real mat1[4][4], Real mat2[4][4]);
void msm6_mult(Real dst[6][6], Real sc, Real src[6][6]);
void vvTm6_mult(Real dst[6], Real src[6], Real mat[6][6]);
void vvm6_mult(Real dst[6], Real src[6], Real mat[6][6]);

void vmv6_mult(Real dst[6], Real mat[6][6], Real src[6]);
void mvv6_mult(Real mat[6][6], Real v1[6], Real v2[6]);
void ms6_mult(Real mat[6][6], Real sc);
void ms6_div(Real mat[6][6], Real sc);
void mmm6_mult(Real dst[6][6], Real m1[6][6], Real m2[6][6]);
void mm6_plus(Real dst[6][6], Real src[6][6]);
void mmm6_subt(Real d[6][6], Real m1[6][6], Real m2[6][6]);

Real line_segment_distance(Real seg1_a[3], Real seg1_b[3],
			   Real seg2_a[3], Real seg2_b[3]);
Real lines_distance_3d(Real L0_a[3], Real L0_b[3],
		       Real L1_a[3], Real L1_b[3]);


#endif
