/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stream.h>
#include <fstream.h>
#include "general.h"
#include "graphic.h"
#include "lsystem.h"
#include "pop.h"
#include "lsystem.h"
#include "build_bricks.h"
#include "algebra.h"



#define STAIR_LIST_LEN 1000
int Stair_ptr = 0;
XZPoint *Stair_list = 0;

int X_loc, Y_loc, Z_loc;
int Stack_ptr;
Stack_struct Stack[STACK_SIZE];


int Num_bricks;
int M_ctr = 0;
int Model[DIM_X][DIM_Y][DIM_Z];
int MInfo[DIM_X][DIM_Y][DIM_Z];

int In_bounds;
int balanced;
int Balanced[DIM_Y];
int balance_pts_x[DIM_X];
int balance_pts_z[DIM_Z];
int Weight_x[DIM_X];
int Weight_y[DIM_Y];
int Weight_z[DIM_Z];
int Extras;

int bal_xpos, bal_xneg;
int bal_zpos, bal_zneg;
int Bal_xpos[DIM_Y], Bal_xneg[DIM_Y];
int Bal_zpos[DIM_Y], Bal_zneg[DIM_Y];
int Cap_xpos[DIM_Y], Cap_xneg[DIM_Y];
int Cap_zpos[DIM_Y], Cap_zneg[DIM_Y];
int Hull_pts[DIM_Y];

int X_max, X_min;
int Y_max, Y_min;
int Z_max, Z_min;
int Xmax[DIM_Y], Xmin[DIM_Y];
int Zmax[DIM_Y], Zmin[DIM_Y];



/*
A build sequence is an array of bytes describing
how to build a Model.
*/


void estimate_footprint(int y, int *num, Real *val, XZPoint *point);
int make_footprint(int y, int *hull_pts, int *area,
		   int footprint[DIM_X][DIM_Z]);
void add_step(int x, int z);



int initialize_variables(void)
{
  int i, j, k;

  In_bounds = TRUE;

  X_loc = DIM_X/2;
  Y_loc = 0;
  Z_loc = DIM_Z/2;

  Stack_ptr = 0;

  balanced = FALSE;
  bal_xpos = 0;
  bal_xneg = 0;
  bal_zpos = 0;
  bal_zneg = 0;

  Num_bricks = 0;
  Extras = 0;

  X_max = 0;
  X_min = DIM_X;
  Y_max = 0;
  Y_min = DIM_Y;
  Z_max = 0;
  Z_min = DIM_Z;
  
  for (i=0; i<DIM_Y; i++) {
    Xmax[i] = 0;
    Xmin[i] = DIM_X;
    Zmax[i] = 0;
    Zmin[i] = DIM_Z;
  }


  M_ctr++;
  if (M_ctr <= 0)
    M_ctr = 1;
  else if (M_ctr > 2147483640)
    M_ctr = 1;
  if (M_ctr == 1) {
    // Must reset matrix so no re-use.
    if (Fitness_func != 2) {
      M_ctr = 1;
      for (i=0; i<DIM_X; i++)
	for (j=0; j<DIM_Y; j++)
	  for (k=0; k<DIM_Z; k++)
	    Model[i][j][k] = 0;

    } else {
      M_ctr = 1;
      for (i=0; i<DIM_X; i++)
	for (j=0; j<DIM_Y; j++)
	    Model[i][j][0] = 0;
    }
  }

  for (i=0; i<DIM_X; i++)
    for (j=0; j<DIM_Y; j++)
      for (k=0; k<DIM_Z; k++)
	Model[i][j][k] = 0;


  for (i=0; i<DIM_X; i++) {
    balance_pts_x[i] = 0;
    Weight_x[i] = 0;
  }

  for (i=0; i<DIM_Y; i++)
    Weight_y[i] = 0;

  for (i=0; i<DIM_Z; i++) {
    balance_pts_z[i] = 0;
    Weight_z[i] = 0;
  }

  init_heading();

  return TRUE;
}


#define X_LEN 3.0
#define Y_LEN 3.0


// Returns type of draw to do:
// 1 - draw block.
// 0 - don't draw block.
// -1 - draw surface of block.
int block_draw_type(int x, int y, int z)
{
  if ((x == Xmin[y]) || (x == Xmax[y]))
    return 1;
  if ((z == Zmin[y]) || (z == Zmax[y]))
    return 1;
  if (y == Y_min)
    return 1;

  if ((Model[x-1][y][z] == M_ctr) &&
      (Model[x+1][y][z] == M_ctr) &&
      (Model[x][y][z-1] == M_ctr) &&
      (Model[x][y][z+1] == M_ctr) &&
      (Model[x][y-1][z] == M_ctr)) {
    if (y == Y_max)
      return -1;
    else if (Model[x][y+1][z] == M_ctr)
      return 0;
  }
  return 1;
}


void draw_structure(int arg)
{
  int x, y, z;
  Real xx, zz;
  double box[3] = {X_LEN, Y_LEN, X_LEN}; // Shape for drawing in GL.

  if (win1 == 0) {
    printf("win1 not defined\n");
    return;
  }


  //  win1->SetColor(col_surf);

  //  win1->translate(0.0, -200.0, 0.0);
  //  win1->drawPlaneY(Real(DIM_X)*X_LEN, -Y_LEN/2.0, Real(DIM_Z)*X_LEN);

  if (Option[8] == 1) {
    win1->SetColor(col_surf3);
    if (Option[10]) {
      win1->drawPlaneY(35.0*X_LEN, -Y_LEN/2.0, 35.0*X_LEN);
    } else {
      win1->drawPlaneY(2.0*Real(DIM_X)*X_LEN, -Y_LEN/2.0-1.0, 2.0*Real(DIM_Z)*X_LEN);
    }

  } else if (Option[8] != 4) {
    //    win1->SetColor(col_surf2);
    //    Real surface[3] = {0.67, 0.58, 0.50};

    //    Real surface[3] = {0.67, 0.605, 0.50};
    //    win1->SetColor(surface);
    //    win1->SetColor(col_surf3);
    win1->SetColor(col_surf5);
    win1->drawPlaneY(25000.0, -Y_LEN/2.0-1.0, 25000.0);
    if (Option[14]) {
      win1->SetLineWidth(2.0);
    } else {
      win1->SetLineWidth(1.0);
    }
    win1->startLine();

    // Stripes on road.
    int ctr = 0;
    for (xx=-10200.0; xx<=10000.0; xx+=50.0) {
      ctr++;
      if ((ctr%5) == 1)
	win1->SetColor(col_stripe1);
      else if ((ctr%5) == 2)
	win1->SetColor(col_stripe2);
      win1->addVertex(xx, -Y_LEN/2.0, -10000.0);
      win1->addVertex(xx, -Y_LEN/2.0, 10000.0);
    }
    ctr = 0;
    for (zz=-10321.0; zz<=10000.0; zz+=50.0) {
      ctr++;
      if ((ctr%5) == 1)
	win1->SetColor(col_stripe1);
      else if ((ctr%5) == 2)
	win1->SetColor(col_stripe2);
      win1->addVertex(-10000, -Y_LEN/2.0, zz);
      win1->addVertex(10000, -Y_LEN/2.0, zz);
    }
    win1->stopLine();
  }




  if (Option[8] != 4 || TRUE) {
    // Draw shadow2.
    if (Option[8] != 4) {
      win1->SetColor(col_grey5); // col_grey3
    } else {
      win1->SetColor(col_grey11);
    }

    for (z=Z_min; z<=Z_max; z++) {
      for (y=Y_min; y<=Y_max; y++) {
	for (x=X_min; x<=X_max; x++) {
	  if (Model[x][y][z] == M_ctr) {
	    win1->startPolygon();
	    win1->addVertex(double(y)*Y_LEN+double(x)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			    -double(Y_LEN)/2.0+0.1,
			    double(y)*0.7*Y_LEN+double(z)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));

	    win1->addVertex(double(y)*Y_LEN+double(x)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			    -double(Y_LEN)/2.0+0.1,
			    double(y)*0.7*Y_LEN+double(z+1)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));

	    win1->addVertex(double(y+1)*Y_LEN+double(x)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			    -double(Y_LEN)/2.0+0.1,
			    double(y+1)*0.7*Y_LEN+double(z+1)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));

	    win1->addVertex(double(y+1)*Y_LEN+double(x+1)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			    -double(Y_LEN)/2.0+0.1,
			    double(y+1)*0.7*Y_LEN+double(z+1)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));

	    win1->addVertex(double(y+1)*Y_LEN+double(x+1)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			    -double(Y_LEN)/2.0+0.1,
			    double(y+1)*0.7*Y_LEN+double(z)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));

	    win1->addVertex(double(y)*Y_LEN+double(x+1)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			    -double(Y_LEN)/2.0+0.1,
			  double(y)*0.7*Y_LEN+double(z)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));

	    win1->stopList();
	  }
	}
      }
    }
  }

  // Draw structure.
  if (Option[8] != 6)
    win1->SetDefaultBox(box, col_side1, col_side3, col_side2);
  else
    win1->SetDefaultBox(box, col_side2, col_side3, col_side1);

  for (z=Z_min; z<=Z_max; z++) {
    for (y=Y_min; y<=Y_max; y++) {
      for (x=X_min; x<=X_max; x++) {
	if (Model[x][y][z] == M_ctr) {
	  win1->push_orientation();
	  win1->translate(double(x)*X_LEN-(X_LEN*Real(DIM_X/2-1.0)),
			  double(y)*Y_LEN,
			  double(z)*X_LEN-(X_LEN*Real(DIM_Z/2-1.0)));
	  win1->drawDefaultBox();
	  win1->pop_orientation();
	}
      }
    }
  }

  // Draw surface.
  win1->SetColor(col_side3);
  y = Y_max;
  for (z=Z_min; z<=Z_max; z++) {
    for (x=X_min; x<=X_max; x++) {
      if (Model[x][y][z] == M_ctr) {
	win1->startPolygon();
	win1->addVertex(double(x+1)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			double(y)*Y_LEN+Y_LEN/2.0,
			double(z+1)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));
	
	win1->addVertex(double(x+1)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			double(y)*Y_LEN+Y_LEN/2.0,
			double(z)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));
	
	win1->addVertex(double(x)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			double(y)*Y_LEN+Y_LEN/2.0,
			double(z)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));

	win1->addVertex(double(x)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			double(y)*Y_LEN+Y_LEN/2.0,
			double(z+1)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));
	win1->stopList();
      }
    }
  }


  win1->SetColor(col_side3);
  for (z=Z_min; z<=Z_max; z++) {
    for (y=Y_min; y<Y_max; y++) {
      for (x=X_min; x<=X_max; x++) {
	if ((Model[x][y][z] == M_ctr) &&
	    (Model[x][y+1][z] != M_ctr)) {
	  win1->startPolygon();
	  win1->addVertex(double(x+1)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			  double(y)*Y_LEN+Y_LEN/2.0,
			  double(z+1)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));
	  
	  win1->addVertex(double(x+1)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			  double(y)*Y_LEN+Y_LEN/2.0,
			  double(z)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));
	
	  win1->addVertex(double(x)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			  double(y)*Y_LEN+Y_LEN/2.0,
			  double(z)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));

	  win1->addVertex(double(x)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			  double(y)*Y_LEN+Y_LEN/2.0,
			  double(z+1)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));
	  win1->stopList();
	}
      }
    }
  }
  
  if (Print[24])
    write_stl("lsystem.stl");
}

/*
	    win1->startPolygon();
	    win1->addVertex(-double(y)*Y_LEN+double(x+1)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			    -double(Y_LEN)/2.0+0.1,
			    -double(y)*0.7*Y_LEN+double(z+1)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));
	    win1->addVertex(-double(y)*Y_LEN+double(x)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			    -double(Y_LEN)/2.0+0.1,
			    -double(y)*0.7*Y_LEN+double(z+1)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));
	    win1->addVertex(-double(y+1)*Y_LEN+double(x)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			    -double(Y_LEN)/2.0+0.1,
			    -double(y+1)*0.7*Y_LEN+double(z+1)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));
	    win1->addVertex(-double(y+1)*Y_LEN+double(x)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			    -double(Y_LEN)/2.0+0.1,
			    -double(y+1)*0.7*Y_LEN+double(z)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));
	    win1->addVertex(-double(y+1)*Y_LEN+double(x+1)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			    -double(Y_LEN)/2.0+0.1,
			    -double(y+1)*0.7*Y_LEN+double(z)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));
	    win1->addVertex(-double(y)*Y_LEN+double(x+1)*X_LEN-(X_LEN*Real(DIM_X/2-0.5)),
			    -double(Y_LEN)/2.0+0.1,
			  -double(y)*0.7*Y_LEN+double(z)*X_LEN-(X_LEN*Real(DIM_Z/2-0.5)));
	    win1->stopList();
*/


void print_structure(void)
{
  int x, y, z;

  if (Print[12]) {
    printf("Top view (@%d):\n", Y_max);
    for (z=Zmin[Y_max]; z<=Zmax[Y_max]; z++) {
      for (x=Xmin[Y_max]; x<=Xmax[Y_max]; x++) {
	if (Model[x][Y_max][z] == M_ctr)
	  printf("X");
	else
	  printf(" ");
      }
      printf("\n");
    }
  }      

  if (Print[17]) {
    printf("Footprint (@%d):\n", Y_min);
    for (z=Zmin[Y_min]; z<=Zmax[Y_min]; z++) {
      for (x=Xmin[Y_min]; x<=Xmax[Y_min]; x++) {
	if (Model[x][Y_min][z] == M_ctr)
	  printf("X");
	else
	  printf(" ");
      }
      printf("\n");
    }
  }



  if (Print[2]) {
    for (y=Y_max; y>=0; y--) {
      for (x=X_min; x<=X_max; x++) {
	if (Model[x][y][DIM_Z/2] == M_ctr)
	  printf("  X");
	else
	  printf("   ");
	  //	  printf("   ");
      }
      cout << endl;
    }
    for (x=X_min; x<=X_max; x++)
      printf("%3d", Weight_x[x]);
    cout << endl;
  }
}


// Future version might be: put_brick(int type);
//  or: put_brick(int length/struct dim);
// returns: TRUE/1 if it successfully places a brick.
// returns: -1 if there is already a brick at the location.
// returns: FALSE/0 if location is out of bounds.
int put_brick(void)
{
  if (!Pen_down)
    return TRUE;

#ifndef OPTIMIZE
  if (Print[11]) {
    printf("%d %d %d\n", X_loc, Y_loc, Z_loc);
  }
#endif

  if ((X_loc<0) || (X_loc >= DIM_X) ||
      (Y_loc<0) || (Y_loc >= DIM_Y) ||
      (Z_loc<0) || (Z_loc >= DIM_Z)) {
#ifndef OPTIMIZE
    if (Print[0] || Print[10])
      printf("Build::exceeded boundaries: %d %d %d\n", X_loc, Y_loc, Z_loc);
#endif
    In_bounds = FALSE;
    return FALSE;
  }

  if (Model[X_loc][Y_loc][Z_loc] == M_ctr) {
    // signify that there is already a brick here.
    Extras++;
    return -1;
  }

  Weight_x[X_loc]++;
  Weight_y[Y_loc]++;
  Weight_z[Z_loc]++;
  Num_bricks++;
  Model[X_loc][Y_loc][Z_loc] = M_ctr;

  if (X_loc < Xmin[Y_loc]) {
    Xmin[Y_loc] = X_loc;
    if (X_loc<X_min)
      X_min = X_loc;
  }
  if (X_loc > Xmax[Y_loc]) {
    Xmax[Y_loc] = X_loc;
    if (X_loc>X_max)
      X_max = X_loc;
  }

  if (Y_loc<Y_min)
    Y_min = Y_loc;
  if (Y_loc>Y_max)
    Y_max = Y_loc;

  if (Z_loc < Zmin[Y_loc]) {
    Zmin[Y_loc] = Z_loc;
    if (Z_loc<Z_min)
      Z_min = Z_loc;
  }
  if (Z_loc>Zmax[Y_loc]) {
    Zmax[Y_loc] = Z_loc;
    if (Z_loc>Z_max)
      Z_max = Z_loc;
  }

  if (Y_loc == 0) {
    balance_pts_x[X_loc]++;
    balance_pts_z[Z_loc]++;
  }

#ifndef OPTIMIZE
  /*
  if (Print[11]) {
    printf("  min:max(%d)  %d:%d(%d:%d)  %d:%d(%d:%d)\n",
	   Y_loc, 
	   X_min, X_max, Xmin[Y_loc], Xmax[Y_loc],
	   Z_min, Z_max, Zmin[Y_loc], Zmax[Y_loc]);
  }
  */
#endif


  return TRUE;
}


// Weight from -inf to start.
int weight_xneg(int start, int all)
{
  int dist, i, wght = 0;

  if (all) {
    i = start;
    dist = 0;
  } else {
    i = start-1;
    dist = 1;
  }

  if (i < 0)
    return 0;

  for (; i>=0; i--) {
    dist++;
    wght += dist * Weight_x[i];
  }

  return wght;
}


// Weight from start to +inf.
int weight_xpos(int start, int all)
{
  int dist, i, wght = 0;

  if (all) {
    i = start;
    dist = 0;
  } else {
    i = start+1;
    dist = 1;
  }

  if (i >= DIM_X)
    return 0;

  for (; i<DIM_X; i++) {
    dist++;
    wght += dist * Weight_x[i];
  }

  return wght;
}



// Weight from -inf to start.
int weight_zneg(int start, int all)
{
  int dist, i, wght = 0;

  if (all) {
    i = start;
    dist = 0;
  } else {
    i = start-1;
    dist = 1;
  }

  if (i < 0)
    return 0;

  for (; i>=0; i--) {
    dist++;
    wght += dist * Weight_z[i];
  }

  return wght;
}


// Weight from start to +inf.
int weight_zpos(int start, int all)
{
  int dist, i, wght = 0;

  if (all) {
    i = start;
    dist = 0;
  } else {
    i = start+1;
    dist = 1;
  }

  if (i >= DIM_Z)
    return 0;

  for (; i<DIM_Z; i++) {
    dist++;
    wght += dist * Weight_z[i];
  }

  return wght;
}


void calc_com(Real *com_x, Real *com_z)
{
  int x, z;
  Real cx = 0.0, cz = 0.0;

  if (Num_bricks == 0) {
    return;
  }

  for (x = X_min; x<=X_max; x++) {
    cx += (Real)(x * Weight_x[x]);
  }

  for (z = Z_min; z<=Z_max; z++) {
    cz += (Real)(z * Weight_z[z]);
  }

  cx /= (Real)Num_bricks;
  cz /= (Real)Num_bricks;
  
  *com_x = cx;
  *com_z = cz;
  //  printf("center of mass: %.2f %.2f\n", cx, cz);
}


int test_balance(void)
{
  int area, num, min_x, max_x, min_z, max_z;
  int cx, cz;
  Real com_x, com_z;
  int footprint[DIM_X][DIM_Z];

  if (Build_type == 10)
    return TRUE;

  if (Y_min > 0) {
    if (Print[10])
      printf("Fail because Y_min > 0: %d.\n", Y_min);
    return FALSE;
  }

  if (Num_bricks == 0)
    // No bricks.
    return TRUE;


  if (Fitness_func == 2)
    // don't test balance for tiles problem.
    return TRUE;

  if (Weight_y[0] == 0) {
    // No support bricks.
    if (Print[10])
      printf("Fail because Weight_y[0] == 0\n");
    return FALSE;
  }

  // Determine footprint.
  make_footprint(0, &num, &area, footprint);
  calc_com(&com_x, &com_z);
  cx = (int)com_x;
  cz = (int)com_z;

  if (Print[15]) {
    printf("footprint %d:%d -> %d (%d)\n", cx, cz,
	   footprint[cx][cz], M_ctr);

    int x, z;
    for (x=X_min; x<=X_max; x++) {
      for (z=Z_min; z<=Z_max; z++) {
	printf("%d", footprint[x][z]);
      }
      printf("\n");
    }

  }

  if (footprint[cx][cz] == M_ctr) {
    balanced = TRUE;
  } else {
    balanced = FALSE;
  }

  //  printf(" fp:%d:%d->%d   bal: %d\n", cx, cz, footprint[cx][cz],
  //	 balanced);
  //  printf("balanced: %d.\n", balanced);
  return balanced;


  for (min_x=0; min_x<DIM_X; min_x++) {
    if (balance_pts_x[min_x])
      break;
  }
  for (max_x=DIM_X; max_x>=0; max_x--) {
    if (balance_pts_x[max_x])
      break;
  }

  for (min_z=0; min_z<DIM_Z; min_z++) {
    if (balance_pts_z[min_z])
      break;
  }
  for (max_z=DIM_Z; max_z>=0; max_z--) {
    if (balance_pts_z[max_z])
      break;
  }


  // Balance on X axis first.
  bal_xneg = weight_xpos(min_x, FALSE) -
    weight_xneg(min_x-1, TRUE);
  if (Print[1])
    printf("X Negative: x:%d,  w-:%d,  w+:%d = %d\n",
	   min_x, weight_xneg(min_x-1, TRUE),
	   weight_xpos(min_x, FALSE), bal_xneg);

  bal_xpos = weight_xneg(max_x, FALSE) -
    weight_xpos(max_x+1, TRUE);
  if (Print[1])
    printf("X Positive: x:%d,  w-:%d,  w+:%d = %d\n",
	   max_x, weight_xneg(max_x, FALSE),
	   weight_xpos(max_x+1, TRUE), bal_xpos);

  if ((bal_xneg < 0) || (bal_xpos < 0)) {
    if (Print[10]) {
      printf("build_bricks :: valid() - fail: x-+:%d:%d.\n", bal_xneg,
	     bal_xpos);
    }

    balanced = FALSE;
    return FALSE;
  }


  // Test balance on Z-axis.
  bal_zneg = weight_zpos(min_z, FALSE) -
    weight_zneg(min_z-1, TRUE);
  if (Print[1])
    printf("Z Negative: z:%d,  w-:%d,  w+:%d = %d\n",
	   min_z, weight_zneg(min_z-1, TRUE),
	   weight_zpos(min_z, FALSE), bal_zneg);

  bal_zpos = weight_zneg(max_z, FALSE) -
    weight_zpos(max_z+1, TRUE);
  if (Print[1])
    printf("Z Positive: z:%d,  w-:%d,  w+:%d = %d\n",
	   max_z, weight_zneg(max_z, FALSE),
	   weight_zpos(max_z+1, TRUE), bal_zpos);

  if ((bal_zneg < 0) || (bal_zpos < 0)) {
    balanced = FALSE;
    if (Print[10]) {
      printf("build_bricks :: valid() - fail: z-+:%d:%d.\n", bal_zneg,
	     bal_zpos);
    }
    return FALSE;
  }

  balanced = TRUE;
  return TRUE;
}



// point[0]: max x, min z
// point[1]: max x, max z
// point[2]: min x, min z
// point[3]: min x, max z
// Makes a rectangle inside the actual footprint.
void estimate_footprint(int y, int *num, Real *val, XZPoint *point)
{
  int x, z;
  int xdiff, zdiff1, zdiff2;
  Real area;

  // Get rough approximation of footprint.
  x=Xmax[y];
  for (z=Zmin[y]; z<=Zmax[y]; z++) {
    if (Model[x][y][z] == M_ctr) {
      point[0].x = x;
      point[0].z = z;
      break;
    }
  }

  for (z=Zmax[y]; z>=Zmin[y]; z--) {
    if (Model[x][y][z] == M_ctr) {
      point[1].x = x;
      point[1].z = z;
      break;
    }
  }

  x=Xmin[y];
  for (z=Zmin[y]; z<=Zmax[y]; z++) {
    if (Model[x][y][z] == M_ctr) {
      point[2].x = x;
      point[2].z = z;
      break;
    }
  }

  for (z=Zmax[y]; z>=Zmin[y]; z--) {
    if (Model[x][y][z] == M_ctr) {
      point[3].x = x;
      point[3].z = z;
      break;
    }
  }

  // Calculate area of footprint.
  xdiff = 1 + point[0].x - point[2].x;

  // Triangle 1
  zdiff1 = 1 + point[1].z - point[0].z;
  area = xdiff * zdiff1;


  // Triagle 2
  zdiff2 = 1 + point[3].z - point[2].z;
  area += xdiff * zdiff2;

  if (xdiff == 1) {
    if (zdiff1 == 1)
      *num = 1;
    else
      *num = 2;
  } else {
    if (zdiff1 == 1) {
      if (zdiff2 == 1)
	*num = 2;
      else
	*num = 3;
    } else {
      if (zdiff2 == 1)
	*num = 3;
      else
	*num = 4;
    }
  }      

#ifndef OPTIMIZE
  if (Print[13]) {
    printf("y:%d  num:%d(%d)  %d:%d %d:%d %d:%d %d:%d  %.1lf\n",
	   y, *num, Weight_y[y],
	   point[0].x, point[0].z,
	   point[1].x, point[1].z,
	   point[2].x, point[2].z,
	   point[3].x, point[3].z,
	   area/2.0);
  }
#endif

  *val = area/2.0;
}





int calc_load_capacity(int y)
{
  int num;
  int min_x, max_x, min_z, max_z;
  Real area;
  XZPoint point[4];

  Cap_xneg[y] = 0;
  Cap_xpos[y] = 0;
  Cap_zneg[y] = 0;
  Cap_zpos[y] = 0;

  if (Num_bricks == 0)
    // No bricks.
    return TRUE;

  if (Fitness_func == 2)
    // don't test balance for tiles problem.
    return TRUE;

  if (Weight_y[y] == 0) {
    // No support bricks.
    if (Print[10])
      printf("Fail because Weight_y[%d] == 0\n", y);
    return FALSE;
  }

  // Determine footprint.
  estimate_footprint(y, &num, &area, point);
  if (point[2].x < point[3].x)
    min_x = point[3].x;
  else
    min_x = point[2].x;

  if (point[0].x > point[1].x)
    max_x = point[1].x;
  else
    max_x = point[0].x;

  if (point[0].z < point[2].z)
    min_z = point[2].z;
  else
    min_z = point[0].z;

  if (point[1].z > point[3].z)
    max_z = point[3].z;
  else
    max_z = point[1].z;

  if (Print[15]) {
    int min_x2, max_x2, min_z2, max_z2;

    for (min_x2=0; min_x2<DIM_X; min_x2++) {
      if (balance_pts_x[min_x2])
	break;
    }
    for (max_x2=DIM_X; max_x2>=0; max_x2--) {
      if (balance_pts_x[max_x2])
	break;
    }

    for (min_z2=0; min_z2<DIM_Z; min_z2++) {
      if (balance_pts_z[min_z2])
	break;
    }
    for (max_z2=DIM_Z; max_z2>=0; max_z2--) {
      if (balance_pts_z[max_z2])
	break;
    }

    printf("x: %d:%d(%d:%d)  z: %d:%d(%d:%d)\n",
	   min_x, max_x, min_x2, max_x2,
	   min_z, max_z, min_z2, max_z2);
  }


  // Balance on X axis first.
  Bal_xneg[y] = weight_xpos(min_x, FALSE) -
    weight_xneg(min_x-1, TRUE);
  if (Print[1])
    printf("X Negative: x:%d,  w-:%d,  w+:%d = %d\n",
	   min_x, weight_xneg(min_x-1, TRUE),
	   weight_xpos(min_x, FALSE), Bal_xneg[y]);

  Bal_xpos[y] = weight_xneg(max_x, FALSE) -
    weight_xpos(max_x+1, TRUE);
  if (Print[1])
    printf("X Positive: x:%d,  w-:%d,  w+:%d = %d\n",
	   max_x, weight_xneg(max_x, FALSE),
	   weight_xpos(max_x+1, TRUE), Bal_xpos[y]);

  if ((Bal_xneg[y] < 0) || (Bal_xpos[y] < 0)) {
    balanced = FALSE;
    return FALSE;
  }


  // Test balance on Z-axis.
  Bal_zneg[y] = weight_zpos(min_z, FALSE) -
    weight_zneg(min_z-1, TRUE);
  if (Print[1])
    printf("Z Negative: z:%d,  w-:%d,  w+:%d = %d\n",
	   min_z, weight_zneg(min_z-1, TRUE),
	   weight_zpos(min_z, FALSE), Bal_zneg[y]);

  Bal_zpos[y] = weight_zneg(max_z, FALSE) -
    weight_zpos(max_z+1, TRUE);
  if (Print[1])
    printf("Z Positive: z:%d,  w-:%d,  w+:%d = %d\n",
	   max_z, weight_zneg(max_z, FALSE),
	   weight_zpos(max_z+1, TRUE), Bal_zpos[y]);


  // Calculate carrying capacities.
  // X min.
  if (min_x <= Xmin[Y_max])
    Cap_xneg[y] = 2 * Bal_xneg[y];
  else
    Cap_xneg[y] = Bal_xneg[y]/(min_x - Xmin[Y_max]);
  // X max.
  if (max_x >= Xmax[Y_max])
    Cap_xpos[y] = 2 * Bal_xpos[y];
  else
    Cap_xpos[y] = Bal_xpos[y]/(Xmax[Y_max]-max_x);
  // Z min.
  if (min_z <= Zmin[Y_max])
    Cap_zneg[y] = 2 * Bal_zneg[y];
  else
    Cap_zneg[y] = Bal_zneg[y]/(min_z - Zmin[Y_max]);
  // Z max.
  if (max_z >= Zmax[Y_max])
    Cap_zpos[y] = 2 * Bal_zpos[y];
  else
    Cap_zpos[y] = Bal_zpos[y]/(Zmax[Y_max]-max_z);

  if (Print[10]) {
    printf("x: %d:%d  z: %d:%d\n", min_x, max_x,
	   min_z, max_z);
    printf("X: %d:%d  Z: %d:%d\n", Xmin[Y_max],
	   Xmax[Y_max], Zmin[Y_max], Zmax[Y_max]);
    printf("Bal X: %d:%d  Z: %d:%d\n", Bal_xneg[y],
	   Bal_xpos[y], Bal_zneg[y],
	   Bal_zpos[y]);
    printf("%d: %d %d %d %d\n",
	   y, Cap_xneg[y], Cap_xpos[y],
	   Cap_zneg[y], Cap_zpos[y]);
  }

  if ((Bal_zneg[y] < 0) || (Bal_zpos[y] < 0)) {
    Balanced[y] = FALSE;
    return FALSE;
  }

  balanced = TRUE;
  return TRUE;
}

int calc_load_capacity2(int y)
{
  int num, area;
  make_footprint(y, &num, &area, 0);
  Hull_pts[y] = num;
  return area;
}


int calc_load_capacity3(int y)
{
  int num, area;
  make_footprint(y, &num, &area, 0);
  Hull_pts[y] = num;

#ifndef OPTIMIZE
  if (Print[13]) {
    printf("%d: #:%d,  area:%d,  W:%d\n", y, num, area, Weight_y[y]);
  }
#endif  
  area -= Weight_y[y];
  if (num == 2) {
    // Special case of line.
    area = int(sqrt(Real(area)));
  }

  return area;
}


// Returns the area enclosed at height y.
// Penalizes the case where its linear.
int calc_load_capacity4(int y)
{
  int num, area;
  make_footprint(y, &num, &area, 0);
  Hull_pts[y] = num;

#ifndef OPTIMIZE
  if (Print[13]) {
    printf("%d: #:%d,  area:%d,  W:%d\n", y, num, area, Weight_y[y]);
  }
#endif  
  if (num == 2) {
    // Special case of line.
    area = int(sqrt(Real(area)));
  }

  return area;
}


Real load_score(void)
{
  int y;
  Real score = 0.0;

  for (y=0; y<Y_max; y++) {
    calc_load_capacity(y);

    if (Cap_xneg[y] > 0)
      score += Cap_xneg[y];
    if (Cap_xpos[y] > 0)
      score += Cap_xpos[y];
    if (Cap_zneg[y] > 0)
      score += Cap_zneg[y];
    if (Cap_zpos[y] > 0)
      score += Cap_zpos[y];

    if (Print[10]) {
      printf("%d: %d %d %d %d -> %.0lf\n",
	     y, Cap_xneg[y], Cap_xpos[y],
	     Cap_zneg[y], Cap_zpos[y], score);
    }

  }

  return sqrt(score);
}


Real load_score2(void)
{
  int y;
  Real score;

  score = 0.0;
  for (y=0; y<Y_max; y++) {
    score += Real(calc_load_capacity2(y));
  }
  return score;
}


void load_score3(Real *strength, Real *penalty)
{
  int y, pen;
  *strength = 1.0;
  *penalty = 0.0;

  for (y=0; y<Y_max; y++) {
    *strength += Real(calc_load_capacity3(y));

    pen = Weight_y[y] - Hull_pts[y];
    if (pen <= 4)
      pen = 0;
    else
      pen -= 4;
    if (Hull_pts[y] < 3)
      pen += 2;
    *penalty += Real(pen);
  }
}


// Returns volume of table - voxels + points on convex hull.
int load_score4(void)
{
  int y, volume;

  volume = 0;
  for (y=0; y<Y_max; y++) {
    volume += calc_load_capacity2(y);
    volume -= Weight_y[y] + Hull_pts[y];
  }

  return volume;
}


// Returns volume of table - voxels + points on convex hull.
// penalizes case when shape is in a line.
int load_score5(void)
{
  int y, volume;

  volume = 0;
  for (y=0; y<Y_max; y++) {
    volume += calc_load_capacity4(y);
    //    volume -= Weight_y[y] + Hull_pts[y];
  }

  return volume;
}


Real fitness_tiles(void)
{
  Real fitness = 0.0;

  
  
  return fitness;
}



int calc_area(int y)
{
  int num, area;
  make_footprint(y, &num, &area, 0);
  Hull_pts[y] = num;
  return area;
}

// Returns the volume of the object (including surface).
int volume_score1(void)
{
  int y, volume;

  volume = 0;
  for (y=0; y<=Y_max; y++) {
    volume += calc_area(y);
  }

  return volume;
}

// Returns the volume of empty spaces (+ hull pts).
int volume_score2(void)
{
  int y, volume;

  volume = 0;
  for (y=0; y<=Y_max; y++) {
    volume += calc_area(y) - Weight_y[y] + Hull_pts[y];
  }

  return volume;
}



Real surface_score(void)
{
  int x, y, z;
  int len, max, surf, surfx, surfz;

  y = Y_max;
  surfx = 0;
  for (x=Xmin[y]; x<=Xmax[y]; x++) {
    len = 0;
    max = 0;
    for (z=Zmin[y]; z<=Zmax[y]; z++) {
      if (Model[x][y][z] == M_ctr) {
	len++;
      } else if (len>0) {
	surfx += len-1;
	if (len > max)
	  max = len;
	len = 0;
      }
    }
    if (len>0)
      surfx += len-1;
  }

  surfz=0;
  for (z=Zmin[y]; z<=Zmax[y]; z++) {
    len = 0;
    max = 0;
    for (x=Xmin[y]; x<=Xmax[y]; x++) {
      if (Model[x][y][z] == M_ctr) {
	len++;
      } else if (len>0) {
	surfz += len-1;
	if (len > max)
	  max = len;
	len = 0;
      }
    }
    if (len>0)
      surfz += len-1;
  }
  surf = surfx * surfz + Weight_y[y];
  //  surf = surfx + surfz + Weight_y[y];

  return surf;
}


inline Real calc_slope(int p1, int p2, XZPoint_CH *points)
{
  return Real(points[p2].z - points[p1].z)/
    Real(points[p2].x - points[p1].x);
}


int make_footprint(int y, int *hull_pts, int *area,
		   int footprint[DIM_X][DIM_Z])
{
  int x, z, zlimit;
  int p, pts_bot, pts_top;
  int min[DIM_X], max[DIM_X];
  int offset;
  XZPoint_CH hull_bot[DIM_X];
  XZPoint_CH hull_top[DIM_X];


  if (footprint != NULL) {
    if (M_ctr == 1) {
      // Initialize footprint matrix.
      for (x=0; x<DIM_X; x++)
	for (z=0; z<DIM_Z; z++)
	  footprint[x][z] = 0;
    }
  }


  // Handle special cases.
  if (Weight_y[y] == 0) {
    *hull_pts = 0;
    *area = 0;
    return FALSE;

  } else if (Weight_y[y] == 1) {
    *hull_pts = 1;
    *area = 1;
    if (footprint != NULL)
      footprint[Xmin[y]][Zmin[y]] = M_ctr;
    return TRUE;
  }


  // Get bottom curve of the convex hull.
  pts_bot = 0;
  x=Xmin[y];
  for (z=Zmin[y]; z<=Zmax[y]; z++) {
    if (Model[x][y][z] == M_ctr) {
      hull_bot[0].x=x;
      hull_bot[0].z=z;
      pts_bot=1;
      break;
    }
  }
  
  zlimit = Zmax[y];
  for (x+=1; x<=Xmax[y]; x++) {
    for (z=Zmin[y]; z<=zlimit; z++) {
      if (Model[x][y][z] == M_ctr) {
	hull_bot[pts_bot].x=x;
	hull_bot[pts_bot].z=z;
	hull_bot[pts_bot].slope = calc_slope(pts_bot-1, pts_bot, hull_bot);
	pts_bot++;
	break;
      }
    }

    if (z>zlimit)
      // No points found.
      continue;

    // Now test slope.
    for (p=pts_bot-2; p>0; p--) {
      if (hull_bot[p].slope > hull_bot[p+1].slope) {
	// p not in convex hull.
	pts_bot--;
	hull_bot[p].x = hull_bot[p+1].x;
	hull_bot[p].z = hull_bot[p+1].z;
	hull_bot[p].slope = calc_slope(p-1, p, hull_bot);
      } else
	break;
    }
  }

  // Make bottom ridge.
  if (pts_bot > 1) {
    for (p=0; p<pts_bot-1; p++) {
      min[hull_bot[p].x] = hull_bot[p].z;
      min[hull_bot[p+1].x] = hull_bot[p+1].z;
      for (x=hull_bot[p].x+1; x<hull_bot[p+1].x; x++) {
	offset = int(hull_bot[p+1].slope * Real(x-hull_bot[p].x));
	min[x] = hull_bot[p].z + offset;
      }
    }
  } else {
    min[hull_bot[0].x] = hull_bot[0].z;
  }



  // Get top curve of the convex hull.
  pts_top = 0;
  x=Xmin[y];
  for (z=Zmax[y]; z>=Zmin[y]; z--) {
    if (Model[x][y][z] == M_ctr) {
      hull_top[0].x=x;
      hull_top[0].z=z;
      pts_top=1;
      break;
    }
  }
  
  zlimit = Zmin[y];
  for (x+=1; x<=Xmax[y]; x++) {
    for (z=Zmax[y]; z>=zlimit; z--) {
      if (Model[x][y][z] == M_ctr) {
	hull_top[pts_top].x=x;
	hull_top[pts_top].z=z;
	hull_top[pts_top].slope = calc_slope(pts_top-1, pts_top, hull_top);
	pts_top++;
	break;
      }
    }

    if (z<zlimit)
      // No points found.
      continue;

    // Now test slope.
    for (p=pts_top-2; p>0; p--) {
      if (hull_top[p].slope < hull_top[p+1].slope) {
	// p+1 not in convex hull.
	pts_top--;
	hull_top[p].x = hull_top[p+1].x;
	hull_top[p].z = hull_top[p+1].z;
	hull_top[p].slope = calc_slope(p-1, p, hull_top);
      } else
	break;
    }
  }

  // Make top ridge.
  if (pts_top > 1) {
    for (p=0; p<pts_top-1; p++) {
      max[hull_top[p].x] = hull_top[p].z;
      max[hull_top[p+1].x] = hull_top[p+1].z;
      for (x=hull_top[p].x+1; x<hull_top[p+1].x; x++) {
	offset = int(hull_top[p+1].slope * Real(x-hull_top[p].x));
	max[x] = hull_top[p].z + offset;
      }
    }
  } else {
    max[hull_top[0].x] = hull_top[0].z;
  }

  // Calc number of points on convex hull.
  int num = 0;
  for (p=0; p<pts_bot-1; p++) {
    if (hull_bot[p].slope == hull_bot[p+1].slope)
      num++;
  }
  int p_bot = pts_bot - num;
  if ((p_bot < 2) && (pts_bot > 1))
    p_bot = 2;

  num = 0;
  for (p=0; p<pts_top-1; p++) {
    if (hull_top[p].slope == hull_top[p+1].slope)
      num++;
  }
  int p_top = pts_top - num;
  if ((p_top < 2) && (pts_top > 1))
    p_top = 2;

  *hull_pts = p_top + p_bot;
  if (hull_bot[0].z == hull_top[0].z)
    *hull_pts -= 1;
  if (pts_top > 1) {
    if (hull_bot[pts_bot-1].z == hull_top[pts_top-1].z)
      *hull_pts -= 1;
  }

  // Calculate area.
  *area = 0;
  for (x=Xmin[y]; x<=Xmax[y]; x++) {
    *area += (max[x]-min[x]) + 1;
  }


  if (footprint != NULL) {
    // Make footprint.
    for (x=Xmin[y]; x<=Xmax[y]; x++) {
      for (z=min[x]; z<=max[x]; z++) {
	footprint[x][z] = M_ctr;
      }
    }
  }

  return TRUE;
}



#define STAIR_CLEARANCE 4

int step_block2(int x, int y, int z, int steps[DIM_X][DIM_Z],
		int footprint[DIM_X][DIM_Z])
{
  if ((x<0) || (x>= DIM_X))
    return FALSE;
  if ((z<0) || (z>= DIM_Z))
    return FALSE;

  if (footprint[x][z] != M_ctr)
    return FALSE;

  if (Model[x][y][z] != M_ctr)
    return FALSE;

  if (steps[x][z] == y+1)
    return FALSE;

  int y_max = y + STAIR_CLEARANCE;
  if (y_max > Y_max)
    y_max = Y_max;
  for (int y2=y+1; y2<=y_max; y2++) {
    if (Model[x][y2][z] == M_ctr)
      return FALSE;
  }

  return TRUE;
}


void build_stair(int y, int steps[DIM_X][DIM_Z], int footprint[DIM_X][DIM_Z])
{
  int i, x, z;

  // Only add a block to list after it has been
  // SET to on.

  for (i=0; i<Stair_ptr; i++) {
    x = Stair_list[i].x;
    z = Stair_list[i].z;

    if (step_block2(x-1, y, z, steps, footprint)) {
      // Add to stair list.
      steps[x-1][z] = y+1;
      add_step(x-1, z);
    }
    if (step_block2(x+1, y, z, steps, footprint)) {
      // Add to stair list.
      steps[x+1][z] = y+1;
      add_step(x+1, z);
    }
    if (step_block2(x, y, z-1, steps, footprint)) {
      // Add to stair list.
      steps[x][z] = y+1;
      add_step(x, z-1);
    }
    if (step_block2(x, y, z+1, steps, footprint)) {
      // Add to stair list.
      steps[x][z+1] = y+1;
      add_step(x, z+1);
    }
  }
}


void add_step(int x, int z)
{
  Stair_list[Stair_ptr].x = x;
  Stair_list[Stair_ptr].z = z;
  Stair_ptr++;
}


// Return TRUE if this block can be used as a step
// in the staircase.
int step_block(int x, int y, int z, int prev_steps[DIM_X][DIM_Z])
{
  int y2, y_max;

  // Is there a block here?
  if (Model[x][y][z] != M_ctr)
    return FALSE;

  // Is there space above?
  y_max = y + STAIR_CLEARANCE;
  if (y_max > Y_max)
    y_max = Y_max;
  for (y2=y+1; y2<=y_max; y2++)
    if (Model[x][y2][z] == M_ctr)
      return FALSE;

  if (y != 0) {
    // Is this beside a lower step?
    if ((prev_steps[x-1][z] == y) && (x>0)) {
      ;
    } else if ((prev_steps[x+1][z] == y) && (x+1<DIM_X)) {
      ;
    } else if ((prev_steps[x][z-1] == y) && (z>0)) {
      ;
    } else if ((prev_steps[x][z+1] == y) && (z+1<DIM_X)) {
      ;
    } else
      // Not adjacent to lower step.
      return FALSE;

  } else {
    if ((Model[x-1][y][z] != M_ctr) && (x>0)) {
      ;
    } else if ((Model[x+1][y][z] != M_ctr) && (x+1<DIM_X)) {
      ;
    } else if ((Model[x][y][z-1] != M_ctr) && (z>0)) {
      ;
    } else if ((Model[x][y][z+1] != M_ctr) && (z+1<DIM_X)) {
      ;
    } else
      // Not adjacent to lower step.
      return FALSE;
  }

  return TRUE;
}


int surface_fitness(int y, int surface[DIM_X][DIM_Z], int footprint[DIM_X][DIM_Z],
		    int *surf_on, int *surf_off)
{
  int x, z, y2;
  int len, max;
  int surf, surf_bad, surfx, surfz;

  y2 = y+1;

  surf = 0;
  surf_bad = 0;
  surfx = 0;
  for (x=Xmin[y]; x<=Xmax[y]; x++) {
    len = 0;
    max = 0;
    for (z=Zmin[y]; z<=Zmax[y]; z++) {
      if (surface[x][z] == y2) {
	if (footprint[x][z] == M_ctr) {
	  len++;
	} else {
	  surf_bad++;
	  if (len>0) {
	    surfx += len-1;
	    if (len > max)
	      max = len;
	    len = 0;
	  }
	}

      } else if (len>0) {
	surfx += len-1;
	if (len > max)
	  max = len;
	len = 0;
      }
    }
    if (len>1)
      surfx += len-1;
  }

  surfz=0;
  for (z=Zmin[y]; z<=Zmax[y]; z++) {
    len = 0;
    max = 0;
    for (x=Xmin[y]; x<=Xmax[y]; x++) {
      if (surface[x][z] == y2) {
	if (footprint[x][z] == M_ctr) {
	  len++;
	} else {
	  surf_bad++;
	  if (len>0) {
	    surfz += len-1;
	    if (len > max)
	      max = len;
	    len = 0;
	  }
	}

      } else if (len>0) {
	surfz += len-1;
	if (len > max)
	  max = len;
	len = 0;
      }
    }
    if (len>1)
      surfz += len-1;
  }
  surf = surfx + surfz - surf_bad;

  if (surf_on != NULL)
    *surf_on = surfx + surfz;
  if (surf_off != NULL)
    *surf_off = surf_bad;

  /*
  if (Print[6]) {
    printf("%d: surf:%d(%d:%d)-%d\n", y, surf, surfx,
	   surfz, surf_bad);
  }
  */

  if (surf <= 0)
    return 0;
  return surf;
}



// Fence:
// must be high, strong, enclose a large area, have
// stoppage property and use a small amount of material.
int fitness_fence1(int fv_len, Real *fitness_vec)
{
  int hull_pts, area;
  int footprint[DIM_X][DIM_Z];

  return make_footprint(0, &hull_pts, &area, footprint);

  // Unroll footprint onto a surface
}



// Only look at one stair per level.
int fitness_staircase2(int fv_len, Real *fitness_vec)
{
  int x, y, z;
  int hull_pts, area;
  int footprint[DIM_X][DIM_Z];
  int num_steps;

  int steps1[DIM_X][DIM_Z];
  int steps2[DIM_X][DIM_Z];
  int surf_on, surf_off;
  int pen, surf_val, surf_prev;
  int fitness;


  if (Print[6]) {
    printf("#:%d\n", Num_bricks);
  }

  if (Stair_list == 0)
    Stair_list = (XZPoint*)malloc(STAIR_LIST_LEN * sizeof(XZPoint));


  fitness = 0;
  make_footprint(0, &hull_pts, &area, footprint);

  // Reset steps_prev array.
  for (x=X_min; x<=X_max; x++) {
    for (z=Z_min; z<=Z_max; z++) {
      steps1[x][z] = 0;
      steps2[x][z] = 0;
    }
  }

  pen = 0;
  surf_prev = DIM_X*DIM_Z;
  for (y=0; y<=Y_max; y++) {
    num_steps = 0;
    Stair_ptr = 0;
    if (y%2) {
      for (x=Xmin[y]; x<=Xmax[y]; x++) {
	for (z=Zmin[y]; z<=Zmax[y]; z++) {
	  if (step_block(x, y, z, steps2)) {
	    if (footprint[x][z] == M_ctr) {
	      num_steps++;
	      steps1[x][z] = y+1;
	      add_step(x, z);
	      x=DIM_X;
	      z=DIM_Z;
	      break; // Stop at first climbing point.
	    }
	  }
	}
      }
      build_stair(y, steps1, footprint);
      surf_val = surface_fitness(y, steps1, footprint, &surf_on, &surf_off);

    } else {
      for (x=Xmin[y]; x<=Xmax[y]; x++) {
	for (z=Zmin[y]; z<=Zmax[y]; z++) {
	  if (step_block(x, y, z, steps1)) {
	    if (footprint[x][z] == M_ctr) {
	      num_steps++;
	      steps2[x][z] = y+1;
	      add_step(x, z);
	      x=DIM_X;
	      z=DIM_Z;
	      break; // Stop at first climbing point.
	    }
	  }
	}
      }
      build_stair(y, steps2, footprint);
      surf_val = surface_fitness(y, steps2, footprint, &surf_on, &surf_off);
    }

    if (num_steps <= 0)
      break;

    //    fitness += Stair_ptr * y;
    surf_val += Stair_ptr;
    if (surf_val < surf_prev)
      surf_prev = surf_val;
    //    if (surf_prev > sqr(num_steps))
    //      surf_prev = sqr(num_steps);

    if (Weight_y[y] > 2*Stair_ptr)
      pen += Weight_y[y] - 2*Stair_ptr;
    if (y==0)
      pen /= 2;

    fitness += surf_prev * y * y;
    if (Print[6]) {
      //      printf("%d: num steps = %d -> %d.\n", y, num_steps, Stair_ptr);
      printf("%d: (%d:%d:%d:%d) - %d -> %d.\n", y, num_steps, 
	     Stair_ptr, surf_val, surf_prev, pen, fitness);
    }
  }

  /*
  pen /= 2;
  if (fitness/2 < pen)
    fitness = fitness/2;
  else
    fitness -= pen;
    */

  fitness_vec[0] = Real(fitness);

  /*
  if (fitness_vec[0]/2.0 > sqrt(Real(Weight_y[0])))
    fitness_vec[0] = fitness_vec[0] - sqrt(Real(Weight_y[0]));
  else
    fitness_vec[0] /= 2.0;
`    */

  if (Print[6]) {
    printf("Staircase #2.  Fitness #3: %.1f\n", fitness_vec[0]);
  }

  return TRUE;
}


// Only look at one stair per level.
int fitness_staircase3(int fv_len, Real *fitness_vec)
{
  int x, y, z;
  int hull_pts, area;
  int footprint[DIM_X][DIM_Z];
  int num_steps;

  int steps1[DIM_X][DIM_Z];
  int steps2[DIM_X][DIM_Z];
  int **steps;
  int **steps_prev;
  int surf_on, surf_off;
  int pen, surf_val, surf_prev;
  int fitness;


  if (Print[6]) {
    printf("#:%d\n", Num_bricks);
  }

  if (Stair_list == 0)
    Stair_list = (XZPoint*)malloc(STAIR_LIST_LEN * sizeof(XZPoint));

  fitness = 0;
  make_footprint(0, &hull_pts, &area, footprint);

  // Reset steps_prev array.
  for (x=X_min; x<=X_max; x++) {
    for (z=Z_min; z<=Z_max; z++) {
      steps1[x][z] = 0;
      steps2[x][z] = 0;
    }
  }

  pen = 0;
  surf_prev = DIM_X*DIM_Z;
  for (y=0; y<=Y_max; y++) {
    num_steps = 0;
    Stair_ptr = 0;
    if (y%2) {
      steps = (int **)steps1;
      steps_prev = (int **)steps2;
    } else {
      steps = (int**)steps2;
      steps_prev = (int**)steps1;
    }

    for (x=Xmin[y]; x<=Xmax[y]; x++) {
      for (z=Zmin[y]; z<=Zmax[y]; z++) {
	//	if (step_block(x, y, z, (int[DIM_X][DIM_Z])steps_prev)) {
	if (step_block(x, y, z, (int(*)[DIM_X])steps_prev)) {
	  if (footprint[x][z] == M_ctr) {
	    num_steps++;
	    if (num_steps == 1) {
	      ((int(*)[DIM_Z])steps)[x][z] = y+1;
	      add_step(x, z);
	    }
	  }
	}
      }
    }

    if (num_steps <= 0)
      break;

    build_stair(y, (int(*)[DIM_X])steps, footprint);
    surface_fitness(y, (int(*)[DIM_X])steps, footprint, &surf_on, &surf_off);


    //    fitness += Stair_ptr * y;
    if (surf_on > sqr(num_steps))
      surf_val = sqr(num_steps);
    else
      surf_val = surf_on;
    surf_val += Stair_ptr;
    if (surf_val < surf_prev)
      surf_prev = surf_val;
    //    if (surf_prev > sqr(num_steps))
    //      surf_prev = sqr(num_steps);

    if (Weight_y[y] > 2*Stair_ptr)
      pen += Weight_y[y] - 2*Stair_ptr;
    pen += surf_off;
    if (y==0)
      pen /= 2;

    fitness += surf_prev * y * y;
    if (Print[6]) {
      //      printf("%d: num steps = %d -> %d.\n", y, num_steps, Stair_ptr);
      printf("%d: (%d:%d:%d:%d) - %d -> %d.\n", y, num_steps, 
	     Stair_ptr, surf_val, surf_prev, pen, fitness);
    }
  }

  /*
  pen /= 2;
  if (fitness/2 < pen)
    fitness = fitness/2;
  else
    fitness -= pen;
    */

  fitness_vec[0] = Real(fitness);

  /*
  if (fitness_vec[0]/2.0 > sqrt(Real(Weight_y[0])))
    fitness_vec[0] = fitness_vec[0] - sqrt(Real(Weight_y[0]));
  else
    fitness_vec[0] /= 2.0;
    */

  if (Print[6]) {
    printf("Staircase #3.  Fitness #4: %.1f\n", fitness_vec[0]);
  }

  return TRUE;
}


int fitness_staircase1(int fv_len, Real *fitness_vec)
{
  int x, y, z;
  int hull_pts, area;
  int footprint[DIM_X][DIM_Z];
  int num_steps;

  int steps1[DIM_X][DIM_Z];
  int steps2[DIM_X][DIM_Z];
  int pen, surf_val, surf_prev;
  int fitness;

  if (Stair_list == 0)
    Stair_list = (XZPoint*)malloc(STAIR_LIST_LEN * sizeof(XZPoint));


  fitness = 0;
  make_footprint(0, &hull_pts, &area, footprint);

  // Reset steps_prev array.
  for (x=X_min; x<=X_max; x++) {
    for (z=Z_min; z<=Z_max; z++) {
      steps1[x][z] = 0;
      steps2[x][z] = 0;
    }
  }

  surf_prev = DIM_X*DIM_Z;
  for (y=0; y<=Y_max; y++) {
    num_steps = 0;
    Stair_ptr = 0;
    if (y%2) {
      for (x=Xmin[y]; x<=Xmax[y]; x++) {
	for (z=Zmin[y]; z<=Zmax[y]; z++) {
	  if (step_block(x, y, z, steps2)) {
	    if (footprint[x][z] == M_ctr) {
	      num_steps++;
	      steps1[x][z] = y+1;
	      add_step(x, z);
	    }
	  }
	}
      }
      build_stair(y, steps1, footprint);
      surf_val = surface_fitness(y, steps1, footprint, 0, 0);

    } else {
      for (x=Xmin[y]; x<=Xmax[y]; x++) {
	for (z=Zmin[y]; z<=Zmax[y]; z++) {
	  if (step_block(x, y, z, steps1)) {
	    if (footprint[x][z] == M_ctr) {
	      num_steps++;
	      steps2[x][z] = y+1;
	      add_step(x, z);
	    }
	  }
	}
      }
      build_stair(y, steps2, footprint);
      surf_val = surface_fitness(y, steps2, footprint, 0, 0);
    }

    if (num_steps <= 0)
      break;

    //    fitness += Stair_ptr * y;
    surf_val += Stair_ptr;
    if (surf_val < surf_prev)
      surf_prev = surf_val;
    if (surf_prev > sqr(num_steps))
      surf_prev = sqr(num_steps);
    pen = Weight_y[y] - Stair_ptr;
    if (pen < 0)
      pen = 0;

    fitness += (surf_prev - pen) * y;

    if (Print[6]) {
      //      printf("%d: num steps = %d -> %d.\n", y, num_steps, Stair_ptr);
      printf("%d: (%d:%d:%d:%d) - %d -> %d.\n", y, num_steps, 
	     Stair_ptr, surf_val, surf_prev, pen, fitness);
    }
  }

  fitness_vec[0] = Real(fitness) / sqrt(Real(Weight_y[0]));
  //  fitness_vec[0] = Real(fitness) *
  //    sqrt(Real(area) / Real(Weight_y[0]));
  //  fitness_vec[0] = Real(fitness);

  if (Print[6]) {
    printf("%d / %d\n", fitness, area);
    printf("Staircase #1.  Fitness #2: %.1f\n", fitness_vec[0]);
  }


  return TRUE;
}


int surface_block(int x, int y, int z, int footprint[DIM_X][DIM_Z])
{
  if ((x<0) || (x>= DIM_X))
    return 0;
  if ((z<0) || (z>= DIM_Z))
    return 0;

  if (Model[x][y][z] != M_ctr)
    return 0;

  if (footprint[x][z] != M_ctr)
    return -1;

  return 1;
}

int surface_score2(int footprint[DIM_X][DIM_Z])
{
  int x, z, y = Y_max;
  int score = 0;

  for (x=Xmin[y]; x<=Xmax[y]; x++) {
    for (z=Zmin[y]; z<=Zmax[y]; z++) {
      if (surface_block(x, y, z, footprint) == 1) {
	score += surface_block(x-1, y, z-1, footprint);
	score += surface_block(x, y, z-1, footprint);
	score += surface_block(x+1, y, z-1, footprint);

	score += surface_block(x-1, y, z, footprint);
	score += surface_block(x+1, y, z, footprint);

	score += surface_block(x-1, y, z+1, footprint);
	score += surface_block(x, y, z+1, footprint);
	score += surface_block(x+1, y, z+1, footprint);

      } else {
	//	score -= 1;
	;
      }
    }
  }
  if (score < 1)
    score = 1;
  return score;
}


Real surface_score3(int footprint[DIM_X][DIM_Z])
{
  int x, y, z;
  int len, max, surf, surfx, surfz;

  y = Y_max;
  surfx = 0;
  for (x=Xmin[y]; x<=Xmax[y]; x++) {
    len = 0;
    max = 0;
    for (z=Zmin[y]; z<=Zmax[y]; z++) {
      // Don't bother checking if not over footpring.
      if ((footprint[x][z] == M_ctr) && (Model[x][y][z] == M_ctr)) {
	len++;
      } else if (len>0) {
	surfx += len-1;
	if (len > max)
	  max = len;
	len = 0;
      }
    }
    if (len>0)
      surfx += len-1;
  }

  surfz=0;
  for (z=Zmin[y]; z<=Zmax[y]; z++) {
    len = 0;
    max = 0;
    for (x=Xmin[y]; x<=Xmax[y]; x++) {
      if ((footprint[x][z] == M_ctr) && (Model[x][y][z] == M_ctr)) {
	len++;
      } else if (len>0) {
	surfz += len-1;
	if (len > max)
	  max = len;
	len = 0;
      }
    }
    if (len>0)
      surfz += len-1;
  }
  surf = surfx * surfz;
  //  surf = surfx * surfz + Weight_y[y];
  //  surf = surfx + surfz + Weight_y[y];
  if (surf < 1)
    surf = 1;
#ifndef OPTIMIZE
  if (Print[6]) {
    printf("Surf#3:  %d * %d + %d = %d  (%d)\n", surfx, surfz, Weight_y[y], surf, y);
  }
#endif

  return Real(surf);
}


int surface_score4(int footprint[DIM_X][DIM_Z])
{
  int x, z, y = Y_max;
  int score = 0;

  for (x=Xmin[y]; x<=Xmax[y]; x++) {
    for (z=Zmin[y]; z<=Zmax[y]; z++) {
      if (surface_block(x, y, z, footprint) == 1) {
	score += surface_block(x-1, y, z-1, footprint);
	score += surface_block(x, y, z-1, footprint);
	score += surface_block(x+1, y, z-1, footprint);

	score += surface_block(x-1, y, z, footprint);
	score += surface_block(x+1, y, z, footprint);

	score += surface_block(x-1, y, z+1, footprint);
	score += surface_block(x, y, z+1, footprint);
	score += surface_block(x+1, y, z+1, footprint);

      } else {
	score -= 1;
      }
    }
  }
  if (score < 1)
    score = 1;
  return score;
}


// Returns number of bricks on the surface that are also
// over the footprint.
int surface_score5(int footprint[DIM_X][DIM_Z])
{
  int x, z, y = Y_max;
  int num = 0;

  for (x=Xmin[y]; x<=Xmax[y]; x++) {
    for (z=Zmin[y]; z<=Zmax[y]; z++) {
      if ((footprint[x][z] == M_ctr) && (Model[x][y][z] == M_ctr)) {
	num++;
      }
    }
  }
  return num;
}


Real surface_score6(int footprint[DIM_X][DIM_Z])
{
  int x, y, z;
  int len, max, surf, surfx, surfz;

  y = Y_max;
  surfx = 0;
  for (x=Xmin[y]; x<=Xmax[y]; x++) {
    len = 0;
    max = 0;
    for (z=Zmin[y]; z<=Zmax[y]; z++) {
      // Don't bother checking if not over footpring.
      if ((footprint[x][z] == M_ctr) && (Model[x][y][z] == M_ctr)) {
	len++;
      } else if (len>0) {
	surfx += len-1;
	if (len > max)
	  max = len;
	len = 0;
      }
    }
    if (len>0)
      surfx += len-1;
  }

  surfz=0;
  for (z=Zmin[y]; z<=Zmax[y]; z++) {
    len = 0;
    max = 0;
    for (x=Xmin[y]; x<=Xmax[y]; x++) {
      if (Model[x][y][z] == M_ctr) {
	len++;
      } else if (len>0) {
	surfz += len-1;
	if (len > max)
	  max = len;
	len = 0;
      }
    }
    if (len>0)
      surfz += len-1;
  }
  surf = surfx * surfz;
  //  surf = surfx * surfz + Weight_y[y];
  //  surf = surfx + surfz + Weight_y[y];
  if (surf < 1)
    surf = 1;
#ifndef OPTIMIZE
  if (Print[6]) {
    printf("Surf#3:  %d * %d + %d = %d  (%d)\n", surfx, surfz, Weight_y[y], surf, y);
  }
#endif

  return Real(surf);
}

int surface_score7(int footprint[DIM_X][DIM_Z])
{
  int x, z, y = Y_max;
  int score = 0, count;

  for (x=Xmin[y]; x<=Xmax[y]; x++) {
    for (z=Zmin[y]; z<=Zmax[y]; z++) {
      if (surface_block(x, y, z, footprint) == 1) {
	count = 0;
	count += surface_block(x-1, y, z-1, footprint);
	count += surface_block(x, y, z-1, footprint);
	count += surface_block(x+1, y, z-1, footprint);

	count += surface_block(x-1, y, z, footprint);
	count += surface_block(x+1, y, z, footprint);

	count += surface_block(x-1, y, z+1, footprint);
	count += surface_block(x, y, z+1, footprint);
	count += surface_block(x+1, y, z+1, footprint);

	if (count == 8)
	  score += 2;
	else if (count > 1)
	  score += 1;

      } else {
	//	score -= 1;
	;
      }
    }
  }
  if (score < 1)
    score = 1;
  return score;
}


int fitness_table19(ModelStruct *model)
{
  int fv_len = model->fv_length;
  int hull_pts, area;
  int footprint[DIM_X][DIM_Z];
  int surf1, bal1;
  Real height, balance;
  Real surf, strength, strength1, penalty;
  Real penalty1, fitness;
  Real *fitness_vec = model->fitness;


  height = (Real)Y_max;

  make_footprint(0, &hull_pts, &area, footprint);
  Hull_pts[0] = hull_pts;


  surf1 = surface_score5(footprint);
  //  surf1 = Weight_y[Y_max];
  //  surf = (Real)surf1;
  //  surf += calc_area(Y_max);
  surf = (Real)Weight_y[Y_max];
  
  strength1 = volume_score1();
  strength = strength1;
  if (strength < 1.0)
    strength = 1.0;


  penalty1 = Num_bricks - Weight_y[Y_max];
  penalty = Num_bricks;


  //  penalty1 = Num_bricks - Weight_y[Y_max] - Hull_pts[0];
  //  penalty1 = Num_bricks - Weight_y[Y_max];
  //  penalty = penalty1;
  //  penalty = pow(penalty, 0.7);

  // *** Try: ****
  //  penalty = sqrt(penalty1);

  if (penalty < 1.0) {
    penalty = 1.0;
  }

  fitness = height * surf * strength / penalty;
  //  fitness = strength + Weight_y[Y_max];
  //  fitness = surf * Y_max * (strength -  penalty);
  //  fitness = surf * strength * height;

  bal1 = minv(bal_xneg, bal_xpos) + minv(bal_zneg, bal_zpos);
  balance = (Real)bal1 / penalty;

  /*
  if (bal1 < (int)penalty) {
    bal1 = 0;
  } else {
    bal1 -= (int)penalty;
  }

  if (bal1 > (int)fitness) {
    bal1 = (int)fitness;
  }
  */

  //  fitness += balance;

  /*
  if (penalty*penalty > fitness) {
    fitness = sqrt(fitness);
  } else {
    fitness /= penalty;
  }
  */
  
  if (!balanced || !In_bounds || (Y_max < 2)) {
    fitness = 0.0;
    if (Print[9]) {
      printf("voxels :: fitness_table19() - fail, bal:%d, in:%d, y_max:%d.\n",
	     balanced, In_bounds, Y_max);
    }

  }

  if (Print[6]) {
    printf("Table #19.\n");
    printf("#:%4d  -:%4d  ",
	   Num_bricks, Extras);
    printf("Balance:   %.0f:%d   x:%d:%d   z:%d:%d\n", balance, bal1, bal_xneg, bal_xpos,
	   bal_zneg, bal_zpos);
    printf("Extremes:  x:%d:%d   z:%d:%d\n", X_min, X_max, Z_min, Z_max);
    printf("surf: %.0f(%d:%d)  strength: %.0lf(%.0lf)  height: %.1f(%d)  pen:%.0lf(%.0lf)\n",
	   surf, Weight_y[Y_max], surf1, strength, strength1, height, Y_max,
	   penalty, penalty1);
    printf("Fitness #19: %.0lf\n", fitness);
  }


  fitness_vec[0] = fitness;
  if (fv_len > 7) { 
    fitness_vec[1] = height;
    fitness_vec[2] = surf;
    fitness_vec[3] = strength;
    fitness_vec[4] = penalty;
    fitness_vec[5] = Num_bricks;

  } else if (fv_len > 1) {
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}

int fitness_table18(ModelStruct *model)
{
  int y;
  int fv_len = model->fv_length;
  int hull_pts, area;
  int footprint[DIM_X][DIM_Z];
  int surf1, bal1;
  Real height, balance;
  Real surf, strength, strength1, penalty;
  Real penalty1, fitness;
  Real *fitness_vec = model->fitness;


  height = (Real)Y_max;

  make_footprint(0, &hull_pts, &area, footprint);
  Hull_pts[0] = hull_pts;


  surf1 = surface_score5(footprint);
  //  surf1 = Weight_y[Y_max];
  //  surf = (Real)surf1;
  //  surf += calc_area(Y_max);
  surf = (Real)Weight_y[Y_max];
  
  strength1 = volume_score1();
  strength = strength1;
  if (strength < 1.0)
    strength = 1.0;


  penalty1 = Num_bricks - Weight_y[Y_max];
  penalty = penalty1 + 1;
  for (y=0; y < Y_max; y++) {
    penalty -= (Real)Hull_pts[y];
  }

  //  penalty1 = Num_bricks - Weight_y[Y_max] - Hull_pts[0];
  //  penalty1 = Num_bricks - Weight_y[Y_max];
  //  penalty = penalty1;
  //  penalty = pow(penalty, 0.7);
  //  penalty = sqrt(penalty1);
  if (penalty < 1.0) {
    penalty = 1.0;
  }

  fitness = height * surf * strength / penalty;
  //  fitness = strength + Weight_y[Y_max];
  //  fitness = surf * Y_max * (strength -  penalty);
  //  fitness = surf * strength * height;

  bal1 = minv(bal_xneg, bal_xpos) + minv(bal_zneg, bal_zpos);
  balance = (Real)bal1 / penalty;

  /*
  if (bal1 < (int)penalty) {
    bal1 = 0;
  } else {
    bal1 -= (int)penalty;
  }

  if (bal1 > (int)fitness) {
    bal1 = (int)fitness;
  }
  */

  //  fitness += balance;

  /*
  if (penalty*penalty > fitness) {
    fitness = sqrt(fitness);
  } else {
    fitness /= penalty;
  }
  */
  
  if (!balanced || !In_bounds || (Y_max < 2)) {
    fitness = 0.0;
    if (Print[9]) {
      printf("voxels :: fitness_table18() - fail, bal:%d, in:%d, y_max:%d.\n",
	     balanced, In_bounds, Y_max);
    }

  }

  if (Print[6]) {
    printf("Table #18.\n");
    printf("#:%4d  -:%4d  ",
	   Num_bricks, Extras);
    printf("Balance:   %.0f:%d   x:%d:%d   z:%d:%d\n", balance, bal1, bal_xneg, bal_xpos,
	   bal_zneg, bal_zpos);
    printf("Extremes:  x:%d:%d   z:%d:%d\n", X_min, X_max, Z_min, Z_max);
    printf("surf: %.0f(%d:%d)  strength: %.0lf(%.0lf)  height: %.1f(%d)  pen:%.0lf(%.0lf)\n",
	   surf, Weight_y[Y_max], surf1, strength, strength1, height, Y_max,
	   penalty, penalty1);
    printf("Fitness #18: %.0lf\n", fitness);
  }


  fitness_vec[0] = fitness;
  if (fv_len > 7) { 
    fitness_vec[1] = height;
    fitness_vec[2] = surf;
    fitness_vec[3] = strength;
    fitness_vec[4] = penalty;
    fitness_vec[5] = Num_bricks;

  } else if (fv_len > 1) {
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}


int fitness_table17(ModelStruct *model)
{
  int fv_len = model->fv_length;
  int hull_pts, area;
  int footprint[DIM_X][DIM_Z];
  Real surf, surf1, strength, strength1, penalty;
  Real fitness;
  Real *fitness_vec = model->fitness;

  //  surf1 = surface_score();
  make_footprint(0, &hull_pts, &area, footprint);
  surf1 = surface_score5(footprint);
  surf = surf1;

  strength1 = load_score5();
  strength = strength1;
  //  strength = pow(strength1, 0.5);
  if (strength < 1.0)
    strength = 1.0;


  penalty = Num_bricks - Weight_y[Y_max];
  if (penalty < 1.0)
    penalty = 1.0;

  fitness = Y_max * surf * strength / penalty;

  if (!balanced || !In_bounds || (Y_max < 2)) {
    fitness = 0.0;
    if (Print[9]) {
      printf("voxels :: fitness_table17() - fail, bal:%d, in:%d, y_max:%d.\n",
	     balanced, In_bounds, Y_max);
    }
  }



  if (Print[6]) {
    printf("Table #17.\n");
    printf("#:%4d  -:%4d  ",
	   Num_bricks, Extras);
    printf("x:%d:%d   z:%d:%d\n", bal_xneg, bal_xpos,
	   bal_zneg, bal_zpos);
    printf("surf: %.0lf(%.0lf:#%d)  strength: %.0lf(%.0lf)  Y_max: %d  pen:%.0lf\n",
	   surf, surf1, Weight_y[Y_max], strength, strength1, Y_max, penalty);
    printf("Fitness #17: %.0lf\n", fitness);
  }

  fitness_vec[0] = fitness;
  if (fv_len > 7) { 
    fitness_vec[1] = Y_max;
    fitness_vec[2] = surf;
    fitness_vec[3] = strength;
    fitness_vec[4] = penalty;
    //    printf("%d: %d:%d %d\n", Y_max, surf, strength, penalty);
  } else if (fv_len > 1) {
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}


int fitness_table16(ModelStruct *model)
{
  int fv_len = model->fv_length;
  int hull_pts, area;
  int footprint[DIM_X][DIM_Z];
  int surf1, bal1;
  Real height, balance;
  Real surf, strength, strength1, penalty;
  Real penalty1, fitness;
  Real *fitness_vec = model->fitness;


  height = (Real)Y_max;

  make_footprint(0, &hull_pts, &area, footprint);
  Hull_pts[0] = hull_pts;


  surf1 = surface_score5(footprint);
  //  surf1 = Weight_y[Y_max];
  //  surf = (Real)surf1;
  //  surf += calc_area(Y_max);

  surf = (Real)Weight_y[Y_max];
  
  strength1 = volume_score1();
  strength = strength1;
  if (strength < 1.0)
    strength = 1.0;


  penalty1 = Num_bricks - Weight_y[Y_max];
  penalty = penalty1;

  /*
  for (y=0; y< Y_max; y++) {
    penalty -= (Real)Hull_pts[y]/2.0;
  }
  */

  //  penalty1 = Num_bricks - Weight_y[Y_max] - Hull_pts[0];
  //  penalty1 = Num_bricks - Weight_y[Y_max];
  //  penalty = penalty1;
  //  penalty = pow(penalty, 0.7);
  //  penalty = sqrt(penalty1);
  if (penalty < 1.0) {
    penalty = 1.0;
  }

  fitness = height * surf * strength / penalty;
  //  fitness = strength + Weight_y[Y_max];
  //  fitness = surf * Y_max * (strength -  penalty);
  //  fitness = surf * strength * height;

  bal1 = minv(bal_xneg, bal_xpos) + minv(bal_zneg, bal_zpos);
  balance = (Real)bal1 / penalty;

  /*
  if (bal1 < (int)penalty) {
    bal1 = 0;
  } else {
    bal1 -= (int)penalty;
  }

  if (bal1 > (int)fitness) {
    bal1 = (int)fitness;
  }
  */

  //  fitness += balance;

  /*
  if (penalty*penalty > fitness) {
    fitness = sqrt(fitness);
  } else {
    fitness /= penalty;
  }
  */
  
  if (!balanced || !In_bounds || (Y_max < 2)) {
    fitness = 0.0;
    if (Print[9]) {
      printf("voxels :: fitness_table16() - fail, bal:%d, in:%d, y_max:%d.\n",
	     balanced, In_bounds, Y_max);
    }

  }

  if (Print[6]) {
    printf("Table #16.\n");
    printf("#:%4d  -:%4d  ",
	   Num_bricks, Extras);
    printf("Balance:   %.0f:%d   x:%d:%d   z:%d:%d\n", balance, bal1, bal_xneg, bal_xpos,
	   bal_zneg, bal_zpos);
    printf("Extremes:  x:%d:%d   z:%d:%d\n", X_min, X_max, Z_min, Z_max);
    printf("surf: %.0f(%d:%d)  strength: %.0lf(%.0lf)  height: %.1f(%d)  pen:%.0lf(%.0lf)\n",
	   surf, Weight_y[Y_max], surf1, strength, strength1, height, Y_max,
	   penalty, penalty1);
    printf("Fitness #16: %.0lf\n", fitness);
  }


  fitness_vec[0] = fitness;
  if (fv_len > 7) { 
    fitness_vec[1] = height;
    fitness_vec[2] = surf;
    fitness_vec[3] = strength;
    fitness_vec[4] = penalty;
    fitness_vec[5] = Num_bricks;

  } else if (fv_len > 1) {
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}


// Param A: height
// Param B: volume/area
// No penalty for being under parameter.
int fitness_table36(ModelStruct *model)
{
  Real height, surf, surf2, strength, strength2, penalty, fitness;
  int fv_len = model->fv_length;
  Real *fitness_vec = model->fitness;
  Real p_height = 4.0 * model->build_args[3];
  Real p_surf = 4.0 * model->build_args[4];
  Real p_strength = 4.0 * model->build_args[4];
  Real ratio1 = 1.0, ratio2 = 1.0, ratio3 = 1.0;


  height = (Real)Y_max;

  surf = (Real)Weight_y[Y_max];
  surf2 = sqrt(surf);
  
  strength = volume_score1();
  if (strength < 1.0)
    strength = 1.0;
  if (height > 1.0) {
    strength2 = strength / height;
  } else {
    strength2 = strength;
  }
  strength2 = sqrt(strength2);

  penalty = Num_bricks - Weight_y[Y_max];
  if (penalty < 1.0) {
    penalty = 1.0;
  }

  /*
  printf("h:%d:%d, s:%d, st:%d:%d\n", (int)height, (int)p_height,
	 (int)surf, (int)strength, (int)p_strength);
  */

  if (!balanced || !In_bounds || (Y_max < 2)) {
    fitness = 0.0;
    if (Print[9]) {
      printf("voxels :: fitness_table36() - fail, bal:%d, in:%d, y_max:%d.\n",
	     balanced, In_bounds, Y_max);
    }

  }  else {
    // Do fitness adjustment for height.
    if (height > p_height) {
      if (height - p_height > p_height) {
	ratio1 = 0.0;
      } else {
	ratio1 = (2.0*p_height - height) / p_height;
	ratio1 *= ratio1;
      }
    }

    // Do fitness adjustment for surface area.
    if (surf2 > p_surf) {
      if (surf2 - p_surf > p_surf) {
	ratio2 = 0.0;
      } else {
	ratio2 = (2.0*p_surf - surf2) / p_surf;
	ratio2 *= ratio2;
      }
    }

    // Do fitness adjustment for strength.
    if (strength2 > p_strength) {
      if (strength2 - p_strength > p_strength) {
	ratio3 = 0.0;
      } else {
	ratio3 = (2.0*p_strength - strength2) / p_strength;
	ratio3 *= ratio3;
      }
    }
    fitness = height * surf * strength / penalty;
    fitness *= ratio1 * ratio2 * ratio3;
  }


  if (Print[6]) {
    printf("Table Fitness #36:  #%4d  %.2f\n", (int)Num_bricks,
	   (float)(height * surf * strength / penalty));
    printf("  (h:%d:%d:%.2f)(a:%d:%d:%d:%.2f)(s:%d:%d:%d:%.2f)/(p:%d) - %.2f\n",
	   (int)height, (int)p_height, (float)ratio1,
	   (int)surf, (int)surf2, (int)p_surf, (float)ratio2,
	   (int)strength, (int)strength2, (int)p_strength, (float)ratio3,
	   (int)penalty, (float)fitness);
  }


  fitness_vec[0] = fitness;
  if (fv_len > 7) { 
    fitness_vec[1] = height;
    fitness_vec[2] = surf;
    fitness_vec[3] = strength;
    fitness_vec[4] = penalty;
    fitness_vec[5] = Num_bricks;

  } else if (fv_len > 1) {
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}


// Param A: height.
// Param B: surface area, volume
// Penalty for being under parameter.
int fitness_table37(ModelStruct *model)
{
  Real height, surf, surf2, strength, strength2, penalty, fitness;
  int fv_len = model->fv_length;
  Real *fitness_vec = model->fitness;
  Real p_height = 4.0 * model->build_args[3];
  Real p_surf = 4.0 * model->build_args[4];
  Real p_strength = 4.0 * model->build_args[4];
  Real ratio1 = 1.0, ratio2 = 1.0, ratio3 = 1.0;


  height = (Real)Y_max;

  surf = (Real)Weight_y[Y_max];
  surf2 = sqrt(surf);

  strength = volume_score1();
  if (strength < 1.0)
    strength = 1.0;
  if (height > 1.0) {
    strength2 = strength / height;
  } else {
    strength2 = strength;
  }
  strength2 = sqrt(strength2);

  penalty = Num_bricks - Weight_y[Y_max];
  if (penalty < 1.0) {
    penalty = 1.0;
  }

  /*
  printf("h:%d:%d, s:%d, st:%d:%d\n", (int)height, (int)p_height,
	 (int)surf, (int)strength, (int)p_strength);
  */

  if (!balanced || !In_bounds || (Y_max < 2)) {
    fitness = 0.0;
    if (Print[9]) {
      printf("voxels :: fitness_table36() - fail, bal:%d, in:%d, y_max:%d.\n",
	     balanced, In_bounds, Y_max);
    }
  }  else {
    // Do fitness adjustment for height.
    if (height < p_height) {
      ratio1 = height / p_height;
    } else if (height > p_height) {
      if (height - p_height > p_height) {
	ratio1 = 0.0;
      } else {
	ratio1 = (2.0*p_height - height) / p_height;
	ratio1 *= ratio1;
      }
    }

    // Do fitness adjustment for surface area.
    if (surf2 < p_surf) {
      ratio2 = surf2 / p_surf;
    } else if (surf2 > p_surf) {
      if (surf2 - p_surf > p_surf) {
	ratio2 = 0.0;
      } else {
	ratio2 = (2.0*p_surf - surf2) / p_surf;
	ratio2 *= ratio2;
      }
    }

    // Do fitness adjustment for strength (volume).
    if (strength2 < p_strength) {
      ratio3 = strength2 / p_strength;
    } else if (strength2 > p_strength) {
      if (strength2 - p_strength > p_strength) {
	ratio3 = 0.0;
      } else {
	ratio3 = (2.0*p_strength - strength2) / p_strength;
	ratio3 *= ratio3;
      }
    }
    fitness = height * surf * strength / penalty;
    fitness *= ratio1 * ratio2 * ratio3;
  }


  if (Print[6]) {
    printf("Table Fitness #37:  #%4d  %.2f\n", (int)Num_bricks,
	   (float)(height * surf * strength / penalty));
    printf("  (h:%d:%d:%.2f)(a:%d:%d:%d:%.2f)(s:%d:%d:%d:%.2f)/(p:%d) - %.2f\n",
	   (int)height, (int)p_height, (float)ratio1,
	   (int)surf, (int)surf2, (int)p_surf, (float)ratio2,
	   (int)strength, (int)strength2, (int)p_strength, (float)ratio3,
	   (int)penalty, (float)fitness);
  }


  fitness_vec[0] = fitness;
  if (fv_len > 7) { 
    fitness_vec[1] = height;
    fitness_vec[2] = surf;
    fitness_vec[3] = strength;
    fitness_vec[4] = penalty;
    fitness_vec[5] = Num_bricks;

  } else if (fv_len > 1) {
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}


// Parametric height only.
int fitness_table38(ModelStruct *model)
{
  Real height, surf, surf2, strength, strength2, penalty, fitness;
  int fv_len = model->fv_length;
  Real *fitness_vec = model->fitness;
  Real p_height = 4.0 * model->build_args[3];
  Real ratio1 = 1.0;


  height = (Real)Y_max+1.0;

  surf = (Real)Weight_y[Y_max];
  surf2 = sqrt(surf);

  strength = volume_score1();
  if (strength < 1.0)
    strength = 1.0;
  if (height > 1.0) {
    strength2 = strength / height;
  } else {
    strength2 = strength;
  }
  strength2 = sqrt(strength2);

  penalty = Num_bricks - Weight_y[Y_max];
  if (penalty < 1.0) {
    penalty = 1.0;
  }

  /*
  printf("h:%d:%d, s:%d, st:%d:%d\n", (int)height, (int)p_height,
	 (int)surf, (int)strength, (int)p_strength);
  */

  if (!balanced || !In_bounds || (Y_max < 2)) {
    fitness = 0.0;
    if (Print[9]) {
      printf("voxels :: fitness_table36() - fail, bal:%d, in:%d, y_max:%d.\n",
	     balanced, In_bounds, Y_max);
    }
  }  else {
    // Do fitness adjustment for height.
    if (height < p_height) {
      ratio1 = height / p_height;
    } else if (height > p_height) {
      if (height - p_height > p_height) {
	ratio1 = 0.0;
      } else {
	ratio1 = (2.0*p_height - height) / p_height;
	ratio1 *= ratio1;
      }
    }

    fitness = height * surf * strength / penalty;
    fitness *= ratio1;
  }


  if (Print[6]) {
    printf("Table Fitness #37:  #%4d  %.2f\n", (int)Num_bricks,
	   (float)(height * surf * strength / penalty));
    printf("  (h:%d:%d:%.2f)(a:%d:%d)(s:%d:%d)/(p:%d) - %.2f\n",
	   (int)height, (int)p_height, (float)ratio1,
	   (int)surf, (int)surf2,
	   (int)strength, (int)strength2,
	   (int)penalty, (float)fitness);
  }


  fitness_vec[0] = fitness;
  if (fv_len > 7) { 
    fitness_vec[1] = height;
    fitness_vec[2] = surf;
    fitness_vec[3] = strength;
    fitness_vec[4] = penalty;
    fitness_vec[5] = Num_bricks;
    fitness_vec[6] = ratio1;

  } else if (fv_len > 1) {
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}


// Parametric volume/area only.
int fitness_table39(ModelStruct *model)
{
  Real height, surf, surf2, strength, strength2, penalty, fitness;
  int fv_len = model->fv_length;
  Real *fitness_vec = model->fitness;
  Real p_surf = 4.0 * model->build_args[3];
  Real p_strength = 4.0 * model->build_args[3];
  Real ratio2 = 1.0, ratio3 = 1.0;


  height = (Real)Y_max;

  surf = (Real)Weight_y[Y_max];
  surf2 = sqrt(surf);

  strength = volume_score1();
  if (strength < 1.0)
    strength = 1.0;
  if (height > 1.0) {
    strength2 = strength / height;
  } else {
    strength2 = strength;
  }
  strength2 = sqrt(strength2);

  penalty = Num_bricks - Weight_y[Y_max];
  if (penalty < 1.0) {
    penalty = 1.0;
  }

  /*
  printf("h:%d:%d, s:%d, st:%d:%d\n", (int)height, (int)p_height,
	 (int)surf, (int)strength, (int)p_strength);
  */

  if (!balanced || !In_bounds || (Y_max < 2)) {
    fitness = 0.0;
    if (Print[9]) {
      printf("voxels :: fitness_table39() - fail, bal:%d, in:%d, y_max:%d.\n",
	     balanced, In_bounds, Y_max);
    }
  }  else {
    // Do fitness adjustment for surface area.
    if (surf2 < p_surf) {
      ratio2 = surf2 / p_surf;
    } else if (surf2 > p_surf) {
      if (surf2 - p_surf > p_surf) {
	ratio2 = 0.0;
      } else {
	ratio2 = (2.0*p_surf - surf2) / p_surf;
	ratio2 *= ratio2;
      }
    }

    // Do fitness adjustment for strength (volume).
    if (strength2 < p_strength) {
      ratio3 = strength2 / p_strength;
    } else if (strength2 > p_strength) {
      if (strength2 - p_strength > p_strength) {
	ratio3 = 0.0;
      } else {
	ratio3 = (2.0*p_strength - strength2) / p_strength;
	ratio3 *= ratio3;
      }
    }
    fitness = height * surf * strength / penalty;
    fitness *= ratio2 * ratio3;
  }


  if (Print[6]) {
    printf("Table Fitness #39:  #%4d  %.2f\n", (int)Num_bricks,
	   (float)(height * surf * strength / penalty));
    printf("  (h:%d)(a:%d:%d:%d:%.2f)(s:%d:%d:%d:%.2f)/(p:%d) - %.2f\n",
	   (int)height,
	   (int)surf, (int)surf2, (int)p_surf, (float)ratio2,
	   (int)strength, (int)strength2, (int)p_strength, (float)ratio3,
	   (int)penalty, (float)fitness);
  }


  fitness_vec[0] = fitness;
  if (fv_len > 7) { 
    fitness_vec[1] = height;
    fitness_vec[2] = surf;
    fitness_vec[3] = strength;
    fitness_vec[4] = penalty;
    fitness_vec[5] = Num_bricks;
    fitness_vec[6] = ratio2*ratio3;

  } else if (fv_len > 1) {
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}


int valid_area(int p_surf)
{
  int y;
  for (y=0; y<=Y_max; y++) {
    if ((Xmax[y] - Xmin[y] >= p_surf) ||
	(Zmax[y] - Zmin[y] >= p_surf)) {
      if (Print[9]) {
	printf("voxels :: valid_area() - at y=%d : %d (%d:%d)(%d:%d).\n",
	       y, p_surf, Xmin[y], Xmax[y], Zmin[y], Zmax[y]);
      }
      return FALSE;
    }
  }
  return TRUE;
}


// Parametric height only.
int fitness_table40(ModelStruct *model)
{
  Real height, surf, surf2, strength, strength2, penalty, fitness;
  int fv_len = model->fv_length;
  Real *fitness_vec = model->fitness;
  Real p_height = 4.0 * model->build_args[3];
  Real ratio1 = 1.0;


  height = (Real)Y_max+1.0;

  surf = (Real)Weight_y[Y_max];
  surf2 = sqrt(surf);

  strength = volume_score1();
  if (strength < 1.0) {
    strength = 1.0;
  }
  //  strength2 = sqrt(strength2);
  strength2 = strength;


  penalty = Num_bricks - Weight_y[Y_max];
  if (penalty < 1.0) {
    penalty = 1.0;
  }

  /*
  printf("h:%d:%d, s:%d, st:%d:%d\n", (int)height, (int)p_height,
	 (int)surf, (int)strength, (int)p_strength);
  */

  if (!balanced || !In_bounds || (Y_max < 2)) {
    fitness = 0.0;
    if (Print[9]) {
      printf("voxels :: fitness_table40() - fail, bal:%d, in:%d, y_max:%d.\n",
	     balanced, In_bounds, Y_max);
    }
  }  else {
    // Do fitness adjustment for height.
    if (height < p_height) {
      ratio1 = height / p_height;
    } else if (height > p_height) {
      ratio1 = 0.0;
    }

    fitness = height * surf * strength / penalty;
    fitness *= ratio1;
  }


  if (Print[6]) {
  printf("Table Fitness #40:  #%4d  %.2f\n", (int)Num_bricks,
	   (float)(height * surf * strength / penalty));
    printf("  (h:%d:%d:%.2f)(a:%d:%d)(s:%d:%d)/(p:%d) - %.2f\n",
	   (int)height, (int)p_height, (float)ratio1,
	   (int)surf, (int)surf2,
	   (int)strength, (int)strength2,
	   (int)penalty, (float)fitness);
  }


  fitness_vec[0] = fitness;
  if (fv_len > 7) { 
    fitness_vec[1] = height;
    fitness_vec[2] = surf;
    fitness_vec[3] = strength;
    fitness_vec[4] = penalty;
    fitness_vec[5] = Num_bricks;
    fitness_vec[6] = ratio1;

  } else if (fv_len > 1) {
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}


// Parametric volume/area only.
int fitness_table41(ModelStruct *model)
{
  Real height, surf, surf2, strength, strength2, penalty, fitness;
  int fv_len = model->fv_length;
  Real *fitness_vec = model->fitness;
  Real p_surf = 4.0 * model->build_args[3];
  Real p_strength = 4.0 * model->build_args[3];
  Real ratio2 = 1.0, ratio3 = 1.0;


  height = (Real)Y_max;
  surf = (Real)Weight_y[Y_max];
  surf2 = sqrt(surf);

  strength = volume_score1();
  if (strength < 1.0)
    strength = 1.0;
  if (height > 1.0) {
    strength2 = strength / height;
  } else {
    strength2 = strength;
  }
  strength2 = sqrt(strength2);

  penalty = Num_bricks - Weight_y[Y_max];
  if (penalty < 1.0) {
    penalty = 1.0;
  }

  /*
  printf("h:%d:%d, s:%d, st:%d:%d\n", (int)height, (int)p_height,
	 (int)surf, (int)strength, (int)p_strength);
  */

  if (!balanced || !In_bounds || (Y_max < 2)) {
    fitness = 0.0;
    if (Print[9]) {
      printf("voxels :: fitness_table41() - fail, bal:%d, in:%d, y_max:%d.\n",
	     balanced, In_bounds, Y_max);
    }

  } else if (!valid_area((int)p_surf)) {
    // Surface/area is too big.
    fitness = 0.0;
    if (Print[9]) {
      printf("voxels :: fitness_table41() - fail, outside of area.\n");
    }

  } else {
    // Do fitness adjustment for surface area.
    if (surf2 < p_surf) {
      ratio2 = surf2 / p_surf;
    } else if (surf2 > p_surf) {
      if (surf2 - p_surf > p_surf) {
	ratio2 = 0.0;
      } else {
	ratio2 = (2.0*p_surf - surf2) / p_surf;
	ratio2 *= ratio2;
      }
    }

    // Do fitness adjustment for strength (volume).
    if (strength2 < p_strength) {
      ratio3 = strength2 / p_strength;
    } else if (strength2 > p_strength) {
      if (strength2 - p_strength > p_strength) {
	ratio3 = 0.0;
      } else {
	ratio3 = (2.0*p_strength - strength2) / p_strength;
	ratio3 *= ratio3;
      }
    }
    fitness = height * surf * strength / penalty;
    fitness *= ratio2 * ratio3;
  }


  if (Print[6]) {
    printf("Table Fitness #41:  #%4d  %.2f\n", (int)Num_bricks,
	   (float)(height * surf * strength / penalty));
    printf("  (h:%d)(a:%d:%d:%d:%.2f)(s:%d:%d:%d:%.2f)/(p:%d) - %.2f\n",
	   (int)height,
	   (int)surf, (int)surf2, (int)p_surf, (float)ratio2,
	   (int)strength, (int)strength2, (int)p_strength, (float)ratio3,
	   (int)penalty, (float)fitness);
  }


  fitness_vec[0] = fitness;
  if (fv_len > 7) { 
    fitness_vec[1] = height;
    fitness_vec[2] = surf;
    fitness_vec[3] = strength;
    fitness_vec[4] = penalty;
    fitness_vec[5] = Num_bricks;
    fitness_vec[6] = ratio2*ratio3;

  } else if (fv_len > 1) {
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}


int fitness_table12(ModelStruct *model)
{
  int y;
  int fv_len = model->fv_length;
  int hull_pts, area;
  int footprint[DIM_X][DIM_Z];
  Real surf, strength, penalty;
  Real fitness;
  Real *fitness_vec = model->fitness;

  make_footprint(0, &hull_pts, &area, footprint);


  surf = Weight_y[Y_max];
  //  surf = surface_score5(footprint); // Bricks must be over footprint.
  if (surf < 2) {
    surf = 0;
  }

  strength = load_score2();
  //  strength1 = load_score5();
  if (strength < 1.0)
    strength = 1.0;
  /*
  else
    strength = sqrt(strength1);
  */

  penalty = Num_bricks - Weight_y[Y_max];
  for (y=0; y<Y_max; y++) {
    penalty -= Hull_pts[y];
  }
  if (penalty > surf / 2) {
    penalty = surf / 2;
  }
  if (penalty <= 1.0) {
    penalty = 1.0;
  }
  
  fitness = surf * strength * Y_max / penalty;
  if (fitness > 1000000.0) {
    fitness = floor(fitness / 10000.0);
    fitness *= 10000.0;
  } else if (fitness > 10000.0) {
    fitness = floor(fitness / 100.0);
    fitness *= 100.0;
  } else if (fitness > 1000.0) {
    fitness = floor(fitness / 10.0);
    fitness *= 10.0;
  } else  if (fitness > 10.0) {
    fitness = floor(fitness);
  }

  if (Y_max < 2)
    fitness = 0.0;


  if (Print[6]) {
    printf("Table #12.\n");
    printf("#:%4d  -:%4d  ",
	   Num_bricks, Extras);
    printf("x:%d:%d   z:%d:%d\n", bal_xneg, bal_xpos,
	   bal_zneg, bal_zpos);
    printf("surf: %.0f  strength: %d:%.0f  Y_max: %d  pen:%.0lf\n",
	   (float)surf, Weight_y[Y_max], (float)strength, Y_max, penalty);
    printf("Fitness #12: %.0lf\n", fitness);
  }

  fitness_vec[0] = fitness;
  if (fv_len > 1) { 
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}




int fitness_table10(ModelStruct *model)
{
  int fv_len = model->fv_length;
  int hull_pts, area;
  int footprint[DIM_X][DIM_Z];
  Real surf, surf1, surf2=0.0, strength, strength1, penalty;
  Real penalty1=0.0, fitness;
  Real *fitness_vec = model->fitness;

  //  surf1 = surface_score();
  make_footprint(0, &hull_pts, &area, footprint);
  surf1 = surface_score5(footprint);
  //  surf1 = Weight_y[Y_max];
  //  surf1 = surface_score6(footprint);
  //  surf2 = surface_score2(footprint);
  //  surf2 = surface_score2(footprint);

  strength1 = load_score5();
  strength = strength1;
  //  strength = pow(strength1, 0.5);
  if (strength < 1.0)
    strength = 1.0;

  surf = surf1;
  //  surf = pow(surf1, 0.9);

  penalty = Num_bricks - Weight_y[Y_max];
  if (penalty < 1.0)
    penalty = 1.0;

  fitness = Y_max * surf * strength / penalty;
  if (fitness > 1000000.0) {
    fitness = floor(fitness / 10000.0);
    fitness *= 10000.0;
  } else if (fitness > 10000.0) {
    fitness = floor(fitness / 100.0);
    fitness *= 100.0;
  } else if (fitness > 1000.0) {
    fitness = floor(fitness / 10.0);
    fitness *= 10.0;
  }

  if (Print[6]) {
    printf("Table #10.\n");
    printf("#:%4d  -:%4d  ",
	   Num_bricks, Extras);
    printf("x:%d:%d   z:%d:%d\n", bal_xneg, bal_xpos,
	   bal_zneg, bal_zpos);
    printf("surf: %.0lf(%.0lf:%.0lf:#%d)  strength: %.0lf(%.0lf)  Y_max: %d  pen:%.0lf(%.0lf)\n",
	   surf, surf1, surf2, Weight_y[Y_max], strength, strength1, Y_max, penalty, penalty1);
    printf("Fitness #10: %.0lf\n", fitness);
  }

  fitness_vec[0] = fitness;
  if (fv_len > 7) { 
    fitness_vec[1] = Y_max;
    fitness_vec[2] = surf;
    fitness_vec[3] = strength;
    fitness_vec[4] = penalty;
    //    printf("%d: %d:%d %d\n", Y_max, surf, strength, penalty);
  } else if (fv_len > 1) {
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}




int fitness_table9(ModelStruct *model)
{
  int fv_len = model->fv_length;
  int hull_pts, area;
  int footprint[DIM_X][DIM_Z];
  Real extras;
  Real surf, surf1, surf2=0.0, strength, strength1, penalty;
  Real penalty1, fitness;
  Real *fitness_vec = model->fitness;

  //  surf1 = surface_score();
  make_footprint(0, &hull_pts, &area, footprint);
  //  surf1 = surface_score5(footprint);
  surf1 = surface_score6(footprint);
  //  surf2 = surface_score2(footprint);
  //  surf2 = surface_score2(footprint);

  strength1 = load_score2();
  penalty1 = Num_bricks - Weight_y[Y_max];
  load_score3(&strength1, &penalty1);
  //  strength = strength1;
  strength = pow(strength1, 0.75);
  strength1 = sqrt(strength1);
  if (strength < 1.0)
    strength = 1.0;
  if (penalty1 <= 1.0) {
    penalty = 1.0;
  } else {
    penalty = penalty1;
    //    penalty = pow(penalty1, 0.95);
  //    penalty2 = pow(penalty1, 0.75);
  }


  //  surf = surf1;
  surf = pow(surf1, 0.9);
  //  surf = sqrt(surf1)*sqrt(surf2);
  //  surf = surf1 + surf2;
  //  surf = surf1 + Weight_y[Y_max];
  //  surf = sqrt(surf1) * sqrt(Weight_y[Y_max]);
  extras = pow(Extras, 0.5);
  if (extras > int(surf/2.0)) {
    surf /= 2.0;
  } else {
    surf = surf - extras;
  }

  fitness = surf * strength * Y_max / penalty;

  if (Print[6]) {
    printf("Table #9.\n");
    printf("#:%4d  -:%4d (%.0f)  ",
	   Num_bricks, Extras, extras);
    printf("x:%d:%d   z:%d:%d\n", bal_xneg, bal_xpos,
	   bal_zneg, bal_zpos);
    printf("surf: %.0lf(%.0lf:%.0lf:#%d)  strength: %.0lf(%.0lf)  Y_max: %d  pen:%.0lf(%.0lf)\n",
	   surf, surf1, surf2, Weight_y[Y_max], strength, strength1, Y_max, penalty, penalty1);
    printf("Fitness #9: %.0lf\n", fitness);
  }

  fitness_vec[0] = fitness;
  if (fv_len > 1) { 
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}



int fitness_table8(ModelStruct *model)
{
  int fv_len = model->fv_length;
  int hull_pts, area;
  int footprint[DIM_X][DIM_Z];
  Real extras;
  Real surf, surf1, surf2, strength, strength1, penalty;
  Real penalty1, fitness;
  Real *fitness_vec = model->fitness;

  surf1 = surface_score();
  make_footprint(0, &hull_pts, &area, footprint);
  //  surf2 = surface_score2(footprint);
  surf2 = surface_score3(footprint);

  penalty1 = Num_bricks - Weight_y[Y_max];
  if (penalty1 <= 1.0) {
    penalty = 1.0;
  } else {
    penalty = pow(penalty1, 0.9);
  //    penalty2 = pow(penalty1, 0.75);
  }

  strength1 = load_score2();
  strength = strength1 / penalty;
  if (strength < 1.0)
    strength = 1.0;

  //  surf = surf1;
  surf = sqrt(surf1)*sqrt(surf2);
  extras = pow(Extras, 0.5);
  if (extras > int(surf/2.0)) {
    surf /= 2.0;
  } else {
    surf = surf - extras;
  }

  fitness = surf * strength * Y_max;
  if (fitness > 1000000.0) {
    fitness = floor(fitness / 10000.0);
    fitness *= 10000.0;
  } else if (fitness > 10000.0) {
    fitness = floor(fitness / 100.0);
    fitness *= 100.0;
  } else if (fitness > 1000.0) {
    fitness = floor(fitness / 10.0);
    fitness *= 10.0;
  }


  if (Print[6]) {
    printf("Table #8.\n");
    printf("#:%4d  -:%4d (%.0f)  ",
	   Num_bricks, Extras, extras);
    printf("x:%d:%d   z:%d:%d\n", bal_xneg, bal_xpos,
	   bal_zneg, bal_zpos);
    printf("surf: %.0lf(%.0lf:%.0lf:#%d)  strength: %.0lf(%.0lf)  Y_max: %d  pen:%.0lf(%.0lf)\n",
	   surf, surf1, surf2, Weight_y[Y_max], strength, strength1, Y_max, penalty, penalty1);
    printf("Fitness #8: %.0lf\n", fitness);
  }

  fitness_vec[0] = fitness;
  if (fv_len > 1) { 
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}


#ifdef OLD
int fitness_table8(ModelStruct *model)
{
  int fv_len = model->fv_length;
  int hull_pts, area;
  int footprint[DIM_X][DIM_Z];
  Real extras;
  Real surf, surf1, surf2, strength, strength2, penalty, penalty2, fitness;
  Real *fitness_vec = model->fitness;

  make_footprint(0, &hull_pts, &area, footprint);
  surf1 = surface_score();
  surf1 = sqrt(surf1);

  surf2 = surface_score3(footprint);
  surf2 = surface_score2(footprint);
  surf2 = sqrt(surf2);

  load_score3(&strength, &penalty);

  if (penalty < 1.0)
    penalty2 = 1.0;
  else
    penalty2 = pow(penalty, 0.9);
  //    penalty2 = pow(penalty, 0.75);
  strength2 = strength / penalty2;
  if (strength2 < 1.0)
    strength2 = 1.0;

  surf = surf1 * surf2;
  extras = pow(Extras, 0.5);
  if (extras > int(surf/2.0)) {
    surf /= 2.0;
  } else {
    surf = surf - extras;
  }

  fitness = surf * strength2 * Y_max;

  if (Print[6]) {
    printf("Table #8.\n");
    printf("#:%4d  -:%4d (%.0f)  ",
	   Num_bricks, Extras, extras);
    printf("x:%d:%d   z:%d:%d\n", bal_xneg, bal_xpos,
	   bal_zneg, bal_zpos);
    printf("surf: %.0lf(%.0lf:%.0lf:#%d)  strength: %.0lf(%.0lf)  Y_max: %d  pen:%.0lf(%.0lf)\n",
	   surf, surf1, surf2, Weight_y[Y_max], strength2, strength, Y_max, penalty2, penalty);
    printf("Fitness #8: %.0lf\n", fitness);
  }

  fitness_vec[0] = fitness;
  if (fv_len > 1) { 
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}
#endif



int fitness_table7(ModelStruct *model)
{
  int fv_len = model->fv_length;
  Real surf, surf2, strength, strength2, pen, pen2, fitness;
  Real *fitness_vec = model->fitness;

  surf = surface_score();
  strength = load_score();
  pen = Num_bricks - Weight_y[Y_max];

  if (pen < 1.0)
    pen2 = 1.0;
  else
    pen2 = pow(pen, 0.75);
  strength2 = strength / pen2;

  if (Extras > int(surf/2.0)) {
    surf2 = surf/2.0;
  } else {
    surf2 = surf - Real(Extras);
  }


  fitness = surf2 * strength2 * Y_max;

  if (Print[6]) {
    printf("Table #7.\n");
    printf("#:%4d  -:%4d\n",
	   Num_bricks, Extras);
    printf("x:%d:%d   z:%d:%d\n", bal_xneg, bal_xpos,
	   bal_zneg, bal_zpos);
    printf("surf: %.0lf(%.0lf)  strength: %.0lf(%.0lf)  Y_max: %d  pen:%.0lf(%.0lf)\n",
	   surf2, surf, strength2, strength, Y_max, pen2, pen);
    printf("Fitness #7: %.0lf\n", fitness);
  }

  fitness_vec[0] = fitness;
  if (fv_len > 1) { 
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}


// First, encourage decent, flat surface.
// Uses old footprint calculation, so surfaces are square.
// But, nice square tables.
// Need stronger penalty for excess material.
int fitness_table6(ModelStruct *model)
{
  int hull_pts, area, y;
  int fv_len = model->fv_length;
  int footprint[DIM_X][DIM_Z];
  Real surf, surf1, surf2, strength, pen, len, fitness;
  Real *fitness_vec = model->fitness;

  make_footprint(0, &hull_pts, &area, footprint);
  surf1 = sqrt(surface_score());
  surf2 = surface_score2(footprint);
  surf = sqrt(surf1 * surf2);
  strength = load_score2();
  pen = Num_bricks - Weight_y[Y_max];
  for (y=0; y<Y_max; y++) {
    pen -= Hull_pts[y];
  }
  len = model->lstring_length;


  if (Print[6]) {
    printf("Table #6\n");
    printf("#blocks: %4d  Xtras:%4d  y:%d\n",
	   Num_bricks, Extras, Y_max);
    printf("x:%d:%d   z:%d:%d\n", bal_xneg, bal_xpos,
	   bal_zneg, bal_zpos);
    printf("surf: %.0lf:%.0lf=%.0lf  strength: %.0lf  pen:%.0lf  len:%.0lf",
	   surf1, surf2, surf, strength, pen, len);
  }

  if (pen > strength)
    pen = strength;
  if (pen < 1.0)
    pen = 1.0;
  len -= Num_bricks;
  if (len < 0.0)
    len = 0.0;
  pen = sqrt(pen) + pow(len, 0.75);

  //  strength -= pen;
  if (strength < 0.0)
    strength = 0.0;
  fitness = surf * strength * Y_max / pen;
  if (fitness > sqr(surf))
    fitness = sqr(surf);
  /*
  if (strength > 1.0)
    fitness = surf * strength * Y_max / pen;
  else
    fitness = surf * Y_max / pen;
    */

  if (Print[6]) {
    printf("  : %.0lf\n", pen);
    printf("Table 6:  Fitness #6: %.0lf\n", fitness);
  }

  fitness_vec[0] = fitness;
  if (fv_len > 1) { 
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}


// First, encourage decent, flat surface.
// Uses old footprint calculation, so surfaces are square.
// But, nice square tables.
// Need stronger penalty for excess material.
int fitness_table5(ModelStruct *model)
{
  int hull_pts, area;
  int fv_len = model->fv_length;
  int footprint[DIM_X][DIM_Z];
  Real surf, surf1, surf2, strength, pen, len, fitness;
  Real *fitness_vec = model->fitness;

  make_footprint(0, &hull_pts, &area, footprint);
  surf1 = sqrt(surface_score());
  surf2 = surface_score2(footprint);
  surf = sqrt(surf1 * surf2);
  strength = load_score();
  pen = Num_bricks - Weight_y[Y_max];
  len = model->lstring_length;

  if (Print[6]) {
    printf("Table #5\n");
    printf("#blocks: %4d  Xtras:%4d  y:%d\n",
	   Num_bricks, Extras, Y_max);
    printf("x:%d:%d   z:%d:%d\n", bal_xneg, bal_xpos,
	   bal_zneg, bal_zpos);
    printf("surf: %.0lf:%.0lf=%.0lf  strength: %.0lf  pen:%.0lf  len:%.0lf",
	   surf1, surf2, surf, strength, pen, len);
  }

  if (pen > strength)
    pen = strength;
  if (pen < 1.0)
    pen = 1.0;
  len -= Num_bricks;
  if (len < 0.0)
    len = 0.0;
  pen = sqrt(pen) + pow(len, 0.75);

  //  strength -= pen;
  if (strength < 0.0)
    strength = 0.0;
  //  fitness = surf * strength / pen;
  fitness = surf * strength * Y_max / pen;
  if (fitness > sqr(surf))
    fitness = sqr(surf);

  if (len > fitness / 2.0)
    fitness /= 2.0;
  else
    fitness -= len;


  /*
  if (strength > 1.0)
    fitness = surf * strength * Y_max / pen;
  else
    fitness = surf * Y_max / pen;
    */

  if (Print[6]) {
    printf("  : %.0lf\n", pen);
    printf("Table 5  Fitness #5: %.0lf\n", fitness);
  }

  fitness_vec[0] = fitness;
  if (fv_len > 1) { 
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}



int fitness_table4(ModelStruct *model)
{
  int i;
  int x, y, z;
  int len, max;
  int surf=0, surf_bad=0, surfx=0, surfz=0;
  int fail = FALSE;
  int num_support=0;
  int fv_len = model->fv_length;

  int hull_pts, area;
  int footprint[DIM_X][DIM_Z];
  Real pen = 1.0;
  Real fp=0.0;
  Real fitness = 0.0;
  Real *fitness_vec = model->fitness;

  if (Print[6]) {
    printf("Table #4\n");
    printf("l:%5d  r:%5d  #blocks:%4d  -:%4d\n",
	   bal_xneg, bal_xpos, Num_bricks, Extras);
  }


  if (Num_bricks < 2) {
    for (i=0; i<fv_len; i++)
      fitness_vec[i] = 0.0;

    if (Print[10])
      printf("Failed because not enough bricks: %d.\n", Num_bricks);

    if (Print[6])
      printf("Fitness = 0.\n");

    return FALSE;
  }


  make_footprint(0, &hull_pts, &area, footprint);
  //  printf("# hull points: %d,  area: %d\n", hull_pts, area);


  y = Y_max;
  if (Y_max > 0) {
    for (x=Xmin[y]; x<=Xmax[y]; x++) {
      len = 0;
      max = 0;
      for (z=Zmin[y]; z<=Zmax[y]; z++) {
	if (Model[x][y][z] == M_ctr) {
	  if (footprint[x][z] == M_ctr) {
	    len++;
	  } else {
	    surf_bad++;
	    if (len>0) {
	      surfx += len-1;
	      if (len > max)
		max = len;
	      len = 0;
	    }
	  }

	} else if (len>0) {
	  surfx += len-1;
	  if (len > max)
	    max = len;
	  len = 0;
	}
      }
      if (len>1)
	surfx += len-1;
    }

    for (z=Zmin[y]; z<=Zmax[y]; z++) {
      len = 0;
      max = 0;
      for (x=Xmin[y]; x<=Xmax[y]; x++) {
	if (Model[x][y][z] == M_ctr) {
	  if (footprint[x][z] == M_ctr) {
	    len++;
	  } else {
	    surf_bad++;
	    if (len>0) {
	      surfz += len-1;
	      if (len > max)
		max = len;
	      len = 0;
	    }
	  }

	} else if (len>0) {
	  surfz += len-1;
	  if (len > max)
	    max = len;
	  len = 0;
	}
      }
      if (len>1)
	surfz += len-1;
    }

    for (i=0; i<y; i++) {
      //      estimate_footprint(i, &num, &val, point);
      make_footprint(i, &hull_pts, &area, 0);

      if (i==0)
	area *= 2;
      fp += Real(area);

      if (Weight_y[i] > 2*hull_pts) {
	//	pen += Real(Weight_y[i]-num);
	//	pen += Real(Weight_y[i])/Real(hull_pts);
	pen += Real(Weight_y[i]-2*hull_pts);
	if (Print[10])
	  printf("      fail at y=%d  (%d:%d) p:%.4lf\n", i,
		 Weight_y[i], hull_pts, pen);
      }
      if (hull_pts < 2)
	pen += 10.0;
      else if (hull_pts < 3)
	pen += 5.0;
      num_support += hull_pts;
    }
    if (fail && Print[10])
      printf("Failed because too many wasted bricks in support.\n");
  }

  //  surf = (surfx+surfz) * (Weight_y[y]-surf_bad);
  //  fitness = fp/pen + (surf-pen)*Y_max;


  //   surf = surfx + surfz + Weight_y[y] - surf_bad;
  //  surf = surfx*surfz + Weight_y[y] - surf_bad;
  //  surf = (surfx+surfz) * (Weight_y[y]-surf_bad);
  surf = (surfx+surfz+1) * (Weight_y[y]-surf_bad);
  fitness = fp;
  fitness *= Real(surf*Y_max);


  if ((fitness > 10.0) && (Y_max < 3))
    fitness = 10.0 + sqrt(fitness);
  else if ((fitness > 100.0) && (Y_max == 3))
    fitness = 100.0 + sqrt(fitness);
  else if ((fitness > 200.0) && (Y_max == 4))
    fitness = 200.0 + sqrt(fitness);
  else if ((fitness > 300.0) && (Y_max == 5))
    fitness = 300.0 + sqrt(fitness);

  //  fitness *= Real(surf)*sqrt(Real(Y_max));

  if (bal_xneg/3 > bal_xpos) {
    if (bal_xpos != 0.0)
      pen *= sqr(Real(bal_xneg/3)/Real(bal_xpos));
    //    fail = TRUE;
    if (Print[10])
      printf("Fail because xneg(%d)/3>xpos(%d) p:%.4lf\n",
	     bal_xneg, bal_xpos, pen);
  } else if (bal_xpos/3 > bal_xneg) {
    if (bal_xneg != 0.0)
      pen *= sqr(Real(bal_xpos/3)/Real(bal_xneg));
    //    fail = TRUE;
    if (Print[10])
      printf("Fail because xpos(%d)/3>xneg(%d) p:%.4lf\n",
	     bal_xpos, bal_xneg, pen);
  }

  if (bal_zneg/3 > bal_zpos) {
    if (bal_zpos != 0.0)
      pen *= sqr(Real(bal_zneg/3)/Real(bal_zpos));
    //    fail = TRUE;
    if (Print[10])
      printf("Fail because zneg(%d)/3>zpos(%d) p:%.4lf\n",
	     bal_zneg, bal_zpos, pen);

  } else if (bal_zpos/3 > bal_zneg) {
    if (bal_zneg != 0.0)
      pen *= sqr(Real(bal_zpos/3)/Real(bal_zneg));
    //    fail = TRUE;
    if (Print[10])
      printf("Fail because zpos(%d)/3>zneg(%d) p:%.4lf\n",
	     bal_zpos, bal_zneg, pen);
  }

  if ((bal_xneg==0) && (bal_xpos==0)) {
    //    fail = TRUE;
    if (fitness > 1.0)
      fitness = 1.0;
    //    if (Print[10])
    //      printf("xneg & xpos == 0.\n");
    //      printf("Fail because xneg & xpos == 0.\n");
  }

  if ((bal_zneg==0) && (bal_zpos==0)) {
    //    fail = TRUE;
    if (fitness > 1.0)
      fitness = 1.0;
    //    if (Print[10])
    //      printf("zneg & zpos == 0.\n");
    //      printf("Fail because zneg & zpos == 0.\n");
  }
  

  if (Print[6]) {
    printf("x:%d:%d   z:%d:%d\n", bal_xneg, bal_xpos,
	   bal_zneg, bal_zpos);
    printf("y-max:%d, y-weight:%d  surf:%d(%d:%d) bad:%d #suprt:%d\n",
	   Y_max, Weight_y[y], surf, surfx, surfz, surf_bad, num_support);
    printf("fp:%.0lf  pen:%.2lf   f:%.0lf\n",
	   fp, pen, fitness);
  }

  if (y==0) {
    if (Print[10])
      printf("Fail because y==0.\n");
    fitness = 0.0;
  } else if (y==1) {
    if (fitness > 1.0)
      fitness = 1.0;
  }

  /*
  if (bal_xneg < bal_xpos)
    fitness *= sqrt(bal_xneg);
  else
    fitness *= sqrt(bal_xpos);
  if (bal_zneg < bal_zpos)
    fitness *= sqrt(bal_zneg);
  else
    fitness *= sqrt(bal_zpos);
    */

  if (bal_xneg > 0)
    pen -= sqrt(bal_xneg);
  //    fitness *= sqrt(bal_xneg);
  
  if (bal_xpos > 0)
    pen -= sqrt(bal_xpos);
  //    fitness *= sqrt(bal_xpos);

  if (bal_zneg > 0)
    pen -= sqrt(bal_zneg);
  //    fitness *= sqrt(bal_zneg);

  if (bal_zpos > 0)
    pen -= sqrt(bal_zpos);
  //    fitness *= sqrt(bal_zpos);

  if (pen > 1.0)
    fitness /= pen;

  if (fitness < 0.0) {
    fitness = 0.0;
  } else {
    fitness = sqrt(fitness);
    if (fitness > 1000000.0) {
      fitness = floor(fitness / 10000.0);
      fitness *= 10000.0;
    } else if (fitness > 10000.0) {
      fitness = floor(fitness / 100.0);
      fitness *= 100.0;
    } else if (fitness > 1000.0) {
      fitness = floor(fitness / 10.0);
      fitness *= 10.0;
    }
  }

  if (fv_len < 5) {
    fitness_vec[0] = fitness;
    if (Print[6]) {
      printf("Fitness #1: %.0lf\n", fitness);
    }
    if (fv_len > 1) { 
      fitness_vec[1] = Num_bricks;
    }

  } else {
    fitness_vec[0] = fitness;
    fitness_vec[1] = Num_bricks;
    fitness_vec[2] = Real(surf);
    fitness_vec[3] = fp;
    fitness_vec[4] = Real(Y_max);
    fitness_vec[5] = 0.0;

    if (pen > 0.0)
      pen = -pen;
    fitness_vec[4] = pen;

    if (Print[6]) {
      printf("Fitness #1: %d %d %d %d\n", surf, int(fp), Y_max,
	     int(pen));
    }
  }

  return TRUE;
}





int fitness_table3(ModelStruct *model)
{
  int i;
  int x, y, z;
  int len, max, num, surf, surfx, surfz;
  int fail = FALSE;
  int num_support;
  int fv_len = model->fv_length;
  Real pen = 1.0;
  Real fp = 0, val;
  Real fitness = 0.0;
  Real *fitness_vec = model->fitness;
  XZPoint point[4];

  if (Print[6]) {
    printf("Table #3\n");
    printf("l:%5d  r:%5d  #:%4d  -:%4d\n",
	   bal_xneg, bal_xpos, Num_bricks, Extras);
  }

  y = Y_max;
  surf = 0;

  /*
  for (x=Xmin[y]; x<=Xmax[y]; x++)
    for (z=Zmin[y]; z<=Zmax[y]; z++) {
      if (Model[x][y][z] == m_ctr)
	surf++;
    }
    */

  surfx = 0;
  for (x=Xmin[y]; x<=Xmax[y]; x++) {
    len = 0;
    max = 0;
    for (z=Zmin[y]; z<=Zmax[y]; z++) {
      if (Model[x][y][z] == M_ctr) {
	len++;
      } else if (len>0) {
	//	surf += len*len;
	surfx += len-1;
	if (len > max)
	  max = len;
	len = 0;
      }
    }
    //    surf += len*len;
    if (len>0)
      surfx += len-1;
  }

  surfz=0;
  for (z=Zmin[y]; z<=Zmax[y]; z++) {
    len = 0;
    max = 0;
    for (x=Xmin[y]; x<=Xmax[y]; x++) {
      if (Model[x][y][z] == M_ctr) {
	len++;
      } else if (len>0) {
	//	surf += len*len;
	surfz += len-1;
	if (len > max)
	  max = len;
	len = 0;
      }
    }
    //    surf += len*len;
    if (len>0)
      surfz += len-1;
  }
  surf = surfx * surfz + Weight_y[y];

  num_support = 0;
  for (i=0; i<y; i++) {
    estimate_footprint(i, &num, &val, point);
    /*
    if (Weight_y[i] > num) {
      val = Real(num) * (val - Real(Weight_y[i]-num));
    }
    */
      //	val /= sqrt(Real(Weight_y[i]-num));
	//	val /= Real(Weight_y[i]-num);

    if (i==0)
      val *= 2.0;
    fp += val * num;

    //    pen += (4 - num);

    //    if (val <= 1.0) // should be less than 1.
    //    if (Weight_y[i] / num > 2) {
    if (Weight_y[i] > num) {
      //      pen *= sqr(Real(Weight_y[i])/Real(num*2));
      pen += Real(Weight_y[i]-num);
      if (Print[10])
	printf("      fail at y=%d  (%d:%d) p:%.4lf\n", i,
	       Weight_y[i], num, pen);
      //      fail = TRUE;
    }

    //      fp += val/Real(Weight_y[i]-num);
    num_support += num;
  }
  if (fail && Print[10])
    printf("Failed because too many wasted bricks in support.\n");


  fitness = fp;
  fitness *= Real(surf*y);

  //  fitness *= sqrt(estimate_footprint(0));
  //  penalty = absv(bal_xneg-bal_xpos) + absv(bal_zneg-bal_zpos);
  //  penalty = Num_bricks - surf - num_support;

  if (bal_xneg/3 > bal_xpos) {
    if (bal_xpos != 0.0)
      pen *= sqr(Real(bal_xneg/3)/Real(bal_xpos));
    //    fail = TRUE;
    if (Print[10])
      printf("Fail because xneg(%d)/3>xpos(%d) p:%.4lf\n",
	     bal_xneg, bal_xpos, pen);
  } else if (bal_xpos/3 > bal_xneg) {
    if (bal_xneg != 0.0)
      pen *= sqr(Real(bal_xpos/3)/Real(bal_xneg));
    //    fail = TRUE;
    if (Print[10])
      printf("Fail because xpos(%d)/3>xneg(%d) p:%.4lf\n",
	     bal_xpos, bal_xneg, pen);
  }

  if (bal_zneg/3 > bal_zpos) {
    if (bal_zpos != 0.0)
      pen *= sqr(Real(bal_zneg/3)/Real(bal_zpos));
    //    fail = TRUE;
    if (Print[10])
      printf("Fail because zneg(%d)/3>zpos(%d) p:%.4lf\n",
	     bal_zneg, bal_zpos, pen);

  } else if (bal_zpos/3 > bal_zneg) {
    if (bal_zneg != 0.0)
      pen *= sqr(Real(bal_zpos/3)/Real(bal_zneg));
    //    fail = TRUE;
    if (Print[10])
      printf("Fail because zpos(%d)/3>zneg(%d) p:%.4lf\n",
	     bal_zpos, bal_zneg, pen);
  }

  if ((bal_xneg==0) && (bal_xpos==0)) {
    //    fail = TRUE;
    if (fitness > 1.0)
      fitness = 1.0;
    //    if (Print[10])
    //      printf("xneg & xpos == 0.\n");
    //      printf("Fail because xneg & xpos == 0.\n");
  }

  if ((bal_zneg==0) && (bal_zpos==0)) {
    //    fail = TRUE;
    if (fitness > 1.0)
      fitness = 1.0;
    //    if (Print[10])
    //      printf("zneg & zpos == 0.\n");
    //      printf("Fail because zneg & zpos == 0.\n");
  }
  

  if (Print[6]) {
    printf("x:%d:%d   z:%d:%d\n", bal_xneg, bal_xpos,
	   bal_zneg, bal_zpos);
    printf("y:%d, #:%d  surf:%d  #suprt:%d  fp:%.0lf  pen:%.2lf   f:%.0lf\n",
	   y, Weight_y[y], surf, num_support, fp, pen, fitness);
    //    printf("%.0lf\n", fitness);
  }

  if (y==0) {
    if (Print[10])
      printf("Fail because y==0.\n");
    fitness = 0.0;
  } else if (y==1) {
    if (fitness > 1.0)
      fitness = 1.0;
  }

  /*
  if (bal_xneg < bal_xpos)
    fitness *= sqrt(bal_xneg);
  else
    fitness *= sqrt(bal_xpos);
  if (bal_zneg < bal_zpos)
    fitness *= sqrt(bal_zneg);
  else
    fitness *= sqrt(bal_zpos);
    */

  if (bal_xneg > 0)
    pen -= sqrt(bal_xneg);
  //    fitness *= sqrt(bal_xneg);
  
  if (bal_xpos > 0)
    pen -= sqrt(bal_xpos);
  //    fitness *= sqrt(bal_xpos);

  if (bal_zneg > 0)
    pen -= sqrt(bal_zneg);
  //    fitness *= sqrt(bal_zneg);

  if (bal_zpos > 0)
    pen -= sqrt(bal_zpos);
  //    fitness *= sqrt(bal_zpos);

  if (pen > 1.0)
    fitness /= pen;

  if (fitness < 0.0)
    fitness = 0.0;

  if (Print[6]) {
    printf("Fitness: %.0lf\n", fitness);
  }

  fitness_vec[0] = fitness;
  if (fv_len > 1) { 
    fitness_vec[1] = Num_bricks;
  }
  return TRUE;
}


// formerly "fitness_table2"
int fitness_table11(ModelStruct *model)
{
  int i;
  int x, y, z;
  int len, max, num;
  int surf=0, surfx=0, surfz=0;
  int fail = FALSE;
  int num_support=0;
  int fv_len = model->fv_length;
  Real pen = 1.0;
  Real fp=0.0, val;
  Real fitness = 0.0;
  Real *fitness_vec = model->fitness;
  XZPoint point[4];

  if (Print[6]) {
    printf("Table #11\n");
    printf("l:%5d  r:%5d  #:%4d  -:%4d\n",
	   bal_xneg, bal_xpos, Num_bricks, Extras);
  }


  if (Num_bricks < 2) {
    for (i=0; i<fv_len; i++)
      fitness_vec[i] = 0.0;

    if (Print[10])
      printf("Failed because not enough bricks: %d.\n", Num_bricks);

    if (Print[6])
      printf("Fitness = 0.\n");

    return FALSE;
  }

  y = Y_max;
  if (Y_max > 0) {
    for (x=Xmin[y]; x<=Xmax[y]; x++) {
      len = 0;
      max = 0;
      for (z=Zmin[y]; z<=Zmax[y]; z++) {
	if (Model[x][y][z] == M_ctr) {
	  len++;
	} else if (len>0) {
	  surfx += len-1;
	  if (len > max)
	    max = len;
	  len = 0;
	}
      }
      if (len>1)
	surfx += len-1;
    }

    for (z=Zmin[y]; z<=Zmax[y]; z++) {
      len = 0;
      max = 0;
      for (x=Xmin[y]; x<=Xmax[y]; x++) {
	if (Model[x][y][z] == M_ctr) {
	  len++;
	} else if (len>0) {
	  surfz += len-1;
	  if (len > max)
	    max = len;
	  len = 0;
	}
      }
      if (len>1)
	surfz += len-1;
    }
    surf = surfx*surfz + Weight_y[y];

    for (i=0; i<y; i++) {
      estimate_footprint(i, &num, &val, point);

      if (i==0)
	val *= 2.0;
      fp += val * num;

      if (Weight_y[i] > num) {
	//	pen += Real(Weight_y[i]-num);
	pen += Real(Weight_y[i])/Real(num);
	if (Print[10])
	  printf("      fail at y=%d  (%d:%d) p:%.4lf\n", i,
		 Weight_y[i], num, pen);
	//      fail = TRUE;
      }
      num_support += num;
    }
    if (fail && Print[10])
      printf("Failed because too many wasted bricks in support.\n");
  }

  fitness = fp;
  fitness *= Real(surf*Y_max);
  if ((fitness > 10.0) && (Y_max < 3))
    fitness = 10.0 + sqrt(fitness);
  else if ((fitness > 100.0) && (Y_max == 3))
    fitness = 100.0 + sqrt(fitness);
  else if ((fitness > 200.0) && (Y_max == 4))
    fitness = 200.0 + sqrt(fitness);
  else if ((fitness > 300.0) && (Y_max == 5))
    fitness = 300.0 + sqrt(fitness);

  //  fitness *= Real(surf)*sqrt(Real(Y_max));

  if (bal_xneg/3 > bal_xpos) {
    if (bal_xpos != 0.0)
      pen *= sqr(Real(bal_xneg/3)/Real(bal_xpos));
    //    fail = TRUE;
    if (Print[10])
      printf("Fail because xneg(%d)/3>xpos(%d) p:%.4lf\n",
	     bal_xneg, bal_xpos, pen);
  } else if (bal_xpos/3 > bal_xneg) {
    if (bal_xneg != 0.0)
      pen *= sqr(Real(bal_xpos/3)/Real(bal_xneg));
    //    fail = TRUE;
    if (Print[10])
      printf("Fail because xpos(%d)/3>xneg(%d) p:%.4lf\n",
	     bal_xpos, bal_xneg, pen);
  }

  if (bal_zneg/3 > bal_zpos) {
    if (bal_zpos != 0.0)
      pen *= sqr(Real(bal_zneg/3)/Real(bal_zpos));
    //    fail = TRUE;
    if (Print[10])
      printf("Fail because zneg(%d)/3>zpos(%d) p:%.4lf\n",
	     bal_zneg, bal_zpos, pen);

  } else if (bal_zpos/3 > bal_zneg) {
    if (bal_zneg != 0.0)
      pen *= sqr(Real(bal_zpos/3)/Real(bal_zneg));
    //    fail = TRUE;
    if (Print[10])
      printf("Fail because zpos(%d)/3>zneg(%d) p:%.4lf\n",
	     bal_zpos, bal_zneg, pen);
  }

  if ((bal_xneg==0) && (bal_xpos==0)) {
    //    fail = TRUE;
    if (fitness > 1.0)
      fitness = 1.0;
    //    if (Print[10])
    //      printf("xneg & xpos == 0.\n");
    //      printf("Fail because xneg & xpos == 0.\n");
  }

  if ((bal_zneg==0) && (bal_zpos==0)) {
    //    fail = TRUE;
    if (fitness > 1.0)
      fitness = 1.0;
    //    if (Print[10])
    //      printf("zneg & zpos == 0.\n");
    //      printf("Fail because zneg & zpos == 0.\n");
  }
  

  if (Print[6]) {
    printf("x:%d:%d   z:%d:%d\n", bal_xneg, bal_xpos,
	   bal_zneg, bal_zpos);
    printf("y-max:%d, y-weight:%d  surf:%d(%d:%d)  #suprt:%d\n",
	   Y_max, Weight_y[y], surf, surfx, surfz, num_support);
    printf("fp:%.0lf  pen:%.2lf   f:%.0lf\n",
	   fp, pen, fitness);
  }

  if (y==0) {
    if (Print[10])
      printf("Fail because y==0.\n");
    fitness = 0.0;
  } else if (y==1) {
    if (fitness > 1.0)
      fitness = 1.0;
  }

  /*
  if (bal_xneg < bal_xpos)
    fitness *= sqrt(bal_xneg);
  else
    fitness *= sqrt(bal_xpos);
  if (bal_zneg < bal_zpos)
    fitness *= sqrt(bal_zneg);
  else
    fitness *= sqrt(bal_zpos);
    */

  if (bal_xneg > 0)
    pen -= sqrt(bal_xneg);
  //    fitness *= sqrt(bal_xneg);
  
  if (bal_xpos > 0)
    pen -= sqrt(bal_xpos);
  //    fitness *= sqrt(bal_xpos);

  if (bal_zneg > 0)
    pen -= sqrt(bal_zneg);
  //    fitness *= sqrt(bal_zneg);

  if (bal_zpos > 0)
    pen -= sqrt(bal_zpos);
  //    fitness *= sqrt(bal_zpos);

  if (pen > 1.0)
    fitness /= pen;

  if (fitness < 0.0)
    fitness = 0.0;
  else
    fitness = sqrt(fitness);

  if (fv_len < 5) {
    fitness_vec[0] = fitness;
    if (fv_len > 1) { 
      fitness_vec[1] = Num_bricks;
    }
    if (Print[6]) {
      printf("Fitness #2: %.0lf\n", fitness);
    }

  } else {
    fitness_vec[0] = fitness;
    if (fv_len > 1) { 
      fitness_vec[1] = Num_bricks;
    }
    fitness_vec[2] = Real(surf);
    fitness_vec[3] = fp;
    fitness_vec[4] = Real(Y_max);
    fitness_vec[5] = 0.0;

    if (pen > 0.0)
      pen = -pen;
    fitness_vec[6] = pen;

    if (Print[6]) {
      printf("Fitness #11: %d %d %d %d\n", surf, int(fp), Y_max,
	     int(pen));
    }
  }

  return TRUE;
}



int fitness_table(ModelStruct *model)
{
  int fv_len = model->fv_length;
  Real surf, surf2, strength, strength2, pen, fitness;
  Real *fitness_vec = model->fitness;

  surf = surface_score();
  strength = load_score();
  pen = Num_bricks - Weight_y[Y_max];

  if (pen < 1.0)
    pen = 1.0;

  if (Extras > int(surf/2.0)) {
    surf2 = surf/2.0;
  } else {
    surf2 = surf - Real(Extras);
  }

  if (pen > strength) {
    strength2 = 1.0 / (pen - strength);
  } else {
    strength2 = strength-pen;
  }
  if (strength2 < 0.0)
    strength2 = 0.0;

  fitness = surf2 * strength2 * Y_max;

  /*
  if (strength > 1.0)
    fitness = surf * strength * Y_max / pen;
  else
    fitness = surf * Y_max / pen;
    */

  if (Print[6]) {
    printf("Table.\n");
    printf("#:%4d  -:%4d\n",
	   Num_bricks, Extras);
    printf("x:%d:%d   z:%d:%d\n", bal_xneg, bal_xpos,
	   bal_zneg, bal_zpos);
    printf("surf: %.0lf(%.0lf)  strength: %.0lf(%.0lf)  Y_max: %d  pen:%.0lf\n",
	   surf2, surf, strength2, strength, Y_max, pen);
    printf("Fitness #4: %.0f\n", (float)fitness);
  }

  fitness_vec[0] = fitness;
  if (fv_len > 1) { 
    fitness_vec[1] = Num_bricks;
  }

  return TRUE;
}




int fitness_brick_blind_watchmaker(int fv_len, Real *fitness_vec)
{
  if (Y_max < 0) {
    if (Print[10])
      printf("Fail because Y_max == 0.\n");
    fitness_vec[0] = 0.0;
    return FALSE;

  } else if (Num_bricks < 2) {
    if (Print[10])
      printf("Fail because number of bricks < 3.\n");
    fitness_vec[0] = 0.0;
    return FALSE;
  }

  fitness_vec[0] = 1.0;
  if (fv_len > 1) { 
    fitness_vec[1] = Num_bricks;
  }
  return TRUE;
}


Real fitness_bridge(void)
{
  int i, j;
  int total, max, current;
  int sum;
  int left0, right0, tot0, sum0, crnt0, max0;
  Real fitness;

  for (i=DIM_Y-1; i>=0; i--) {
    if (Weight_y[i] > 0)
      break;
  }

  sum = 0;
  total = 0;
  max = 0;
  current = 0;
  for (j=0; j<DIM_X; j++) {
    if (Model[j][i][0] > 0) {
      total++;
      current++;
    } else if (current > 1) {
      sum += sqr(current-2);
      if (current > max)
	max = current;
      current = 0;
    }
  }

  sum += sqr((max-2));
  fitness = pow(Real(bal_xneg), 0.5);
  fitness += pow(Real(bal_xpos), 0.5);
  //  fitness -= Real(Extras / Num_bricks);
  fitness += Real(sum*Real(i) + total*i);

  if (Print[6]) {
    printf("l:%5d  r:%5d  #:%4d  -:%4d\n",
	   bal_xneg, bal_xpos, Num_bricks, Extras);
    printf("y:%2d  total:%3d  sum:%3d  max:%3d  -> %5.2f\n", i, total, sum, max, fitness);
  }

  sum0 = 0;
  tot0 = 0;
  max0 = 0;
  crnt0 = 0;
  left0 = DIM_X;
  right0 = 0;
  for (j=0; j<DIM_X; j++) {
    if (Model[j][0][0] > 0) {
      tot0++;
      crnt0++;
      if (j<left0)
	left0 = j;
      if (j> right0)
	right0 = j;
    } else if (crnt0 > 1) {
      sum0 += sqr(crnt0-2);
      if (crnt0 > max0)
	max0 = crnt0;
      crnt0 = 0;
    }
  }
  sum0 += sqr((max0-2));

  fitness -= Real(sum0*Real(i) + tot0*i);
  fitness /= sqrt(Real(Num_bricks));
  fitness *= sqrt(Real(right0 - left0));

  if (i<=2)
    fitness = 0.0;

  if (Print[6]) {
    printf("y:0  total:%3d  sum:%3d  max:%3d  -> %5.2f\n",
	   tot0, sum0, max0, fitness);
  }

  if (fitness < 0.0)
    fitness = 0.0;

  return fitness;
}



int fitness_struct(ModelStruct *model)
{
  if (Fitness_func == 1) {
    return fitness_table4(model);
    
  } else if (Fitness_func == 2) {
    return fitness_staircase1(model->fv_length, model->fitness);

  } else if (Fitness_func == 3) {
    return fitness_staircase2(model->fv_length, model->fitness);

  } else if (Fitness_func == 4) {
    return fitness_table(model);
    //    return fitness_staircase3(model->fv_length, model->fitness);

  } else if (Fitness_func == 5) {
    return fitness_table5(model);

  } else if (Fitness_func == 6) {
    return fitness_table6(model);

  } else if (Fitness_func == 7) {
    return fitness_table7(model);

  } else if (Fitness_func == 8) {
    return fitness_table8(model);

  } else if (Fitness_func == 9) {
    return fitness_table9(model);

  } else if (Fitness_func == 10) {
    return fitness_table10(model);

  } else if (Fitness_func == 11) {
    return fitness_table11(model);

  } else if (Fitness_func == 12) {
    return fitness_table12(model);

  } else if (Fitness_func == 16) {
    return fitness_table16(model);

  } else if (Fitness_func == 17) {
    return fitness_table17(model);

  } else if (Fitness_func == 18) {
    return fitness_table18(model);

  } else if (Fitness_func == 19) {
    return fitness_table19(model);

  } else if (Fitness_func == 36) { // Multiple test single L-system.
    return fitness_table36(model);

  } else if (Fitness_func == 37) { // Multiple test single L-system.
    return fitness_table37(model);

  } else if (Fitness_func == 38) { // Multiple test single L-system.
    return fitness_table38(model);

  } else if (Fitness_func == 39) { // Multiple test single L-system.
    return fitness_table39(model);

  } else if (Fitness_func == 40) { // Multiple test single L-system.
    return fitness_table40(model);

  } else if (Fitness_func == 41) { // Multiple test single L-system.
    return fitness_table41(model);

  } else if (Fitness_func == 15) {
    model->fitness[0] = fitness_tiles();
    return TRUE;
  }
  return FALSE;
}



Real calc_fitness_3d_plant(void)
{
  int x, y, z;
  Real fitness = 0.0;

  for (x=X_min; x<=X_max; x++)
    for (y=Y_min; y<=Y_max; y++)
      for (z=Z_min; z<=Z_max; z++)
	if (Model[x][y][z] == M_ctr) {
	  if (x == X_min)
	    fitness += Real(y);
	  else if (Model[x-1][y][z] != M_ctr)
	    fitness += Real(y);

	  // X pos.
	  if (x == X_max)
	    fitness += Real(y);
	  else if (Model[x+1][y][z] != M_ctr)
	    fitness += Real(y);

	  // Y neg.
	  if (y == Y_min)
	    fitness += Real(y);
	  else if (Model[x][y-1][z] != M_ctr)
	    fitness += Real(y);

	  // Y pos.
	  if (y == Y_max)
	    fitness += Real(y);
	  else if (Model[x][y+1][z] != M_ctr)
	    fitness += Real(y);

	  // Z neg.
	  if (z == Z_min)
	    fitness += Real(y);
	  else if (Model[x][y][z-1] != M_ctr)
	    fitness += Real(y);

	  // Z pos.
	  if (z == Z_max)
	    fitness += Real(y);
	  else if (Model[x][y][z+1] != M_ctr)
	    fitness += Real(y);
	}

  fitness += Real(bal_xneg + bal_xpos + bal_zneg + bal_zpos);

  if (Print[6])
    printf("#:%d  x-:%d, x+:%d\n", Num_bricks,
	   bal_xneg, bal_xpos);

  return fitness;
}


Real calc_fitness_3d(void)
{
  int i;
  Real fitness = 0.0;

  if (Print[6])
    printf("#:%d  x-:%d, x+:%d\n", Num_bricks,
	   bal_xneg, bal_xpos);

  for (i=0; i<DIM_Y; i++)
    fitness += Real(i*Weight_y[i]);

  fitness += Real(bal_xneg + bal_xpos + bal_zneg + bal_zpos);

  if (Print[6])
    printf("x:%d:%d  z:%d:%d\n", bal_xneg, bal_xpos,
	   bal_zneg, bal_zpos);

  return fitness;
}


Real calc_fitness_plant2(void)
{
  int x, y, z;
  Real fitness = 0.0;

  for (x=X_min; x<=X_max; x++)
    for (y=Y_min; y<=Y_max; y++)
      for (z=Z_min; z<=Z_max; z++)
	if (Model[x][y][z] == M_ctr) {
	  if (x == X_min)
	    fitness += 1.0;
	  else if (Model[x-1][y][z] != M_ctr)
	    fitness += 1.0;

	  // X pos.
	  if (x == X_max)
	    fitness += 1.0;
	  else if (Model[x+1][y][z] != M_ctr)
	    fitness += 1.0;

	  // Y neg.
	  if (y == Y_min)
	    fitness += 1.0;
	  else if (Model[x][y-1][z] != M_ctr)
	    fitness += 1.0;

	  // Y pos.
	  if (y == Y_max)
	    fitness += 1.0;
	  else if (Model[x][y+1][z] != M_ctr)
	    fitness += 1.0;

	  // Z neg.
	  if (z == Z_min)
	    fitness += 1.0;
	  else if (Model[x][y][z-1] != M_ctr)
	    fitness += 1.0;

	  // Z pos.
	  if (z == Z_max)
	    fitness += 1.0;
	  else if (Model[x][y][z+1] != M_ctr)
	    fitness += 1.0;
	}

  fitness += Real(bal_xneg + bal_xpos + bal_zneg + bal_zpos);
  fitness /= sqrt(Real(Num_bricks));

  if (Print[6])
    printf("#:%d  x-:%d, x+:%d\n", Num_bricks,
	   bal_xneg, bal_xpos);

  return fitness;
}


int push_stack(void)
{
  int x, y;

  if (!Pen_down) {
    Pen_down = TRUE;
    put_brick();
  }

  if (Stack_ptr == STACK_SIZE) {
    if (Print[0] || Print[10])
      printf("Build:: stack size exceeded.\n");
    return FALSE;
  }
  Stack[Stack_ptr].x = X_loc;
  Stack[Stack_ptr].y = Y_loc;
  Stack[Stack_ptr].z = Z_loc;
  Stack[Stack_ptr].pen_down = Pen_down;
  for (x=0; x<3; x++)
    for (y=0; y<3; y++)
      Stack[Stack_ptr].heading[x][y] = Heading[x][y];
  Stack_ptr++;
  return TRUE;
}


int pop_stack(void)
{
  int x, y;

  if (Stack_ptr-- == 0)
    return FALSE;

  X_loc = Stack[Stack_ptr].x;
  Y_loc = Stack[Stack_ptr].y;
  Z_loc = Stack[Stack_ptr].z;
  Pen_down = Stack[Stack_ptr].pen_down;
  for (x=0; x<3; x++)
    for (y=0; y<3; y++)
      Heading[x][y] = Stack[Stack_ptr].heading[x][y];
  return TRUE;
}


int move_up(void)
{
  X_loc += Heading[0][2];
  Y_loc += Heading[1][2];
  Z_loc += Heading[2][2];
  return put_brick();
}

int move_down(void)
{
  X_loc -= Heading[0][2];
  Y_loc -= Heading[1][2];
  Z_loc -= Heading[2][2];
  return put_brick();
}

int move_left(void)
{
  X_loc += Heading[0][1];
  Y_loc += Heading[1][1];
  Z_loc += Heading[2][1];
  return put_brick();
}

int move_right(void)
{
  X_loc -= Heading[0][1];
  Y_loc -= Heading[1][1];
  Z_loc -= Heading[2][1];
  return put_brick();
}

int move_forward(void)
{
  X_loc += Heading[0][0];
  Y_loc += Heading[1][0];
  Z_loc += Heading[2][0];
  return put_brick();
}

int move_backward(void)
{
  X_loc -= Heading[0][0];
  Y_loc -= Heading[1][0];
  Z_loc -= Heading[2][0];
  return put_brick();
}

int place_straight(void)
{
  if (!move_forward())
    return FALSE;
  if (!move_forward())
    return FALSE;
  if (!move_forward())
    return FALSE;
  return TRUE;
}

int place_bent(void)
{
  if (!move_forward())
    return FALSE;
  if (!move_forward())
    return FALSE;
  turn_up();
  if (!move_forward())
    return FALSE;
  return TRUE;
}


// Build structure using parameterized relative coordinates.
// A direction is specified, the heading is set to that
// direction and moves forward the specified number of
// units/repititions.
int build_structure_param_rel2(int length, LString *command,
			       Real *args)
{
  int i, j, reps;

  initialize_variables();
  X_loc += int(args[0]);
  Y_loc += int(args[1]);
  Z_loc += int(args[2]);


  if (command[0].symbol == TERM_PENDOWN) {
    //    cout << "Putting initial brick" << endl;
    Pen_down = TRUE;
    put_brick();
  } else
    Pen_down = FALSE;
  Pen_down = TRUE;
  put_brick();

  for (i=0; i<length; i++) {
    if (command[i].symbol == 0) {
      if (Print[0])
	printf("Build:: unexpected early termination of command sequence @:%d(!%d)\n", i, length);
      return FALSE;
    } else if (command[i].symbol == TERM_PENUP) {
      Pen_down = FALSE;

    } else if (command[i].symbol == TERM_PENDOWN) {
      Pen_down = TRUE;

    } else if (command[i].symbol == TERM_PUSH) {
      if (!push_stack())
	return FALSE;

    } else if (command[i].symbol == TERM_POP) {
      if (!pop_stack())
	return FALSE;

    } else if (command[i].symbol == TERM_UP) {
      turn_up();
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	if (!move_forward())
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_DOWN) {
      turn_down();
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	if (!move_forward())
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_LEFT) {
      turn_left();
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	if (!move_forward())
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_RIGHT) {
      turn_right();
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	if (!move_forward())
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_FORWARD) {
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	if (!move_forward())
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_BACKWARD) {
      turn_right();
      turn_right();
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	if (!move_forward())
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_TURN_UP) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_up();

    } else if (command[i].symbol == TERM_TURN_DOWN) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_down();

    } else if (command[i].symbol == TERM_TURN_LEFT) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_left();

    } else if (command[i].symbol == TERM_TURN_RIGHT) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_right();

    } else if (command[i].symbol == TERM_TURN_CLOCKWISE) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_clockwise();

    } else if (command[i].symbol == TERM_TURN_CNTR_CLOCKWISE) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_cntr_clockwise();

    } else {
#ifndef OPTIMIZE
	if (Print[0])
	  printf("Build:: unknown command: (%c:%d)\n", 
		 command[i].symbol, int(command[i].symbol));
#endif
    }     
  }

  return TRUE;
}

// Build structure using parameterized relative coordinates.
int build_structure_param_rel(int length, LString *command,
			      Real *args)
{
  int i, j, reps;

  initialize_variables();
  X_loc += int(args[0]);
  Y_loc += int(args[1]);
  Z_loc += int(args[2]);


  if (command[0].symbol == TERM_PENDOWN) {
    //    cout << "Putting initial brick" << endl;
    Pen_down = TRUE;
    put_brick();
  } else
    Pen_down = FALSE;
  Pen_down = TRUE;
  put_brick();

  for (i=0; i<length; i++) {
    if (command[i].symbol == 0) {
      if (Print[0])
	printf("Build:: unexpected early termination of command sequence @:%d(!%d)\n", i, length);
      return FALSE;
    } else if (command[i].symbol == TERM_PENUP) {
      Pen_down = FALSE;

    } else if (command[i].symbol == TERM_PENDOWN) {
      Pen_down = TRUE;

    } else if (command[i].symbol == TERM_PUSH) {
      if (!push_stack())
	return FALSE;

    } else if (command[i].symbol == TERM_POP) {
      if (!pop_stack())
	return FALSE;

    } else if (command[i].symbol == TERM_UP) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++) {
	if (!move_up())
	  return FALSE;
	turn_up();
      }

    } else if (command[i].symbol == TERM_DOWN) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++) {
	if (!move_down())
	  return FALSE;
	turn_down();
      }

    } else if (command[i].symbol == TERM_LEFT) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++) {
	if (!move_left())
	  return FALSE;
	turn_left();
      }

    } else if (command[i].symbol == TERM_RIGHT) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++) {
	if (!move_right())
	  return FALSE;
	turn_right();
      }

    } else if (command[i].symbol == TERM_FORWARD) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++) {
	if (!move_forward())
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_BACKWARD) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++) {
	if (!move_backward())
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_TURN_UP) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_up();

    } else if (command[i].symbol == TERM_TURN_DOWN) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_down();

    } else if (command[i].symbol == TERM_TURN_LEFT) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_left();

    } else if (command[i].symbol == TERM_TURN_RIGHT) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_right();

    } else if (command[i].symbol == TERM_TURN_CLOCKWISE) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_clockwise();

    } else if (command[i].symbol == TERM_TURN_CNTR_CLOCKWISE) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_cntr_clockwise();

    } else {
#ifndef OPTIMIZE
	if (Print[0])
	  printf("Build:: unknown command: (%c:%d)\n", 
		 command[i].symbol, int(command[i].symbol));
#endif
    }     
  }

  return TRUE;
}


// Build structure using parameterized absolute coordinates.
// Stops building in a direction when it encounters an
// existing brick.
int build_structure_param_obstruct(int length, LString *command, 
				   Real *args)
{
  int i, j, reps;
  int ret_val, obstructed = FALSE;

  initialize_variables();
  X_loc += int(args[0]);
  Y_loc += int(args[1]);
  Z_loc += int(args[2]);

  if (command[0].symbol == TERM_PENDOWN) {
    //    cout << "Putting initial brick" << endl;
    Pen_down = TRUE;
    put_brick();
  } else
    Pen_down = FALSE;
  Pen_down = TRUE;
  put_brick();

  for (i=0; i<length; i++) {
    if (obstructed) {
      while ((command[i].symbol != TERM_POP) &&
	     (i<length)) {
	i++;
      };
      if (i >= length)
	break;
      obstructed = FALSE;
    }

    if (Print[13]) {
      printf("%d: %c(%d)\n", i, command[i].symbol,
	     int(command[i].arg));
    }

    if (command[i].symbol == 0) {
      if (Print[0])
	printf("Build:: unexpected early termination of command sequence @:%d(!%d)\n", i, length);
      return FALSE;
    } else if (command[i].symbol == TERM_PENUP) {
      Pen_down = FALSE;

    } else if (command[i].symbol == TERM_PENDOWN) {
      Pen_down = TRUE;

    } else if (command[i].symbol == TERM_PUSH) {
      if (!push_stack())
	return FALSE;

    } else if (command[i].symbol == TERM_POP) {
      if (!pop_stack())
	return FALSE;

    } else if (command[i].symbol == TERM_UP) {
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	ret_val = move_up();
	if (ret_val == -1)
	  obstructed = TRUE;
	else if (ret_val == FALSE)
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_DOWN) {
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	ret_val = move_down();
	if (ret_val == -1)
	  obstructed = TRUE;
	else if (ret_val == FALSE)
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_LEFT) {
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	ret_val = move_left();
	if (ret_val == -1)
	  obstructed = TRUE;
	else if (ret_val == FALSE)
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_RIGHT) {
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	ret_val = move_right();
	if (ret_val == -1)
	  obstructed = TRUE;
	else if (ret_val == FALSE)
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_FORWARD) {
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	ret_val = move_forward();
	if (ret_val == -1)
	  obstructed = TRUE;
	else if (ret_val == FALSE)
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_BACKWARD) {
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	ret_val = move_backward();
	if (ret_val == -1)
	  obstructed = TRUE;
	else if (ret_val == FALSE)
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_TURN_UP) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_up();

    } else if (command[i].symbol == TERM_TURN_DOWN) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_down();

    } else if (command[i].symbol == TERM_TURN_LEFT) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_left();

    } else if (command[i].symbol == TERM_TURN_RIGHT) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_right();

    } else if (command[i].symbol == TERM_TURN_CLOCKWISE) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_clockwise();

    } else if (command[i].symbol == TERM_TURN_CNTR_CLOCKWISE) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_cntr_clockwise();

    } else {
#ifndef OPTIMIZE
	if (Print[0])
	  printf("Build:: unknown command: (%c:%d)\n", 
		 command[i].symbol, int(command[i].symbol));
#endif
    }     
  }

  return TRUE;
}


// Build structure using parameterized absolute coordinates.
int build_structure_param(int length, LString *command,
			  Real *args)
{
  int i, j, reps;


  //  printf("build_bricks :: build_structure_param()\n");

  initialize_variables();
  X_loc += int(args[0]);
  Y_loc += int(args[1]);
  Z_loc += int(args[2]);

  //  printf("%d:%d:%d\n", X_loc, Y_loc, Z_loc);

  if (command[0].symbol == TERM_PENDOWN) {
    //    cout << "Putting initial brick" << endl;
    Pen_down = TRUE;
    put_brick();
  } else
    Pen_down = FALSE;
  Pen_down = TRUE;
  put_brick();

  for (i=0; i<length; i++) {
    if (Print[13]) {
      printf("%d: %c(%d)\n", i, command[i].symbol,
	     int(command[i].arg));
    }

    if (command[i].symbol == 0) {
      if (Print[0])
	printf("Build:: unexpected early termination of command sequence @:%d(!%d)\n", i, length);
      return FALSE;
    } else if (command[i].symbol == TERM_PENUP) {
      Pen_down = FALSE;

    } else if (command[i].symbol == TERM_PENDOWN) {
      Pen_down = TRUE;

    } else if (command[i].symbol == TERM_PUSH) {
      if (!push_stack())
	return FALSE;

    } else if (command[i].symbol == TERM_POP) {
      if (!pop_stack())
	return FALSE;

    } else if (command[i].symbol == TERM_UP) {
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	if (!move_up())
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_DOWN) {
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	if (!move_down())
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_LEFT) {
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	if (!move_left())
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_RIGHT) {
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	if (!move_right())
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_FORWARD) {
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	if (!move_forward())
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_BACKWARD) {
      reps = int(command[i].arg);
      for (j=0; j<reps; j++) {
	if (!move_backward())
	  return FALSE;
      }

    } else if (command[i].symbol == TERM_TURN_UP) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_up();

    } else if (command[i].symbol == TERM_TURN_DOWN) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_down();

    } else if (command[i].symbol == TERM_TURN_LEFT) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_left();

    } else if (command[i].symbol == TERM_TURN_RIGHT) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_right();

    } else if (command[i].symbol == TERM_TURN_CLOCKWISE) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_clockwise();

    } else if (command[i].symbol == TERM_TURN_CNTR_CLOCKWISE) {
      reps = int(command[i].arg) % 4;
      for (j=0; j<reps; j++)
	turn_cntr_clockwise();

    } else {
#ifndef OPTIMIZE
	if (Print[0])
	  printf("Build:: unknown command: (%c:%d)\n", 
		 command[i].symbol, int(command[i].symbol));
#endif
    }     
  }

  return TRUE;
}


#define SCALE 4

void write_stl_xneg(FILE *file, int x, int y, int z)
{
  //  printf("xneg: %3d %3d %3d\n", x, y, z);
  x *= SCALE;
  y *= SCALE;
  z *= SCALE;
  
  fprintf(file, " facet normal -1.0 0.0 0.0\n");
  fprintf(file, "  outer loop\n");
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y), Real(z));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y), Real(z+SCALE));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y+SCALE), Real(z));
  fprintf(file, "  endloop\n");
  fprintf(file, " endfacet\n");

  fprintf(file, " facet normal -1.0 0.0 0.0\n");
  fprintf(file, "  outer loop\n");
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y+SCALE), Real(z+SCALE));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y+SCALE), Real(z));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y), Real(z+SCALE));
  fprintf(file, "  endloop\n");
  fprintf(file, " endfacet\n");
}

void write_stl_xpos(FILE *file, int x, int y, int z)
{
  //  printf("xpos: %3d %3d %3d\n", x, y, z);
  x *= SCALE;
  y *= SCALE;
  z *= SCALE;

  fprintf(file, " facet normal 1.0 0.0 0.0\n");
  fprintf(file, "  outer loop\n");
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y), Real(z));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y+SCALE), Real(z));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y), Real(z+SCALE));
  fprintf(file, "  endloop\n");
  fprintf(file, " endfacet\n");

  fprintf(file, " facet normal 1.0 0.0 0.0\n");
  fprintf(file, "  outer loop\n");
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y+SCALE), Real(z+SCALE));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y), Real(z+SCALE));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y+SCALE), Real(z));
  fprintf(file, "  endloop\n");
  fprintf(file, " endfacet\n");
}


void write_stl_yneg(FILE *file, int x, int y, int z)
{
  //  printf("yneg: %3d %3d %3d\n", x, y, z);
  x *= SCALE;
  y *= SCALE;
  z *= SCALE;

  fprintf(file, " facet normal 0.0 -1.0 0.0\n");
  fprintf(file, "  outer loop\n");
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y), Real(z));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y), Real(z));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y), Real(z+SCALE));
  fprintf(file, "  endloop\n");
  fprintf(file, " endfacet\n");

  fprintf(file, " facet normal 0.0 -1.0 0.0\n");
  fprintf(file, "  outer loop\n");
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y), Real(z+SCALE));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y), Real(z+SCALE));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y), Real(z));
  fprintf(file, "  endloop\n");
  fprintf(file, " endfacet\n");
}

void write_stl_ypos(FILE *file, int x, int y, int z)
{
  //  printf("ypos: %3d %3d %3d\n", x, y, z);
  x *= SCALE;
  y *= SCALE;
  z *= SCALE;

  fprintf(file, " facet normal 0.0 1.0 0.0\n");
  fprintf(file, "  outer loop\n");
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y+SCALE), Real(z));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y+SCALE), Real(z+SCALE));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y+SCALE), Real(z));
  fprintf(file, "  endloop\n");
  fprintf(file, " endfacet\n");

  fprintf(file, " facet normal 0.0 1.0 0.0\n");
  fprintf(file, "  outer loop\n");
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y+SCALE), Real(z+SCALE));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y+SCALE), Real(z));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y+SCALE), Real(z+SCALE));
  fprintf(file, "  endloop\n");
  fprintf(file, " endfacet\n");
}


void write_stl_zneg(FILE *file, int x, int y, int z)
{
  //  printf("zneg: %3d %3d %3d\n", x, y, z);
  x *= SCALE;
  y *= SCALE;
  z *= SCALE;

  fprintf(file, " facet normal 0.0 0.0 -1.0\n");
  fprintf(file, "  outer loop\n");
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y), Real(z));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y+SCALE), Real(z));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y), Real(z));
  fprintf(file, "  endloop\n");
  fprintf(file, " endfacet\n");

  fprintf(file, " facet normal 0.0 0.0 -1.0\n");
  fprintf(file, "  outer loop\n");
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y+SCALE), Real(z));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y), Real(z));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y+SCALE), Real(z));
  fprintf(file, "  endloop\n");
  fprintf(file, " endfacet\n");
}

void write_stl_zpos(FILE *file, int x, int y, int z)
{
  //  printf("zpos: %3d %3d %3d\n", x, y, z);
  x *= SCALE;
  y *= SCALE;
  z *= SCALE;

  fprintf(file, " facet normal 0.0 0.0 1.0\n");
  fprintf(file, "  outer loop\n");
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y), Real(z+SCALE));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y), Real(z+SCALE));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y+SCALE), Real(z+SCALE));
  fprintf(file, "  endloop\n");
  fprintf(file, " endfacet\n");

  fprintf(file, " facet normal 0.0 0.0 1.0\n");
  fprintf(file, "  outer loop\n");
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y+SCALE), Real(z+SCALE));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x), Real(y+SCALE), Real(z+SCALE));
  fprintf(file, "   vertex %f %f %f\n",
	  Real(x+SCALE), Real(y), Real(z+SCALE));
  fprintf(file, "  endloop\n");
  fprintf(file, " endfacet\n");
}



void write_stl(char *fname)
{
  int ctr;
  int x, y, z;
  int xx, yy, zz;
  FILE *file = fopen(fname, "w");
  char model2[DIM_X][DIM_Y][DIM_Z];

  // Note: default is hollow.

  // Make Model2 a copy of Model.
  for (x=X_min; x<=X_max; x++) {
    for (y=Y_min; y<=Y_max; y++)
      for (z=Z_min; z<=Z_max; z++)
	if (Model[x][y][z] == M_ctr)
	  model2[x][y][z] = M_ctr;
	else
	  model2[x][y][z] = 0;
      }

  if (!Print[25]) {
    // Option to NOT make hollow.
    // Remove interior cubes.
    for (x=X_min+1; x<X_max; x++)
      for (y=Y_min+1; y<Y_max; y++)
	for (z=Z_min+1; z<Z_max; z++)
	  if (Model[x][y][z] == M_ctr) {
	    ctr = 0;
	    // printf("test: %d,%d,%d->1\n", x, y, z);
	    for (xx=x-1; xx<=x+1; xx++)
	      for (yy=y-1; yy<=y+1; yy++)
		for (zz=z-1; zz<=z+1; zz++) {
		  if (Model[xx][yy][zz] == M_ctr) {
		    ctr++;
		    // printf("%d,%d,%d->1\n", xx, yy, zz);
		  } else {
		    xx = DIM_X;
		    yy = DIM_Y;
		    zz = DIM_Z;
		    break;
		  }
		}
	    if (ctr == 27) {
	      model2[x][y][z] = 0;
	    }
	  }
  }

  // Write cubes to file.
  fprintf(file, "# Structure evolved by G.S. Hornby\n");
  for (x=X_min; x<=X_max; x++) {
    for (y=Y_min; y<=Y_max; y++) {
      for (z=Z_min; z<=Z_max; z++) {
	if (model2[x][y][z] == M_ctr) {
	  // X neg.
	  if (x == X_min)
	    write_stl_xneg(file, x, y, z);
	  else if (model2[x-1][y][z] != M_ctr)
	    write_stl_xneg(file, x, y, z);

	  // X pos.
	  if (x == X_max)
	    write_stl_xpos(file, x, y, z);
	  else if (model2[x+1][y][z] != M_ctr)
	    write_stl_xpos(file, x, y, z);

	  // Y neg.
	  if (y == Y_min)
	    write_stl_yneg(file, x, y, z);
	  else if (model2[x][y-1][z] != M_ctr)
	    write_stl_yneg(file, x, y, z);

	  // Y pos.
	  if (y == Y_max)
	    write_stl_ypos(file, x, y, z);
	  else if (model2[x][y+1][z] != M_ctr)
	    write_stl_ypos(file, x, y, z);

	  // Z neg.
	  if (z == Z_min)
	    write_stl_zneg(file, x, y, z);
	  else if (model2[x][y][z-1] != M_ctr)
	    write_stl_zneg(file, x, y, z);

	  // Z pos.
	  if (z == Z_max)
	    write_stl_zpos(file, x, y, z);
	  else if (model2[x][y][z+1] != M_ctr)
	    write_stl_zpos(file, x, y, z);
	}
      }
    }
  }

  fclose(file);
}
