/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#ifndef BUILD_BRICKS_HEADER_FILE
#define BUILD_BRICKS_HEADER_FILE

#include "general.h"

class Graphic;
class Individual;
class Pop;
class Deme;

//#define DIM_X 40
//#define DIM_Y 40
//#define DIM_Z 40

#define DIM_X 40
#define DIM_Y 40
#define DIM_Z 40

//define DIM_X 100
//define DIM_Y 50
//define DIM_Z 100

typedef struct tXZPoint
{
  int x, z;
} XZPoint;

typedef struct tXZPoint_CH
{
  int x, z;
  Real slope;
} XZPoint_CH;

typedef struct tStack
{
  int x;
  int y;
  int z;
  int pen_down;
  int heading[3][3];
} Stack_struct;


#define STACK_SIZE 100
extern int X_loc, Y_loc, Z_loc;
extern int Stack_ptr;
extern Stack_struct Stack[STACK_SIZE];


extern int Num_bricks;
extern int M_ctr;
extern int Model[DIM_X][DIM_Y][DIM_Z];

extern int balanced;
extern int balance_pts_x[DIM_X];
extern int balance_pts_z[DIM_Z];
extern int Weight_x[DIM_X];
extern int Weight_y[DIM_Y];
extern int Weight_z[DIM_Z];
extern int Extras;

extern int bal_xpos, bal_xneg;
extern int bal_zpos, bal_zneg;
extern int Bal_xpos[DIM_Y], Bal_xneg[DIM_Y];
extern int Bal_zpos[DIM_Y], Bal_zneg[DIM_Y];
extern int Cap_xpos[DIM_Y], Cap_xneg[DIM_Y];
extern int Cap_zpos[DIM_Y], Cap_zneg[DIM_Y];

void draw_structure(int arg);
void print_structure(void);

void write_stl(char *fname);

Real surface_score(void);
int surface_score2(int footprint[DIM_X][DIM_Z]);
Real surface_score3(int footprint[DIM_X][DIM_Z]);
int surface_score4(int footprint[DIM_X][DIM_Z]);
int surface_score5(int footprint[DIM_X][DIM_Z]);
Real surface_score6(int footprint[DIM_X][DIM_Z]);


int fitness_fence1(int fv_len, Real *fitness_vec);
int fitness_staircase2(int fv_len, Real *fitness_vec);
int fitness_staircase3(int fv_len, Real *fitness_vec);
int fitness_staircase1(int fv_len, Real *fitness_vec);
int fitness_table10(ModelStruct *model);
int fitness_table9(ModelStruct *model);
int fitness_table8(ModelStruct *model);
int fitness_table7(ModelStruct *model);
int fitness_table6(ModelStruct *model);
int fitness_table5(ModelStruct *model);
int fitness_table4(ModelStruct *model);
int fitness_table3(ModelStruct *model);
int fitness_table2(ModelStruct *model);
int fitness_table(ModelStruct *model);
Real fitness_bridge(void);
Real fitness_tiles(void);
int fitness_brick_blind_watchmaker(int fv_len, Real *fitness_vec);
int fitness_struct(ModelStruct *model);

int test_balance(void);
int build_structure_param_rel2(int length, LString *command, Real *args);
int build_structure_param_rel(int length, LString *command, Real *args);
int build_structure_param_obstruct(int length, LString *command, Real *args);
int build_structure_param(int length, LString *command, Real *args);

#endif
