/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stream.h>
#include <fstream.h>
#include <limits.h>
#include "deme.h"
#include "pop.h"
#include "individual.h"
#include "general.h"
#include "stats.h"
#include "prod_body.h"
#include "multi_obj.h"


FILE *mut_pheno_file = 0;
FILE *recomb_pheno_file = 0;
char mut_pheno_file_name[50];
char recomb_pheno_file_name[50];


Deme :: Deme(int nindivids)
{
  deme_id = -1;

  cycle_index = 0;
  structure = STRUCT_STEADY_STATE;
  structure = STRUCT_CANONICAL;
  nghbr_size = 10;
  nghbr_type = NGHBR_NONE;
  max_age = MAX_AGE;
  adaptive_mut = FALSE;
  mutate_scale = 1.0;
  prev_scale = mutate_scale;
  num_individs = nindivids;
  replace_size = 1;
  tournament_size = 3;
  elitism = 0;
  pressure = 0.1;
  recomb_rate = 0.5;
  sample_individ = NULL;

  if ((Option[18]) && (mut_pheno_file == 0)) {
    if (Config.filename_stem[0] != 0) {
      int i;
      for (i=0; i<20; i++) {
	mut_pheno_file_name[i] = Config.filename_stem[i];
	if (mut_pheno_file_name[i] == 0)
	  break;
      }
      mut_pheno_file_name[i++] = '_';
      mut_pheno_file_name[i++] = 'm';
      mut_pheno_file_name[i++] = 'u';
      mut_pheno_file_name[i++] = 't';
      mut_pheno_file_name[i++] = '_';
      mut_pheno_file_name[i++] = 'p';
      mut_pheno_file_name[i++] = 'h';
      mut_pheno_file_name[i++] = 'e';
      mut_pheno_file_name[i++] = 'n';
      mut_pheno_file_name[i++] = 'o';
      mut_pheno_file_name[i++] = '.';
      mut_pheno_file_name[i++] = 'd';
      mut_pheno_file_name[i++] = 'a';
      mut_pheno_file_name[i++] = 't';
      mut_pheno_file_name[i] = 0;
      mut_pheno_file = fopen(mut_pheno_file_name, "w");
      fprintf(mut_pheno_file, "# mutation: gen p-fit dist1 edist1 c-fit dist2 edist2 m-fit\n");
      printf("Opening <%s> to write pheno data.\n", mut_pheno_file_name);

    } else {
      mut_pheno_file = fopen("mut_pheno.dat", "w");
      printf("Opening >mut_pheno.dat< to write mutation pheno data.\n");
    }

    if ((Option[18] > 1) && (recomb_pheno_file == 0)) {
      if (Config.filename_stem[0] != 0) {
	int i;
	for (i=0; i<20; i++) {
	  recomb_pheno_file_name[i] = Config.filename_stem[i];
	  if (recomb_pheno_file_name[i] == 0)
	    break;
	}
	recomb_pheno_file_name[i++] = '_';
	recomb_pheno_file_name[i++] = 'r';
	recomb_pheno_file_name[i++] = 'e';
	recomb_pheno_file_name[i++] = 'c';
	recomb_pheno_file_name[i++] = '_';
	recomb_pheno_file_name[i++] = 'p';
	recomb_pheno_file_name[i++] = 'h';
	recomb_pheno_file_name[i++] = 'e';
	recomb_pheno_file_name[i++] = 'n';
	recomb_pheno_file_name[i++] = 'o';
	recomb_pheno_file_name[i++] = '.';
	recomb_pheno_file_name[i++] = 'd';
	recomb_pheno_file_name[i++] = 'a';
	recomb_pheno_file_name[i++] = 't';
	recomb_pheno_file_name[i] = 0;
	recomb_pheno_file = fopen(recomb_pheno_file_name, "w");
	fprintf(recomb_pheno_file, "# recombination: gen p-fit dist1 edist1 c-fit dist2 edist2 r-fit\n");
	printf("Opening <%s> to write recombination pheno data.\n", recomb_pheno_file_name);
	
      } else {
	recomb_pheno_file = fopen("recomb_pheno.dat", "w");
	printf("Opening >recomb_pheno.dat< to write pheno data.\n");
      }
    }
  }
}


void Deme :: create(void)
{
  int i;
  double sum;
  Real max_prob = 0.0;

  max_fitness = -DBL_MAX;
  min_children = (int*)malloc(num_individs*sizeof(int));

  select_prob = (double *)malloc(num_individs*sizeof(double));
  select_prob[0] = pressure;
  for (i=1; i<num_individs; i++)
    select_prob[i] = pressure*pow(1.0-pressure, (double)i);

  sum = 0.0;
  for (i=0; i<num_individs; i++)
    sum += select_prob[i];

  for (i=0; i<num_individs; i++)
    select_prob[i] = select_prob[i]*(double)num_individs/sum;

  /*
  for (i=0; i<num_individs; i++)
    printf("%2d  %5.3f\n", i, select_prob[i]);
    */

  for (i=0; i<num_individs; i++) {
    if (select_prob[i] > 1.0) {
      min_children[i] = (int)select_prob[i];
      select_prob[i] -= (double)min_children[i];
    } else {
      /*
      if (max_prob == 0.0) {
	max_prob = select_prob[i];
	for (j=0; j<i; j++)
	  select_prob[j] = max_prob;
      }
      */
      min_children[i] = 0;
    }
  }

  max_prob = 0.0;
  for (i=0; i<num_individs; i++)
    max_prob += select_prob[i];

  for (i=0; i<num_individs; i++)
    select_prob[i] /= max_prob;


  if (Print[16]) {
    for (i=0; i<num_individs; i++)
      printf("%3d: %2d  %10.8f  %f\n", i, min_children[i], select_prob[i],
	     select_prob[i]);
  }

  for (i=1; i<num_individs; i++)
    select_prob[i] += select_prob[i-1];

  age = (int *)malloc(num_individs * sizeof(int));

  if (HOF_SIZE > 0) {
    hof = (Individual**)malloc(HOF_SIZE*sizeof(Individual*));
    for (i=0; i<HOF_SIZE; i++)
      hof[i] = new Individual();
  }

  individ = (Individual**)malloc(num_individs*sizeof(Individual*));
  individ_new = (Individual**)malloc(num_individs*sizeof(Individual*));
  for (i=0; i<num_individs; i++) {
    individ[i] = new Individual(sample_individ);
    individ_new[i] = new Individual(sample_individ);
  }
}


void Deme :: initialize(void)
{
  int i;
  long int num_evals = 0;

  if (Config.num_trials == 1) {
    printf("Needed: %d.\n", num_individs);
  }
  for (i=0; i<num_individs; i++) {
    if (!individ[i]->created) {
      individ[i]->init();
    }
    individ[i]->make_random();

    if (!individ_new[i]->created) {
      individ_new[i]->init();
    }
    individ_new[i]->make_random();
  }

  if (structure != STRUCT_BLIND_WATCHMAKER) {
    FILE *file;
#ifdef NETWORK
    file = fopen("first.net", "w");
#elif TREE
    file = fopen("first.prg", "w");
#else
    file = fopen("first.ind", "w");
#endif
    individ[0]->write(file);
    fclose(file);
  }

  for (i=0; i<num_individs; i++) {
    num_evals++;
    individ[i]->evaluate(TRUE);
    if (Print[8]) {
      individ[i]->print_history_full();
      cout << endl;
    }

    while (individ[i]->fitness_fail()) {
      num_evals++;
      individ[i]->make_random();
      individ[i]->evaluate(TRUE);
      if (Print[8]) {
	printf("\n");
	individ[i]->print_history_full();
	cout << endl;
      }
    }

    if (Option[16] == 2) {
      Config.do_graphics = TRUE;
      individ[i]->evaluate(TRUE);
      individ[i]->evaluate(TRUE);
      Config.do_graphics = FALSE;
    }

    if (Config.num_trials == 1) {
      printf("%d (%d : %ld) ", i, num_individs, num_evals);
      individ[i]->print_fitness();
      printf("\n");
    }

    if (i==0) {
      FILE *file;
#ifdef NETWORK
      file = fopen("first.net", "w");
#elif TREE
      file = fopen("first.prg", "w");
#else
      file = fopen("first.ind", "w");
#endif
      if (file == NULL) {
	printf("Error creating file 'first.ind'.\n");
      } else {
	individ[0]->write(file);
	fclose(file);
      }
    }
  }

  if (Config.num_trials == 1) {
    printf("\n");
  }
}

void Deme :: initialize(Individual *ind_seed)
{
  int i;


  for (i=0; i<num_individs; i++) {
    individ[i]->init();
    individ[i]->duplicate(ind_seed);
#ifdef NETWORK
    for (int j=0; j<10; j++)
      individ[i]->mutate();
#else
    individ[i]->mutate_init();
#endif

    individ_new[i]->init();
    individ_new[i]->duplicate(ind_seed);
    individ_new[i]->evaluate();
  }


  if (structure != STRUCT_BLIND_WATCHMAKER) {
    FILE *file;
#ifdef NETWORK
    file = fopen("first.net", "w");
#elif TREE
    file = fopen("first.prg", "w");
#else
    file = fopen("first.ind", "w");
#endif
    if (file == NULL) {
      printf("Error creating file 'first.ind'.\n");
    } else {
      individ[0]->write(file);
      fclose(file);
    }
  }


  for (i=0; i<num_individs; i++) {
    if (structure == STRUCT_BLIND_WATCHMAKER) {
      while (1) {
	individ[i]->duplicate(ind_seed);
	individ[i]->evaluate(FALSE);
	if (!individ[i]->fitness_fail())
	  break;
      }	
    }

    if (Print[8]) {
      individ[i]->print_history_full();
      cout << endl;
    }

    if (i==0) {
      FILE *file;
#ifdef NETWORK
      file = fopen("first.net", "w");
#elif TREE
      file = fopen("first.prg", "w");
#else
      file = fopen("first.ind", "w");
#endif
      individ[0]->write(file);
      fclose(file);
    }

    individ[i]->init_hist();
  }
}


void Deme :: initialize_simple(void)
{
  int i;

  for (i=0; i<num_individs; i++) {
    individ[i]->init();
    individ[i]->make_random();
    individ_new[i]->init();
    individ_new[i]->make_random(); // necessary?
  }
}

void Deme :: initialize_simple(Individual *ind_seed)
{
  int i, j;

  for (i=0; i<num_individs; i++) {
    individ[i]->init();
    individ[i]->duplicate(ind_seed);

    //    individ[i]->make_random();
    for (j=0; j<15; j++)
      individ[i]->mutate();
    individ[i]->init_hist();

    individ_new[i]->init();
    individ_new[i]->make_random(); // necessary?
  }
}


void Deme :: update_best_vec(double *pop_best_vec)
{
  for (int i=0; i<Num_fit_dims; i++) {
    pop_best_vec[i] = best_vec[i];
  }
}


void Deme :: update_stats(int print_stats)
{
  int i;

  //  print_stats = TRUE;

  avg_fitness = 0.0;
  max_fitness = -DBL_MAX;
  for (i=0; i<num_individs; i++) {
    avg_fitness += individ[i]->get_fitness();
    if (individ[i]->get_fitness() >= max_fitness) {
      individ[i]->get_fitness(best_vec);
      max_fitness = individ[i]->get_fitness();
      max_index = i;
    }
  }
  avg_fitness = avg_fitness / (double)num_individs;

  if (structure != STRUCT_BLIND_WATCHMAKER) {
    if (print_stats) {
      /*
      printf("%3d: %4.0f  %4.0f ", pop->generation,
	     avg_fitness, max_fitness);
      individ[max_index]->print_history_self2();
      individ[max_index]->print_history();
      */

      printf("%3d: <%.0f> ", pop->generation,
	     avg_fitness);
      individ[max_index]->print_history_self();
      individ[max_index]->print_history();

      printf("  ");

      /*
      test_random_ind();

      printf("  ");
      */

      print_time();
      cout << endl;
    }
  }

  if (pop->log) {
    fprintf(pop->log, "%3d: %4.0f %4.0f ", pop->generation,
	   avg_fitness, max_fitness);
    individ[max_index]->fprint_history_self2(pop->log);
    individ[max_index]->fprint_history(pop->log);
    fprintf(pop->log, "\n");
  }
}


void Deme :: update_stats_vector(Stats *stats_vector)
{
  int i;
  Real fitness_vec[Num_fit_dims];

  for (i=0; i<num_individs; i++) {
    individ[i]->get_fitness(fitness_vec);
    stats_vector->multi_add(fitness_vec);
    stats_vector->multi_again();
  }
}


int ind_compare(const void *elt1, const void *elt2)
{
  Individual *ind1, *ind2;

  
  ind1 = *(Individual**)elt1;
  ind2 = *(Individual**)elt2;

  if (ind1->get_fitness() < ind2->get_fitness())
    return 1;
  else if (ind1->get_fitness() > ind2->get_fitness())
    return -1;

  return 0;
}


Individual **individs;

int ind_compare2(const void *elt1, const void *elt2)
{
  int index1, index2;
  Individual *ind1, *ind2;

  index1 = *(int*)elt1;
  index2 = *(int*)elt2;
  ind1 = individs[index1];
  ind2 = individs[index2];

  if (ind1->get_fitness() < ind2->get_fitness())
    return 1;
  else if (ind1->get_fitness() > ind2->get_fitness())
    return -1;

  return 0;
}


int Deme :: get_index(void)
{
  int bot, mid, top;
  double val, total = select_prob[num_individs-1];

  bot = 0;
  top = num_individs-1;
  mid = top/2;

  val = random_dbl()*total;
  while (bot != top) {
    mid = (top-bot)/2 + bot;
    if (select_prob[mid] < val) {
      if (bot == mid)
	return top;
      bot = mid;
	continue;
    } else if (mid == bot) {
      //	  cout << "got "<<mid<<"\n" << endl;
      return mid;
    } else if (select_prob[mid-1] < val) {
      //	  cout << "got "<<mid<<"\n" << endl;
      return mid;
    } else {
      top = mid;
      continue;
    }
  }

  return bot; // not used.
}


void Deme :: display(int index)
{
  individ[index]->display();
}


Real Deme :: evaluate(int index)
{
  return individ[index]->evaluate();
}


int Deme :: reevaluate(void)
{
  int i;

  for (i=0; i<num_individs; i++)
    if ((age[i] >= max_age) && (max_age != 0)) {
      //      individ[i]->set_creation(CREATE_REEVALUATED);
      individ[i]->evaluate();
      if (Print[8]) {
	individ[i]->print_history_full();
      }
      age[i] = 0;
      return TRUE;
    }
  return FALSE;
}

int Deme :: reevaluate_test(void)
{
  int i;

  for (i=0; i<num_individs; i++)
    if ((age[i] >= max_age) && (max_age != 0)) {
      age[i] = 0;
      return i;
    }
  return -1;
}


int Deme :: recombination_test(void)
{
  if (random_dbl() < recomb_rate)
    return TRUE;
  return FALSE;
}


void Deme :: post_evaluate(void)
{
  int i;
  Real rec_rate, mut_rate;

  num_rec_better = 0;
  num_mut_better = 0;

  for (i=0; i<num_individs; i++) {
    if (individ[i]->evaluated != 2) {
      if (individ[i]->create_mutate()) {
	if (individ[i]->better() == 1)
	  num_mut_better++;
      } else {
	if (individ[i]->better() == 1)
	  num_rec_better++;
      }
    }
  }
  if (Print[14])
    cout << endl;

  mut_rate = Real(num_mut_better) / Real(num_mutate);
  rec_rate = Real(num_rec_better) / Real(num_recomb);
  if ((absv(mut_rate - rec_rate) > 0.025 ) && adaptive_mut) {
    if ((mut_rate > rec_rate) && (mutate_scale <= prev_scale)) {
      prev_scale = mutate_scale;
      mutate_scale *= 2.0;
    } else if ((mut_rate < rec_rate) && (mutate_scale >= prev_scale)) {
      prev_scale = mutate_scale;
      mutate_scale /= 2.0;
    } else
      prev_scale = mutate_scale;
  }
}


int Deme :: mutate_bw(int p1, int c1)
{
  num_mutate++;
  individ[c1]->duplicate(individ[p1]);
  individ[c1]->mutate();
  individ[c1]->evaluate();

#ifndef OPTIMIZE
  if (Print[8]) {
    individ[c1]->print_history_full();
    cout << endl;
  }
#endif

#ifndef OPTIMIZE2  
  Op_used(0);
  Op_used(OP_MUTATE);
#endif

  if (individ[c1]->better() == 1) {
#ifndef OPTIMIZE2
    Op_update(TRUE);
#endif
    return TRUE;
  }

#ifndef OPTIMIZE2
  Op_update(FALSE);
#endif
  if (individ[c1]->get_fitness() >
      individ[p1]->get_fitness()/10.0) {
    return TRUE;
  }

  return FALSE;
}

Individual *ind_pheno = 0;
Individual *ind_pheno2 = 0;

#define MAX_TRIES 2

int Deme :: mutate(int p1, int c1)
{
  int j, k;
  int dist1, dist2, edit_dist1, edit_dist2, num;
  int successful = FALSE;
  int successful2 = FALSE;
  int num_prod;
  int max_bodies;
  int max_prod_len;
  int prod_len;


  for (j=0; j<MAX_TRIES; j++) {
    num_mutate++;
    individ_new[c1]->duplicate(individ[p1]);
    //    individ_new[c1]->write(stdout);
    individ_new[c1]->mutate();
    individ_new[c1]->evaluate();

#ifndef OPTIMIZE2
    Op_used(0);
    Op_used(OP_MUTATE);

    if (individ_new[c1]->better() == 1)
      Op_update(TRUE);
    else
      Op_update(FALSE);
#endif

#ifndef OPTIMIZE
    if (Print[8]) {
      individ_new[c1]->print_history_full();
      cout << endl;
    }
#endif

    if (individ_new[c1]->get_fitness() >
	       individ[p1]->get_fitness()/10.0) {
      successful = TRUE;
      break;
    }
  }

  if (Option[18]) {
    if (ind_pheno == 0) {
      num_prod = Num_productions;
      max_bodies = Max_bodies;
      max_prod_len = Max_prod_length;
      prod_len = Prod_length;

      Num_productions = 1;
      Max_bodies = 1;
      Max_prod_length = MAX_STRING_LENGTH;
      Prod_length = MAX_STRING_LENGTH;

      ind_pheno = new Individual();
      ind_pheno->init_pheno(TRUE);

      ind_pheno2 = new Individual();
      ind_pheno2->init_pheno(TRUE);

      Num_productions = num_prod;
      Max_bodies = max_bodies;
      Max_prod_length = max_prod_len;
      Prod_length = prod_len;
    }

    for (j=0; j<MAX_TRIES; j++) {
      if (!ind_pheno->duplicate_pheno(individ[p1])) {
	FILE *file;
	file = fopen("invalid.ind", "w");
	individ[p1]->write(file);
	fclose(file);
	successful2 = FALSE;
	continue;
      }


      successful2 = TRUE;
	
      // Compares mutation performance between generative and its direct equivalent.
      num_prod = Num_productions;
      max_bodies = Max_bodies;
      max_prod_len = Max_prod_length;
      prod_len = Prod_length;
	
      Num_productions = 1;
      Max_bodies = 1;
      Max_prod_length = MAX_STRING_LENGTH;
      Prod_length = MAX_STRING_LENGTH;

      // Decide on number of mutations.
      if (random_int(2)) {
	num = random_int(4) + random_int(4);
      } else {
	num = random_int(7);
      }
      for (k=0; k<num; k++) {
	ind_pheno->mutate_pheno();
      }
      ind_pheno->evaluate();

      Num_productions = num_prod;
      Max_bodies = max_bodies;
      Max_prod_length = max_prod_len;
      Prod_length = prod_len;

      if (ind_pheno->get_fitness() >
	  individ[p1]->get_fitness()/10.0) {
	break;
      }
    }
	  
    //	individ_new[c1]->set_fitness(6, dist2); // distance from parent.
    //	individ_new[c1]->set_fitness(7, ind_pheno->get_fitness()); // parent's fitness
    
    if (successful2) {
      dist1 = individ_new[c1]->phenotype_dist(individ[p1]);	
      dist2 = ind_pheno->phenotype_dist(individ[p1]);

      if (Option[18] >= 3) {
	edit_dist1 = individ_new[c1]->edit_distance(individ[p1]);	
	edit_dist2 = ind_pheno->edit_distance(individ[p1]);
      } else {
	edit_dist1 = -1;
	edit_dist2 = -1;
      }

      fprintf(mut_pheno_file, "%d %.2f %d %d %.2f %d %d %.2f\n",
	      pop->generation,
	      individ[p1]->get_fitness(),
	      dist1, edit_dist1,
	      individ_new[c1]->get_fitness(),
	      dist2, edit_dist2,
	      ind_pheno->get_fitness());
    }
  }

  return successful;
}

int Deme :: mutate_ess(int p1, int c1)
{
  int j;

  for (j=0; j<5; j++) {
    num_mutate++;
    individ_new[0]->duplicate(individ[p1]);
    individ_new[0]->mutate();
    individ_new[0]->evaluate();

#ifndef OPTIMIZE2
    Op_used(0);
    Op_used(OP_MUTATE);

    if (individ_new[0]->better() == 1)
      Op_update(TRUE);
    else
      Op_update(FALSE);
#endif

#ifndef OPTIMIZE
    if (Print[8]) {
      individ_new[0]->print_history_full();
      cout << endl;
    }
#endif

    if (individ_new[0]->replace(individ[p1]))
      return TRUE;
    if (individ_new[0]->replace(individ[c1]))
      return TRUE;
  }

  return FALSE;
}


int Deme :: recombine(int p1, int p2, int c1)
{
  int j;
  int dist1, dist2, edit_dist1, edit_dist2;
  int successful = FALSE;
  int successful2 = FALSE;
  int num_prod;
  int max_bodies;
  int max_prod_len;
  int prod_len;


  for (j=0; j<MAX_TRIES; j++) {
    num_recomb++;
    individ_new[c1]->duplicate(individ[p1]);
    //    individ_new[c1]->write(stdout);
    individ_new[c1]->recombine(individ[p2]);
    individ_new[c1]->evaluate();

#ifndef OPTIMIZE2
    Op_used(0);
    Op_used(OP_RECOMB);

    if (individ_new[c1]->better() == 1)
      Op_update(TRUE);
    else
      Op_update(FALSE);
#endif

#ifndef OPTIMIZE
    if (Print[8]) {
      individ_new[c1]->print_history_full();
      cout << endl;
    }
#endif

    if (individ_new[c1]->get_fitness() >
	individ[p1]->get_fitness()/10.0) {
      successful = TRUE;
      break;
    }
  }

  if (Option[18] > 1) {
    if (ind_pheno == 0) {
      num_prod = Num_productions;
      max_bodies = Max_bodies;
      max_prod_len = Max_prod_length;
      prod_len = Prod_length;

      Num_productions = 1;
      Max_bodies = 1;
      Max_prod_length = MAX_STRING_LENGTH;
      Prod_length = MAX_STRING_LENGTH;

      ind_pheno = new Individual();
      ind_pheno->init_pheno(TRUE);

      ind_pheno2 = new Individual();
      ind_pheno2->init_pheno(TRUE);

      Num_productions = num_prod;
      Max_bodies = max_bodies;
      Max_prod_length = max_prod_len;
      Prod_length = prod_len;
    }

    for (j=0; j<MAX_TRIES; j++) {
      if (!ind_pheno->duplicate_pheno(individ[p1])) {
	FILE *file;
	file = fopen("invalid.ind", "w");
	individ[p1]->write(file);
	fclose(file);
	successful2 = FALSE;
	continue;
      }
      if (!ind_pheno2->duplicate_pheno(individ[p2])) {
	FILE *file;
	file = fopen("invalid.ind", "w");
	individ[p1]->write(file);
	fclose(file);
	successful2 = FALSE;
	continue;
      }

      successful2 = TRUE;
	
      // Compares mutation performance between generative and its direct equivalent.
      num_prod = Num_productions;
      max_bodies = Max_bodies;
      max_prod_len = Max_prod_length;
      prod_len = Prod_length;
	
      Num_productions = 1;
      Max_bodies = 1;
      Max_prod_length = MAX_STRING_LENGTH;
      Prod_length = MAX_STRING_LENGTH;

      ind_pheno->recombine_pheno(ind_pheno2);
      ind_pheno->evaluate();

      Num_productions = num_prod;
      Max_bodies = max_bodies;
      Max_prod_length = max_prod_len;
      Prod_length = prod_len;

      if (ind_pheno->get_fitness() >
	  individ[p1]->get_fitness()/10.0) {
	break;
      }
    }
	  
    //	individ_new[c1]->set_fitness(6, dist2); // distance from parent.
    //	individ_new[c1]->set_fitness(7, ind_pheno->get_fitness()); // parent's fitness
    
    if (successful2) {
      dist1 = individ_new[c1]->phenotype_dist(individ[p1]);
      dist2 = ind_pheno->phenotype_dist(individ[p1]);

      if (Option[18] >= 3) {
	edit_dist1 = individ_new[c1]->edit_distance(individ[p1]);
	edit_dist2 = ind_pheno->edit_distance(individ[p1]);
      } else {
	edit_dist1 = -1;
	edit_dist2 = -1;
      }
      
      fprintf(recomb_pheno_file, "%d %.2f %d %d %.2f %d %d %.2f\n",
	      pop->generation,
	      individ[p1]->get_fitness(),
	      dist1, edit_dist1,
	      individ_new[c1]->get_fitness(),
	      dist2, edit_dist2,
	      ind_pheno->get_fitness());
    }
  }

  return successful;
}


int Deme :: recombine_ess(int p1, int p2, int c1)
{
  int j;

  for (j=0; j<5; j++) {
    num_recomb++;
    individ_new[0]->duplicate(individ[p1]);
    individ_new[0]->recombine(individ[p2]);
    individ_new[0]->evaluate();

#ifndef OPTIMIZE2
    Op_used(0);
    Op_used(OP_RECOMB);

    if (individ_new[0]->better() == 1)
      Op_update(TRUE);
    else
      Op_update(FALSE);
#endif

#ifndef OPTIMIZE
    if (Print[8]) {
      individ_new[0]->print_history_full();
      cout << endl;
    }
#endif

    if (individ_new[0]->replace(individ[p1]) ||
	individ_new[0]->replace(individ[p2]) ||
	individ_new[0]->replace(individ[c1]))
      return TRUE;
  }

  return FALSE;
}


void Deme :: swap_populations(void)
{
  Individual **ind_tmp;

  ind_tmp = individ_new;
  individ_new = individ;
  individ = ind_tmp;
}


// Make a new generation for Blind Watchmaker.
// Parents are those individuals where selected[] == TRUE.
// All other individuals are 
void Deme :: make_new_generation(int evolve_type, int *selected)
{
  int p1, p2, i;
  int num_parents;
  int parent[num_individs];

  num_parents = 0;
  for (i=0; i<num_individs; i++) {
    if (selected[i])
      parent[num_parents++] = i;
  }

  if (num_parents == 0) {
    initialize();
    return;
  }

  for (i=0; i<num_individs; i++) {
    if (evolve_type == 2) {
      if (selected[i]) {
	individ_new[i]->duplicate(individ[i]);
	individ_new[i]->evaluate();
	continue;
      }
    }

    // Create a new child.
    while (TRUE) {
      if (recombination_test() && (num_parents > 1)) {
	// Recombination.
	p1 = random_int(num_parents);
	p2 = random_int(num_parents-1);
	if (p2 == p1)
	  p2 = num_parents-1;

	//	printf("r: %2d + %2d -> %2d\n", parent[p1], parent[p2], i);
	recombine(parent[p1], parent[p2], i);
	/*
	if (!recombine(parent[p1], parent[p2], i)) {
	  individ_new[i]->duplicate(individ[parent[p1]]);
	};
	*/

      } else {
	// Mutation.
	p1 = random_int(num_parents);
	//	printf("m:      %2d -> %2d  ", parent[p1], i);
	//	mutate(parent[p1], i);
	/*
	if (!mutate(parent[p1], i)) {
	  individ_new[i]->duplicate(individ[parent[p1]]);
	};
	*/
	//	printf(": %.1lf\n", individ_new[i]->get_fitness());
      }

      if (individ_new[i]->get_fitness() > 0.0)
	break;
    }
  }	

  swap_populations();
}



void Deme :: test_random_ind(void)
{
  int index, dist, dist1, dist2;

  index = random_int(num_individs);

  individ_new[index]->duplicate(individ[index]);
  individ_new[index]->mutate();
  individ_new[index]->evaluate();

  dist1 = individ[index]->phenotype_distance(individ_new[index]);
  dist2 = individ[index]->phenotype_distance2(individ_new[index]);
  if (dist1 > dist2)
    dist = dist1;
  else
    dist = dist2;

  printf("%d %d %d %.2f %.2f", dist, dist1, dist2,
	 individ[index]->get_fitness(),
	 individ_new[index]->get_fitness());
}



Real previous_best = 0.0;

void Deme :: make_generation_canonical(void)
{
  int i, j, ctr, tmp;
  int parent[num_individs];
  int order[num_individs];


  for (i=0; i<num_individs; i++)
    order[i] = i;

  //  qsort((void*)individ, num_individs, sizeof(Individual*),
  //	ind_compare);

  individs = individ;
  qsort((void*)order, num_individs, sizeof(int),
	ind_compare2);

  /*
  for (i=0; i<num_individs; i++)
    printf("%3d:%3d:%.3f  ", i, order[i], individ[order[i]]->get_fitness());
  printf("\n");
  */

  if (Option[16] == 1) {
    if (previous_best < individ[order[0]]->get_fitness()) {
      previous_best = individ[order[0]]->get_fitness();
      Config.do_graphics = TRUE;
      individ[order[0]]->evaluate(TRUE);
      individ[order[0]]->evaluate(TRUE);
      Config.do_graphics = FALSE;
    }
  }



  for (ctr=0; ctr<elitism; ctr++)
    parent[ctr] = order[ctr];

  for (i=0; i<num_individs; i++) {
    if (ctr == num_individs)
      break;
    for (j=0; j<min_children[i]; j++)
      parent[ctr++] = order[i];
  }

  for (i=ctr; i<num_individs; i++)
    parent[i] = order[get_index()];

  // Randomize order.
  for (i=elitism; i<num_individs; i++) {
    j = random_int(num_individs-elitism)+elitism;
    tmp = parent[i];
    parent[i] = parent[j];
    parent[j] = tmp;
  }    

  /*
  for (i=0; i<num_individs; i++)
    printf("%3d:%3d:%3d:%.3f\n", i, parent[i], order[i], individ[parent[i]]->get_fitness());
  printf("\n");
  */

  for (i=0; i<elitism; i++) {
    individ_new[i]->duplicate(individ[parent[i]]);
    individ_new[i]->evaluate(); // re-evaluate individuals.
  }

  for (i=elitism; i<num_individs-1; i+=2) {
    if (recombination_test()) {
      // Recombine.
      recombine(parent[i], parent[i+1], i);

#ifndef TREE
      recombine(parent[i+1], parent[i], i+1);
#endif

    } else {
      mutate(parent[i], i);
      mutate(parent[i+1], i+1);
    }
  }

  if (i == num_individs-1) {
    mutate(parent[i], i);
  }

  swap_populations();
}



int **fitnesses = 0;


void Deme :: make_generation_multi_objective(void)
{
  int i, ctr;
  int p1, p2;
  //  int fitnesses[num_individs][2];
  float exp_childs[num_individs];
  Real fit_vec[10];


  //******************************************************
  // Start of selection
  //******************************************************

  if (fitnesses == 0) {
    fitnesses = (int**)malloc(num_individs*sizeof(int*));
    for (i=0; i<num_individs; i++) {
      fitnesses[i] = (int*)malloc(5*sizeof(int));
    }
  }

  if (elitism <= 0) {
    elitism = num_individs / 2;
  }


  for (i=0; i<elitism; i++) {
    individ[i]->get_fitness(fit_vec);
    fitnesses[i][0] = int(fit_vec[1]);
    fitnesses[i][1] = int(fit_vec[2]);
    fitnesses[i][2] = int(fit_vec[3]);
    fitnesses[i][3] = -int(fit_vec[4]);
    fitnesses[i][4] = int(fit_vec[0]);
  }


  //  sort_multi_objective(elitism, 4, (int**)fitnesses, exp_childs, num_individs/2);
  sort_multi_objective(elitism, 5, (int**)fitnesses, exp_childs, num_individs/2);

  ctr = 0;
  for (i=0; i<elitism; i++) {
    if (exp_childs[i] > 0) {
      if (random_real() < exp_childs[i]) {
	individ[ctr]->duplicate(individ[i]);
	ctr++;
	//	printf("%d: exp
      }
    }
  }

  printf("%3d: ", ctr);

  elitism = ctr + num_individs/2;
  if (elitism >= num_individs) {
    elitism = num_individs;
  }

  for (i=ctr; i<elitism; i++) {
    if (recombination_test()) {
      // Recombine.
      p1 = random_int(ctr);
      p2 = random_int(ctr-1);
      if (p2 == p1) {
	p2 = ctr - 1;
      }
      recombine(p1, p2, i);

    } else {
      p1 = random_int(ctr);
      mutate(p1, i);
    }
    individ[i]->duplicate(individ_new[i]);
  }

  //  swap_populations();
}


void Deme :: make_generation_blind_watchmaker(void)
{
  int i;
  int p1, p2, c1;

  // assume all fitness are 0.

  for (i=0; i<num_individs; i++)
    individ[i]->set_fitness(0.0);

  c1 = 0;
  while (1) {
    if (recombination_test() && (num_individs > 1)) {
      p1 = random_int(num_individs);
      p2 = random_int(num_individs-1);
      if (p2 == p1)
	p2 = num_individs-1;
      recombine(p1, p2, c1);

    } else {
      p1 = random_int(num_individs);
      mutate(p1, c1);
    }

    if (individ_new[c1]->get_fitness() == -2.0)
      break;
    
    if (individ_new[c1]->better())
      c1++;
    if (c1 >= pop->num_individs)
      break;
  }

  if (c1 == 0)
    make_generation_canonical();
  else
    num_individs = c1+1;
  printf("Next generation.  %d individuals.\n", num_individs);
}


void Deme :: make_generation_dc(void)
{
  int i;
  Individual *tmp;

  for (i=0; i<num_individs; i++) {
    int p1, p2;

    p1 = cycle_index;
    if (++cycle_index >= num_individs)
      cycle_index = 0;
    individ_new[0]->duplicate(individ[p1]);
    age[p1]++;

    if (recombination_test()) {
      p2 = random_index(p1);
      age[p2]++;
	  
      num_recomb++;
      individ_new[0]->recombine(individ[p2]);
      individ_new[0]->evaluate();

    } else {
      num_mutate++;
      individ_new[0]->mutate(mutate_scale);
      individ_new[0]->evaluate();
    }

    if (Print[8]) {
      printf("%2d: ", p1);
      individ_new[0]->print_history_full();
      cout << "\n";
    }
      
    //    if (individ_new[0]->better() > 0) {
    if (individ_new[0]->get_fitness() >=
	individ[p1]->get_fitness()) {
      if (individ_new[0]->get_creation() == CREATE_MUTATE)
	num_mut_better++;
      else
	num_rec_better++;

      age[p1] = 0;
      tmp = individ_new[0];
      individ_new[0] = individ[p1];
      individ[p1] = tmp;
    }
  }
}


int Deme :: random_index(int index)
{
  int base, index2;

  switch (nghbr_type) {
  case NGHBR_NONE:
    index2 = random_int(num_individs);
    if (index2 == index)
      index2 = num_individs - 1;
    break;

  case NGHBR_DEME:
    base = index - (index % nghbr_size);
    index2 = base + random_int(nghbr_size);
    if (index2 == index) {
      /*
      // Select outside of deme.
      base = random_int(num_individs / nghbr_size - 1);
      if (base == index / nghbr_size) 
	base = num_individs / nghbr_size - 1;
      index2 = base + random_int(nghbr_size);
      */
      index2 = random_int(num_individs - 1);
      if (index2 == index)
	index2 = num_individs - 1;
    }
    break;

  case NGHBR_NEAR:
    index2 = index + random_int(nghbr_size - 1) + 1;
    index2 = index2 % num_individs;
    break;

  default:
    index2 = random_int(num_individs);
    if (index2 == index)
      index2 = num_individs - 1;
    break;
  }

  return index2;
}

/*
  if (cycle_index == num_individs)
    cycle_index = 0;
  ind[1] = random_int(num_individs - 1);
  ind[2] = random_int(num_individs - 2);
  
  if (ind[2] == ind[1])
    ind[2] = num_individs - 2;
  if (ind[1] == ind[0])
    ind[1] = num_individs - 1;
  else if (ind[2] == ind[0])
    ind[2] = num_individs - 1;
    */

int Deme :: tournament(void)
{
  int i, j, tmp1, tmp2;
  int ind[tournament_size];
  int p1, p2;
  int ind1, ind2, ind3;
  Real min1, min2, min3;

  ind[0] = cycle_index++;
  if (cycle_index == num_individs)
    cycle_index = 0;
  for (i=1; i<tournament_size; ) {
    ind[i] = random_index(ind[0]);
    for (j=1; j<i; j++) {
      if (ind[i] == ind[j])
	break;
    }
    if (j == i)
      i++;
  }

  /*
  for (i=0; i<tournament_size; i++)
    cout << ind[i] << "  ";
  cout << endl;
  */

  // Mix.
  for (i=0; i<tournament_size; i++) {
    tmp1 = random_int(tournament_size);
    tmp2 = ind[i];
    ind[i] = ind[tmp1];
    ind[tmp1] = tmp2;
  }


  if (replace_size == 1) {
    ind1 = ind[0];
    min1 = individ[ind1]->get_fitness();
    for (i=1; i<tournament_size; i++) {
      if (individ[ind[i]]->get_fitness() < min1) {
	ind1 = ind[i];
	min1 = individ[ind1]->get_fitness();
      }
    }
  } else {
    if (individ[ind[0]]->get_fitness() < individ[ind[1]]->get_fitness()) {
      ind1 = ind[0];
      ind2 = ind[1];
    } else {
      ind1 = ind[1];
      ind2 = ind[0];
    }
    min1 = individ[ind1]->get_fitness();
    min2 = individ[ind2]->get_fitness();
    
    for (i=2; i<tournament_size; i++) {
      ind3 = ind[i];
      min3 = individ[ind3]->get_fitness();

      if (min3 < min2) {
	min2 = min3;
	ind2 = ind3;

	if (min2 < min1) {
	  min2 = min1;
	  ind2 = ind1;
	  min1 = min3;
	  ind1 = ind3;
	}
      }
    }

    /*
    printf("1: %d %4.1f  2: %d %4.1f\n", ind1, min1,
	   ind2, min2);
	   */

    if (random_int(2) && (min1 > 0.0))
      ind1 = ind2;
  }

  if (ind1 == ind[0]) {
    p1 = ind[1];
    p2 = ind[2];
  } else if (ind1 == ind[1]) {
    p1 = ind[0];
    p2 = ind[2];
  } else {
    p1 = ind[0];
    p2 = ind[1];
  }

  individ[ind1]->duplicate(individ[p1]);
  age[ind1] = 0;
  age[p1]++;

  if (recombination_test()) {
    // Recombination
    age[p2]++;
    individ[ind1]->recombine(individ[p2]);
  } else {
    // Mutation
    individ[ind1]->mutate(mutate_scale);
  }

  if (Print[8]) {
    individ[p1]->print_history_full();
    cout << "\n";
  }
  return ind1;
}


int Deme :: next_individual_init(void)
{
  if (cycle_index >= num_individs)
    return -1;

  if (individ[cycle_index]->get_fitness() <= 0.0) {
    individ[cycle_index]->make_random();
    return cycle_index;
  }

  if (++cycle_index >= num_individs)
    return -1;

  return cycle_index;
}

int Deme :: next_individual(void)
{
  int next_ind = -1;

  switch (structure)
    {
    case STRUCT_CANONICAL:
      if (cycle_index == num_individs) {
	// Create next generation.
	make_next_generation();
	cycle_index = 0;
      }
      next_ind = cycle_index++;
      break;

    case STRUCT_NEIGHBOR:


      break;

    case STRUCT_STEADY_STATE:
      next_ind = reevaluate_test();
      if (next_ind < 0)
	next_ind = tournament();
      break;

    case STRUCT_HILL_CLIMB: 
      if (individ[1]->get_fitness() > individ[0]->get_fitness()) {
	max_fitness = individ[1]->get_fitness();
	individ[0]->duplicate(individ[1]);
	next_ind = 0;
	age[0] = 0;
      } else {
	if ((++age[0] >= max_age) && (max_age != 0)) {
	  next_ind = 0;
	  age[0] = 0;
	} else {
	  individ[1]->duplicate(individ[0]);
	  next_ind = 1;
	}
      }
      break;
    }

  return next_ind;
}


void Deme :: make_generation_elitist_ss(void)
{
  int i;
  int p1, p2, c1;

  for (i=0; i<num_individs; i++) {
    p1 = cycle_index;
    if (++cycle_index >= num_individs)
      cycle_index = 0;

    if (recombination_test()) {
      p2 = random_index(p1);
      do {
	c1 = random_index(p1);
      } while (c1 == p2);
	  
      recombine_ess(p1, p2, c1);
      if (individ_new[0]->better(individ[c1]))
	individ[c1]->duplicate(individ_new[0]);
      else if (individ_new[0]->better(individ[p2]))
	individ[p2]->duplicate(individ_new[0]);
      else if (individ_new[0]->better(individ[p1]))
	individ[p1]->duplicate(individ_new[0]);

    } else {
      c1 = random_index(p1);
      mutate_ess(p1, c1);
      if (individ_new[0]->better(individ[c1]))
	individ[c1]->duplicate(individ_new[0]);
      else if (individ_new[0]->better(individ[p1]))
	individ[p1]->duplicate(individ_new[0]);
    }
  }
}


void Deme :: make_next_generation(void)
{
  int i, ind;
  Individual *tmp, *a;

  num_recomb = 0;
  num_mutate = 0;

  switch (structure)
    {
    case STRUCT_TEST_ONE:
      individ[0]->evaluate();
      individ[1]->duplicate(individ[0]);
      individ[1]->evaluate();

      break;

    case STRUCT_CANONICAL:
      make_generation_canonical();
      break;

    case STRUCT_MULTI_OBJECTIVE:
      make_generation_multi_objective();
      break;

    case STRUCT_BLIND_WATCHMAKER:
      make_generation_blind_watchmaker();
      break;

    case STRUCT_DETERMINISTIC_CROWDING:
      make_generation_dc();
      break;

    case STRUCT_NEIGHBOR:


      break;

    case STRUCT_ELITIST_STEADY_STATE:
      make_generation_elitist_ss();
      break;


    case STRUCT_STEADY_STATE2:
      for (i=0; i<num_individs; i++) {
	if (!reevaluate()) {
	  int p1, p2;

	  p1 = cycle_index;
	  if (++cycle_index >= num_individs)
	    cycle_index = 0;
	  individ_new[0]->duplicate(individ[p1]);
	  age[p1]++;

	  if (recombination_test()) {
	    p2 = random_index(p1);
	    age[p2]++;
	  
	    num_recomb++;
	    individ_new[0]->recombine(individ[p2]);
	    individ_new[0]->evaluate();
	  } else {
	    num_mutate++;
	    individ_new[0]->mutate(mutate_scale);
	    individ_new[0]->evaluate();
	  }

	  if (Print[8]) {
	    individ_new[0]->print_history_full();
	    cout << "\n";
	  }
	  
	  if (individ_new[0]->better()) {
	    if (individ_new[0]->get_creation() == CREATE_MUTATE)
	      num_mut_better++;
	    else
	      num_rec_better++;

	    age[p1] = 0;
	    tmp = individ_new[0];
	    individ_new[0] = individ[p1];
	    individ[p1] = tmp;
	  }
	}
      }

      break;

    case STRUCT_STEADY_STATE:
      for (i=0; i<num_individs; i++) {
	if (!reevaluate()) {
	  ind = tournament();
	  individ[ind]->evaluate();
	  if (Print[8]) {
	    individ[ind]->print_history_full();
	    cout << endl;
	  }

	  if (individ[ind]->get_creation() == CREATE_MUTATE) {
	    num_mutate++;
	    if (individ[ind]->better())
	      num_mut_better++;
	  } else {
	    num_recomb++;
	    if (individ[ind]->better())
	      num_rec_better++;
	  }
	}
      }

      break;

    case STRUCT_HILL_CLIMB: 
      individ[1]->duplicate(individ[0]);
      individ[1]->mutate();
      individ[1]->evaluate();

      if (individ[1]->get_fitness() > individ[0]->get_fitness())
	{
	  max_fitness = individ[1]->get_fitness();
	  a = individ[0];
	  individ[0] = individ[1];
	  individ[1] = a;
	  age[0] = 0;
	}
      else
	{
	  if ((++age[0] >= max_age) && (max_age != 0))
	    {
	      cout << "\n" << individ[0]->get_fitness()<<" Re-evaluate\n" << endl;
	      individ[0]->evaluate();
	      age[0] = 0;
	    }
	}
      break;
    }
}


void Deme :: mutate_all(void)
{
  int i;

  for (i=0; i<num_individs; i++)
    individ[i]->mutate();
}

Individual *Deme :: get_individ(int index)
{
  if ((index < 0) || (index >= num_individs))
    return 0;

  return individ[index];
}


/***** Print functions *****/

void Deme :: info(void)
{
  int i;

  for (i=0; i<num_individs; i++)
    individ[i]->info();
}

void Deme :: print_all(void)
{
  int i;

  for (i=0; i<num_individs; i++) {
    individ[i]->print_history_self();
    cout << " ";
  }
}

void Deme :: print_gen_results(void)
{
  int i;
  Real rec_rate, mut_rate;

  for (i=0; i<num_individs; i++) {
    individ[i]->print_history_full();
    cout << endl;
  }


  if (num_mutate != 0)
    mut_rate = Real(num_mut_better) / Real(num_mutate);
  else
    mut_rate = -1.0;

  if (num_recomb != 0)
    rec_rate = Real(num_rec_better) / Real(num_recomb);
  else
    rec_rate = -1.0;

  printf("#mut:%3d:%3d, #rec:%3d:%3d\n", num_mut_better,
	 num_mutate, num_rec_better, num_recomb);
  printf("Mutate %%: %5.3f,  Recomb %%: %5.3f,  >%c<\n",
	 mut_rate, rec_rate, ((mut_rate > rec_rate) ? 'm' : 'r'));
}



/***** File I/O *****/

void Deme :: write_best(FILE *file)
{
  //  cout << "write best: " << max_index << ", ";
  //  individ[max_index]->print_history_self();
  //  individ[max_index]->print_history();
  //  cout << endl;
  individ[max_index]->write(file);
}

void Deme :: write_best(char *fname)
{
  char filename[100];
  FILE *file;

  if (deme_id < 0)
#ifdef NETWORK
    sprintf(filename, "%s.net", fname);
  else
    sprintf(filename, "%s_%d.net", fname, deme_id);
#elif TREE
    sprintf(filename, "%s.prg", fname);
  else
    sprintf(filename, "%s_%d.prg", fname, deme_id);
#else
    sprintf(filename, "%s.ind", fname);
  else
    sprintf(filename, "%s_%d.ind", fname, deme_id);
#endif

  file = fopen(filename, "w");
  write_best(file);
  fclose(file);
}


void Deme :: write_best_new(FILE *file)
{
  int i, index = 0;
  double fitness = -DBL_MAX;

  for (i=0; i<num_individs; i++) {
    if ((age[i] == 0) && (individ[i]->get_fitness() > fitness)) {
      fitness = individ[i]->get_fitness();
      index = i;
    }
  }

  cout << "Write best new:: index: " << index;
  cout << "  fitness: " << fitness << endl;
  individ[index]->write(file);
}

void Deme :: write_best_new(char *fname)
{
  char filename[100];
  FILE *file;

  if (deme_id < 0)
#ifdef NETWORK
    sprintf(filename, "%s.net", fname);
#elif TREE
    sprintf(filename, "%s.prg", fname);
#else
    sprintf(filename, "%s.ind", fname);
#endif
  else
#ifdef NETWORK
    sprintf(filename, "%s_%d.net", fname, deme_id);
#elif TREE
    sprintf(filename, "%s_%d.prg", fname, deme_id);
#else
    sprintf(filename, "%s_%d.ind", fname, deme_id);
#endif

  file = fopen(filename, "w");
  write_best_new(file);
  fclose(file);  
}



int Deme :: read_individ(FILE *file)
{
  Individual *ind = new Individual();

  if (ind->read(file) == FALSE)
    return FALSE;
  individ[num_individs++] = ind;

  return TRUE;
}


int Deme :: write_individ(int index, char *fname)
{
  FILE *file;

  file = fopen(fname, "w");
  if (file == NULL) {
    printf("Error creating file '%s'.\n", fname);
    return FALSE;
  }

  individ[index]->write(file);
  fclose(file);
  return TRUE;
}


void Deme :: write(FILE *file)
{
  int i;

  fprintf(file, "%d %d %d %d %d %d %d %d %d %.4f\n",
	       num_individs, cycle_index, structure,
	       nghbr_size, nghbr_type, max_age,
	       tournament_size, replace_size, elitism,
	       pressure);

  for (i=0; i<num_individs; i++) {
    fprintf(file, "%d\n", age[i]);
    individ[i]->write(file);
  }
}


void Deme :: write(char *fname)
{
  char filename[100];
  FILE *file;

  if (deme_id < 0)
    sprintf(filename, "%s.deme", fname);
  else
    sprintf(filename, "%s_%d.deme", fname, deme_id);

  file = fopen(filename, "w");
  write(file);
  fclose(file);  
}


int Deme :: read(FILE *file)
{
  int i, num;

  num = fscanf(file, "%d %d %d %d %d %d %d %d %d %lf\n",
	       &num_individs, &cycle_index, &structure,
	       &nghbr_size, &nghbr_type, &max_age,
	       &tournament_size, &replace_size, &elitism,
	       &pressure);
  if (num != 10) {
      cout << "Deme: error in reading structure\n" << endl;
      return FALSE;
  }


  for (i=0; i<num_individs; i++) {
    num = fscanf(file, "%d\n", &(age[i]));
    if (num != 1) { 
      cout << "Deme: error in reading age for individual " << i << endl;
      return FALSE;
    }      
    if (individ[i]->read(file) == FALSE)
      return FALSE;
  }

  for (i=0; i<num_individs; i++)
    individ_new[i]->init();

  return TRUE;
}

int Deme :: read(char *fname)
{
  int result;
  char filename[100];
  FILE *file;

  if (deme_id < 0)
#ifdef NETWORK
    sprintf(filename, "%s.net", fname);
#elif TREE
    sprintf(filename, "%s.prg", fname);
#else
    sprintf(filename, "%s.ind", fname);
#endif
  else
#ifdef NETWORK
    sprintf(filename, "%s_%d.net", fname, deme_id);
#elif TREE
    sprintf(filename, "%s_%d.prg", fname, deme_id);
#else
    sprintf(filename, "%s_%d.ind", fname, deme_id);
#endif

  file = fopen(filename, "r");
  result = read(file);
  fclose(file);  

  return result;
}

