/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/


#ifndef DEME_HEADER_FILE
#define DEME_HEADER_FILE

#include "general.h"

class Individual;
class Pop;
class Stats;

#define HOF_SIZE 0


class Deme
{
public:
  int deme_id;
  int structure;
  int num_individs;
  int max_age;
  int *age;
  int cycle_index;
  Individual **individ;
  Individual **individ_new;
  Individual **hof; // Hall of Fame (saves best individual from each generation).
  int max_index;
  int elitism;
  int tournament_size;
  int replace_size;
  int nghbr_size;
  int nghbr_type;

  int adaptive_mut;
  int num_recomb, num_mutate;
  int num_rec_better, num_mut_better;
  double mutate_scale;
  double prev_scale;
  double recomb_rate;

  double max_fitness;
  double avg_fitness;
  double best_vec[10];
  double pressure;
  double *select_prob;
  int *min_children;

  int num_inputs, num_hidden, num_outputs;

  Individual *sample_individ;
  Pop *pop;
  
  /******************************/


  Deme(int nindivids);
  void create(void);
  void initialize(void);
  void initialize(Individual *ind_seed);
  void initialize_simple(void);
  void initialize_simple(Individual *ind_seed);
  int get_index(void);
  void display(int index);
  Real evaluate(int index);
  int reevaluate(void);
  int reevaluate_test(void);
  void post_evaluate(void);
  int recombination_test(void);
  int mutate_bw(int p1, int c1);
  int mutate(int p1, int c1);
  int mutate_ess(int p1, int c1);
  int recombine(int p1, int p2, int c1);
  int recombine_ess(int p1, int p2, int c1);
  void swap_populations(void);
  void make_new_generation(int evolve_type, int *selected);
  void test_random_ind(void);
  void make_generation_canonical(void);
  void make_generation_multi_objective(void);
  void make_generation_blind_watchmaker(void);
  void make_generation_dc(void);
  void make_generation_elitist_ss(void);
  void make_next_generation(void);
  int random_index(int index);
  int tournament(void);
  int next_individual_init(void);
  int next_individual(void);
  void update_best_vec(Real *pop_best_vec);
  void update_stats(int print_stats);
  void update_stats_vector(Stats *stats_vector);
  void mutate_all(void);

  Individual *get_individ(int index);

  void info(void);
  void print_all(void);
  void print_gen_results(void);

  void write_best(FILE *file);
  void write_best(char *fname);
  void write_best_new(FILE *file);
  void write_best_new(char *fname);
  int write_individ(int index, char *fname);

  void write(FILE *file);
  void write(char *fname);
  int read(FILE *file);
  int read_individ(FILE *file);
  int read(char *fname);
};

#endif
