/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/


#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <math.h>
#include <stream.h>
#include "general.h"


/**********************************/
// Mersenne Twister defs.

typedef unsigned long uint32;
void seedMT(uint32 seed);
uint32 reloadMT(void);
uint32 randomMT(void);

/**********************************/




int Print[50];
int Option[50];

int Max_operator;
int Last_op;
int Last_op2;
int Operation[100];
int Op_better[100];
int Operation_hist[100];
int Op_better_hist[100];
int Draw_arg[10] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

#define OP_STACK_SIZE 20
int Op_stack[OP_STACK_SIZE];
int Op_stack_ptr = 0;
int Num_fit_dims = 1;

Real Rec_delta = 1.0;

Configuration Config;

void print_vec2(char *string, Real vec[2])
{
  cout << string << vec[0] << " " << vec[1] << endl;
}


unsigned int random_int(int max)
{
  /*
  unsigned int val;
  val = (randomMT() % max);
  printf("random_int: %d\n", val);
  return val;
  */
  return (randomMT() % max);
}

unsigned int random_integer(void)
{
  return randomMT();
}


Real random_real(void)
{
  /*
  unsigned int i_val;
  float f_val;

  i_val = random_int(RAND_MAX);
  f_val = (float(i_val) / float(RAND_MAX));
  printf("radnom_real: %f  (%d / %d)\n", f_val, i_val, RAND_MAX);
  return f_val;
  */
  return (Real(random_int(RAND_MAX)) / Real(RAND_MAX));
  //  return (float(randomMT()) / float(RAND_MAX));
  //  return (((float) rand ()) / ((float) RAND_MAX));
}


double random_dbl(void)
{
  return random_real();
}


/* Generate a random variable with a normal distribution */
/* N(mu, sigma^2) */
double random_norm(double mu, double sigma)
{
   double u, v, x, y;

   u = random_dbl();
   v = random_dbl();

   if (u == 0.0) {
     u = 1.0;
     printf("Error: random_norm: u = 0.0!\n");
   }

   x = sqrt(-2.0 * log(u)) * cos(2.0 * PI * v);
   y = sigma * x + mu;

   return y;
}


static long state1[32] = {
  3,
  0x9a319039, 0x32d9c024, 0x9b663182, 0x5da1f342,
  0x7449e56b, 0xbeb1dbb0, 0xab5c5918, 0x946554fd,
  0x8c2e680f, 0xeb3d799f, 0xb11ee0b7, 0x2d436b86,
  0xda672e2a, 0x1588ca88, 0xe369735d, 0x904f35f7,
  0xd7158fd6, 0x6fa6f051, 0x616e6b96, 0xac94efdc,
  0xde3b81e0, 0xdf0a6fb5, 0xf103bc02, 0x48f340fb,
  0x36413f93, 0xc622c298, 0xf5a42ab8, 0x8a88d77b,
  0xf5ad9d0e, 0x8999220b, 0x27fb47b9
};



void init_random(int val)
{
   unsigned seed = 1;
   int n = 128;
   struct timeval tp;
   struct timezone tzp;

   initstate(seed, (char*)state1, n);
   setstate((char*)state1);
 
   /* Initialize random number generator */
   if (val < 0) {
     gettimeofday(&tp, &tzp);
     srandom(tp.tv_usec);
     srand48(tp.tv_usec);
     seedMT(tp.tv_usec);

   } else {
     srandom(val);
     srand48(val);
     seedMT(val);
   }
}


void print_time(void)
{
  int sec, min, hour;
  struct timeval tp;
  struct timezone tzp;
  gettimeofday(&tp, &tzp);

  tp.tv_sec -= 4*3600; // change to EST.
  hour = (tp.tv_sec/3600) % 24;
  min = (tp.tv_sec/60) % 60;
  sec = tp.tv_sec % 60;

  printf("(%d:%2d:%2d)", hour, min, sec);
}


Real recombine_real(Real val1, Real val2)
{
  Real diff;

  diff = Rec_delta * (val1 - val2);
  return reduce(val1 + diff * (random_dbl() * 2.0 - 1.0));
}


#define INTEGER_RANGE 100.0

Real recombine_int(Real val1, Real val2)
{
  
  if (val1 == val2)
    return val1;

  int diff = int(val1 - val2);
  diff = random_int(2 * diff + 1) - diff;

  val1 += Real(diff);
  if (val1 > INTEGER_RANGE)
    val1 = INTEGER_RANGE;
  else if (val1 < 0.0)
    val1 = 0.0;
  return val1;
}

int Op_used(int op_num)
{
  if (Op_stack_ptr >= OP_STACK_SIZE)
    return FALSE;

  Op_stack[Op_stack_ptr++] = op_num;
  if (Print[31])
    printf("operation: %d (%d)\n", op_num, Op_stack_ptr);
  return TRUE;
}

void Op_update(int better)
{
  int i;

  for (i=0; i<Op_stack_ptr; i++)
    Operation[Op_stack[i]]++;
  if (better) {
    for (i=0; i<Op_stack_ptr; i++)
      Op_better[Op_stack[i]]++;
  }
  Op_stack_ptr = 0;
  if (Print[31])
    printf("variation done.\n\n");
}
