/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#ifndef GENERAL_HEADER_FILE
#define GENERAL_HEADER_FILE

#include "values.h"

#ifdef __cplusplus
extern "C" {
#endif

#define FALSE 0
#define TRUE 1

typedef double Real;
  /* typedef float Real; */


typedef struct tConfiguration
{
  int start_frame;
  int frame_rate;
  int do_graphics;
  int draw_history;
  int save_graphics;
  int num_trials;
  char filename_stem[255];
  char filename_error[255];
} Configuration;



#define PI 3.1415927
#define RAD 57.29578
#define DEG2RAD (PI / 180.0)
#define RAD2DEG (180.0 / PI)


// long int random(void);
double drand48(void);
// int srandom(unsigned int seed);
void srand48(long seedval);



#define absv(val) (((val) >= 0.0) ? (val) : -(val))
#define sqr(val) (val)*(val)
#define maxv(val1, val2) (((val1) > (val2)) ? (val1) : (val2))
#define minv(val1, val2) (((val1) < (val2)) ? (val1) : (val2))
  //#define random_int(max) (((int)random() >> 4) % (max))
  //#define random_dbl() ((Real)drand48())
#define reduce(value) (floor((value) * 1.0E+8 + 0.5) / 1.0E+8)
#define reduce2(value) (floor((value) * 1.0E+5 + 0.5) / 1.0E+5)

unsigned int random_int(int max);
Real random_real(void);
double random_dbl(void);

extern int Print[50];
extern int Option[50];
extern int Draw_arg[10];

extern int Max_operator;
extern int Operation[100];
extern int Op_better[100];
extern int Operation_hist[100];
extern int Op_better_hist[100];
extern Real Rec_delta;
extern Configuration Config;
extern int Num_fit_dims;

void print_vec2(char *string, Real vec[2]);
double random_norm(double mu, double sigma);
void init_random(int val);

void print_time(void);
Real recombine_real(Real val1, Real val2);
Real recombine_int(Real val1, Real val2);

int Op_used(int op_num);
void Op_update(int better);

#ifdef __cplusplus
}
#endif

#endif
