/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/


#include <GL/gl.h>
#include <GL/glx.h>             /* this includes the necessary X headers */
#include <GL/glu.h>
#include <X11/keysym.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "general.h"
#include "graphic.h"
#include "deme.h"



#define VIEW_X 300
#define VIEW_Y 300


int face[6][4] = {
    {0, 1, 2, 3},
    {3, 2, 6, 7},
    {7, 6, 5, 4},
    {4, 5, 1, 0},
    {1, 2, 6, 5},
    {0, 4, 7, 3},
};


// Dark: 44, 38, 32
// Middle: 94, 79, 64
// Light: 145, 120, 96
// Surface: 186, 163, 141

// LMiddle: 104, 89, 77
// LLight: 165, 141, 121


double col_black[3] = {0.0, 0.0, 0.0}, col_white[3] = {1.0, 1.0, 1.0};
double col_red[3], col_blue2[3], col_green[3];
double col_grey1[3] = {0.2, 0.2, 0.2}, col_grey2[3];
//double col_grey3[3] = {0.5, 0.41, 0.35};
double col_grey4[3] = {140.0/255.0, 116.0/255.0, 98.0/255.0};
//double col_grey3[3] = {166.0/255.0, 142.0/255.0, 124.0/255.0};
double col_grey3[3] = {161.0/255.0, 138.0/255.0, 120.0/255.0};
double col_grey5[3] = {150.0/255.0, 131.0/255.0, 117.0/255.0};
//double col_grey6[3] = {142.0/255.0, 118.0/255.0, 103.0/255.0};
double col_grey6[3] = {142.0/255.0, 118.0/255.0, 103.0/255.0};
double col_grey7[3] = {194.0/255.0, 152.0/255.0, 118.0/255.0};
double col_grey10[3] = {150.0/255.0, 150.0/255.0, 150.0/255.0};
double col_grey11[3] = {0.9, 0.9, 0.9};
// 156, 115, 90
// 168, 130, 106

double col_blue[3] = {0.0, 0.0/255.0, 255.0/255.0}; // pure blue
double col_blue3[3] = {0.0, 100.0/255.0, 240.0/255.0}; // nice blue.
double col_stripe1[3] = {44.0/255.0, 38.0/255.0, 32.0/255.0}; //dark brown-grey
double col_stripe2[3] = {126.0/255.0, 111.0/255.0, 87.0/255.0}; // light-medium brown-grey

//double col_side1[3] = {146.0/255.0, 121.0/255.0, 97.0/255.0}; // medium brown-grey
//double col_side2[3] = {199.0/255.0, 176.0/255.0, 157.0/255.0}; // light brown-grey
////double col_side2[3] = {189.0/255.0, 166.0/255.0, 147.0/255.0}; // light brown-grey
//double col_side3[3] = {44.0/255.0, 38.0/255.0, 32.0/255.0}; //dark brown-grey

double col_side1[3] = {165.0/255.0, 141.0/255.0, 121.0/255.0}; // medium brown-grey
double col_side2[3] = {104.0/255.0, 89.0/255.0, 77.0/255.0}; // light brown-grey
double col_side3[3] = {44.0/255.0, 38.0/255.0, 32.0/255.0}; //dark brown-grey


double col_back_grnd[3] = {228.0/255.0, 230.0/255.0, 240.0/255.0}; // light blue
//double col_back_grnd[3] = {188.0/255.0, 190.0/255.0, 230.0/255.0}; // light blue

double col_surf[3] = {100.0/255.0, 111.0/255.0, 136.0/255.0}; // dark blue
//double col_surf[3] = {100.0/255.0, 101.0/255.0, 196.0/255.0}; // dark blue
//double col_surf[3] = {84.0/255.0, 94.0/255.0, 178.0/255.0}; // dark blue
//double col_surf2[3] = {156.0/255.0, 133.0/255.0, 111.0/255.0}; // medium brown-grey
double col_surf2[3] = {166.0/255.0, 143.0/255.0, 121.0/255.0}; // medium brown-grey
//double col_surf3[3] = {190.0/255.0, 157.0/255.0, 130.0/255.0}; // medium brown-grey
double col_surf3[3] = {200.0/255.0, 162.0/255.0, 130.0/255.0}; // medium brown-grey
double col_surf4[3] = {216.0/255.0, 193.0/255.0, 171.0/255.0}; // medium brown-grey
//double col_surf5[3] = {199.0/255.0, 173.0/255.0, 155.0/255.0}; // medium brown-grey
double col_surf5[3] = {199.0/255.0, 177.0/255.0, 149.0/255.0}; // medium brown-grey
double col_surf_print[3] = {200.0/255.0, 162.0/255.0, 130.0/255.0}; // medium brown-grey

// Real surface[3] = {0.8, 0.7, 0.6}; // nice beige


double col_side1_ref[3] = {106.0/255.0, 91.0/255.0, 67.0/255.0}; // medium brown-grey
double col_side2_ref[3] = {156.0/255.0, 139.0/255.0, 123.0/255.0}; // light brown-grey
double col_side3_ref[3] = {52.0/255.0, 40.0/255.0, 29.0/255.0}; //dark brown-grey

double col_blue2_ref[3] = {95.0/255.0, 96.0/255.0, 178.0/255.0}; // light blue
double col_blue3_ref[3] = {22.0/255.0, 23.0/255.0, 140.0/255.0}; // dark blue
//double col_green2[3] = {22.0/255.0, 140.0/255.0, 23.0/255.0};
//double col_green2[3] = {12.0/255.0, 210.0/255.0, 43.0/255.0};
double col_green2[3] = {2.0/255.0, 203.0/255.0, 50.0/255.0};
//double col_green2[3] = {72.0/255.0, 190.0/255.0, 73.0/255.0};
//double col_green2[3] = {52.0/255.0, 190.0/255.0, 53.0/255.0};
//double col_green2[3] = {32.0/255.0, 190.0/255.0, 33.0/255.0};

double col_green3[3] = {10.0/255.0, 160.0/255.0, 53.0/255.0};
//double col_green3[3] = {20.0/255.0, 190.0/255.0, 63.0/255.0};

//double col_green4[3] = {5.0/255.0, 197.0/255.0, 65.0/255.0};
//double col_green4[3] = {80.0/255.0, 210.0/255.0, 80.0/255.0};
//double col_green4[3] = {6.0/255.0, 195.0/255.0, 0.0/255.0};
double col_green4[3] = {6.0/255.0, 190.0/255.0, 0.0/255.0};
double col_green5[3] = {0.0/255.0, 102.0/255.0, 29.0/255.0};
double col_green6[3] = {0.0/255.0, 150.0/255.0, 16.0/255.0}; // 0, 142, 16

// Light green: 25,200,25;  0,200,48
// 28,186,78, 12,199,71
// 12,196,135; 10,195,120
// 87,207,41
// 0,166,50
// 15,200,70

void fatalError(char *message)
{
    fprintf(stderr, "glxsimple: %s\n", message);
    exit(1);
}


int hsv2rgb(double h, double s, double v, double *r,
	       double *g, double *b)
{
    double f, p, q, t;
    int   i, iret;

    iret = 0;                                   /* preset iret as 0 */

    /* check hsv values to see if valid */
    if (s < 0.0 || s > 1.0 || v < 0.0 || v > 1.0) return(-1);
    if (h < 0.0 || h > 360.0) return(-1);

    /* the colour is on the black-and-white center line */
    if (s == 0.0)                               
            *r = *g = *b = v;
    else {                                      /* Chromatic colour, hue */ 
        if (h == 360.0)                         /* 360 is equivalent to 0 */
            h = 0;
        h = h / 60;                             /* h is now in [0, 6) */
        i = ((int)h > h) ? (int)h - 1 : (int)h; /* largest integer <= h */
        f = h - i;                              /* f is the fraction part of h */
        p = v * (1 - s);
        q = v * (1 - s * f);
        t = v * (1 - s * (1 - f));
        switch (i) {
        case 0: 
             *r = v;
             *g = t;
             *b = p;
             break;
        case 1:
             *r = q;
             *g = v;
             *b = p;
             break;
        case 2:
             *r = p;
             *g = v;
             *b = t;
             break;
        case 3:
             *r = p;
             *g = q;
             *b = v;
             break;
        case 4:
             *r = t;
             *g = p;
             *b = v;
             break;
        case 5:
             *r = v;
             *g = p;
             *b = q;
             break;
        }
    }
    return (iret);
}       


#define TITLE "Evolving Creatures"


Graphic :: Graphic(int x, int y, char *title)
{
  int             dummy;
  XVisualInfo    *vi;
  Colormap        cmap;
  XSetWindowAttributes swa;

  // Initialize some variables.
  xsize = x;
  ysize = y;

  needRedraw = GL_FALSE;
  recalcModel = GL_TRUE;

  frame_num = 20;
  save_frame = FALSE;
  save_movie = FALSE;
  save_movie_frame = FALSE;

  init_sphere = FALSE;
  pixels = 0;
  doubleBuffer = GL_TRUE;

  x_translate = -15.0;
  y_translate = 0.0;
  z_translate = 0.0;
  Zoom = 0.60;

  xAngle = 0.0;
  yAngle = 0.0;
  zAngle = 0.0;

  xAngle = -4.0;
  yAngle = 17.0;
  zAngle = -1.0;
  
  /*
  if (Option[16]) {
    y_translate = -25.0;
    //    xAngle = -3.0;
    zAngle = 3.0;
    Zoom = 0.88;
    Zoom = 4.0;
    yAngle = 10.0;
  }
  */

  /*
  Zoom = 0.76;
  x_translate = -5.0;
  y_translate = -40.0;
  z_translate = 0.0;
  */
  /*
  Zoom = 0.80;
  xAngle = -4.0;
  yAngle = 10.0;
  zAngle = -1.0;
  x_translate = -25.0;
  y_translate = -25.0;
  z_translate = 0.0;
  */
  Zoom = 0.7; // 0.72
  xAngle = -4.0; // -4.0
  yAngle = 15.0; // 17
  zAngle = -1.0; // -1
  x_translate = -10.0;
  y_translate = -10.0;
  z_translate = -5.0;

  /*
  Zoom = 5.8;
  xAngle = -6.0;
  yAngle = 15.0;
  zAngle = -3.0;
  x_translate = -70.0;
  y_translate = 10.0;
  z_translate = -20.0;
  */

  /*
  xAngle = 0.0;
  yAngle = -90.0;
  zAngle = 0.0;
  x_translate = -10.0;
  y_translate = -75.0; // -60
  z_translate = 0.0;
  Zoom = 2.06; // 1.44
  */

  
  snglBuf[0] = GLX_RGBA;
  snglBuf[1] = GLX_DEPTH_SIZE;
  snglBuf[2] = 16;
  snglBuf[3] = None;

  dblBuf[0] = GLX_RGBA;
  dblBuf[1] = GLX_DEPTH_SIZE;
  dblBuf[2] = 16;
  dblBuf[3] = GLX_DOUBLEBUFFER;
  dblBuf[4] = None;
  

  /*** (1) open a connection to the X server ***/

  dpy = XOpenDisplay(NULL);
  if (dpy == NULL)
    fatalError("could not open display");

  /*** (2) make sure OpenGL's GLX extension supported ***/

  if (!glXQueryExtension(dpy, &dummy, &dummy))
    fatalError("X server has no OpenGL GLX extension");

  /*** (3) find an appropriate visual ***/

  /* find an OpenGL-capable RGB visual with depth buffer */
  vi = glXChooseVisual(dpy, DefaultScreen(dpy), dblBuf);
  if (vi == NULL) {
    vi = glXChooseVisual(dpy, DefaultScreen(dpy), snglBuf);
    if (vi == NULL)
      fatalError("no RGB visual with depth buffer");
    doubleBuffer = GL_FALSE;
  }

  //  if (vi->class != TrueColor)  // <- depending on definition of __cplusplus
  if (vi->c_class != TrueColor)
    fatalError("TrueColor visual required for this program");

  /*** (4) create an OpenGL rendering context  ***/

  /* create an OpenGL rendering context */
  /* no sharing of display lists */
  /* direct rendering if possible */
  cx = glXCreateContext(dpy, vi, None, GL_TRUE);

  if (cx == NULL)
    fatalError("could not create rendering context");
  
  /*** (5) create an X window with the selected visual ***/

  /* create an X colormap since probably not using default visual */
  cmap = XCreateColormap(dpy, RootWindow(dpy, vi->screen), vi->visual, AllocNone);
  swa.colormap = cmap;
  swa.border_pixel = 0;
  swa.event_mask = KeyPressMask | KeyReleaseMask |
    ExposureMask | ButtonPressMask | StructureNotifyMask;
  win = XCreateWindow(dpy, RootWindow(dpy, vi->screen), 0, 0,
		      xsize, ysize, 0, vi->depth,
		      InputOutput, vi->visual, CWBorderPixel | CWColormap | CWEventMask, &swa);
  XSetStandardProperties(dpy, win, title, title,
			 None, NULL, 0, NULL);

  /*** (6) bind the rendering context to the window ***/
  
  glXMakeCurrent(dpy, win, cx);
    
  /*** (7) request the X window to be displayed on the screen ***/

  XMapWindow(dpy, win);

  hsv2rgb(110.0, 0.15, 0.32, &col_grey2[0], &col_grey2[1], &col_grey2[2]);
  hsv2rgb(345.0, 1.0, 1.0, &col_red[0], &col_red[1], &col_red[2]);
  hsv2rgb(105.0, 0.7, 1.0, &col_green[0], &col_green[1], &col_green[2]);
  hsv2rgb(240.0, 1.0, 1.0, &col_blue2[0], &col_blue2[1], &col_blue2[2]);
}


void Graphic :: Initialize(double xrot, double yrot, double xdist,
			   double ydist, double zdist)
{
  /* establish initial viewport */
  /* pedantic, full window size is default viewport */
  glViewport(0, 0, xsize, ysize);

  GLfloat mat_specular[] = {1.0, 1.0, 1.0};
  GLfloat mat_shininess[] = {50.0};
  GLfloat light_position[] = {1.0, 1.0, 1.0, 0.0};

  glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);

  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LESS);
  glClearDepth(1.0);

  /*  The default z mapping  */
  glDepthRange (-2000.0, 2000.0);
  //    glClearColor(1.0, 1.0, 1.0, 1.0);
  //  glClearColor(0.63, 0.7, 0.625, 0.65);
  //  glClearColor(col_grey1[0], col_grey1[1], col_grey1[2], 1.0);
  //  glClearColor(col_back_grnd[0], col_back_grnd[1], col_back_grnd[2], 1.0);
  glClearColor(col_black[0], col_black[1], col_black[2], 1.0);
  //  glClearColor(col_white[0], col_white[1], col_white[2], 1.0);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();


  //  gluPerspective(60.0, 1.0, 1.0, 2000.0);
  gluPerspective(45.0, (GLfloat)xsize / (GLfloat)ysize, 1.0, 2000.0);
  if (ydist == 0.0)
    ydist = -0.3;
  if (zdist == 0.0)
    zdist = -5.5;
  if (xrot != 90.0) {
    RotX(xrot);
    RotY(yrot);
    glTranslated(xdist, ydist, zdist);
  } else {
    glTranslated(xdist, ydist, zdist);
    RotX(xrot);
    RotY(yrot);
  }  

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  if (doubleBuffer)
    glXSwapBuffers(dpy, win); /* buffer swap does implicit glFlush */
  else
    glFlush(); /* explicit flush for single buffered case */
}



void Graphic :: Initialize(double xrot, double yrot, double zrot,
			   double xdist, double ydist, double zdist)
{
  /* establish initial viewport */
  /* pedantic, full window size is default viewport */
  glViewport(0, 0, xsize, ysize);

#ifdef LIGHTING
  GLfloat mat_specular[] = {1.0, 1.0, 1.0};
  GLfloat mat_shininess[] = {50.0};
  GLfloat light_position[] = {1000.0, 1000.0, 0.0, 0.0};

  glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
  glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
  glLightfv(GL_LIGHT0, GL_POSITION, light_position);

  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
#endif

  //  glDepthFunc(GL_LESS);
  glDepthFunc(GL_LEQUAL);
  glClearDepth(1.0);
  glEnable(GL_DEPTH_TEST);

  /*  The default z mapping  */
  glDepthRange (-2000.0, 2000.0);

  //    glClearColor(1.0, 1.0, 1.0, 1.0);
  //  glClearColor(0.63, 0.7, 0.625, 0.65);
  //  glClearColor(col_grey1[0], col_grey1[1], col_grey1[2], 1.0);
  //  glClearColor(col_back_grnd[0], col_back_grnd[1], col_back_grnd[2], 1.0);
  glClearColor(col_black[0], col_black[1], col_black[2], 1.0);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();


  //  gluPerspective(60.0, 1.0, 1.0, 2000.0);
  gluPerspective(45.0, (GLfloat)xsize / (GLfloat)ysize, 1.0, 2000.0);
  if (ydist == 0.0)
    ydist = -0.3;
  if (zdist == 0.0)
    zdist = -5.5;

  /*
  xAngle = 0.0;
  yAngle = -90.0;
  zAngle = 0.0;
  x_translate = -10.0;
  y_translate = -85.0;
  z_translate = 0.0;
  Zoom = 1.9;
  */

  if (xrot != 90.0) {
    RotZ(zrot);
    RotY(yrot);
    RotX(xrot);
    glTranslated(xdist, ydist, zdist);
  } else {
    glTranslated(xdist, ydist, zdist);
    RotX(xrot);
    RotY(yrot);
    RotZ(zrot);
  }  

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  if (doubleBuffer)
    glXSwapBuffers(dpy, win); /* buffer swap does implicit glFlush */
  else
    glFlush(); /* explicit flush for single buffered case */
}


void Graphic :: SetClearColor(Real r, Real g, Real b, Real a)
{
  glClearColor(r, g, b, a);
}

void Graphic :: RotX(double angle)
{
  glRotated(angle, 0.1, 0.0, 0.0);
}

void Graphic :: RotY(double angle)
{
  glRotated(angle, 0.0, 0.1, 0.0);
}

void Graphic :: RotZ(double angle)
{
  glRotated(angle, 0.0, 0.0, 1.0);
}

void Graphic :: RotateX(double angle)
{
   xAngle += angle;
}

void Graphic :: RotateY(double angle)
{
   yAngle += angle;
}

void Graphic :: RotateZ(double angle)
{
   zAngle += angle;
}



/**************************************************/

void Graphic :: rotate(double angle, double *vec)
{
   glRotated(angle, vec[0], vec[1], vec[2]);
}

void Graphic :: translate(double x, double y, double z)
{
   glTranslated(x, y, z);
}

void Graphic :: set_transformation(double matrix[4][4])
{
   /*
   int i, j;
   double flip[4][4];
   for (i=0; i<4; i++)
      for (j=0; j<4; j++)
	 flip[i][j] = matrix[j][i];
	 */
   glMultMatrixd((double*)matrix);
}

void Graphic :: push_orientation(void)
{
   glPushMatrix();
}

void Graphic :: pop_orientation(void)
{
   glPopMatrix();
}


/**********************************************/


void Graphic :: ScalePrism(pBox box, double val)
{
   int i, j;

   for (i=0; i<8; i++)
      for (j=0; j<3; j++)
	 box_coord(box, i, j) *= val;
}


pBox Graphic :: MakeBox(double scale)
{
   pBox box;

   box = (pBox)malloc(sizeof(struct tBox));

   box_3color(box, 0) = 1.0;
   box_3color(box, 1) = 0.1;
   box_3color(box, 2) = 1.0;

   box_coord(box, 0, 0) = 0.0;
   box_coord(box, 0, 1) = 0.0;
   box_coord(box, 0, 2) = 0.0;

   box_coord(box, 1, 0) = 0.0;
   box_coord(box, 1, 1) = 0.0;
   box_coord(box, 1, 2) = 1.0;

   box_coord(box, 2, 0) = 0.0;
   box_coord(box, 2, 1) = 1.0;
   box_coord(box, 2, 2) = 0.0;

   box_coord(box, 3, 0) = 0.0;
   box_coord(box, 3, 1) = 1.0;
   box_coord(box, 3, 2) = 1.0;

   box_coord(box, 4, 0) = 1.0;
   box_coord(box, 4, 1) = 1.0;
   box_coord(box, 4, 2) = 0.0;

   box_coord(box, 5, 0) = 1.0;
   box_coord(box, 5, 1) = 1.0;
   box_coord(box, 5, 2) = 1.0;

   box_coord(box, 6, 0) = 1.0;
   box_coord(box, 6, 1) = 0.0;
   box_coord(box, 6, 2) = 0.0;

   box_coord(box, 7, 0) = 1.0;
   box_coord(box, 7, 1) = 0.0;
   box_coord(box, 7, 2) = 1.0;
   ScalePrism(box, scale);

   return box;
}


/*************************************************/


void Graphic :: drawPlaneY(double x, double y, double z)
{
   glBegin(GL_QUAD_STRIP);
   glVertex3d(-0.5*x, y, -0.5*z);
   glVertex3d(-0.5*x, y, 0.5*z);
   glVertex3d(0.5*x, y, -0.5*z);
   glVertex3d(0.5*x, y, 0.5*z);
   glEnd();
}

void Graphic :: drawPlaneZ(double x, double y, double z)
{
   glBegin(GL_QUAD_STRIP);
   glVertex3d(-0.2+x, -0.2+y, z);
   glVertex3d(-0.2+x, 0.2+y, z);
   glVertex3d(0.2+x, -0.2+y, z);
   glVertex3d(0.2+x, 0.2+y, z);
   glEnd();
}


void Graphic :: drawPrismOL(pBox box)
{
   int i;

   glColor3dv(box_color1(box));
   glBegin(GL_QUAD_STRIP);
   for (i = 0; i < 8; i++)
      glVertex3dv(box_pt(box, i));
   glEnd();

   glColor3dv(box_color2(box));

   glBegin(GL_QUAD_STRIP);
      glVertex3dv(box_pt(box, 2));
      glVertex3dv(box_pt(box, 4));
      glVertex3dv(box_pt(box, 0));
      glVertex3dv(box_pt(box, 6));
      glVertex3dv(box_pt(box, 1));
      glVertex3dv(box_pt(box, 7));
      glVertex3dv(box_pt(box, 3));
      glVertex3dv(box_pt(box, 5));
   glEnd();

   glColor3dv(col_black);
   glBegin(GL_LINE_STRIP);
   glVertex3dv(box_pt(box, 0));
   glVertex3dv(box_pt(box, 1));
   glVertex3dv(box_pt(box, 3));
   glVertex3dv(box_pt(box, 2));
   glVertex3dv(box_pt(box, 0));
   glEnd();

   glBegin(GL_LINE_STRIP);
   glVertex3dv(box_pt(box, 4));
   glVertex3dv(box_pt(box, 5));
   glVertex3dv(box_pt(box, 7));
   glVertex3dv(box_pt(box, 6));
   glVertex3dv(box_pt(box, 4));
   glEnd();

   glBegin(GL_LINE);
   glVertex3dv(box_pt(box, 0));
   glVertex3dv(box_pt(box, 4));
   glVertex3dv(box_pt(box, 1));
   glVertex3dv(box_pt(box, 5));
   glVertex3dv(box_pt(box, 2));
   glVertex3dv(box_pt(box, 6));
   glVertex3dv(box_pt(box, 3));
   glVertex3dv(box_pt(box, 7));
   glEnd();
}

void Graphic :: drawPrismOL(pBox box, double *colorX,
			    double *colorY, double *colorZ)
{
   glColor3dv(colorX);
   glBegin(GL_QUADS);
   glVertex3dv(box_pt(box, 0));
   glVertex3dv(box_pt(box, 1));
   glVertex3dv(box_pt(box, 3));
   glVertex3dv(box_pt(box, 2));

   glVertex3dv(box_pt(box, 4));
   glVertex3dv(box_pt(box, 5));
   glVertex3dv(box_pt(box, 7));
   glVertex3dv(box_pt(box, 6));

   glColor3dv(colorY);
   glVertex3dv(box_pt(box, 0));
   glVertex3dv(box_pt(box, 1));
   glVertex3dv(box_pt(box, 7));
   glVertex3dv(box_pt(box, 6));

   glVertex3dv(box_pt(box, 2));
   glVertex3dv(box_pt(box, 3));
   glVertex3dv(box_pt(box, 5));
   glVertex3dv(box_pt(box, 4));

   glColor3dv(colorZ);
   glVertex3dv(box_pt(box, 0));
   glVertex3dv(box_pt(box, 2));
   glVertex3dv(box_pt(box, 4));
   glVertex3dv(box_pt(box, 6));

   glVertex3dv(box_pt(box, 1));
   glVertex3dv(box_pt(box, 3));
   glVertex3dv(box_pt(box, 5));
   glVertex3dv(box_pt(box, 7));
   glEnd();

   
   // Outline
   glColor3dv(col_black);
   glBegin(GL_LINES);
   // -x
   glVertex3dv(box_pt(box, 0));
   glVertex3dv(box_pt(box, 1));
   glVertex3dv(box_pt(box, 1));
   glVertex3dv(box_pt(box, 3));
   glVertex3dv(box_pt(box, 3));
   glVertex3dv(box_pt(box, 2));
   glVertex3dv(box_pt(box, 2));
   glVertex3dv(box_pt(box, 0));

   // +x
   glVertex3dv(box_pt(box, 4));
   glVertex3dv(box_pt(box, 5));
   glVertex3dv(box_pt(box, 5));
   glVertex3dv(box_pt(box, 7));
   glVertex3dv(box_pt(box, 7));
   glVertex3dv(box_pt(box, 6));
   glVertex3dv(box_pt(box, 6));
   glVertex3dv(box_pt(box, 4));

   // lines connecting +x:-x
   glVertex3dv(box_pt(box, 0));
   glVertex3dv(box_pt(box, 6));
   glVertex3dv(box_pt(box, 1));
   glVertex3dv(box_pt(box, 7));
   glVertex3dv(box_pt(box, 3));
   glVertex3dv(box_pt(box, 5));
   glVertex3dv(box_pt(box, 2));
   glVertex3dv(box_pt(box, 4));
   glEnd();
}

void Graphic :: drawPrism(pBox box, double *colorX,
			  double *colorY, double *colorZ)
{
   glColor3dv(colorX);
   glBegin(GL_QUADS);
   glVertex3dv(box_pt(box, 0));
   glVertex3dv(box_pt(box, 1));
   glVertex3dv(box_pt(box, 3));
   glVertex3dv(box_pt(box, 2));

   glVertex3dv(box_pt(box, 4));
   glVertex3dv(box_pt(box, 5));
   glVertex3dv(box_pt(box, 7));
   glVertex3dv(box_pt(box, 6));

   glColor3dv(colorY);
   glVertex3dv(box_pt(box, 0));
   glVertex3dv(box_pt(box, 1));
   glVertex3dv(box_pt(box, 7));
   glVertex3dv(box_pt(box, 6));

   glVertex3dv(box_pt(box, 2));
   glVertex3dv(box_pt(box, 3));
   glVertex3dv(box_pt(box, 5));
   glVertex3dv(box_pt(box, 4));

   glColor3dv(colorZ);
   glVertex3dv(box_pt(box, 0));
   glVertex3dv(box_pt(box, 2));
   glVertex3dv(box_pt(box, 4));
   glVertex3dv(box_pt(box, 6));

   glVertex3dv(box_pt(box, 1));
   glVertex3dv(box_pt(box, 3));
   glVertex3dv(box_pt(box, 5));
   glVertex3dv(box_pt(box, 7));
   glEnd();
}


void Graphic :: drawPrism(pBox box)
{
   int i;

   glColor3dv(box_color1(box));
   glBegin(GL_QUAD_STRIP);
   for (i = 0; i < 8; i++)
      glVertex3dv(box_pt(box, i));
   glEnd();

   glColor3dv(box_color2(box));

   glBegin(GL_QUAD_STRIP);
      glVertex3dv(box_pt(box, 2));
      glVertex3dv(box_pt(box, 4));
      glVertex3dv(box_pt(box, 0));
      glVertex3dv(box_pt(box, 6));
      glVertex3dv(box_pt(box, 1));
      glVertex3dv(box_pt(box, 7));
      glVertex3dv(box_pt(box, 3));
      glVertex3dv(box_pt(box, 5));
   glEnd();
}

void Graphic :: drawPoint(double x, double y, double z)
{
  glVertex3d(x, y, z);
}

void Graphic :: SetLineWidth(double w)
{
  glLineWidth(w);
}

void Graphic :: startLine(void)
{
  glBegin(GL_LINES);
}

void Graphic :: startPolygon(void)
{
  glBegin(GL_POLYGON);
}

// Assumes we are in lines mode.
#define VERTEX_SIZE 1.0
void Graphic :: starLocation(double x, double y)
{
  glVertex3d(x+VERTEX_SIZE, y+VERTEX_SIZE, 0.0);
  glVertex3d(x-VERTEX_SIZE, y-VERTEX_SIZE, 0.0);
  glVertex3d(x+VERTEX_SIZE, y-VERTEX_SIZE, 0.0);
  glVertex3d(x-VERTEX_SIZE, y+VERTEX_SIZE, 0.0);
}


void Graphic :: starLocation(double x, double y, double z)
{
  glVertex3d(x+VERTEX_SIZE, y+VERTEX_SIZE, z);
  glVertex3d(x-VERTEX_SIZE, y-VERTEX_SIZE, z);
  glVertex3d(x+VERTEX_SIZE, y-VERTEX_SIZE, z);
  glVertex3d(x-VERTEX_SIZE, y+VERTEX_SIZE, z);

  glVertex3d(x+VERTEX_SIZE, y, z+VERTEX_SIZE);
  glVertex3d(x-VERTEX_SIZE, y, z-VERTEX_SIZE);
  glVertex3d(x+VERTEX_SIZE, y, z-VERTEX_SIZE);
  glVertex3d(x-VERTEX_SIZE, y, z+VERTEX_SIZE);

  glVertex3d(x, y+VERTEX_SIZE, z+VERTEX_SIZE);
  glVertex3d(x, y-VERTEX_SIZE, z-VERTEX_SIZE);
  glVertex3d(x, y+VERTEX_SIZE, z-VERTEX_SIZE);
  glVertex3d(x, y-VERTEX_SIZE, z+VERTEX_SIZE);
}



#define VERTEX_SIZE_BIG 2.0
void Graphic :: starLocationBig(double x, double y, double z)
{
  glVertex3d(x+VERTEX_SIZE_BIG, y+VERTEX_SIZE_BIG, z);
  glVertex3d(x-VERTEX_SIZE_BIG, y-VERTEX_SIZE_BIG, z);
  glVertex3d(x+VERTEX_SIZE_BIG, y-VERTEX_SIZE_BIG, z);
  glVertex3d(x-VERTEX_SIZE_BIG, y+VERTEX_SIZE_BIG, z);

  glVertex3d(x+VERTEX_SIZE_BIG, y, z+VERTEX_SIZE_BIG);
  glVertex3d(x-VERTEX_SIZE_BIG, y, z-VERTEX_SIZE_BIG);
  glVertex3d(x+VERTEX_SIZE_BIG, y, z-VERTEX_SIZE_BIG);
  glVertex3d(x-VERTEX_SIZE_BIG, y, z+VERTEX_SIZE_BIG);

  glVertex3d(x, y+VERTEX_SIZE_BIG, z+VERTEX_SIZE_BIG);
  glVertex3d(x, y-VERTEX_SIZE_BIG, z-VERTEX_SIZE_BIG);
  glVertex3d(x, y+VERTEX_SIZE_BIG, z-VERTEX_SIZE_BIG);
  glVertex3d(x, y-VERTEX_SIZE_BIG, z+VERTEX_SIZE_BIG);
}

void Graphic :: starShadowLocation(double x, double y, double z)
{
  glVertex3d(x+VERTEX_SIZE, y, z);
  glVertex3d(x-VERTEX_SIZE, y, z);

  glVertex3d(x+VERTEX_SIZE, y, z+VERTEX_SIZE);
  glVertex3d(x-VERTEX_SIZE, y, z-VERTEX_SIZE);
  glVertex3d(x+VERTEX_SIZE, y, z-VERTEX_SIZE);
  glVertex3d(x-VERTEX_SIZE, y, z+VERTEX_SIZE);

  glVertex3d(x, y, z+VERTEX_SIZE);
  glVertex3d(x, y, z-VERTEX_SIZE);
}

void Graphic :: starShadowLocationBig(double x, double y, double z)
{
  glVertex3d(x+VERTEX_SIZE_BIG, y, z);
  glVertex3d(x-VERTEX_SIZE_BIG, y, z);

  glVertex3d(x+VERTEX_SIZE_BIG, y, z+VERTEX_SIZE_BIG);
  glVertex3d(x-VERTEX_SIZE_BIG, y, z-VERTEX_SIZE_BIG);
  glVertex3d(x+VERTEX_SIZE_BIG, y, z-VERTEX_SIZE_BIG);
  glVertex3d(x-VERTEX_SIZE_BIG, y, z+VERTEX_SIZE_BIG);

  glVertex3d(x, y, z+VERTEX_SIZE_BIG);
  glVertex3d(x, y, z-VERTEX_SIZE_BIG);
}


// Assumes we are in lines mode.
void Graphic :: addVertex(double x, double y)
{
  //  glVertex2d(x, y);
  glVertex3d(x, y, 0.0);
}

// Assumes we are in lines mode.
void Graphic :: addVertex(double x, double y, double z)
{
  //  glVertex2d(x, y);
  glVertex3d(x, y, z);
}

void Graphic :: addVertex3v(double vec[3])
{
  glVertex3dv(vec);
}


void Graphic :: stopLine(void)
{
  glEnd();
}

void Graphic :: stopList(void)
{
  glEnd();
}


void Graphic :: drawLine(double x1, double y1, double x2, double y2)
{
  glColor3f(0.0, 0.0, 0.0);
  glBegin(GL_LINES);
  glVertex3d(x1, y1, 0.0);
  glVertex3d(x2, y2, 0.0);
  glEnd();
}

#define V_JOINT 5.0
void Graphic :: drawJoint(void)
{
   glColor3f(0.0, 0.0, 0.0);
   glBegin(GL_LINES);
   glVertex3f(0.0, 0.0, V_JOINT);
   glVertex3f(0.0, 0.0, -V_JOINT);
   glVertex3f(V_JOINT, 0.0, 0.0);
   glVertex3f(-V_JOINT, 0.0, 0.0);
   glVertex3f(0.0, V_JOINT, 0.0);
   glVertex3f(0.0, -V_JOINT, 0.0);
   /*   glVertex3f(0.0, 0.0, 0.0);
   glVertex3dv(vec); */
   glEnd();
}



void Graphic :: drawBox(double *dim, double *color1, double *color2)
{
   struct tBox box;

   /*   printf("%.4f %.4f %.4f\n", dim[0], dim[1], dim[2]);
	printf("%.4f %.4f %.4f\n", color1[0], color1[1], color1[2]); */
   box_color1((&box)) = color1;
   box_color2((&box)) = color2;

   box_coord((&box), 0, 0) = -dim[0]/2.0;
   box_coord((&box), 0, 1) = -dim[1]/2.0;
   box_coord((&box), 0, 2) = -dim[2]/2.0;

   box_coord((&box), 1, 0) = -dim[0]/2.0;
   box_coord((&box), 1, 1) = -dim[1]/2.0;
   box_coord((&box), 1, 2) = dim[2]/2.0;

   box_coord((&box), 2, 0) = -dim[0]/2.0;
   box_coord((&box), 2, 1) = dim[1]/2.0;
   box_coord((&box), 2, 2) = -dim[2]/2.0;

   box_coord((&box), 3, 0) = -dim[0]/2.0;
   box_coord((&box), 3, 1) = dim[1]/2.0;
   box_coord((&box), 3, 2) = dim[2]/2.0;

   box_coord((&box), 4, 0) = dim[0]/2.0;
   box_coord((&box), 4, 1) = dim[1]/2.0;
   box_coord((&box), 4, 2) = -dim[2]/2.0;

   box_coord((&box), 5, 0) = dim[0]/2.0;
   box_coord((&box), 5, 1) = dim[1]/2.0;
   box_coord((&box), 5, 2) = dim[2]/2.0;

   box_coord((&box), 6, 0) = dim[0]/2.0;
   box_coord((&box), 6, 1) = -dim[1]/2.0;
   box_coord((&box), 6, 2) = -dim[2]/2.0;

   box_coord((&box), 7, 0) = dim[0]/2.0;
   box_coord((&box), 7, 1) = -dim[1]/2.0;
   box_coord((&box), 7, 2) = dim[2]/2.0;

   drawPrism((&box));
}


void Graphic :: drawBoxOL(double *dim, double *color1, double *color2)
{
   struct tBox box;

   /*   printf("%.4f %.4f %.4f\n", dim[0], dim[1], dim[2]);
	printf("%.4f %.4f %.4f\n", color1[0], color1[1], color1[2]); */
   box_color1((&box)) = color1;
   box_color2((&box)) = color2;

   box_coord((&box), 0, 0) = -dim[0]/2.0;
   box_coord((&box), 0, 1) = -dim[1]/2.0;
   box_coord((&box), 0, 2) = -dim[2]/2.0;

   box_coord((&box), 1, 0) = -dim[0]/2.0;
   box_coord((&box), 1, 1) = -dim[1]/2.0;
   box_coord((&box), 1, 2) = dim[2]/2.0;

   box_coord((&box), 2, 0) = -dim[0]/2.0;
   box_coord((&box), 2, 1) = dim[1]/2.0;
   box_coord((&box), 2, 2) = -dim[2]/2.0;

   box_coord((&box), 3, 0) = -dim[0]/2.0;
   box_coord((&box), 3, 1) = dim[1]/2.0;
   box_coord((&box), 3, 2) = dim[2]/2.0;

   box_coord((&box), 4, 0) = dim[0]/2.0;
   box_coord((&box), 4, 1) = dim[1]/2.0;
   box_coord((&box), 4, 2) = -dim[2]/2.0;

   box_coord((&box), 5, 0) = dim[0]/2.0;
   box_coord((&box), 5, 1) = dim[1]/2.0;
   box_coord((&box), 5, 2) = dim[2]/2.0;

   box_coord((&box), 6, 0) = dim[0]/2.0;
   box_coord((&box), 6, 1) = -dim[1]/2.0;
   box_coord((&box), 6, 2) = -dim[2]/2.0;

   box_coord((&box), 7, 0) = dim[0]/2.0;
   box_coord((&box), 7, 1) = -dim[1]/2.0;
   box_coord((&box), 7, 2) = dim[2]/2.0;

   drawPrismOL((&box));
}


void Graphic :: drawBoxOL(double *dim, double *colorX,
			  double *colorY, double *colorZ)
{
   struct tBox box;

   /*   printf("%.4f %.4f %.4f\n", dim[0], dim[1], dim[2]);
	printf("%.4f %.4f %.4f\n", color1[0], color1[1], color1[2]); */

   box_coord((&box), 0, 0) = -dim[0]/2.0;
   box_coord((&box), 0, 1) = -dim[1]/2.0;
   box_coord((&box), 0, 2) = -dim[2]/2.0;

   box_coord((&box), 1, 0) = -dim[0]/2.0;
   box_coord((&box), 1, 1) = -dim[1]/2.0;
   box_coord((&box), 1, 2) = dim[2]/2.0;

   box_coord((&box), 2, 0) = -dim[0]/2.0;
   box_coord((&box), 2, 1) = dim[1]/2.0;
   box_coord((&box), 2, 2) = -dim[2]/2.0;

   box_coord((&box), 3, 0) = -dim[0]/2.0;
   box_coord((&box), 3, 1) = dim[1]/2.0;
   box_coord((&box), 3, 2) = dim[2]/2.0;

   box_coord((&box), 4, 0) = dim[0]/2.0;
   box_coord((&box), 4, 1) = dim[1]/2.0;
   box_coord((&box), 4, 2) = -dim[2]/2.0;

   box_coord((&box), 5, 0) = dim[0]/2.0;
   box_coord((&box), 5, 1) = dim[1]/2.0;
   box_coord((&box), 5, 2) = dim[2]/2.0;

   box_coord((&box), 6, 0) = dim[0]/2.0;
   box_coord((&box), 6, 1) = -dim[1]/2.0;
   box_coord((&box), 6, 2) = -dim[2]/2.0;

   box_coord((&box), 7, 0) = dim[0]/2.0;
   box_coord((&box), 7, 1) = -dim[1]/2.0;
   box_coord((&box), 7, 2) = dim[2]/2.0;

   drawPrismOL((&box), colorX, colorY, colorZ);
}


int Made_default = FALSE;
struct tBox Default_box;
double *ColorX, *ColorY, *ColorZ;

void Graphic :: SetDefaultBox(double *dim, double *colorX,
			  double *colorY, double *colorZ)
{
  if (Made_default)
    return;

  ColorX = colorX;
  ColorY = colorY;
  ColorZ = colorZ;

  box_coord((&Default_box), 0, 0) = -dim[0]/2.0;
  box_coord((&Default_box), 0, 1) = -dim[1]/2.0;
  box_coord((&Default_box), 0, 2) = -dim[2]/2.0;

  box_coord((&Default_box), 1, 0) = -dim[0]/2.0;
  box_coord((&Default_box), 1, 1) = -dim[1]/2.0;
  box_coord((&Default_box), 1, 2) = dim[2]/2.0;

  box_coord((&Default_box), 2, 0) = -dim[0]/2.0;
  box_coord((&Default_box), 2, 1) = dim[1]/2.0;
  box_coord((&Default_box), 2, 2) = -dim[2]/2.0;

  box_coord((&Default_box), 3, 0) = -dim[0]/2.0;
  box_coord((&Default_box), 3, 1) = dim[1]/2.0;
  box_coord((&Default_box), 3, 2) = dim[2]/2.0;

  box_coord((&Default_box), 4, 0) = dim[0]/2.0;
  box_coord((&Default_box), 4, 1) = dim[1]/2.0;
  box_coord((&Default_box), 4, 2) = -dim[2]/2.0;

  box_coord((&Default_box), 5, 0) = dim[0]/2.0;
  box_coord((&Default_box), 5, 1) = dim[1]/2.0;
  box_coord((&Default_box), 5, 2) = dim[2]/2.0;

  box_coord((&Default_box), 6, 0) = dim[0]/2.0;
  box_coord((&Default_box), 6, 1) = -dim[1]/2.0;
  box_coord((&Default_box), 6, 2) = -dim[2]/2.0;

  box_coord((&Default_box), 7, 0) = dim[0]/2.0;
  box_coord((&Default_box), 7, 1) = -dim[1]/2.0;
  box_coord((&Default_box), 7, 2) = dim[2]/2.0;

  glNewList(1, GL_COMPILE);
  if (Print[29])
    drawPrismOL(&Default_box, ColorX, ColorY, ColorZ);
  else
    drawPrism(&Default_box, ColorX, ColorY, ColorZ);
  glEndList();
  Made_default = TRUE;
}


void Graphic :: drawDefaultBox(void)
{
  if (!Made_default)
    return;
  glCallList(1);
}


void Graphic :: drawRect(double x1, double y1, double x2, double y2)
{
  glRectd(x1, y1, x2, y2);
}



/*=============================================================================
**=============================================================================
**
** Function: drawSphere()
**
** Description:
**
** Draws the sphere at the correct location, scale and orientation and using
** the desired colors. The sphere will either be drawn as wire frame or
** as a flat shaded model depending on the current mode specified by the
** interface.
**
**=============================================================================
**===========================================================================*/

#define X 0
#define Y 1
#define Z 2
                                
void Graphic :: drawSphere(double radius, double col1[3], double col2[3])
{
  int toggle = FALSE;
  int i, j;                         /* General purpose indices              */
  float v[4][3];                    /* Holds vertices of current polygon    */
  float n[4][3];
  double angle;                     /* The angle in the XZ plane            */
  double angleY;                    /* The angle in the XY plane            */
  double angleStep = 2*M_PI/10.0;   /* Determines number of polygons in sphr*/
  double x, y, z;                   /* Holds the current x,y,z coordinates  */



  /*==========================================================================
   * Position the object, rotate it and scale it. Note that these transforms
   * have to be done in this order. If we were to place scale before rotate
   * then we would not get the correct results. Move them around and convince
   * yourself that this is the case.
   *=========================================================================*/

  glPushMatrix();
  glScalef(radius, radius, radius);

  /*==========================================================================
   * Now its time to compute the vertices of the sphere. This uses the
   * standard method of computing the x,y,z coordinates. If you don't 
   * understand, then remember that the equation of a unit CIRCLE is given
   * by x = cos(angle), z = sin(angle), the circle being in the x,z plane.
   * A sphere can then be imaged as a set of these circles extending up
   * the Y axis. However the circles radii must get smaller as we move
   * towards the top of the sphere - i.e as Y increases. This is done by
   * scaling the x and z coordiantes by sine(angleY). See below...
   *
   * Note that the code below computes quadrilateral polygons and is therefore
   * not the most efficient or even the most exact. OpenGL has support for
   * triangles and triangle meshes which you might want to consider also.
   *=========================================================================*/

  angleY = 0;
  for (i = 0; i <=4 ; i++) {
    angle = 0.0;
    for (j = 0; j <=10; j++) {

      /*====================================================================
       * Compute the first point. We have 30 in each direction. This is
       * a fairly detailed sphere, and you might want to lower this number
       * for speed. Just remember to change the angleStep calculation above
       * too.
       *===================================================================*/

      x = cos(angle) * sin(angleY);
      z = sin(angle) * sin(angleY) ;
      y = cos(angleY);
      
      /*===================================================================
       * The normals for the sphere are easy as they are just the same
       * as the hit point. Draw a sphere if you can't see this.
       *==================================================================*/

      v[0][X] = x; v[0][Y] = y; v[0][Z] = z;
      n[0][X] = x; n[0][Y] = y; n[0][Z] = z;
         
      /*====================================================================
       * Compute second vertex and normal
       *===================================================================*/

      angle = angle + angleStep;

      x = cos(angle) * sin(angleY);
      z = sin(angle) * sin(angleY) ;
      y = cos(angleY);

      v[1][X] = x; v[1][Y] = y; v[1][Z] = z;
      n[1][X] = x; n[1][Y] = y; n[1][Z] = z;

      /*====================================================================
       * Compute third vertex and normal
       *===================================================================*/

      angleY = angleY + angleStep;
         
      x = cos(angle) * sin(angleY);
      z = sin(angle) * sin(angleY) ;
      y = cos(angleY);

      v[2][X] = x; v[2][Y] = y; v[2][Z] = z;
      n[2][X] = x; n[2][Y] = y; n[2][Z] = z;
         
      /*====================================================================
       * Compute fourth vertex and normal
       *===================================================================*/

      angle = angle-angleStep;

      x = cos(angle) * sin(angleY);
      z = sin(angle) * sin(angleY) ;
      y = cos(angleY);

      v[3][X] = x; v[3][Y] = y; v[3][Z] = z;
      n[3][X] = x; n[3][Y] = y; n[3][Z] = z;

      angleY = angleY- angleStep;

      /*====================================================================
       * Now we draw the polygon defined by these four vertices. First 
       * though we check what mode we are in. If we are in wire frame mode
       * then we simply want to use the closedline geometry, otherwise
       * we can to use the filled polygon geometry. Check the GL manual
       * pages if you are unclear on these function calls.
       *===================================================================*/
      
      glBegin(GL_POLYGON);
        
      if (toggle)
	glColor3dv(col1);
      else
	glColor3dv(col2);
      toggle = !toggle;

	   
      /*===================================================================
       * Finally if we are in quick render mode we must specify the
       * normals for each vertex.
       *==================================================================*/

      glVertex3fv(v[0]);
      glVertex3fv(v[1]);
      glVertex3fv(v[2]);
      glVertex3fv(v[3]);
      
      glEnd();

      angle = angle + angleStep;
    }
    angleY = angleY + angleStep;
  }

  /*==========================================================================
   * This undoes the scale that we issued. This means that the next object
   * will not be scaled.
   *=========================================================================*/
   
  glPopMatrix();
  glPopMatrix();
}



void Graphic :: drawCylinder(double radius, double col1[3], double col2[3])
{

}



/**************************************************/

void Graphic :: SetColor(double col[3])
{
  glColor3dv(col);
}


// try: 0.2 0.4 0.65  // a nice blue
void Graphic :: SetColor (int col)
{
  if (col == BLACK)
    glColor3dv(col_black);
  else if (col == WHITE)
    glColor3dv(col_white);
  if (col == RED)
    glColor3dv(col_red);
  if (col == GREEN)
    glColor3dv(col_green);
  if (col == BLUE)
    glColor3dv(col_blue);
}


void Graphic :: Reshape(GLsizei w, GLsizei h)
{
  /*  prepare for and then  */ 
  /*  define the projection  */
  /*  transformation  */
  glViewport(0, 0, w, h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-1.0, 1.0, -1.0, 1.0, 0.0, 20.0);
}


void Graphic :: Redraw(void (*update_func)(int), int param)
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   
  update_func(param);

  if (doubleBuffer)
    glXSwapBuffers(dpy, win); /* buffer swap does implicit glFlush */
  else
    glFlush(); /* explicit flush for single buffered case */
}


void Graphic :: StartDrawing(void)
{
  glXMakeCurrent(dpy, win, cx);
  //  XMapWindow(dpy, win);

  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  RotX(xAngle);
  RotY(yAngle);
  RotZ(zAngle);

  glTranslated(x_translate, y_translate, z_translate);

  glScaled(Zoom, Zoom, Zoom);
}

void Graphic :: StartDrawing(int clear)
{
  glXMakeCurrent(dpy, win, cx);
  //  XMapWindow(dpy, win);

  if (clear)
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  else
    glXSwapBuffers(dpy, win); /* buffer swap does implicit glFlush */
    
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}

void Graphic :: DoneDrawing(void)
{
  if (save_frame) {
    WriteWindow("window.ppm");
    printf("..written.\n");
    save_frame = FALSE;

  } else if (save_movie) {
    char name[20];
    printf("Frame: %d,  ", frame_num);
    sprintf(&name[0], "movie%04d.ppm", frame_num++);
    WriteWindow(name);
  }

  if (doubleBuffer)
    glXSwapBuffers(dpy, win); /* buffer swap does implicit glFlush */
  else
    glFlush(); /* explicit flush for single buffered case */
}



// Write window to file
void Graphic :: WriteWindow(char *name)
{
  FILE *stream;

  printf("WriteWindow(%s)  %dx%d\n", name, xsize, ysize);

  if (pixels == 0) {
    pixels = (unsigned char *)malloc(3*xsize*ysize*
				     sizeof(unsigned char));
    pixels2 = (unsigned char *)malloc(3*xsize*ysize*
				      sizeof(unsigned char));
  }
  
  //  if (doubleBuffer)
  //    glXSwapBuffers(dpy, win);
  glReadPixels(0, 0, xsize, ysize, GL_RGB, GL_UNSIGNED_BYTE,
  	       pixels);
  //  if (doubleBuffer)
  //    glXSwapBuffers(dpy, win);

  for (int y=0; y<ysize; y++) {
    for (int x=0; x<xsize; x++) {
      pixels2[3*((ysize-y-1)*xsize+x)] = pixels[3*(y*xsize+x)];
      pixels2[3*((ysize-y-1)*xsize+x)+1] = pixels[3*(y*xsize+x)+1];
      pixels2[3*((ysize-y-1)*xsize+x)+2] = pixels[3*(y*xsize+x)+2];
    }
  }

  stream = fopen(name, "w");
  fprintf(stream, "P6\n%i %i\n255\n", xsize, ysize);
  fwrite(pixels2, 1, xsize*ysize*3, stream);

  fclose(stream);
}


/**************************************************/

int Graphic :: MainLoop(void (*update_func)(int))
{
  bool escape = FALSE;
  char buffer[1];
  int rc;
  int inc_time = FALSE;
  int ret_val = -1;
  KeySym keysym;
  XEvent event;

   /*** (9) dispatch X events ***/
  /*
  while (XPending(dpy)) {
    XNextEvent(dpy, &event);
  }
  */

  while (XPending(dpy)) {
     XNextEvent(dpy, &event);
     switch (event.type) {
     case KeyRelease:
       rc = XLookupString((XKeyEvent *) &event, buffer,
			  1, &keysym, NULL);
       switch (keysym) {
       case XK_Escape:
	 escape = TRUE;
	 ret_val = -2;
	 break;

       case 65434:
	 z_translate += 5.0;
	 recalcModel = GL_TRUE;
	 break;

       case 65436:
	 z_translate -= 5.0;
	 recalcModel = GL_TRUE;
	 break;

       case 65431:
	 y_translate += 5.0;
	 recalcModel = GL_TRUE;
	 break;

       case 65433:
	 y_translate -= 5.0;
	 recalcModel = GL_TRUE;
	 break;

       case 65430:
	 x_translate -= 5.0;
	 recalcModel = GL_TRUE;
	 break;

       case 65432:
	 x_translate += 5.0;
	 recalcModel = GL_TRUE;
	 break;

       case XK_m:
	 Draw_arg[2] = TRUE;
	 recalcModel = GL_TRUE;
	 break;

       case XK_h:
	 Draw_arg[3] = TRUE;
	 recalcModel = GL_TRUE;
	 break;

       case XK_i:
	 Draw_arg[4] = TRUE;
	 recalcModel = GL_TRUE;
	 break;

       case XK_space:
	 Draw_arg[0] = TRUE;
	 Draw_arg[1] = TRUE;
	 recalcModel = GL_TRUE;
	 if (save_movie_frame)
	   save_movie = TRUE;
	 break;

       case XK_Return:
	 ret_val = 1;
	 //	 recalcModel = GL_TRUE;
	 break;
       case XK_Down:
	 RotateX(1.0);
	 recalcModel = GL_TRUE;
	 break;
       case XK_Up:
	 RotateX(-1.0);
	 recalcModel = GL_TRUE;
	 break;
       case XK_Right:
	 RotateY(1.0);
	 recalcModel = GL_TRUE;
	 break;
       case XK_Left:
	 RotateY(-1.0);
	 recalcModel = GL_TRUE;
	 break;

       case XK_a:
	 /*	       world_bend(1.0); */
	 //	 needRedraw = GL_TRUE;
	 break;

       case XK_c:
	 if (save_movie_frame) {
	   printf("Movie stopped.\n");
	   save_movie_frame = FALSE;
	 } else {
	   printf("Movie started...\n");
	   save_movie_frame = TRUE;
	 }
	 break;

       case XK_s:
	 /*	       world_bend(-1.0); */
	 //	 needRedraw = GL_TRUE;
	 break;

       case XK_q:
	 RotateZ(1.0);
	 recalcModel = GL_TRUE;
	 break;
	 /*	       world_twist(1.0); */
	 //	 needRedraw = GL_TRUE;
	 break;
       case XK_w:
	 RotateZ(-1.0);
	 recalcModel = GL_TRUE;
	 break;
	 /*	       world_twist(-1.0); */
	 //	 needRedraw = GL_TRUE;

       case XK_p:
	 save_frame = TRUE;
	 break;

	case XK_z:
	  Zoom -= 0.02;
	  needRedraw = GL_TRUE;
	  break;

	case XK_x:
	  Zoom += 0.02;
	  needRedraw = GL_TRUE;
	  break;

       default:
	 printf("pressed: %ld\n", keysym);
	 printf("angle : x %.1f  y %.1f  z %.1f ;  translate : x %.1f  y %.1f  z %.1f   zoom: %lf\n",
		xAngle,	yAngle, zAngle, x_translate, y_translate, z_translate, Zoom);
       }
       break;
     
     case ButtonPress:
       recalcModel = GL_TRUE;
       switch (event.xbutton.button)
	 {
	 case 1: xAngle += 10.0; break;
	 case 2: yAngle += 10.0; break;
	 case 3: zAngle += 10.0; break;
	 }
       break;
     case ConfigureNotify:
       glViewport(0, 0, event.xconfigure.width, event.xconfigure.height);
       /* fall through... */
     case Expose:
       needRedraw = GL_TRUE;
       break;
     }
   }

   /* loop to compress events */;
   if (recalcModel) {
     recalcModel = GL_FALSE;
     needRedraw = GL_TRUE;
   }

   if ((Option[16]) && (save_movie_frame)) {
     save_movie = TRUE;
   }

   if (needRedraw || save_frame) {
     // Redraw();
     StartDrawing();
     update_func(0);
     DoneDrawing();
     save_movie = FALSE;
  
     inc_time = FALSE;
     needRedraw = GL_FALSE;
     Draw_arg[0] = FALSE;
     Draw_arg[1] = FALSE;
     Draw_arg[2] = FALSE;
     Draw_arg[3] = FALSE;
   }

   return ret_val;
}



int Graphic :: RunViewer(int time, void (*update_func)(int))
{
  int ret_val = 0;

  needRedraw = GL_TRUE;

  if (time <= 0) {
    do {
      ret_val = MainLoop(update_func);
    } while (ret_val == -1);

  } else {
    for (int i=0; i<time; i++) {
      ret_val = MainLoop(update_func);
      if (ret_val == -2)
	break;
    }
  }
  return ret_val;
}



void Graphic :: draw_bw(int *selected, Deme *deme)
{
  int i, j;

  glXMakeCurrent(dpy, win, cx);
  //  XMapWindow(dpy, win);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glMatrixMode(GL_MODELVIEW);

  glLoadIdentity();
  glScaled(0.2, 0.2, 0.2);

  for (i=0; i<4; i++) {
    for (j=0; j<4; j++) {
      glPushMatrix();
      glTranslated(143.0*j,143.0*i, 0.0);
      if (selected[4*i+j]) {
	// draw blue square around individual.
	glColor3dv(col_blue);
      } else {
	// draw red square around individual.
	glColor3dv(col_red);
      }
      glLineWidth(3.0);
      glBegin(GL_LINE_STRIP);
      glVertex3d(-70.0, -45.0, 0.0);
      glVertex3d(70.0, -45.0, 0.0);
      glVertex3d(70.0, 95.0, 0.0);
      glVertex3d(-70.0, 95.0, 0.0);
      glVertex3d(-70.0, -45.0, 0.0);
      glEnd();
      glLineWidth(1.0);

      RotX(xAngle + i*15.0);
      RotY(yAngle + 15.0 - 4.0*j);
      RotZ(zAngle);

      glTranslated(x_translate, y_translate, z_translate);

      glScaled(Zoom, Zoom, Zoom);
      deme->display(4*i+j);
      glPopMatrix();
    }
  }

  Draw_arg[1] = FALSE;
  DoneDrawing();
}


void Graphic :: BlindWatchMaker(Deme *deme)
{
  int i, i_loc, j_loc;
  int evolve_type = 2;
  int selected[16];

  bool escape = FALSE;
  char buffer[1];
  int rc;
  int ret_val = -1;
  KeySym keysym;
  XEvent event;
  char fname_saved[300], fname_best[300];

  Zoom = 0.88;
  for (i=0; i<16; i++)
    selected[i] = FALSE;
  
  if (Config.filename_stem[0] != 0) {
    sprintf(fname_saved, "%s_saved.ind", Config.filename_stem);
    sprintf(fname_best, "%s_best.ind", Config.filename_stem);
  } else {
    sprintf(fname_saved, "saved.ind");
    sprintf(fname_best, "best.ind");
  }

  while (!escape) {
    /*** (9) dispatch X events ***/
    while (XPending(dpy)) {
      XNextEvent(dpy, &event);
      switch (event.type) {
      case KeyRelease:
	rc = XLookupString((XKeyEvent *) &event, buffer,
			   1, &keysym, NULL);
	switch (keysym) {
	case XK_Escape:
	  escape = TRUE;
	  ret_val = -2;
	  break;
	  
       case 65434:
	 z_translate += 5.0;
	 recalcModel = GL_TRUE;
	 break;

       case 65436:
	 z_translate += 5.0;
	 recalcModel = GL_TRUE;
	 break;

       case 65431:
	 y_translate += 5.0;
	 recalcModel = GL_TRUE;
	 break;

       case 65433:
	 y_translate -= 5.0;
	 recalcModel = GL_TRUE;
	 break;

       case 65430:
	 x_translate -= 5.0;
	 recalcModel = GL_TRUE;
	 break;

       case 65432:
	 x_translate += 5.0;
	 recalcModel = GL_TRUE;
	 break;

	case XK_space:
	  Draw_arg[1] = TRUE;
	  ret_val = 0;
	  recalcModel = GL_TRUE;
	  break;
	case XK_Return:
	  ret_val = 1;
	  //	 recalcModel = GL_TRUE;
	  break;
	case XK_Down:
	  RotateX(1.0);
	  recalcModel = GL_TRUE;
	  break;
	case XK_Up:
	  RotateX(-1.0);
	  recalcModel = GL_TRUE;
	  break;
	case XK_Right:
	  RotateY(1.0);
	  recalcModel = GL_TRUE;
	  break;
	case XK_Left:
	  RotateY(-1.0);
	  recalcModel = GL_TRUE;
	  break;

	case XK_1:
	  printf("Parents are replaces.\n");
	  evolve_type = 1;
	  break;

	case XK_2:
	  printf("Parents are not replaced.\n");
	  evolve_type = 2;
	  break;

	case XK_u:
	  deme->swap_populations();
	  needRedraw = GL_TRUE;
	  break;

	case XK_a:
	  /*	       world_bend(1.0); */
	  //	 needRedraw = GL_TRUE;
	  break;

	case XK_m:
	  // Mutate first selected.
	  for (i_loc=0; i_loc<16; i_loc++) {
	    if (selected[i_loc])
	      break;
	  }
	  for (j_loc = i_loc+1; j_loc<16; j_loc++) {
	    if (!selected[j_loc])
	      break;
	  }
	  if ((j_loc < 16) && (i_loc < 16)) {
	    deme->mutate_bw(i_loc, j_loc);
	  }
	  
	  needRedraw = GL_TRUE;
	  break;

	case XK_s:
	  for (i=0; i<16; i++) {
	    if (selected[i]) {
	      if (deme->write_individ(i, fname_saved)) {
		printf("Individual #%d is saved as '%s'.\n", i, fname_saved);
	      } else {
		printf("Error writing.  Individual not saved.\n");
	      }
	      break;
	    }
	  }
	  break;

	case XK_q:
	  RotateZ(1.0);
	  recalcModel = GL_TRUE;
	  break;
	  /*	       world_twist(1.0); */
	  //	 needRedraw = GL_TRUE;
	  break;
	case XK_w:
	  RotateZ(-1.0);
	  recalcModel = GL_TRUE;
	  break;

	case XK_p:
	  save_frame = TRUE;
	  break;

	case XK_z:
	  Zoom -= 0.02;
	  needRedraw = GL_TRUE;
	  break;

	case XK_x:
	  Zoom += 0.02;
	  needRedraw = GL_TRUE;
	  break;


	default:
	  //	  printf("pressed: %ld\n", keysym);
	  //	  printf("x %.1f  y %.1f  z %.1f\n", xAngle,
	  //		 yAngle, zAngle);
	  break;
	}
	break;
	
      case ButtonPress:
	needRedraw = GL_TRUE;
	switch (event.xbutton.button) {
	case 1:
	  // Top Left = 0:0.
	  //	  printf("x:%d, y:%d\n", event.xbutton.x,
	  //		 event.xbutton.y);
	  i_loc = 3-event.xbutton.y / 250;
	  j_loc = event.xbutton.x / 250;
	  if (selected[4*i_loc+j_loc])
	    selected[4*i_loc+j_loc] = FALSE;
	  else
	    selected[4*i_loc+j_loc] = TRUE;
	  break;

	case 2:
	  break;

	case 3: 
	  for (i=0; i<16; i++) {
	    if (selected[i]) {
	      deme->write_individ(i, fname_best);
	      break;
	    }
	  }
	  deme->make_new_generation(evolve_type, selected);
	  if (evolve_type == 1) {
	    for (i=0; i<16; i++)
	      selected[i] = FALSE;
	  }
	  break;
	}
	break;

      case ConfigureNotify:
	glViewport(0, 0, event.xconfigure.width, event.xconfigure.height);
	/* fall through... */
      case Expose:
	needRedraw = GL_TRUE;
	break;
      }
    }
    
    /* loop to compress events */;
    if (recalcModel) {
      recalcModel = GL_FALSE;
      needRedraw = GL_TRUE;
    }
    if (needRedraw || save_frame) {
      // Redraw();
      draw_bw(selected, deme);
      needRedraw = GL_FALSE;
    }
  }
}
