/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#ifndef GRAPHIC_HEADER_FILE
#define GRAPHIC_HEADER_FILE

#include <GL/gl.h>
#include <GL/glx.h>             /* this includes the necessary X headers */

#ifndef GENERAL_HEADER_FILE
#include "general.h"
#endif

class Deme;

#define BOX_PTS 8
#define SPHERE_PTS 38

#define BLACK 0
#define WHITE 1
#define RED 2
#define BLUE 3
#define GREEN 4

#define box_color1(b) b->color1
#define box_color2(b) b->color2
#define box_3color(b, i) b->color1[i]
#define box_coord(box, i, j) box->vertex[i][j]
#define box_pt(box, i) box->vertex[i]


extern double col_black[3], col_white[3];
extern double col_blue[3], col_green[3], col_red[3];
extern double col_grey1[3], col_grey2[3], col_grey3[3], col_grey4[3], col_grey5[3], col_grey6[3], col_grey10[3], col_grey11[3];
extern double col_blue2[3], col_blue3[3], col_blue2_ref[3], col_blue3_ref[3];
extern double col_green2[3], col_green3[3], col_green4[3], col_green5[3];
extern double col_green6[3];

extern double col_stripe1[3], col_stripe2[3];
extern double col_side1[3], col_side2[3], col_side3[3];
extern double col_back_grnd[3], col_surf[3], col_surf2[3], col_surf3[3];
extern double  col_surf4[3], col_surf5[3], col_surf_print[3];



// The vertices define a box.
// 0  -X -Y -Z
// 1  -X -Y +Z
// 2  -X +Y -Z
// 3  -X +Y +Z
// 4  +X +Y -Z
// 5  +X +Y +Z
// 6  +X -Y -Z
// 7  +X -Y +Z
typedef struct tBox
{
  double *color1;
  double *color2;
  double vertex[8][3];
} *pBox;


int hsv2rgb(double h, double s, double v, double *r, 
	    double *g, double *b);


class Graphic {
public:
  int xsize;  // Window xsize
  int ysize;  // Window ysize

  int init_sphere;
  unsigned char *pixels, *pixels2;
  int frame_num;
  int save_movie;
  int save_movie_frame;
  int save_frame;

  int snglBuf[4]; // = {GLX_RGBA, GLX_DEPTH_SIZE, 16, None};
  int dblBuf[5]; // = {GLX_RGBA, GLX_DEPTH_SIZE, 16, GLX_DOUBLEBUFFER, None};

  pBox            box;
  Display        *dpy;
  Window          win;
  GLboolean       doubleBuffer;
  GLXContext      cx;

  // Tranlsation
  GLdouble x_translate;
  GLdouble y_translate;
  GLdouble z_translate;

  GLdouble Zoom;
  // Angles of rotation about the coordinate axis
  GLdouble xAngle;  // Scene rotation about the X axis
  GLdouble yAngle;  // Scene rotation about the Y axis
  GLdouble zAngle;  // Scene rotation about the Z axis
  
  GLboolean needRedraw;
  GLboolean recalcModel;


  Graphic(int x = 300, int y = 300, char title[30] = "Evolutionary Robotics");
  void Initialize(double xrot, double yrot, double xdist, double ydist, double zdist);
  void Initialize(double xrot, double yrot, double zrot, double xdist, double ydist, double zdist);
  void SetClearColor(Real r, Real g, Real b, Real a);

  void RotX(double angle);
  void RotY(double angle);
  void RotZ(double angle);
  void RotateX(double angle);
  void RotateY(double angle);
  void RotateZ(double angle);

  void rotate(double angle, double *vec);
  void translate(double x, double y, double z);
  void push_orientation(void);
  void pop_orientation(void);
  void set_transformation(double matrix[4][4]);

  void ScalePrism(pBox box, double val);
  pBox MakeBox(double scale);


  void drawPlaneY(double x, double y, double z);
  void drawPlaneZ(double x, double y, double z);
  void drawPrism(pBox box);
  void drawPrismOL(pBox box);
  void drawPrism(pBox box, double *colorX, double *colorY, double *colorZ);
  void drawPrismOL(pBox box, double *colorX, double *colorY, double *colorZ);
  void drawPoint(double x, double y, double z);
  void SetLineWidth(double w);
  void startLine(void);
  void startPolygon(void);
  void addVertex(double x, double y);
  void addVertex(double x, double y, double z);
  void addVertex3v(double vec[3]);
  void starLocation(double x, double y);
  void starLocation(double x, double y, double z);
  void starLocationBig(double x, double y, double z);
  void starShadowLocation(double x, double y, double z);
  void starShadowLocationBig(double x, double y, double z);
  void stopLine(void);
  void stopList(void);
  void drawLine(double x1, double y1, double x2, double y2);
  void drawJoint(void);
  void drawBox(double *dim, double *color1, double *color2);
  void drawBoxOL(double *dim, double *color1, double *color2);
  void drawBoxOL(double *dim, double *colorX, double *colorY, double *colorZ);
  void SetDefaultBox(double *dim, double *colorX, double *colorY, double *colorZ);
  void drawDefaultBox(void);
  void drawRect(double x1, double y1, double x2, double y2);
  void drawSphere(double radius, double col1[3], double col2[3]);
  void drawCylinder(double radius, double col1[3], double col2[3]);

  void SetColor(int col);
  void SetColor(double col[3]);
  void Reshape(GLsizei w, GLsizei h);
  void Redraw(void (*update_func)(int), int param);
  void WriteWindow(char *name);
  void StartDrawing(void);
  void StartDrawing(int clear);
  void DoneDrawing(void);
  int MainLoop(void (*update_func)(int));
  int RunViewer(int time, void (*update_func)(int));

  void draw_bw(int *selected, Deme *deme);
  void BlindWatchMaker(Deme *deme);
};

#endif
