/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stream.h>
#include <iostream.h>
#include <fstream.h>
#include "general.h"
#include "lsystem.h"
#include "ind_lsys.h"
#include "prod_body.h"

#define IND_VERSION 8

extern int Max_joints; // from bars2d.cc

extern Real evaluate_ind(int init, int size, Individual *ind);

int Num_arguments = MAX_PROD_VARS;
int Num_constants = MAX_CONSTANTS;
int Num_functions = MAX_FUNCTIONS;
int Num_productions = NUM_PRODUCTIONS;
int Max_bodies = MAX_PROD_BODIES;
int Num_operations = 10;

int Num_processed;

/* Notes:
  If we allow branches to modify parameters like productions,
     eg [abcdfs](n0+2, n1*2, ...)
  then we get a heirarchy of modules.
  This is useful if we want the l-system to generate a set of
  build rules instead of a genotype.
 */



#define MAX_PROCESSED 100000
//char lstring[20100];
#define BONUS_LENGTH 110
LString Lstring[MAX_STRING_LENGTH + BONUS_LENGTH];
Tape Tape1[MAX_STRING_LENGTH + BONUS_LENGTH];
Tape Tape2[MAX_STRING_LENGTH + BONUS_LENGTH];
int Created_tape = FALSE;

int process_rule(Real *var, Tape *tape_dest, Prod_Head phead);
int tape_2_string(char end_type, char end_symbol,
		  int *t, Tape *tape, int *s, char *string);
Real table_fitness(Real p_height, Real p_surf, Individual *ind);



void print_tape_simple(Tape *tape)
{
  int i = 0;

  cout << "-t: ";
  while ((tape[i].type != PBC_TYPE_NONE) &&
	 (tape[i].symbol != TAPE_END)) {
    if (tape[i].type == PBC_TYPE_TERMINAL) {
      printf("%c(%.1f) ", tape[i].symbol,
	     tape[i].arg[0]);

    } else if (tape[i].type == PBC_TYPE_BLOCK) {
      if (tape[i].symbol == SYMBOL_BLK_END) {
	printf("}(%d) ", int(tape[i].arg[0]));
      } else if (tape[i].symbol == SYMBOL_POP) {
	printf("] ");
      } else {
	printf("%c ", tape[i].symbol);
      }

    } else if (tape[i].type == PBC_TYPE_PRODUCTION) {
      printf("P%d(", tape[i].symbol);
      for (int j=0; j<Num_arguments; j++) {
	printf("%d", int(tape[i].arg[j]));
	if (j < Num_arguments-1)
	  printf(",");
      }
      printf(") ");
    } else {
      printf("? ");
    }

    i++;
    if (!(i%20))
      printf("\n  ");

    if (i>=MAX_STRING_LENGTH)
      break;
  }

  cout << "  : " << i << endl;
}

void print_tape(Tape *tape)
{
  int i = 0;

  cout << "-t: ";
  while ((tape[i].type != PBC_TYPE_NONE) &&
	 (tape[i].symbol != TAPE_END)) {
    if (tape[i].type == PBC_TYPE_TERMINAL) {
      printf("%c(%.1lf) ", tape[i].symbol,
	     tape[i].arg[0]);

    } else if (tape[i].type == PBC_TYPE_BLOCK) {
      if (tape[i].symbol == SYMBOL_BLK_END) {
	printf("}(%.1lf) ", tape[i].arg[0]);
      } else if (tape[i].symbol == SYMBOL_POP) {
	printf("] ");
      } else {
	printf("%c", tape[i].symbol);
      }

    } else if (tape[i].type == PBC_TYPE_PRODUCTION) {
      printf("P%d(", tape[i].symbol);
      for (int j=0; j<Num_arguments; j++) {
	printf("%.1lf", tape[i].arg[j]);
	if (j < Num_arguments-1)
	  printf(", ");
      }
      printf(") ");
    } else {
      printf("? ");
    }

    i++;
    if (!(i%20))
      printf("\n  ");

    if (i>=MAX_STRING_LENGTH)
      break;
  }

  cout << "  : " << i << endl;
}

void print_string_detailed(LString *string)
{
  int i = 0;

  cout << "-s: ";
  while (string[i].symbol != 0) {
    if (string[i].symbol == SYMBOL_PUSH)
      printf("[");
    else if (string[i].symbol == SYMBOL_POP)
      printf("] ");
    else {
      printf("%c(%.1lf) ", string[i].symbol,
	     string[i].arg);
    }

    i++;
    if (!(i%20))
      printf("\n  ");

    if (i>=MAX_STRING_LENGTH)
      break;
  }  
  printf(" : %d (%d)\n", i, Num_processed);
}   

void print_string_english(LString *string)
{
  int i = 0;

  while (string[i].symbol != 0) {
    if (string[i].symbol == SYMBOL_PUSH) {
      printf("[");
    } else if (string[i].symbol == SYMBOL_POP) {
      printf("] ");
    } else if (string[i].symbol == SYMBOL_PUSH_EDGE) {
      printf("(");
    } else if (string[i].symbol == SYMBOL_POP_EDGE) {
      printf(") ");
    } else if (string[i].symbol == SYMBOL_BLK_START) {
      printf("\\{ ");
    } else if (string[i].symbol == SYMBOL_BLK_END) {
      printf("\\}(%d) ", (int)string[i].arg);
    } else if (string[i].symbol == TERM_FORWARD) {
      if (Terminal_set <= 6) {
	printf("forward(%.1lf) ", string[i].arg);
      } else {
	printf("forward ");
      }

    } else if (string[i].symbol == TERM_BACKWARD) {
      if (Terminal_set <= 6) {
	printf("backward(%.1lf) ", string[i].arg);
      } else {
	printf("backward ");
      }
    } else if (string[i].symbol == TERM_TURN_UP) {
      printf("up(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_TURN_DOWN) {
      printf("down(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_TURN_LEFT) {
      printf("left(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_TURN_RIGHT) {
      printf("right(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_TURN_CLOCKWISE) {
      printf("clockwise(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_TURN_CNTR_CLOCKWISE) {
      printf("counter-clockwise(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_OFFSET_INCREASE) {
      printf("increase-offset(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_OFFSET_DECREASE) {
      printf("decrease-offset(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_JOINT_REVOLUTE) {
      printf("revolute1(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_JOINT_REVOLUTE2) {
      printf("revolute2(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_JOINT_TWIST) {
      printf("twist90(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_JOINT_TWIST180) {
      printf("twist180(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_ADD_OUTPUT) {
      printf("add-output(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_ADD_INPUT) {
      printf("add-input(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_SPLIT_LINK) {
      printf("split(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_DUPLICATE_LINK) {
      printf("duplicate(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_MAKE_LOOP) {
      printf("loop(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_MERGE_NEURONS) {
      printf("merge(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_NEXT_LINK) {
      printf("next(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_PARENT_NEURON) {
      printf("parent(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_REVERSE_LINK) {
      printf("reverse(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_INCREASE_WEIGHT) {
      printf("increase-weight(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_DECREASE_WEIGHT) {
      printf("decrease-weight(%.1lf) ", string[i].arg);
    } else if (string[i].symbol == TERM_SET_OUTPUT_FUNCTION) {
      printf("set-function(%.1lf) ", string[i].arg);

    } else {
      printf("!%c(%.1lf)! ", string[i].symbol,
	     string[i].arg);
    }
    i++;
    if (!(i%5))
      printf("\n");

    if (i>=MAX_STRING_LENGTH)
      break;
  }  
  printf("\n%d (%d)\n", i, Num_processed);
}   

void print_tape_english(Tape *tape)
{
  int i = 0, j;

  printf("\nprint_tape_english() :: \n");

  while ((tape[i].type != PBC_TYPE_NONE) || (tape[i].symbol != TAPE_END)) {
    if (tape[i].type == PBC_TYPE_PRODUCTION) {
      printf("P%d", tape[i].symbol);
      if (Num_arguments == 0) {
	printf(" ");
      } else {
	printf("(");
	for (j=0; j<Num_arguments; j++) {
	  printf("%.2f", (float)tape[i].arg[j]);
	  if (j < Num_arguments-1)
	    printf(",");
	}
	printf(") ");
      }

    } else if (tape[i].symbol == SYMBOL_PUSH) {
      printf("[");
    } else if (tape[i].symbol == SYMBOL_POP) {
      printf("] ");
    } else if (tape[i].symbol == SYMBOL_PUSH_EDGE) {
      printf("(");
    } else if (tape[i].symbol == SYMBOL_POP_EDGE) {
      printf(") ");
    } else if (tape[i].symbol == SYMBOL_BLK_START) {
      printf("\\{ ");
    } else if (tape[i].symbol == SYMBOL_BLK_END) {
      printf("\\}(%d) ", (int)tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_FORWARD) {
      if (Terminal_set <= 6) {
	printf("forward(%.1lf) ", tape[i].arg[0]);
      } else {
	printf("forward ");
      }

    } else if (tape[i].symbol == TERM_BACKWARD) {
      if (Terminal_set <= 6) {
	printf("backward(%.1lf) ", tape[i].arg[0]);
      } else {
	printf("backward ");
      }
    } else if (tape[i].symbol == TERM_TURN_UP) {
      printf("up(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_TURN_DOWN) {
      printf("down(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_TURN_LEFT) {
      printf("left(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_TURN_RIGHT) {
      printf("right(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_TURN_CLOCKWISE) {
      printf("clockwise(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_TURN_CNTR_CLOCKWISE) {
      printf("counter-clockwise(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_OFFSET_INCREASE) {
      printf("increase-offset(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_OFFSET_DECREASE) {
      printf("decrease-offset(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_JOINT_REVOLUTE) {
      printf("revolute1(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_JOINT_REVOLUTE2) {
      printf("revolute2(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_JOINT_TWIST) {
      printf("twist90(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_JOINT_TWIST180) {
      printf("twist180(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_ADD_OUTPUT) {
      printf("add-output(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_ADD_INPUT) {
      printf("add-input(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_SPLIT_LINK) {
      printf("split(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_DUPLICATE_LINK) {
      printf("duplicate(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_MAKE_LOOP) {
      printf("loop(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_MERGE_NEURONS) {
      printf("merge(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_NEXT_LINK) {
      printf("next(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_PARENT_NEURON) {
      printf("parent(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_REVERSE_LINK) {
      printf("reverse(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_INCREASE_WEIGHT) {
      printf("increase-weight(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_DECREASE_WEIGHT) {
      printf("decrease-weight(%.1lf) ", tape[i].arg[0]);
    } else if (tape[i].symbol == TERM_SET_OUTPUT_FUNCTION) {
      printf("set-function(%.1lf) ", tape[i].arg[0]);

    } else {
      printf("!%c:%d(%.1lf)! ", tape[i].symbol, tape[i].type,
	     tape[i].arg[0]);
    }
    i++;
    /*
    if (!(i%5))
      printf("\n");
    */
    if (i>=MAX_STRING_LENGTH)
      break;
  }  
  printf("\n   :: %d (%d)\n\n", i, Num_processed);
}   

void print_string(LString *string)
{
  int i = 0, ctr=0;

  cout << "-s: ";
  while (string[i].symbol != 0) {
    if (string[i].symbol == SYMBOL_PUSH) {
      ctr++;
      printf("[");
    } else if (string[i].symbol == SYMBOL_POP) {
      ctr++;
      printf("]");
    } else {
      ctr++;
      printf("%c", string[i].symbol);
    }

    i++;
    if (!(ctr%30))
      printf("\n  ");

    if (i>=MAX_STRING_LENGTH)
      break;
  }  
  printf(" : %d (%d)\n", i, Num_processed);
}   

void print_string_old(LString *string)
{
  int i = 0, ctr=0;

  cout << "-s: ";
  while (string[i].symbol != 0) {
    if (string[i].symbol == SYMBOL_PUSH) {
      ctr++;
      printf("[");
    } else if (string[i].symbol == SYMBOL_POP) {
      ctr++;
      printf("]");
    } else {
      for (int j=0; j<int(string[i].arg); j++) {
	printf("%c", string[i].symbol);
	ctr++;
	if (!(ctr%30))
	  printf("\n  ");
      }
      i++;
      if (i>=MAX_STRING_LENGTH)
	break;
      continue;
    }

    i++;
    if (!(ctr%30))
      printf("\n  ");

    if (i>=MAX_STRING_LENGTH)
      break;
  }  
  printf(" : %d (%d)\n", i, Num_processed);
}   


void clear_string(LString *string)
{
  for (int i=0; i<MAX_STRING_LENGTH; i++) {
    string[i].symbol = 0;
    string[i].arg = 0.0;
  }
}

void clear_tape(Tape *tape)
{
  for (int i=0; i<MAX_STRING_LENGTH; i++) {
    tape[i].type = PBC_TYPE_NONE;
    tape[i].symbol = 0;
    for (int j=0; j<Num_arguments; j++)
      tape[i].arg[j] = 0.0;
  }
}


// Returns the number of symbols added.
int process_rule(Real *arg, Tape *tape_dest, Prod_Head phead)
{
  int i;

  // Statistics stuff.
  phead->used++;

  for (i=0; i<Num_arguments; i++) {
    if (phead->used == 1) {
      phead->min[i] = arg[i];
      phead->max[i] = arg[i];
    } else {
      if (arg[i] < phead->min[i])
	phead->min[i] = arg[i];
      else if (arg[i] > phead->max[i])
	phead->max[i] = arg[i];
    }
  }


  // Process the rule.
  for (i=0; i<phead->num_bodies; i++) {
    if (phead->body[i]->test_cond(arg)) {
      if (Print[3])
	printf("%d:", i);
      return phead->body[i]->write_body(arg, tape_dest);
    }
  }
  return 0;
}



//////////////////////////////////////////




Individual :: Individual(void)
{
  init_hist(&hist_self);
  init_hist(&hist_parent1);
  init_hist(&hist_parent2);
  created = FALSE;
  evaluated = FALSE;
  iterations = 4;
  version = IND_VERSION;
}

Individual :: Individual(Individual *sample)
{
  init_hist(&hist_self);
  init_hist(&hist_parent1);
  init_hist(&hist_parent2);
  created = FALSE;
  evaluated = FALSE;
  version = IND_VERSION;

  if (sample != 0)
    iterations = sample->iterations;
}

void Individual :: init(int init_h)
{
  int i, j;


  if (!Created_tape) {
    Created_tape = TRUE;

    int num_args = Num_arguments;
    if (num_args == 0)
      num_args = 1;
    for (i=0; i<MAX_STRING_LENGTH+BONUS_LENGTH; i++) {
      Tape1[i].arg = (Real*)malloc(num_args * sizeof(Real));
      if (Tape1[i].arg == 0) {
	printf("Ind_lsys :: init() malloc for tape 1 failed.\n");
	return;
      }
	
      Tape2[i].arg = (Real*)malloc(num_args * sizeof(Real));
      if (Tape2[i].arg == 0) {
	printf("Ind_lsys :: init() malloc for tape 1 failed.\n");
	return;
      }
    }
  }

  if (!created) {
    created = TRUE;
    length = -1;
    if (Option[10]) {
      lstring = (LString*)malloc(MAX_STRING_LENGTH*sizeof(LString));
      if (lstring == 0) {
	printf("Ind_lsys :: init() malloc for lstring failed.\n");
	return;
      }

    } else {
      //      lstring = &(Lstring[0]);      
      lstring = (LString*)malloc(MAX_STRING_LENGTH*sizeof(LString));

    }
    if (Option[10])
      model.model_id = -1;
    else
      model.model_id = 0;
    model.lstring = lstring;
    if (Num_fit_dims == 1) {
      model.fv_length = 1;
    } else if (Num_fit_dims == 4) {
      model.fv_length = 4;
    } else {
      model.fv_length = Num_fit_dims;
    }
    model.fitness = (Real*)malloc(Num_fit_dims * sizeof(Real));
    model.build_args = (Real*)malloc(MAX_CONSTANTS * sizeof(Real));
    model.offset[0] = 0.0;
    model.offset[1] = 0.0;
    model.offset[2] = 0.0;

    rules = (Prod_Head*)malloc(Num_productions * sizeof(Prod_Head));
    for (i=0; i<Num_productions; i++) {
      rules[i] = (Prod_Head)malloc(sizeof(Production_Head));
      if (rules[i] == 0) {
	printf("Ind_lsys :: init() malloc for rules[%d] failed.\n", i);
	return;
      }

      rules[i]->max = (Real*)malloc(Num_arguments * sizeof(Real));
      if (rules[i]->max == 0) {
	printf("Ind_lsys :: init() malloc for rules[%d]->max failed.\n", i);
	return;
      }

      rules[i]->min = (Real*)malloc(Num_arguments * sizeof(Real));
      if (rules[i]->min == 0) {
	printf("Ind_lsys :: init() malloc for rules[%d]->min failed.\n", i);
	return;
      }

      rules[i]->body = (Prod_Body**)malloc(Max_bodies * sizeof(Prod_Body*));
      if (rules[i]->body == 0) {
	printf("Ind_lsys :: init() malloc for rules[%d]->body failed.\n", i);
	return;
      }

      for (j=0; j<Max_bodies; j++) {
	rules[i]->body[j] = new Prod_Body();
	rules[i]->body[j]->phead = rules[i];
	rules[i]->body[j]->constant = &(constant[0]);
	rules[i]->body[j]->parent = this;
      }
    }
  }

  if (init_h) {
    init_hist(&hist_self);
    init_hist(&hist_parent1);
    init_hist(&hist_parent2);
  }
  created = TRUE;
}


void Individual :: init_pheno(int init_h)
{
  int i, j;


  if (!Created_tape) {
    Created_tape = TRUE;

    int num_args = Num_arguments;
    if (num_args == 0)
      num_args = 1;
    for (i=0; i<MAX_STRING_LENGTH+BONUS_LENGTH; i++) {
      Tape1[i].arg = (Real*)malloc(num_args * sizeof(Real));
      if (Tape1[i].arg == 0) {
	printf("Ind_lsys :: init() malloc for tape 1 failed.\n");
	return;
      }
	
      Tape2[i].arg = (Real*)malloc(num_args * sizeof(Real));
      if (Tape2[i].arg == 0) {
	printf("Ind_lsys :: init() malloc for tape 1 failed.\n");
	return;
      }
    }
  }

  if (!created) {
    created = TRUE;
    length = -1;
    if (Option[10]) {
      lstring = (LString*)malloc((MAX_STRING_LENGTH+BONUS_LENGTH)*sizeof(LString));
      if (lstring == 0) {
	printf("Ind_lsys :: init() malloc for lstring failed.\n");
	return;
      }

    } else {
      //      lstring = &(Lstring[0]);      
      lstring = (LString*)malloc((MAX_STRING_LENGTH+BONUS_LENGTH)*sizeof(LString));

    }
    if (Option[10])
      model.model_id = -1;
    else
      model.model_id = 0;
    model.lstring = lstring;
    if (Num_fit_dims == 1) {
      model.fv_length = 1;
    } else {
      //      model.fv_length = 2;
      model.fv_length = Num_fit_dims;
    }
    model.fitness = (Real*)malloc(Num_fit_dims * sizeof(Real));
    model.build_args = (Real*)malloc(MAX_CONSTANTS * sizeof(Real));
    model.offset[0] = 0.0;
    model.offset[1] = 0.0;
    model.offset[2] = 0.0;

    rules = (Prod_Head*)malloc(1 * sizeof(Prod_Head));
    for (i=0; i<1; i++) {
      rules[i] = (Prod_Head)malloc(sizeof(Production_Head));
      if (rules[i] == 0) {
	printf("Ind_lsys :: init() malloc for rules[%d] failed.\n", i);
	return;
      }

      rules[i]->max = (Real*)malloc(Num_arguments * sizeof(Real));
      if (rules[i]->max == 0) {
	printf("Ind_lsys :: init() malloc for rules[%d]->max failed.\n", i);
	return;
      }

      rules[i]->min = (Real*)malloc(Num_arguments * sizeof(Real));
      if (rules[i]->min == 0) {
	printf("Ind_lsys :: init() malloc for rules[%d]->min failed.\n", i);
	return;
      }

      rules[i]->body = (Prod_Body**)malloc(1 * sizeof(Prod_Body*));
      if (rules[i]->body == 0) {
	printf("Ind_lsys :: init() malloc for rules[%d]->body failed.\n", i);
	return;
      }

      
      int tmp = Prod_length;
      Prod_length = MAX_STRING_LENGTH+BONUS_LENGTH;
      for (j=0; j<1; j++) {
	rules[i]->body[j] = new Prod_Body();
	rules[i]->body[j]->phead = rules[i];
	rules[i]->body[j]->constant = &(constant[0]);
	rules[i]->body[j]->parent = this;
      }
      Prod_length = tmp;
    }
  }

  if (init_h) {
    init_hist(&hist_self);
    init_hist(&hist_parent1);
    init_hist(&hist_parent2);
  }
  created = TRUE;
}


void Individual :: init(void)
{
  init(TRUE);
}


void Individual :: init_hist(history *hist)
{
  int i;

  hist->index = -1;
  hist->how_created = CREATE_INIT;
  hist->num_mutate = 0;
  hist->num_recombine = 0;
  hist->delta = 0.0;

  for (i=0; i<Num_fit_dims; i++) {
    hist->fitness[i] = 0.0;
    hist->fitness_prev[i] = 0.0;
  }

  for (i=0; i<Num_operations; i++)
    hist->op_count[i] = 0;
}

void Individual :: init_hist(void)
{
  int i;

  hist_self.index = -1;
  hist_self.how_created = CREATE_INIT;
  hist_self.num_mutate = 0;
  hist_self.num_recombine = 0;
  hist_self.delta = 0.0;
  hist_self.fitness_prev[0] = 0.0;

  for (i=0; i<Num_operations; i++)
    hist_self.op_count[i] = 0;
}


// Repairs a rule by increasing (by 1) the number
// of production bodies that are used.
// Adjusts truth conditions in doing so.
// Returns -1 if rule is not used.
// Returns -2 if all bodies are used, and may
//   modify a truth condition of one of the bodies.
int Individual :: repair_rule(Prod_Head ph)
{
  int b, num;
  int b_unused, b_used;
  int unused[Max_bodies];

  if (ph->used == 0)
    return -1;

#ifndef OPTIMIZE2
  op_used(OP_UNKNOWN);
#endif
#ifndef OPTIMIZE
  if (Print[9])
    printf("Repair rule\n");
#endif

  // First count the number of bodies used (to see if
  // at least one is used).
  num = 0;
  for (b=0; b<ph->num_bodies; b++) {
    if (ph->body[b]->used > 0)
      num++;
  }

  if (num == 0) {
    // If no body is used, set condition of last one
    // to always True.
    num = ph->num_bodies - 1;
    if (num == -1) {
      num = 0;
      ph->num_bodies = 1;
      ph->body[0]->initialize();
    }

    ph->body[num]->set_cond_true();
    ph->body[num]->delete_string();
    return num;
  }    


  // Find un-used bodies.
  num = 0;
  for (b=0; b<ph->num_bodies; b++) {
    if (ph->body[b]->used == 0)
      unused[num++] = b;
  }

  if (num == 0) {
    // All bodies are used.
    // May want to adjust truth conditions.
#ifndef OPTIMIZE
    if (Print[9])
      printf(" fail - all bodies are used\n");
#endif
    return -2;
  }

  // Select an unused body to repair.
  b_unused = unused[random_int(num)];
  if (ph->body[b_unused]->tested == 0) {
    // Never tested, decrease index to a used body.
    b_used = b_unused--;
    while (ph->body[b_used]->used == 0)
      b_used--;

  } else {      
    if (random_int(2) ||
	((b_unused == 0) && (b_unused < ph->num_bodies-1))) {
      // Increase index to a used body.
      b_used = b_unused + 1;
      while ((ph->body[b_used]->used == 0) &&
	     (b_used < ph->num_bodies))
	b_used++;

    } else {
      // Decrease index to a used body.
      b_used = b_unused - 1;
      while ((ph->body[b_used]->used == 0) &&
	     (b_used > 0))
	b_used--;
      if (ph->body[b_used]->used == 0) {
	// Actually, not used, so must increase to find used body.
	// Increase index to a used body.
	b_used = b_unused + 1;
	while ((ph->body[b_used]->used == 0) &&
	       (b_used < ph->num_bodies))
	  b_used++;
      }
    }
  }

  if ((b_used == ph->num_bodies) &&
      (ph->body[b_used]->used == 0)) {
    // Last body and is not used.
    ph->body[b_used]->set_cond_true();
    ph->body[b_used]->delete_string();
    return b_used;
  }
  
  // Make the unused rule a duplicate of the used one.
  ph->body[b_unused]->duplicate(ph->body[b_used]);

  // Adjust truth conditions.
  if (b_used < b_unused) {
    ph->body[b_used]->make_cond_smaller();
    ph->body[b_unused]->make_cond_larger();

  } else {    
    ph->body[b_unused]->make_cond_smaller();
    ph->body[b_used]->make_cond_larger();
  }

  return b_unused;
}


int Individual :: repair_rule(int i)
{
  return repair_rule(rules[i]);
}


void Individual :: repair_init(void)
{
  int i, j;

  return;

  for (i=0; i<Num_productions; i++) {
    if (rules[i]->used == 0) {
      for (j=0; j<rules[i]->num_bodies; j++)
	rules[i]->body[j]->delete_string();
    }
  }
}



int Individual :: create_mutate(void)
{
  if (hist_self.how_created == CREATE_MUTATE)
    return TRUE;

  return FALSE;
}

// Returns 1 if better fitness than parent(2)
// returns 0 if equal/inbetween parent(s)
// returns -1 if strictly less than parent(s).
int Individual :: better(void)
{
  if (create_mutate()) {
    if (hist_self.fitness[0] > hist_parent1.fitness[0])
      return 1;
    else if (hist_self.fitness[0] == hist_parent1.fitness[0])
      return 0;
    else
      return -1;
  } else {
    if ((hist_self.fitness[0] > hist_parent1.fitness[0]) &&
	(hist_self.fitness[0] > hist_parent2.fitness[0]))
      return 1;
    else if ((hist_self.fitness[0] == hist_parent1.fitness[0]) ||
	     (hist_self.fitness[0] == hist_parent2.fitness[0]))
      return 0;
    else if ((hist_self.fitness[0] > hist_parent1.fitness[0]) ||
	     (hist_self.fitness[0] > hist_parent2.fitness[0]))
      return 0;
    else
      return -1;
  }
}


// Returns TRUE iff this individual is not worse than
// the other individual for all fitnesses.
int Individual :: better(Individual *other)
{
  int i;

  for (i=0; i<Num_fit_dims; i++) {
    if (get_fitness(i) < other->get_fitness(i))
      return FALSE;
  }
  return TRUE;
}

int Individual :: replace(Individual *other)
{
  return better(other);
}


// Returns TRUE if individual fails its evaluation.
int Individual :: fitness_fail(void)
{
  if (Num_fit_dims < 5) {
    return (get_fitness(0) <= 0.0);
  }

  return (get_fitness(0) <= 0.0);

  if (!Option[10]) {
    if (get_fitness(2) < 2) {
      return TRUE;
    }
  }

  return FALSE;
}


Real Individual :: get_fitness(void)
{
  return hist_self.fitness[0];
}

void Individual :: set_fitness(Real val)
{
  hist_self.fitness[0] = val;
}


Real Individual :: get_fitness(int index)
{
  return hist_self.fitness[index];
}

void Individual :: get_fitness(Real *fitness_vector)
{
  for (int i=0; i<Num_fit_dims; i++)
    fitness_vector[i] = hist_self.fitness[i];
}


void Individual :: set_fitness(Real *val)
{
  for (int i=0; i<Num_fit_dims; i++)
    hist_self.fitness[i] = val[i];
}


void Individual :: set_fitness(int index, Real val)
{
  hist_self.fitness[index] = val;
}


void Individual :: add_fitness(Real val)
{
  hist_self.fitness[0] += val;
}

void Individual :: div_fitness(Real val)
{
  hist_self.fitness[0] /= val;
}


void Individual :: set_creation(int type)
{
  hist_self.how_created = type;
}

int Individual :: get_creation(void)
{
  return hist_self.how_created;
}


void Individual :: duplicate_hist(history *hist_from, history *hist_to)
{
  int i;

  hist_to->index = hist_from->index;
  hist_to->operation = hist_from->operation;
  hist_to->how_created = hist_from->how_created;
  hist_to->num_mutate = hist_from->num_mutate;
  hist_to->num_recombine = hist_from->num_recombine;
  hist_to->delta = hist_from->delta;

  hist_to->val1 = hist_from->val1;
  for (i=0; i<Num_operations; i++)
    hist_to->op_count[i] = hist_from->op_count[i];

  for (i=0; i<Num_fit_dims; i++) {
    hist_to->fitness[i] = hist_from->fitness[i];
    hist_to->fitness_prev[i] = hist_from->fitness[i];
  }
}


void Individual :: print_fitness(void)
{
  printf("%6.1f", hist_self.fitness[0]);
}



void Individual :: print_history2(history *hist)
{
  if (hist->how_created == CREATE_INIT)
    cout << "[-:";
  else if (hist->how_created == CREATE_MUTATE)
    cout << "[m:";
  else if (hist->how_created == CREATE_RECOMBINE)
    cout << "[r:";
  else 
    cout << "[?:";

  printf("%3d:%3d]", hist->num_mutate, hist->num_recombine);
}


void Individual :: fprint_history2(history *hist, FILE *log)
{
  if (hist->how_created == CREATE_INIT)
    fprintf(log, "[-:");
  else if (hist->how_created == CREATE_MUTATE)
    fprintf(log, "[m:");
  else if (hist->how_created == CREATE_RECOMBINE)
    fprintf(log, "[r:");
  else 
    fprintf(log, "[?:");

  fprintf(log, "%3d:%3d]", hist->num_mutate, hist->num_recombine);
}

void Individual :: print_history(history *hist)
{
  if (Num_fit_dims == 1) {
    printf("%6.1f ", hist->fitness[0]);
  } else {
    for (int i=0; i<Num_fit_dims; i++) {
      printf("%.0f ", hist->fitness[i]);
    }
  }
  print_history2(hist);
}

void Individual :: print_history_self(void)
{
  print_history(&hist_self);
}

void Individual :: print_history_self2(void)
{
  print_history2(&hist_self);
}

void Individual :: fprint_history_self2(FILE *log)
{
  fprint_history2(&hist_self, log);
}

void Individual :: print_history(void)
{
  int i;

  if (Num_operations == 0)
    return;

  printf("[%3d", hist_self.op_count[0]);

  for (i=1; i<Num_operations; i++) {
    printf(":%3d", hist_self.op_count[i]);
  }
  printf("]");
}

void Individual :: fprint_history(FILE *log)
{
  int i;

  if (Num_operations == 0)
    return;

  fprintf(log, "[%3d", hist_self.op_count[0]);

  for (i=1; i<Num_operations; i++) {
    fprintf(log, ":%3d", hist_self.op_count[i]);
  }
  fprintf(log, "]");
}

void Individual :: print_results(void)
{
  if (hist_self.how_created == CREATE_MUTATE) {
    if (hist_self.fitness[0] > hist_parent1.fitness[0])
      cout << "  Higher!";
    else if (hist_self.fitness[0] == hist_parent1.fitness[0])
      cout << "  Same.";
  } else if (hist_self.how_created == CREATE_RECOMBINE) {
    if ((hist_self.fitness[0] > hist_parent1.fitness[0]) &&
	(hist_self.fitness[0] > hist_parent2.fitness[0]))
      cout << "  Highest!";
    else if ((hist_self.fitness[0] == hist_parent1.fitness[0]) ||
	     (hist_self.fitness[0] == hist_parent2.fitness[0]))
      cout << "  Same.";
    else if ((hist_self.fitness[0] > hist_parent1.fitness[0]) ||
	     (hist_self.fitness[0] > hist_parent2.fitness[0]))
      cout << "  Middle.";
  }
}


void Individual :: print_history_full(void)
{
  Real diff, val;

  if (evaluated == 2) {
    printf("Evl                        ");
    val = get_fitness();
    diff = val - hist_self.fitness_prev[0];
    set_fitness(hist_self.fitness_prev[0]);
    print_history(&hist_self);
    set_fitness(val);
    printf(" -> ");
    printf("%6.1f", val + diff);
    printf(" diff: %6.1f ", 2.0*diff);
    //    print_history();
    return;
  } else if (hist_self.how_created == CREATE_INIT) {
    print_history(&hist_self);
    return;
  } else if (hist_self.how_created == CREATE_MUTATE) {
    cout << "Mut ";
  } else if (hist_self.how_created == CREATE_RECOMBINE) {
    cout << "Rec ";
  }

  //  if (hist_self.operation == OP_INIT)
  //    return;

  if (hist_self.how_created == CREATE_MUTATE) {
    
    printf("            %6.3f   ", hist_self.delta);
    print_history(&hist_parent1);
  } else {
    print_history(&hist_parent1);
    cout << " + ";
    print_history(&hist_parent2);
  }

  cout << " -> ";
  print_history(&hist_self);
  print_history();
  print_results();
}


void Individual :: info(void)
{
  print_history();
}



int Individual :: randomize_rule(int r)
{
  int i, len;


  for (i=0; i<rules[r]->num_bodies; i++) {
    rules[r]->body[i]->make_random_condition();
    len = random_int(3)+1;
    rules[r]->body[i]->make_random_sequence(0, len);
    rules[r]->body[i]->length = len;
  }
  return TRUE;
}


// At most two bodies are usable.
// This allows room to grow.
void Individual :: make_random(void)
{
  int i, j, num;

  if (!created) {
    cout << "Individual not created." << endl;
    init();
  }


  // Randomize the constants.
  // ifdef INTEGER_VALUE
  if (Option[5] < 0) {
    // No productions, hence at most one iteration.
    constant[0].value = 1.0;
    
  } else {
    if (Option[11]) {
      constant[0].value = Real(Option[11] + random_int(5));
    } else {
      constant[0].value = Real(10 + random_int(8));
      //      constant[0].value = Real(15);
    }
  }


  for (i=1; i<Num_arguments+1; i++) {
    constant[i].value = Real(random_int(6))+2.0;
  }


  if (Option[17] == 0) {
    for (; i<Num_constants; i++) {
      constant[i].value = Real(random_int(10)) + 10.0;
    }

    if (Model_type == 1) {
      //    constant[i].value = 4.0;
      constant[Num_arguments+2].value = 0.0; // Set y value to 0.0.
    }
    constant[Num_arguments+1].value = 14.0; // Set y value to 0.0.
    constant[Num_arguments+3].value = 14.0; // Set y value to 0.0.

  } else {
    for (; i<Num_constants; i++) {
      constant[i].value = 0.0;
    }
  }

  // Need to add production that maps X->X
  for (i=0; i<Num_productions; i++) {
    if (Num_arguments == 0)
      rules[i]->num_bodies = 1;
    else
      rules[i]->num_bodies = Max_bodies;
  }


  num = Num_productions;
  //  Num_productions = 5;

  // First production is special.
  // It MUST call another production, unless number of productions is <= 1.
  // First body...
  while (1) {
    rules[0]->body[0]->make_random();
    if (rules[0]->body[0]->num_productions() > 0) {
      break;
    } else if (Num_productions <= 1) {
      break;
    }
  }

  // Last body...
  if (rules[0]->num_bodies > 1) {
    // Make middle rules duplicate of first.
    for (j=1; j<rules[0]->num_bodies-1; j++)
      rules[0]->body[j]->duplicate(rules[0]->body[0]);

    j = rules[0]->num_bodies-1;
    while (1) {
      rules[0]->body[j]->make_random();
      if (rules[0]->body[j]->num_productions() > 0)
	break;
      else if (Num_productions <= 1)
	break;
    }
  }
  rules[0]->body[rules[0]->num_bodies-1]->set_cond_true();


  for (i=0; i<Num_productions; i++) {
    rules[i]->body[0]->make_random();
    if (rules[i]->num_bodies > 1) {
      for (j=1; j<rules[i]->num_bodies-1; j++)
	rules[i]->body[j]->duplicate(rules[i]->body[0]);
      rules[i]->body[j]->make_random();
      rules[i]->body[j]->set_cond_greater_zero();
    }
  }

  Num_productions = num;
  for (; i<Num_productions; i++) {
    rules[i]->body[0]->make_random();
    if (rules[i]->num_bodies > 1) {
      for (j=1; j<rules[0]->num_bodies-1; j++)
	rules[i]->body[j]->duplicate(rules[i]->body[0]);
      rules[i]->body[j]->make_random();
      rules[i]->body[j]->set_cond_true();
    }
  }


  evaluated = FALSE;
}


// Make this individual a duplicate of ind.
void Individual :: duplicate(Individual *ind)
{
  int i, j;

  if (!created) {
    init();
  }

  evaluated = ind->evaluated;
  duplicate_hist(&(ind->hist_self), &hist_self);
  duplicate_hist(&(ind->hist_parent1), &hist_parent1);
  duplicate_hist(&(ind->hist_parent2), &hist_parent2);

  if (evaluated && (Option[10])) {
    length = ind->length;
    for (i=0; i<length; i++) {
      lstring[i].symbol = ind->lstring[i].symbol;
      lstring[i].arg = ind->lstring[i].arg;
    }
  }

  iterations = ind->iterations;
  for (i=0; i<Num_constants; i++) {
    constant[i].value = ind->constant[i].value;
    constant[i].used = ind->constant[i].used;
  }

  for (i=0; i<Num_productions; i++) {
    rules[i]->used = ind->rules[i]->used;
    rules[i]->num_bodies = ind->rules[i]->num_bodies;
    if (rules[i]->num_bodies > Max_bodies) {
      printf("ind_lsys :: duplicate() - error, num-bodies:%d > Max_bodies: %d.",
	     rules[i]->num_bodies, Max_bodies);
    }

    for (j=0; j<Num_arguments; j++) {
      rules[i]->max[j] = ind->rules[i]->max[j];
      rules[i]->min[j] = ind->rules[i]->min[j];
    }

    for (j=0; j<rules[i]->num_bodies; j++)
      rules[i]->body[j]->duplicate(ind->rules[i]->body[j]);
  }
}


// Duplicates the phenotype of ind into this ones genotype.
int Individual :: duplicate_pheno(Individual *ind)
{
  int i;

  if (!created) {
    init_pheno(TRUE);
  }

  if ((ind->length >= MAX_STRING_LENGTH) || (ind->length < 0)) {
    // Not a valid string, so cannot duplicate.
    return FALSE;
  }

  evaluated = ind->evaluated;
  duplicate_hist(&(ind->hist_self), &hist_self);
  duplicate_hist(&(ind->hist_parent1), &hist_parent1);
  duplicate_hist(&(ind->hist_parent2), &hist_parent2);

  if (evaluated && (Option[10])) {
    length = ind->length;
    for (i=0; i<length; i++) {
      lstring[i].symbol = ind->lstring[i].symbol;
      lstring[i].arg = ind->lstring[i].arg;
    }
  }

  for (i=0; i<Num_constants; i++) {
    constant[i].value = ind->constant[i].value;
    constant[i].used = ind->constant[i].used;
  }
  iterations = 1;
  constant[0].value = 1;

  rules[0]->used = 1;
  rules[0]->body[0]->tested = 1;
  rules[0]->num_bodies = 1;
  rules[0]->body[0]->duplicate(ind->length, ind->lstring);
  if (!rules[0]->body[0]->valid()) {
    printf("ind_lsys :: duplicate_pheno() - Invalid rule: %d.\n", ind->length);
    /*
    for (i=0; i<ind->length; i++) {
      printf("%3d: ", i);
      print_lchar(i, ind->lstring);
      printf(" ");
      rules[0]->body[0]->fprint_char(stdout, i);
      printf("\n");
    }
    */

    return FALSE;
  }
  return TRUE;
}



void Individual :: op_used(int op_num)
{
  Op_used(op_num + 3);

  if (op_num < Num_operations) {
    hist_self.op_count[op_num]++;
    hist_self.operation = op_num;
  }
}


// Returns the length of the L-system.
// Length is measured by the length of each used production body
// plus 2 for each used conditional plus 1 for each used production head.
int Individual :: system_length(void)
{
  int r, b;
  int length = 0;

  if ((Num_productions == 1) && (Num_arguments == 0))
    return rules[0]->body[0]->length;

  for (r=0; r<Num_productions; r++) {
    if (is_rule_used(r)) {
      // This production rule is used.
      length += 1;
      for (b=0; b<rules[r]->num_bodies; b++) {
	if (rules[r]->body[b]->used > 0) {
	  // This production body is used.
	  length += 2+rules[r]->body[b]->length;
	}
      }
    }
  }
  return length;
}


int Individual :: duplicate_used_bodies(int r)
{
  int i, b, num;
  int usd[Max_bodies];

  if (!rules[r]->used > 0)
    return -2;
  if (rules[r]->num_bodies <= 1)
    return -2;

  num = 0;
  for (b=0; b<rules[r]->num_bodies; b++) {
    if (rules[r]->body[b]->used > 0)
      usd[num++] = b;
  }

  if (num == 0)
    return -2;


  for (i=0; i<rules[r]->num_bodies; i++) {
    if (rules[r]->body[i]->used == 0) {
      b = usd[random_int(num)];
      rules[r]->body[i]->duplicate(rules[r]->body[b]);
    }
  }
  return TRUE;
}


// Returns the index of the body to mutate,
// or, -1 if no body to mutate.
// Copies an existing body over an empty body.
int Individual :: mutate_rule_duplicate(int r)
{
  int i, bp, bc;

  if (!rules[r]->used > 0)
    return -2;
  if (rules[r]->num_bodies <= 1)
    return -2;

  // In duplicating.  Duplicate just the string?
  // Or the condition?

  bp = random_used_body(r);
  if (bp == -1)
    return -2;

  bc = random_unused_body(r);
  if (bc == -1)
    return -2;

  rules[r]->body[bc]->duplicate_string(rules[r]->body[bp]);

#ifndef OPTIMIZE
  if (Print[9]) {
    printf("Mutate production rule ::P%d  %d->%d\n",
	   r, bp, bc);
  }
#endif

  if (rules[r]->body[bc]->tested == 0) {
    for (i=bc-1; i>=0; i--) {
      if (rules[r]->body[i]->make_cond_smaller()) {
	break;
      }
    }

  } else if (!rules[r]->body[bc]->usable_cond()) {
    rules[r]->body[bc]->mutate_cond();

  } else {
    rules[r]->body[bc]->make_cond_larger();
  }

  return bp;
}

// Returns the index of the body to mutate,
// or, -1 if no body to mutate.
// Creates a new body.
int Individual :: mutate_rule_new(int r)
{
  int i, bc;

  bc = random_unused_body(r);
  if (bc == -1)
    return -2;

  rules[r]->body[bc]->delete_string();

#ifndef OPTIMIZE
  if (Print[9]) {
    printf("Mutate production rule new ::P%d  %d\n",
	   r, bc);
  }
#endif

  if (rules[r]->body[bc]->tested == 0) {
    for (i=bc-1; i>=0; i--) {
      if (rules[r]->body[i]->make_cond_smaller()) {
	rules[r]->body[bc]->tested = 1;
	break;
      }
    }

  } else if (!rules[r]->body[bc]->usable_cond()) {
    rules[r]->body[bc]->mutate_cond();

  } else {
    rules[r]->body[bc]->make_cond_larger();
  }

  if (rules[r]->body[bc]->tested)
    return bc;
  return -1;
}





// Returns the index of the body to mutate,
// -1 if no body to mutate.
// or -2 if failure.
int Individual :: mutate_rule(int r)
{
  int i, bp, bc;
  Real choice;
  Real op_prob[3];

  if (!rules[r]->used > 0)
    return -2;
  if (rules[r]->num_bodies <= 1)
    return -2;

  op_prob[0] = 0.35;
  op_prob[1] = 0.35;
  if (Num_productions > 1) {
    op_prob[2] = 0.3;
  } else {
    op_prob[2] = 0.0;
  }

  for (i=1; i<=2; i++) {
    op_prob[i] += op_prob[i-1];
  }

  choice = random_real() * op_prob[2];

  if (choice <= op_prob[0]) {
    return mutate_rule_duplicate(r);
  } else if (choice <= op_prob[1]) {
    return mutate_rule_new(r);
  } else if (choice <= op_prob[2]) {
    return encapsulate(r);
  }

  printf("ind_lsys :: mutate_rule() - error, invalid choice: %f\n", float(choice));
  return -2;

  // In duplicating.  Duplicate just the string?
  // Or the condition?

  bp = random_used_body(r);
  if (bp == -1)
    return -1;

  if (bp == 0) {
    bc = 1;
  } else if (bp == rules[r]->num_bodies-1) {
    bc = bp-1;
  } else {
    if (random_int(2))
      bc = bp-1;
    else
      bc = bp+1;
  }

  rules[r]->body[bc]->duplicate_string(rules[r]->body[bp]);

#ifndef OPTIMIZE
  if (Print[9]) {
    printf("Mutate production rule ::P%d  %d->%d\n",
	   r, bp, bc);
  }
#endif

  if (rules[r]->body[bc]->tested == 0) {
    rules[r]->body[bp]->make_cond_smaller();
    return bp;
  } else if (!rules[r]->body[bc]->usable_cond()) {
    rules[r]->body[bc]->mutate_cond();
  }

  if (random_int(2))
    return bp;
  return bc;
}


void Individual :: clear_rule(int r)
{
  int i;

  for (i=0; i<rules[r]->num_bodies; i++)
    rules[r]->body[i]->delete_string();
}


int Individual :: random_used_body_p2(int r)
{
  int num, b;
  int usd[Max_bodies];

  if (rules[r]->used == 0)
    return 0;

  num = 0;
  for (b=0; b<rules[r]->num_bodies; b++) {
    if (rules[r]->body[b]->used > 0)
      usd[num++] = b;
  }

  if (num == 0)
    return -1;

  return usd[random_int(num)];
}


int Individual :: random_used_body(int r)
{
  int num, b;
  int usd[Max_bodies];

  if (r == -1)
    return -1;

  if (rules[r]->used == 0)
    return -1;

  num = 0;
  for (b=0; b<rules[r]->num_bodies; b++) {
    if (rules[r]->body[b]->used > 0)
      usd[num++] = b;
  }

  if (num == 0)
    return -1;

  return usd[random_int(num)];
}


int Individual :: random_unused_body(int r)
{
  int num, b;
  int unusd[Max_bodies];

  if (r == -1)
    return -1;

  if (rules[r]->used == 0)
    return -1;

  num = 0;
  for (b=0; b<rules[r]->num_bodies; b++) {
    if (rules[r]->body[b]->used == 0)
      unusd[num++] = b;
  }

  if (num == 0)
    return -1;

  return unusd[random_int(num)];
}


int Individual :: is_rule_used(int r)
{
  for (int i=0; i<rules[r]->num_bodies; i++)
    if (rules[r]->body[i]->used > 0)
      return TRUE;
  return FALSE;
}


int Individual :: random_used_rule(void)
{
  int num, r;
  int usd[Num_productions];

  num = 0;
  for (r=0; r<Num_productions; r++) {
    if (is_rule_used(r))
      usd[num++] = r;
  }

  if (num == 0)
    return 0;

  return usd[random_int(num)];
}


int Individual :: random_unused_rule(void)
{
  int num, r;
  int unused[Num_productions];

  num = 0;
  for (r=0; r<Num_productions; r++) {
    if (!is_rule_used(r))
      unused[num++] = r;
  }

  if (num == 0)
    return -1;

  return unused[random_int(num)];
}


int Individual :: random_used_constant(void)
{
  int i, num = 0;
  int l[Num_constants];

  for (i=0; i<Num_constants; i++) {
    if (constant[i].used > 0)
      l[num++] = i;
  }

  if (num == 0)
    return -1;
  
  return l[random_int(num)];
}


int Individual :: mutate_constant(void)
{
  int k;

  k = random_used_constant();
  if (k == -1)
    return FALSE;


#ifndef OPTIMIZE2
  op_used(OP_MUT_CONST);
#endif
#ifndef OPTIMIZE
  if (Print[9])
    printf("mutate constant %d: %.1lf <%d> -> ", k, 
	   constant[k].value, constant[k].used);
#endif

  if (random_int(2))
    constant[k].value += 1.0;
  else
    constant[k].value -= 1.0;
  if (constant[k].value == 0.0)
    constant[k].value = 2.0;


#ifndef OPTIMIZE
  if (Print[9])
    printf("%.1lf\n\n", constant[k].value);
#endif

  return TRUE;
}


int Individual :: encapsulate(int r)
{
  int i, b=-1, u, loc1, loc2;

  if (Num_productions < 2)
    return -2;

  u = random_unused_rule();
  if (u == -1)
    return -2;

  for (i=0; i<5; i++) {
    b = random_used_body(r);
    if (b == -1) {
      b = 0;
      loc1 = 0;
      loc2 = 0;
      continue;
    }
    rules[r]->body[b]->random_block(&loc1, &loc2);
    if (loc1 < loc2)
      break;
  }

  if (i == 5) {
    // Failed to find a valid subblock.
    return -2;
  }

  for (i=0; i<rules[r]->num_bodies; i++) {
    rules[u]->body[i]->duplicate_cond(rules[r]->body[i]);
  }


  //  printf("%d - %d (%d)\n", loc1, loc2, rules[r]->body[b]->length);
  /*
  if ((loc1 == 0) && (loc2+1 == rules[r]->body[b]->length)) {
    // block is entire string, already "encapsulated".
    return -2;
  }
  */

  if (rules[r]->body[b]->is_block(loc1)) {
    if (rules[r]->body[b]->string[loc1].match == loc2) {
      if (random_int(2)) {
	// sometimes take block delimiters, sometimes not.
	loc1++;
	loc2--;
      }
    }
  }

  
  if (loc2 <= loc1)
    return -2; // did not find a block, or block is single character.

  //  printf("P%d: (%d - %d)\n", r, loc1, loc2);
  //  rules[r]->body[b]->fprint(stdout);

  // Copy block from used body to unused body.
  for (i=loc1; i<=loc2; i++) {
    rules[u]->body[b]->duplicate_pchar(i-loc1, i, rules[r]->body[b]);
  }
  rules[u]->body[b]->length = loc2 - loc1 + 1;
  if (!rules[u]->body[b]->match_parens()) {
    printf("Ind_lsys :: encapsulate() - error, parens do not match.\n");
    rules[u]->body[b]->fprint(stdout);
    exit(-1);
  }
  // Copy this body to other bodies.
  for (i=0; i<rules[u]->num_bodies; i++) {
    if (i==b)
      continue;
    rules[u]->body[i]->duplicate_string(rules[u]->body[b]);
  }


  // replace first character with production call to unused rule.
  rules[r]->body[b]->string[loc1].type = PBC_TYPE_PRODUCTION;
  rules[r]->body[b]->string[loc1].symbol = u;
  
  // Set input to this production: Pu(n0, n1, n2, .... ni).
  for (i=0; i<Num_arguments; i++) {
    //    rules[r]->body[b]->string[loc1].param[i].op = PBC_OP_NONE;
    rules[r]->body[b]->string[loc1].param[i].arg1.type = PBC_TYPE_VARIABLE;
    rules[r]->body[b]->string[loc1].param[i].arg1.symbol = i;

    if (random_int(2)) {
      rules[r]->body[b]->string[loc1].param[i].op = PBC_OP_SUB;
      rules[r]->body[b]->string[loc1].param[i].arg2.type = PBC_TYPE_VALUE;
      rules[r]->body[b]->string[loc1].param[i].arg2.value = 0.0;
    } else {
      rules[r]->body[b]->string[loc1].param[i].op = PBC_OP_DIV;
      rules[r]->body[b]->string[loc1].param[i].arg2.type = PBC_TYPE_VALUE;
      rules[r]->body[b]->string[loc1].param[i].arg2.value = 1.0;
    }
  }

  // Delete extra characters in body.
  for (i=loc1+1; i<=loc2; i++) {
    rules[r]->body[b]->delete_single_symbol(loc1+1);
  }
  if (!rules[r]->body[b]->match_parens()) {
    printf("Ind_lsys :: encapsulate() - error, parens do not match.\n");
    rules[r]->body[b]->fprint(stdout);
    exit(-1);
  }

  /*
  rules[r]->body[b]->fprint(stdout);
  printf("P%d:\n", u);
  rules[u]->body[b]->fprint(stdout);
  printf("\n");
  */

  /*
  printf("P%d:\n", r);
  for (i=0; i<rules[r]->num_bodies; i++) {
    rules[r]->body[i]->fprint(stdout);
  }
  printf("P%d:\n", u);
  for (i=0; i<rules[r]->num_bodies; i++) {
    rules[u]->body[i]->fprint(stdout);
  }
  printf("\n");
  */

  return b;
}


/******************************
MUTATION
1. mutate a production body.
 (see prod_body.cc)
 - add a sequence
 - delete a sequence
 - reorder
 - mutate condition

2. mutate a rule.
 - copy one body over another (within the same rule),
   adjust its condition and mutate the body.

3. mutate a constant.

********************************/

int Individual :: mutate(Real alpha)
{
  int r, b = -1;

  evaluated = FALSE;

  // Can also mutate constants and functions.

  duplicate_hist(&hist_self, &hist_parent1);
  init_hist(&hist_parent2);
  hist_self.num_mutate++;
  hist_self.how_created = CREATE_MUTATE;

  if (!random_int(4))
    mutate_constant();

  r = random_used_rule();
  if (!random_int(3)) {
    b = mutate_rule(r);
    if (b == -1) {
      return TRUE;
    } else if (b == -2) {
      b = random_used_body(r);
    }
    if (random_int(2)) {
      return TRUE;
    }

  } else {
    b = random_used_body(r);
  }

  /*
  } else {
    if ((Num_productions > 1) && (!random_int(7))) {
      encapsulate(r);
      r = random_used_rule();
    }

    b = random_used_body(r);
    if (b == -1) {
      return TRUE;
    }
  }
  */

  if (b == -1)
    return FALSE;


  while (!rules[r]->body[b]->mutate())
    // Continue mutating until a mutation happens.
    ;
    
  return TRUE;
}


int Individual :: mutate(void)
{
  return mutate(1.0);
}


int Individual :: mutate_pheno(void)
{
  int r, b = -1;

  evaluated = FALSE;

  // Can also mutate constants and functions.

  duplicate_hist(&hist_self, &hist_parent1);
  init_hist(&hist_parent2);
  hist_self.num_mutate++;
  hist_self.how_created = CREATE_MUTATE;

  if (!random_int(4))
    mutate_constant();

  r = 0;
  b = 0;

  while (!rules[r]->body[b]->mutate_pheno())
    // Continue mutating until a mutation happens.
    ;
    
  return TRUE;
}



int Individual :: mutate_init(void)
{

  return TRUE;
}


int Individual :: recombine_constants(Individual *parent2)
{
  int i, num;

#ifndef OPTIMIZE2
  op_used(OP_REC_CONST);
#endif

#ifndef OPTIMIZE
  if (Print[9]) {
    printf("recombine constants:\n");
    printf("P2: ");
    for (i=0; i<Num_constants; i++)
      printf("%4.1lf ", parent2->constant[i].value);
    printf("\nP1: ");
    for (i=0; i<Num_constants; i++)
      printf("%4.1lf ", constant[i].value);
    printf("\n");
  }
#endif

  num = 0;
  for (i=1; i<Num_constants; i++) {
    if ((constant[i].used > 0) && (parent2->constant[i].used > 0)) {
      num++;
      constant[i].value = Real(recombine_int(int(constant[i].value),
					   int(parent2->constant[i].value)));
    }
  }

  if (num == 0)
    return FALSE;

#ifndef OPTIMIZE
  if (Print[9]) {
    printf("C : ");
    for (i=0; i<Num_constants; i++)
      printf("%4.1lf ", constant[i].value);
    printf("\n\n");
  }
#endif

  return TRUE;
}


// Insert some part of a body in parent 2 into/over a body in parent 1.
int Individual :: recombine_productions(Individual *parent2)
{
  int b1, b2;
  int r1, r2;

  r1 = random_used_rule();
  if (rules[r1]->num_bodies == 1) {
    b1 = 0;
  } else {
    b1 = random_used_body(r1);
    if (b1 == -1)
      return FALSE;
  }

  r2 = r1;
  if (parent2->rules[r2]->num_bodies == 1) {
    b2 = 0;
  } else {
    b2 = random_int(parent2->rules[r2]->num_bodies);
  }

  if ((rules[r1]->body[b1]->length == 0) &&
      (parent2->rules[r2]->body[b2]->length == 0)) {
    return FALSE;
  }

#ifndef OPTIMIZE
  if (Print[9]) {
    printf("Recombine:  %d->%d, %d->%d\n", r1, b1, r2, b2);
    cout << "P2: ";
    parent2->rules[r2]->body[b2]->fprint(stdout);
    cout << "P1: ";
    rules[r1]->body[b1]->fprint(stdout);
  }
#endif

  return rules[r1]->body[b1]->recombine(parent2->rules[r2]->body[b2]);
}


// Swap production bodies from a rule in parent2 over a body in the same
// rule of parent1.
int Individual :: recombine_rule(Individual *parent2)
{
  int i, b;
  int r1, r2;

  r2 = parent2->random_used_rule();
  r1 = r2;
  if (r2 < 0)
    return -1;
#ifndef OPTIMIZE2
  op_used(OP_REC_RULE);
#endif


  for (i=random_int(3)+1; i>0; i--) {

    r2 = parent2->random_used_rule();
    r1 = r2;
  
    // Copy entire rule.
    for (b=0; b<rules[r1]->num_bodies; b++)
      rules[r1]->body[b]->duplicate(parent2->rules[r2]->body[b]);  
  }

  /*
  if (random_int(2)) {
    // Copy just the string (successor).
    b1 = random_int(rules[r1]->num_bodies);
    b2 = random_int(parent2->rules[r2]->num_bodies);
    rules[r1]->body[b1]->duplicate(parent2->rules[r2]->body[b2]);

  } else {
    // Copy entire rule.
    b1 = random_int(rules[r1]->num_bodies);
    b2 = random_int(parent2->rules[r2]->num_bodies);
    rules[r1]->body[b1]->duplicate_string(parent2->rules[r2]->body[b2]);
  }
  */

  return r1;
}


// Combine rules from both this individual and parent2.
int Individual :: join(Individual *parent2)
{

  return TRUE;
}


/*******************************
RECOMBINATION
1. recombine parts of production bodies.
2. swap production bodies (special case of 1).
3. swap production rules.
4. recombine constants?
5. recombine conditions?
`6. join production rules?
********************************/
int Individual :: recombine(Individual *parent2)
{
  evaluated = FALSE;

  // History for this individual.
  duplicate_hist(&hist_self, &hist_parent1);
  duplicate_hist(&(parent2->hist_self), &hist_parent2);
  hist_self.num_recombine++;
  hist_self.how_created = CREATE_RECOMBINE;


  int i=0;
  if (random_int(3)) {
    // 66% recombine productions.
    for (i=0; i<5; i++) {
      if (recombine_productions(parent2))
	// Continually recombine until a change is made.
	return TRUE;
    }

  } else {
    int r = recombine_rule(parent2);
    if (r<0)
      return FALSE;
    if (random_int(2)) {
      return TRUE; // Stop here half the time.
    }
    int b = mutate_rule(r);
    if (b == -2) {
      return FALSE;
    }
    if (b != -1)
      while (!rules[r]->body[b]->mutate())
	;
    return TRUE;
  }

  return FALSE;
}


int Individual :: recombine_pheno(Individual *parent2)
{
  int result;

  evaluated = FALSE;

  // Can also mutate constants and functions.

  duplicate_hist(&hist_self, &hist_parent1);
  duplicate_hist(&(parent2->hist_self), &hist_parent2);
  hist_self.num_recombine++;
  hist_self.how_created = CREATE_RECOMBINE;

  if (!random_int(4))
    mutate_constant();

  if ((rules[0]->body[0]->length == 0) && 
      (parent2->rules[0]->body[0]->length == 0)) {
    // No material to recombine.
    return FALSE;
  }

  result = rules[0]->body[0]->recombine(parent2->rules[0]->body[0]);
  if (result == -1) {
    printf("ind_lsys :: recombine_pheno() - error recombining.\n");
    exit(-1);
    return FALSE;
  }
    
  return TRUE;
}



Real read_real(char *line)
{
  Real val;

  if ((line[0] != '-') &&
      ((line[0] < '0') || (line[0] > '9'))) {
    // not a number.
    printf("read_real() - error reading value at: %s.\n", line);
    return -9999.9;
  }

  int i = 0;
  while (i != ' ')
    i++;

  char prev = line[i];
  line[i] = 0;
  val = atof(line);
  line[i] = prev;
  return val;
}


int Individual :: read_history(history *hist, FILE *file)
{
  int i, j, num;
  char line[1000];

  if (!fgets(line, 1000, file)) {
    cout << "Individual :: read_history() - error, empty fitness line." << endl;
    return FALSE;
  }

  j = 0;
  for (i=0; i<Num_fit_dims; i++) {
    hist->fitness[i] = read_real(&(line[j]));
    while (line[j] != ' ')
      j++;
    j++;
  }

  num = fscanf(file, "%d %d %d %d %d\n", &(hist->index),
	       &(hist->how_created), &(hist->operation),
	       &(hist->num_mutate), &(hist->num_recombine));
  if (num != 5) {
    cout << "Read history:: Error reading line 2." << endl;
    return FALSE;
  }    

  if (!fgets(line, 1000, file)) {
    cout << "Individual :: read_history() - error, empty operations line." << endl;
    return FALSE;
  }

  j = 0;
  for (i=0; i<Num_operations; i++) {
    hist->op_count[i] = read_integer(&(line[j]));
    while (line[j] != ' ')
      j++;
    j++;
  }

  return TRUE;
}


int Individual :: read(FILE *file)
{
  int i, j, num;
  int max_prod;
  int index;

  Number_trials = 1;
  Max_joints = 1000;
  
  num = fscanf(file, "Version: %d\n", &version);
  if (num != 1) {
   cout << "Individual: error read version: #" << version << ".\n" << endl;
    return FALSE;

  } else if (version < 7) {
    cout << "Individual: error version #" << version << " not supported.\n" << endl;
    return FALSE;
  } else if (version != IND_VERSION) {
    cout << "Individual: Using version #" << version << endl;
  }    


  num = fscanf(file, "%d %d\n", &Num_fit_dims, &Num_operations);
  if (num != 2) {
    cout << "Individual :: error in reading config row 1.\n";
    return FALSE;
  }

  num = fscanf(file, "%d %d %d %d %d\n", &Num_productions,
	       &Num_terminals, &Num_arguments,
	       &Num_constants, &Num_functions);
  if (num != 5) {
    cout << "Individual: error in reading config row 2.\n" << endl;
    return FALSE;
  }


  if (version == 6) {
    int rule_set;
    num = fscanf(file, "%d %d %d %d\n", &Fitness_func,
		 &iterations, &rule_set, &max_prod);
    if (num != 4) {
      cout << "Individual: error in reading config row 1.\n" << endl;
      return FALSE;
    }
    Max_bodies = 2;
    Terminal_set = rule_set;
    Build_type = 1;

  } else if (version == 7) {
    num = fscanf(file, "%d %d %d %d %d %d\n", &Fitness_func,
		 &iterations, &Build_type, &Terminal_set, &Max_bodies,
		 &max_prod);
    if (num != 6) {
      cout << "Individual: error in reading config row 3.\n" << endl;
      return FALSE;
    }

    if (Build_type == 10) {
      Model_type = 2;
      Build_type = 1;

    } else if (Build_type == 11) {
      Model_type = 3;
      Build_type = 1;

    } else {
      Model_type = 1;
    }

  } else {
    num = fscanf(file, "%d %d %d %d %d %d %d\n", &Fitness_func,
		 &iterations, &Build_type, &Terminal_set,
		 &Model_type, &Max_bodies, &max_prod);
    if (num != 7) {
      cout << "Individual: error in reading config row 3.\n" << endl;
      return FALSE;
    }
  }

  if (Print[29]) {
    printf("Fitness func: %d\n", Fitness_func);
    printf("iterations: %d\n", iterations);
    printf("Build type: %d\n", Build_type);
    printf("Model type: %d\n", Model_type);
    printf("Terminal set: %d\n", Terminal_set);
    printf("Max bodies: %d\n", Max_bodies);
    printf("Max productions: %d\n", max_prod);
  }

  set_terminal_set(Terminal_set);


  // Initialize.
  init(FALSE);

  if (!read_history(&hist_self, file)) {
    cout << "Individ: error in history.\n" << endl;
    return FALSE;
  }


  // Get terminals.
  char terminals[52];
  if (!fgets(terminals, 51, file)) {
    printf("Read() - error, could not read terminals.\n");
    return FALSE;
  }

  for (i=0; i<Num_terminals; i++)
    Terminal[i] = terminals[i];


  // Get constants.
  if (version == 7) {
    for (i=0; i<Num_constants; i++) {
      if (!fscanf(file, "%lf  <%d>\n", &constant[i].value,
		  &constant[i].used)) {
	printf("Read() - error, could not read constant[%d].\n", i);
	return FALSE;
      }
    }
    constant[Num_arguments+1].value = 0.0;
    constant[Num_arguments+2].value = 0.0;
    constant[Num_arguments+3].value = 0.0;

  } else {
    for (i=0; i<Num_constants; i++) {
      if (!fscanf(file, "k%d = %lf  <%d>\n", &index, &constant[i].value,
		  &constant[i].used)) {
	printf("Read() - error, could not read constant[%d].\n", i);
	return FALSE;
      } else if (index != i) {
	printf("Read() - error, invalid constant index: %d:%d.\n", i, index);
	return FALSE;
      }
    }
  }

    
  // Read production rules.
  for (i=0; i<Num_productions; i++) {
    // Read production head.
    num = fscanf(file, "P%d:  %d  <%d>\n", 
		 &index,
		 &(rules[i]->num_bodies),
		 &(rules[i]->used));
    if ((num != 3) || (i != index)) {
      cout << "Individual: error in reading rule head:" << i << "\n" << endl;
      printf("P%d:  %d  (%d)\n", 
	     i,
	     rules[i]->num_bodies,
	     rules[i]->used);  
      return FALSE;
    }

    if (Num_arguments > 0) {
      // Read argument min/max stats.
      char line[1000];
      if (!fgets(line, 1000, file)) {
	printf("Ind_lsys :: error reading arguments line for body P%d.\n", i);
	return FALSE;
      }
      int loc = 0;
      while (line[loc] == ' ')
	loc++;
      for (j=0; j<Num_arguments; j++) {
	rules[i]->min[j] = read_real(&line[loc]);
	while (line[loc] != ':')
	  loc++;
	loc++;
	rules[i]->max[j] = read_real(&line[loc]);
	while ((line[loc] != ' ') && (line[loc] != '\n'))
	  loc++;
	loc += 2;
      }
    }


    /*
    for (j=0; j<Num_arguments; j++) {
      num = fscanf(file, "     %lf:%lf\n",
		   &(rules[i]->min[j]),
		   &(rules[i]->max[j]));
    }
    */

    // Read production bodies.
    for (j=0; j<rules[i]->num_bodies; j++) {
      if (!rules[i]->body[j]->read(file)) {
	cout << "Individual: error in reading P"<<i<<" body:" << j << endl;
	return FALSE;
      }
    }
    fscanf(file, "\n");
  }

  printf("f:%d  t-set:%d  build:%d  model:%d  args:%d  bod:%d  prod:%d\n",
	 Fitness_func, Terminal_set, Build_type, Model_type,
	 Num_arguments, Max_bodies, Num_productions);


  return TRUE;
}


void Individual :: write_history(history *hist, FILE *file)
{
  int i;

  for (i=0; i<Num_fit_dims; i++)
    fprintf(file, "%.4f ", hist->fitness[i]);
  fprintf(file, "\n");

  fprintf(file, "%d %d %d %d %d\n", hist->index,
	  hist->how_created, hist->operation,
	  hist->num_mutate, hist->num_recombine);

  for (i=0; i<Num_operations; i++)
    fprintf(file, "%d ", hist->op_count[i]);
  fprintf(file, "\n");
}


void Individual :: write(FILE *file)
{
  int i, j;
  char terminals[50];

  if (file == NULL) {
    printf("Ind_lsys :: write() - error saving individual.  FILE is null.\nThis may have been caused by inability to create the file.\n");
    return;
  }

  if (!created) {
    cout << "Ind_lsys:: error writing - not created!" << endl;
    return;
  }

  fprintf(file, "Version: %d\n", IND_VERSION);
  fprintf(file, "%d %d\n", Num_fit_dims, Num_operations);
  fprintf(file, "%d %d %d %d %d\n", Num_productions,
	  Num_terminals, Num_arguments, Num_constants,
	  Num_functions);
  fprintf(file, "%d %d %d %d %d %d %d\n", Fitness_func,
	  iterations, Build_type, Terminal_set, Model_type,
	  Max_bodies, MAX_PROD_LENGTH);


  // After init info.
  write_history(&hist_self, file);
  for (i=0; i<Num_terminals; i++)
    terminals[i] = Terminal[i];
  terminals[i] = 0;
  fprintf(file, "%s\n", terminals);
  
  for (i=0; i<Num_constants; i++)
    fprintf(file, "k%d = %.4lf  <%d>\n", i,
	    constant[i].value,
	    constant[i].used);
  
  for (i=0; i<Num_productions; i++) {
    fprintf(file, "P%d:  %d  <%d>\n", 
	    i,
	    rules[i]->num_bodies,
	    rules[i]->used);

    fprintf(file, "     ");
    for (j=0; j<Num_arguments; j++) {
      fprintf(file, "%.1lf:%.1lf  ",
	      rules[i]->min[j],
	      rules[i]->max[j]);
    }
    if (Num_arguments > 0)
      fprintf(file, "\n");

    for (j=0; j<rules[i]->num_bodies; j++) {
      if (!rules[i]->body[j]->fprint(file)) {
	printf("ind_lsys() :: write() - error writing rule %d, body %d.\n",
	       i, j);
      }
    }
    fprintf(file, "\n");
  }
}


void Individual :: write_pheno(FILE *file)
{
  int num_prod = Num_productions;
  int max_bodies = Max_bodies;

  Num_productions = 1;
  Max_bodies = 1;

  write(file);

  Num_productions = num_prod;
  Max_bodies = max_bodies;
}



int Individual :: process_tape(Tape *tape_src,
			       Tape *tape_dest)
{
  int src = 0, dst = 0;

  while (tape_src[src].symbol != TAPE_END) {
    if (tape_src[src].type == PBC_TYPE_PRODUCTION) {
      // Handle production
#ifndef OPTIMIZE
      if (Print[3])
	printf("(P%d):", tape_src[src].symbol);
      //      	printf("%P%d(%.1f) ", tape_src[src].symbol,
      //      	       tape_src[src].arg);
#endif

      dst += process_rule(tape_src[src].arg,
			  &(tape_dest[dst]), 
			  rules[tape_src[src].symbol]);
      if (Print[3])
	printf("!%d ", tape_src[src].symbol);

    } else {
      // Handle non-production.
      tape_dest[dst].type = tape_src[src].type;
      tape_dest[dst].symbol = tape_src[src].symbol;
      tape_dest[dst].arg[0] = tape_src[src].arg[0];
      dst++;

#ifndef OPTIMIZE
      if (Print[3]) {
	printf("%c(%.f) ", tape_src[src].symbol,
	       tape_src[src].arg[0]);
      }
#endif
    }
    src++;

    if (dst >= MAX_STRING_LENGTH)
      break;
  }

  tape_dest[dst].type = PBC_TYPE_NONE;
  tape_dest[dst].symbol = TAPE_END;

#ifndef OPTIMIZE
  if (Print[3])
    cout << "  : Length: " << dst << endl;
  if (Print[27])
    print_tape_simple(tape_dest);

  if (Print[29]) {
    print_tape_english(tape_dest);
  }
#endif

  return dst;
}


void Individual :: reset_used(void)
{
  int i, j;

  for (i=0; i<Num_constants; i++)
    constant[i].used = 0;

  for (i=0; i<Num_productions; i++) {
    rules[i]->used = 0;
    for (j=0; j<Num_arguments; j++) {
      rules[i]->max[j] = 0.0;
      rules[i]->min[j] = 0.0;
    }

    for (j=0; j<rules[i]->num_bodies; j++) {
      rules[i]->body[j]->used = 0;
      rules[i]->body[j]->tested = 0;
    }
  }
}

int Individual :: start_tape(Real *arg, Tape *tape_dest)
{
  int t;

  t = process_rule(arg, tape_dest, rules[0]);
  tape_dest[t].type = PBC_TYPE_NONE;
  tape_dest[t].symbol = TAPE_END;

#ifndef OPTIMIZE
  if (Print[3])
    cout << endl;

  if (Print[4])
    print_tape(tape_dest);

  if (Print[29]) {
    printf("Starting args: ");
    for (int i=0; i<Num_arguments; i++) {
      printf("%.2f  ", (float)arg[i]);
    }
    printf("\n");
    print_tape_english(tape_dest);
  }
#endif

  return t;
}




// Also, in making tape, 5F should translate to:
// 5{F}, so further tape iterations correctly
// duplicate all parts of F, not just first symbol.
// at each [ or { should recursively call itself
// that way it will correctly handle nested blocks
int tape_2_string(char end_type, char end_symbol, int *t,
		  Tape *tape, int *s, LString *string)
{
  char symbol, type;
  int i, mult_val, s2;

#ifndef OPTIMIZE
  if (Print[18])
    printf("\n");
#endif

  while (1) {
    if (Num_processed++ > MAX_PROCESSED) {
      if (Print[10])
	printf("Exceeded maximum processing length: %d\n", Num_processed);
      return FALSE;
    }

    if (*s >= MAX_STRING_LENGTH) {
      string[0].symbol = 0;
      if (Print[10])
	printf("Exceeded maximum data length: %d\n", *s);
      return FALSE;
    }

    type = tape[*t].type;
    symbol = tape[*t].symbol;

    if ((type == end_type) && (symbol == end_symbol)) {
      // End of block.
      *t += 1;
      if (type == PBC_TYPE_BLOCK) {
	if ((symbol == SYMBOL_POP) || (symbol == SYMBOL_POP_EDGE)) {
	string[*s].symbol = symbol;
	*s += 1;
	}
      }

      if (Print[18])
	printf("!");
      return TRUE;

    } else if (type == PBC_TYPE_PRODUCTION) {
      // Production.
      if (Print[18])
	printf("P%d ", symbol);
      *t += 1;
      continue;

    } else if ((type == PBC_TYPE_BLOCK) &&
	       (symbol == SYMBOL_BLK_START)) {
      // This allows handling of nested blocks.
      // Must do multiple times.
      if (Print[18])
	printf("%c", symbol);

      mult_val = int(tape[*t].arg[0]);

      if (mult_val > 0) {
	int t2 = *t + 1;
	for (i=0; i<mult_val; i++) {
	  *t = t2;
	  if (!tape_2_string(PBC_TYPE_BLOCK, SYMBOL_BLK_END, 
			     t, tape, s, string))
	    return FALSE;
	}

      } else {
	s2 = *s;
	*t += 1;
	if (!tape_2_string(PBC_TYPE_BLOCK, SYMBOL_BLK_END, 
			   t, tape, &s2, string))
	  return FALSE;
      }

    } else if ((type == PBC_TYPE_BLOCK) &&
	       (symbol == SYMBOL_PUSH)) {
      // This allows handling of nested blocks.
      if (Print[18])
	printf("%c", symbol);

      string[*s].symbol = symbol;
      *s += 1;
      *t += 1;
      if (!tape_2_string(PBC_TYPE_BLOCK, SYMBOL_POP, 
			 t, tape, s, string))
	return FALSE;

    } else if ((type == PBC_TYPE_BLOCK) &&
	       (symbol == SYMBOL_PUSH_EDGE)) {
      // This allows handling of nested blocks.
      if (Print[18])
	printf("%c", symbol);

      string[*s].symbol = symbol;
      *s += 1;
      *t += 1;
      if (!tape_2_string(PBC_TYPE_BLOCK, SYMBOL_POP_EDGE, 
			 t, tape, s, string))
	return FALSE;

    } else if (type == PBC_TYPE_TERMINAL) {
      // Build terminal.
      if (Print[18])
	printf("%c", symbol);
      string[*s].symbol = symbol;
      string[*s].arg = tape[*t].arg[0];
      if (string[*s].arg < -DBL_MAX) {
	string[*s].arg = -10000000000.0;
      } else if (string[*s].arg > DBL_MAX) {
	string[*s].arg = 10000000000.0;
      }	
      *s += 1;
      *t += 1;

    } else {
      printf("\nInd_lsys::tape_2_string - error processing tape: ->(%d:%d) end=(%d:%d) @(%d:%d)\n",
	     type, symbol, end_type, end_symbol, *t, *s);
      //      while (1) ;
      //      exit(-1);
      return FALSE;
    }
  }
}



int Individual :: tape_to_string(Tape *tape, LString *string)
{
  int t=0, s=0;

  Num_processed = 0;

  if (!tape_2_string(PBC_TYPE_NONE, TAPE_END,
		     &t, tape, &s, string)) {
    // Save individual if error.
    //    FILE *file = fopen("error_t2c.ind", "w");
    //    write(file);
    //    fclose(file);

    //    exit(-1);

    string[0].symbol = TAPE_END;
  }

  string[s].symbol = 0;

  if (string[0].symbol == TAPE_END) {
    s = 0;
  }

  if (Print[19])
    printf("t2c : length %d (processed: %d)\n", s, Num_processed);
  //  if (Print[3])
  //    cout << "  : " << s << endl;

  if (Print[5])
    print_string(string);

  model.lstring_length = s;
  model.new_string = TRUE;

  return s;
}


void measure_phenotype(int *count, int length, LString *string)
{
  int i, j;

  for (i=0; i<Num_terminals+2; i++) {
    count[i] = 0;
  }

  for (i=0; i<length; i++) {
    if (string[i].symbol == TERM_PUSH) {
      count[Num_terminals]++;

    } else if (string[i].symbol == TERM_POP) {
      count[Num_terminals+1]++;

    } else {
      for (j=0; j<Num_terminals; j++) {
	if (string[i].symbol == Terminal[j]) {
	  count[j]++;
	}
      }
    }
  }
}


int Individual :: edit_distance(Individual *ind2)
{
  int i, j;
  int val;
  //  int val2, val3;
  int *tmp;
  int length1, length2;
  static int *dist = 0;
  static int *dist_prev = 0;


  length1 = length;
  length2 = ind2->length;


  if (dist == 0) {
    dist = (int*)malloc((10+MAX_STRING_LENGTH)*sizeof(int)); // dist[i, ].
    dist_prev = (int*)malloc((10+MAX_STRING_LENGTH)*sizeof(int)); // dist[i-1, ].
  }

  // Initialize two distance arrays.
  for (i=0; i<=length2; i++) {
    dist_prev[i] = i;
  }


  // Do the computation.
  for (i=1; i<=length1; i++) {
    dist[0] = i;

    for (j=1; j<=length2; j++) {
      //  m[i-1,j-1] + if s1[i]=s2[j] then 0 else 1 fi,
      val = dist_prev[j-1];
      if (lstring[i-1].symbol != ind2->lstring[j-1].symbol) {
	val += 1;
      } else if ((lstring[i-1].symbol == TERM_PUSH) ||
		 (lstring[i-1].symbol == TERM_POP)) {
	;
      } else if (lstring[i-1].arg != ind2->lstring[j-1].arg) {
	val += 1;
      }

      if (dist_prev[j]+1 < val) {
	val = dist_prev[j]+1;
      }

      if (dist[j-1]+1 < val) {
	val = dist[j-1]+1;
      }

      dist[j] = val;
    }

    tmp = dist;
    dist = dist_prev;
    dist_prev = tmp;
  }

  val = dist_prev[length2];
  free(dist);
  free(dist_prev);
  dist = 0;
  dist_prev = 0;

  return val;
}



int Individual :: phenotype_distance(Individual *ind2)
{
  int i;
  int dist;
  int count1[Num_terminals+2];
  int count2[Num_terminals+2];

  measure_phenotype(count1, length, lstring);
  measure_phenotype(count2, ind2->length, ind2->lstring);

  dist = 0;
  for (i=0; i<Num_terminals+2; i++) {
    dist += absv((count1[i] - count2[i]));
  }

  return dist;  
}


int Individual :: phenotype_distance2(Individual *ind2)
{
  int i;
  int dist;

  if (length != ind2->length)
    return -1;

  dist = 0;
  for (i=0; i<length; i++) {
    if (lstring[i].symbol != ind2->lstring[i].symbol) {
      dist++;

    } else if ((lstring[i].symbol == TERM_PUSH) ||
	       (lstring[i].symbol == TERM_POP)) {
      ;

    } else if (lstring[i].arg != ind2->lstring[i].arg) {
      dist++;
    }

    if (Print[37]) {
      printf("%d:  %c(%.2f)?%c(%.2f)<%d>\n", i,
	     lstring[i].symbol,
	     lstring[i].arg,
	     ind2->lstring[i].symbol,
	     ind2->lstring[i].arg,
	     dist);
    }
  }
  return dist;  
}



int Individual :: phenotype_dist(Individual *ind2)
{
  int dist;

  dist = phenotype_distance(ind2);
  /*
  if (dist <= 0)
    dist = phenotype_distance2(ind2);
  */

  return dist;
}


void Individual :: display(void)
{
  FILE *file1 = fopen("last.ind", "w");
  write(file1);
  fclose(file1);

  if (length < MAX_STRING_LENGTH) {
    if (Print[5])
      print_string(lstring);
    if (Model_type == 10)
      build_structure(&model);
    display_structure(model.model_id);

  } else {
    //    if (Print[10])
    printf("Tape too long: %d\n", length);
  }
}


Real table_fitness(Real p_height, Real p_strength, Individual *ind)
{
  // uses fitness_table16();
  p_height = 4.0 * p_height;
  p_strength = p_strength * p_strength * p_height * 25.0;

  ModelStruct *model = &ind->model;
  Real height = model->fitness[1] + 1.0;
  Real surf = model->fitness[2];
  Real strength = model->fitness[3];
  Real penalty = model->fitness[4];
  Real ratio1 = 1.0, ratio3 = 1.0;


  if (penalty < 1.0) {
    penalty = 1.0;
  }

  Real fitness = height * surf * strength / penalty;
  
  if (fitness > 0.0) {
    // Do fitness adjustment for height.
    if (height < p_height) {
      ratio1 = height / p_height;
    } else if (height > p_height) {
      //      ratio1 = p_height / height;
      if (height - p_height > p_height) {
	ratio1 = 0.0;
      } else {
	ratio1 = (2.0*p_height - height) / p_height;
	ratio1 *= ratio1;
      }
    }

    // Do fitness adjustment for strength.
    if (strength < p_strength) {
      ratio3 = strength / p_strength;
    } else if (strength > p_strength) {
      if (strength - p_strength > p_strength) {
	ratio3 = 0.0;
      } else {
	ratio3 = (2.0*p_strength - strength) / p_strength;
	ratio3 *= ratio3;
      }
    }
    fitness *= ratio1 * ratio3;
  }

  if (Print[6]) {
    printf("actual: %.2f - (h:%d:%d:%.2f)(a:%d)(s:%d:%d:%.2f) - calc'd: %.2f\n\n",
	   (float)model->fitness[0],
	   (int)height, (int)p_height, (float)ratio1,
	   (int)surf,
	   (int)strength, (int)p_strength,  (float)ratio3,(float)fitness);
  }
  return fitness;
}

Real table_fitness3(Real p_height, Real p_surf, Individual *ind)
{
  // uses fitness_table16();
  ModelStruct *model = &ind->model;
  Real height = model->fitness[1];
  Real surf = model->fitness[2];
  Real p_surf2 = p_surf*p_surf * 25.0;
  p_height = 5.0 * p_height - 1.0;
  Real p_strength = p_height * p_surf2;
  Real strength = model->fitness[3];
  Real  penalty = model->fitness[4];
  Real ratio1 = 1.0, ratio2 = 1.0, ratio3 = 1.0;


  if (penalty < 1.0) {
    penalty = 1.0;
  }

  Real fitness = height * surf * strength / penalty;
  
  if (fitness > 0.0) {
    // Do fitness adjustment for height.
    if (height < p_height) {
      ratio1 = height / p_height;
    } else if (height > p_height) {
      //      ratio1 = p_height / height;
      if (height - p_height > p_height) {
	ratio1 = 0.0;
      } else {
	ratio1 = (2.0*p_height - height) / p_height;
	ratio1 *= ratio1;
      }
    }

    // Do fitness adjustment for area.    
    if (surf < p_surf2) {
      ratio2 = surf / p_surf2;
    } else if (surf > p_surf2) {
      if (surf - p_surf2 > p_surf2) {
	ratio2 = 0.0;
      } else {
	ratio2 = (2.0*p_surf2 - surf) / p_surf2;
	ratio2 *= ratio2;
      }
    }

    // Do fitness adjustment for strength.
    if (strength < p_strength) {
      ratio3 = strength / p_strength;
    } else if (strength > p_strength) {
      if (strength - p_strength > p_strength) {
	ratio3 = 0.0;
      } else {
	ratio3 = (2.0*p_strength - strength) / p_strength;
	ratio3 *= ratio3;
      }
    }
    fitness *= ratio1 * ratio2 * ratio3;
  }

  if (Print[6]) {
    printf("actual: %.2f - (h:%d:%d:%.2f)(a:%d:%d:%.2f)(s:%d:%d:%.2f) - calc'd: %.2f\n\n",
	   (float)model->fitness[0],
	   (int)height, (int)p_height, (float)ratio1,
	   (int)surf, (int)p_surf2, (float)ratio2,
	   (int)strength, (int)p_strength,  (float)ratio3,(float)fitness);
  }
  return fitness;
}

Real table_fitness2(Real p_height, Real p_surf, Individual *ind)
{
  // uses fitness_table16();
  p_height = 4.0 * p_height;

  ModelStruct *model = &ind->model;
  Real height = model->fitness[1] + 1.0;
  Real surf = model->fitness[2];
  Real p_surf2 = p_surf*p_surf * 25.0;
  Real p_strength = p_height * p_surf2;
  Real strength = model->fitness[3];
  Real  penalty = model->fitness[4];
  Real ratio1 = 1.0, ratio2 = 1.0, ratio3 = 1.0;

  if (penalty < 1.0) {
    penalty = 1.0;
  }

  Real fitness = height * surf * strength / penalty;
  
  if (fitness > 0.0) {
    // Do fitness adjustment for height.
    if (height > p_height) {
      //      ratio1 = p_height / height;
      if (height - p_height > p_height) {
	ratio1 = 0.0;
      } else {
	ratio1 = (2.0*p_height - height) / p_height;
	ratio1 *= ratio1;
      }
    }

    // Do fitness adjustment for area.    
    if (surf > p_surf2) {
      if (surf - p_surf2 > p_surf2) {
	ratio2 = 0.0;
      } else {
	ratio2 = (2.0*p_surf2 - surf) / p_surf2;
	ratio2 *= ratio2;
      }
    }

    // Do fitness adjustment for strength.
    if (height < p_height) {
      if (surf < p_surf2) {
	p_strength = surf * height;
      } else {
	p_strength = height * p_surf2;
      }
    } else if (surf < p_surf2) {
      p_strength = surf * p_height;
    }
    if (strength > p_strength) {
      if (strength - p_strength > p_strength) {
	ratio3 = 0.0;
      } else {
	ratio3 = (2.0*p_strength - strength) / p_strength;
	ratio3 *= ratio3;
      }
    }
    fitness *= ratio1 * ratio2 * ratio3;
  }

  if (Print[6]) {
    printf("actual: %.2f - (h:%d:%d:%.2f)(a:%d:%d:%.2f)(s:%d:%d:%.2f) - calc'd: %.2f\n\n",
	   (float)model->fitness[0],
	   (int)height, (int)p_height, (float)ratio1,
	   (int)surf, (int)p_surf2, (float)ratio2,
	   (int)strength, (int)p_strength, (float)ratio3,(float)fitness);
  }
  return fitness;
}


Real Individual :: evaluate_helper(int init, int iterations, Real *argument)
{
  int i, ok = TRUE;
  Real val;


  for (i=0; i<Num_fit_dims; i++)
    model.fitness[i] = 0.0;

  length = start_tape(argument, Tape1);
  if (length >= MAX_STRING_LENGTH) {
    ok = FALSE;
  }

  if (Option[9] && ok) {
    length = tape_to_string(Tape1, lstring);
    printf("Option 9: i=0\n");
    if (build_structure(&model)) {
      evaluate_structure(&model);
    } else {
      //      ok = FALSE;
      ;
    }
  }

  i = 1;
  while(ok) {
    if (i >= iterations)
      break;

    if (Print[29]) {
      printf("Iteration %d.\n", i);
    }
    
    i++;
    length = process_tape(Tape1, Tape2);
    if (length >= MAX_STRING_LENGTH) {
      ok = FALSE;
      break;
    }
    if (Option[9]) {
      printf("Option 9: i=%d\n", i);
      length = tape_to_string(Tape2, lstring);
      if (build_structure(&model)) {
	evaluate_structure(&model);
      } else {
	//	ok = FALSE;
	//	break;
      }
    }

    if (i >= iterations)
      break;

    if (Print[29]) {
      printf("Iteration %d.\n", i);
    }
    
    i++;
    length = process_tape(Tape2, Tape1);
    if (length >= MAX_STRING_LENGTH) {
      ok = FALSE;
      break;
    }
    if (Option[9]) {
      printf("Option 9: i=%d\n", i);
      length = tape_to_string(Tape1, lstring);
      if (build_structure(&model)) {
	evaluate_structure(&model);
      } else {
	//	ok = FALSE;
	//	break;
      }
    }
  }

#ifndef OPTIMIZE
  if (Print[28]) {
    if (i%2)
      print_tape_simple(Tape1);
    else
      print_tape_simple(Tape2);
  }
#endif

  //  if (!Option[9] && ok) {
  if (ok) {
    if (i%2)
      length = tape_to_string(Tape1, lstring);
    else
      length = tape_to_string(Tape2, lstring);

    if (Print[29]) {
      print_string_english(lstring);
    }
    
    if (length >= MAX_STRING_LENGTH) {
      ok = FALSE;
    } else {
      if (build_structure(&model)) {
	evaluate_structure(&model);
      } else {
	ok = FALSE;
      }
      if (model.error) {
	printf("Error evaluating individual.  Saving: 'last.ind'.\n");
	FILE *file1 = fopen("last.ind", "w");
	write(file1);
	fclose(file1);
	exit(-1);
	while (1);
      }
    }
  }

  if (!ok) {
    val = 0.0;
  } else if (length >= MAX_STRING_LENGTH) {
    val = 0.0;
  } else
    val = model.fitness[0];

  if (Num_fit_dims > 7) {
    set_fitness(1, model.fitness[1]);
    set_fitness(2, model.fitness[2]);
    set_fitness(3, model.fitness[3]);
    set_fitness(4, model.fitness[4]);
    set_fitness(5, model.fitness[5]);
  }

  return val;
}


Real Individual :: evaluate(int init)
{
  int i;
  Real val = 0.0, val1;
  Real attr1, attr2, penalty;
  Real argument[Num_arguments];


  /*
  FILE *file1 = fopen(Config.filename_error, "w");
  write(file1);
  fclose(file1);
  */

  for (i=0; i<Num_fit_dims; i++)
    model.fitness[i] = 0.0;

  reset_used();

  iterations = int(constant[0].value);
  constant[0].used++;
  if (iterations < 1)
    iterations = 1;

  for (i=0; i<Num_arguments; i++) {
    argument[i] = constant[i+1].value;
    constant[i+1].used++;
  }

  model.build_args[0] = constant[i+1].value;
  model.build_args[1] = constant[i+2].value;
  model.build_args[2] = constant[i+3].value;
  constant[i+1].used++;
  constant[i+2].used++;
  constant[i+3].used++;


  if ((Build_type == 5) && (Model_type == 3) && (Fitness_func == 12)) {
    // N-parity.
      
    val = 0.0;
      
    /*
      argument[0] = 2.0;
      model.build_args[0] = argument[0];
      val += evaluate_helper(init, iterations, argument);
    */
    argument[0] = 3.0;
    model.build_args[0] = argument[0];
    val += evaluate_helper(init, iterations, argument);
    /*
      argument[0] = 4.0;
      model.build_args[0] = argument[0];
      val += evaluate_helper(init, iterations, argument);
    */
    argument[0] = 5.0;
    model.build_args[0] = argument[0];
    val += evaluate_helper(init, iterations, argument);
    /*
      argument[0] = 6.0;
      model.build_args[0] = argument[0];
      val += evaluate_helper(init, iterations, argument);
    */
    argument[0] = 7.0;
    model.build_args[0] = argument[0];
    val += evaluate_helper(init, iterations, argument);

    /*
    argument[0] = 9.0;
    model.build_args[0] = argument[0];
    val += evaluate_helper(init, iterations, argument);

    argument[0] = 11.0;
    model.build_args[0] = argument[0];
    val += evaluate_helper(init, iterations, argument);

    argument[0] = 13.0;
    model.build_args[0] = argument[0];
    val += evaluate_helper(init, iterations, argument);
    */

  } else if ((Build_type == 5) && (Model_type == 3) && (Fitness_func == 13)) {
    // n-parity.
    val = 0.0;

    /*    
    argument[0] = 4.0;
    model.build_args[0] = argument[0];
    val += evaluate_helper(init, iterations, argument);
    */

    argument[0] = 5.0;
    model.build_args[0] = argument[0];
    val += evaluate_helper(init, iterations, argument);

    /*
    argument[0] = 6.0;
    model.build_args[0] = argument[0];
    val += evaluate_helper(init, iterations, argument);

    argument[0] = 7.0;
    model.build_args[0] = argument[0];
    val += evaluate_helper(init, iterations, argument);
    */

  } else if ((Build_type == 1) && (Model_type == 1) &&  (Fitness_func >= 36)) {
    // parameterized tables.
    if ((Fitness_func == 36) || (Fitness_func == 37)) {
      val = 0.0;
      int fails = 0;

      argument[0] = 5.0;
      argument[1] = 5.0;
      model.build_args[3] = argument[0];
      model.build_args[4] = argument[1];
      val1 = evaluate_helper(init, iterations, argument);
      val += val1;
      if (val1 == 0.0) {
	fails++;
      }

      if (!fails) {
	argument[0] = 10.0;
	argument[1] = 5.0;
	model.build_args[3] = argument[0];
	model.build_args[4] = argument[1];
	val1 = evaluate_helper(init, iterations, argument);
	val += val1;
	if (val1 == 0.0) {
	  fails++;
	}
      }

      if (!fails) {
	argument[0] = 5.0;
	argument[1] = 10.0;
	model.build_args[3] = argument[0];
	model.build_args[4] = argument[1];
	val1 = evaluate_helper(init, iterations, argument);
	val += val1;
	if (val1 == 0.0) {
	  fails++;
	}
      }

      if (!fails) {
	argument[0] = 10.0;
	argument[1] = 10.0;
	model.build_args[3] = argument[0];
	model.build_args[4] = argument[1];
	val1 = evaluate_helper(init, iterations, argument);
	val += val1;
	if (val1 == 0.0) {
	  fails++;
	}
      }

      if (fails) {
	val = 0.0;
      }

    } else if ((Fitness_func == 38) || (Fitness_func == 39) ||
	       (Fitness_func = 40) || (Fitness_func == 41)) {
      // parameterized table height(38,40) or volume(39,41).
      val = 0.0;
      penalty = 1.0;
      int fails = 0;

      argument[0] = 2.0;
      model.build_args[3] = argument[0];
      val1 = evaluate_helper(init, iterations, argument);
      val += val1;
      if (val1 == 0.0) {
	fails++;
      }
      if ((Fitness_func == 38) || (Fitness_func == 40)) {
	attr1 = model.fitness[1];
      } else {
	attr1 = model.fitness[2];
      }
      penalty *= model.fitness[6];
      //      penalty += model.fitness[6];
      //      penalty *= sqrt(model.fitness[6]);
      //      penalty *= (1.0 - model.fitness[6])/2.0;

      if (!fails) {
	argument[0] = 4.0;
	model.build_args[3] = argument[0];
	val1 = evaluate_helper(init, iterations, argument);
	val += val1;
	penalty *= model.fitness[6];
	//	penalty += model.fitness[6];
	//	penalty *= sqrt(model.fitness[6]);
	//	penalty *= (1.0 - model.fitness[6])/2.0;
	if (val1 == 0.0) {
	  fails++;
	} else {
	  if ((Fitness_func == 38) || (Fitness_func == 40)) {
	    attr2 = model.fitness[1];
	  } else {
	    attr2 = model.fitness[2];
	  }
	  if (attr2 <= attr1) {
	    fails++;
	  } else {
	    attr1 = attr2;
	  }
	}
      }

      if (!fails) {
	argument[0] = 6.0;
	model.build_args[3] = argument[0];
	val1 = evaluate_helper(init, iterations, argument);
	val += val1;
	penalty *= model.fitness[6];
	//	penalty += model.fitness[6];
	//	penalty *= sqrt(model.fitness[6]);
	//	penalty *= (1.0 - model.fitness[6])/2.0;
	if (val1 == 0.0) {
	  fails++;
	} else {
	  if ((Fitness_func == 38) || (Fitness_func == 40)) {
	    attr2 = model.fitness[1];
	  } else {
	    attr2 = model.fitness[2];
	  }
	  if (attr2 <= attr1) {
	    fails++;
	  } else {
	    attr1 = attr2;
	  }
	}
      }

      if (!fails) {
	argument[0] = 8.0;
	model.build_args[3] = argument[0];
	val1 = evaluate_helper(init, iterations, argument);
	val += val1;
	penalty *= model.fitness[6];
	//	penalty += model.fitness[6];
	//	penalty *= sqrt(model.fitness[6]);
	//	penalty *= (1.0 - model.fitness[6])/2.0;
	if (val1 == 0.0) {
	  fails++;
	} else {
	  if ((Fitness_func == 38) || (Fitness_func == 40)) {
	    attr2 = model.fitness[1];
	  } else {
	    attr2 = model.fitness[2];
	  }
	  if (attr2 <= attr1) {
	    fails++;
	  } else {
	    attr1 = attr2;
	  }
	}
      }

      if (fails) {
	val = 0.0;
      } else {
	//	val *= ((penalty-1.0)/4.0);
	val *= penalty;
      }
    }

  } else if (Model_type == 4) {
    // sorting programs.

    if (Fitness_func == 3) {
      argument[0] = 7.0; // Size of data set.
      model.build_args[0] = argument[0];
      val = evaluate_helper(init, iterations, argument);

    } else if (Fitness_func == 4) {
      // parametric sorting programs.
      val = 0.0;
      argument[0] = 3.0;
      model.build_args[0] = argument[0];
      val += evaluate_helper(init, iterations, argument);

      argument[0] = 4.0;
      model.build_args[0] = argument[0];
      val += evaluate_helper(init, iterations, argument);

      argument[0] = 5.0;
      model.build_args[0] = argument[0];
      val += evaluate_helper(init, iterations, argument);

      argument[0] = 6.0;
      model.build_args[0] = argument[0];
      val += evaluate_helper(init, iterations, argument);

    } else if (Fitness_func == 5) {
      argument[0] = 8.0; // Size of data set.
      model.build_args[0] = argument[0];
      val = evaluate_helper(init, iterations, argument);

    } else if (Fitness_func == 6) {
      argument[0] = 9.0; // Size of data set.
      model.build_args[0] = argument[0];
      val = evaluate_helper(init, iterations, argument);

    } else {
      model.build_args[0] = 8.0;
      val = evaluate_helper(init, iterations, argument);
    }

  } else {
    val = evaluate_helper(init, iterations, argument);
  }

  if (evaluated) {
    Real val_prev;

    val_prev = get_fitness();
    hist_self.fitness_prev[0] = val_prev;
    set_fitness((val + val_prev)/2.0);
    evaluated = 2;
    /*
    if (val != val_prev) {
      printf("Fitnesses are different!!! %.4lf, now-> %.4lf\n",
	     val_prev, val);

      FILE *file1 = fopen("error1.ind", "w");
      write(file1);
      fclose(file1);
    }
    */

  } else {
    evaluated = 1;
  }

  set_fitness(val);

  if (Num_fit_dims < 4)
    Num_fit_dims = 4;

  if (Num_fit_dims == 6) {
    set_fitness(1, system_length());
    set_fitness(2, length);
    set_fitness(3, model.fitness[1]);
    set_fitness(4, model.fitness[2]);
    set_fitness(5, model.fitness[3]);
    
  } else if (Num_fit_dims > 7) {
    set_fitness(Num_fit_dims-2, system_length());
    set_fitness(Num_fit_dims-1, length);
    /*
    printf("%.0f %.0f %.0f %.0f %.0f %.0f %.0f %.0f\n",
	   hist_self.fitness[0],
	   hist_self.fitness[1],
	   hist_self.fitness[2],
	   hist_self.fitness[3],
	   hist_self.fitness[4],
	   hist_self.fitness[5],
	   hist_self.fitness[6],
	   hist_self.fitness[7]);
    */

  } else if (Num_fit_dims > 3) {
    set_fitness(Num_fit_dims-3, system_length());
    set_fitness(Num_fit_dims-2, length);
    set_fitness(Num_fit_dims-1, model.fitness[1]);

    if (Print[6]) {
      printf("System: %d, String: %d, Comp: %d.\n", system_length(),
	     length, int(model.fitness[1]));
    }
  }

  return model.fitness[0];
}

Real Individual :: evaluate(void) {
  return evaluate(FALSE);
}



/*
    argument[0] = 4.0;
    argument[1] = 4.0;
    model.build_args[3] = argument[0];
    model.build_args[4] = argument[1];
    val1 = evaluate_helper(init, iterations, argument);
    val += val1;
    if (val1 == 0.0) {
      fails++;
    }

    if (!fails) {
      argument[0] = 7.0;
      argument[1] = 4.0;
      model.build_args[3] = argument[0];
      model.build_args[4] = argument[1];
      val1 = evaluate_helper(init, iterations, argument);
      val += val1;
      if (val1 == 0.0) {
	fails++;
      }
    }

    if (!fails) {
      argument[0] = 10.0;
      argument[1] = 4.0;
      model.build_args[3] = argument[0];
      model.build_args[4] = argument[1];
      val1 = evaluate_helper(init, iterations, argument);
      val += val1;
      if (val1 == 0.0) {
	fails++;
      }
    }

    if (!fails) {
      argument[0] = 4.0;
      argument[1] = 7.0;
      model.build_args[3] = argument[0];
      model.build_args[4] = argument[1];
      val1 = evaluate_helper(init, iterations, argument);
      val += val1;
      if (val1 == 0.0) {
	fails++;
      }
    }

    if (!fails) {
      argument[0] = 7.0;
      argument[1] = 7.0;
      model.build_args[3] = argument[0];
      model.build_args[4] = argument[1];
      val1 = evaluate_helper(init, iterations, argument);
      val += val1;
      if (val1 == 0.0) {
	fails++;
      }
    }

    if (!fails) {
      argument[0] = 10.0;
      argument[1] = 7.0;
      model.build_args[3] = argument[0];
      model.build_args[4] = argument[1];
      val1 = evaluate_helper(init, iterations, argument);
      val += val1;
      if (val1 == 0.0) {
	fails++;
      }
    }

    if (!fails) {
      argument[0] = 4.0;
      argument[1] = 10.0;
      model.build_args[3] = argument[0];
      model.build_args[4] = argument[1];
      val1 = evaluate_helper(init, iterations, argument);
      val += val1;
      if (val1 == 0.0) {
	fails++;
      }
    }

    if (!fails) {
      argument[0] = 7.0;
      argument[1] = 10.0;
      model.build_args[3] = argument[0];
      model.build_args[4] = argument[1];
      val1 = evaluate_helper(init, iterations, argument);
      val += val1;
      if (val1 == 0.0) {
	fails++;
      }
    }

    if (!fails) {
      argument[0] = 10.0;
      argument[1] = 10.0;
      model.build_args[3] = argument[0];
      model.build_args[4] = argument[1];
      val1 = evaluate_helper(init, iterations, argument);
      val += val1;
      if (val1 == 0.0) {
	fails++;
      }
    }      

 */

void print_lchar(int i, LString *lstring)
{
  if ((lstring[i].symbol == TERM_PUSH) ||
      (lstring[i].symbol == TERM_POP)) {
    printf("%d:%c", i, lstring[i].symbol);
  } else {	
    //	printf("%c(%d) ", lstring[i].symbol, int(lstring[i].arg));
    printf("%d:%c(%.2f)", i, lstring[i].symbol, lstring[i].arg);
  }
}




#ifdef OLD
Real Individual :: evaluate(int init)
{
  int i, ok = TRUE;
  Real val;
  Real argument[Num_arguments];


  /*
  FILE *file1 = fopen(Config.filename_error, "w");
  write(file1);
  fclose(file1);
  */

  for (i=0; i<Num_fit_dims; i++)
    model.fitness[i] = 0.0;

  reset_used();
  iterations = int(constant[0].value);
  constant[0].used++;
  if (iterations < 1)
    iterations = 1;

  for (i=0; i<Num_arguments; i++) {
    argument[i] = constant[i+1].value;
    constant[i+1].used++;
  }

  
  model.build_args[0] = constant[i+1].value;
  model.build_args[1] = constant[i+2].value;
  model.build_args[2] = constant[i+3].value;
  constant[i+1].used++;
  constant[i+2].used++;
  constant[i+3].used++;

  length = start_tape(argument, Tape1);
  if (length >= MAX_STRING_LENGTH) {
    ok = FALSE;
  }
  if (Option[9] && ok) {
    length = tape_to_string(Tape1, lstring);
    printf("Option 9: i=0\n");
    if (build_structure(&model)) {
      evaluate_structure(&model);
    } else {
      //      ok = FALSE;
      ;
    }
  }

  i = 1;
  while(ok) {
    if (i >= iterations)
      break;
    
    i++;
    length = process_tape(Tape1, Tape2);
    if (length >= MAX_STRING_LENGTH) {
      ok = FALSE;
      break;
    }
    if (Option[9]) {
      printf("Option 9: i=%d\n", i);
      length = tape_to_string(Tape2, lstring);
      if (build_structure(&model)) {
	evaluate_structure(&model);
      } else {
	//	ok = FALSE;
	//	break;
      }
    }

    if (i >= iterations)
      break;

    i++;
    length = process_tape(Tape2, Tape1);
    if (length >= MAX_STRING_LENGTH) {
      ok = FALSE;
      break;
    }
    if (Option[9]) {
      printf("Option 9: i=%d\n", i);
      length = tape_to_string(Tape1, lstring);
      if (build_structure(&model)) {
	evaluate_structure(&model);
      } else {
	//	ok = FALSE;
	//	break;
      }
    }
  }

#ifndef OPTIMIZE
  if (Print[28]) {
    if (i%2)
      print_tape_simple(Tape1);
    else
      print_tape_simple(Tape2);
  }
#endif

  //  if (!Option[9] && ok) {
  if (ok) {
    if (i%2)
      length = tape_to_string(Tape1, lstring);
    else
      length = tape_to_string(Tape2, lstring);
    
    if (length >= MAX_STRING_LENGTH) {
      ok = FALSE;
    } else {
      if (build_structure(&model)) {
	evaluate_structure(&model);
      } else {
	ok = FALSE;
      }
      if (model.error) {
	printf("Error evaluating individual.  Saving: 'last.ind'.\n");
	FILE *file1 = fopen("last.ind", "w");
	write(file1);
	fclose(file1);
	exit(-1);
	while (1);
      }
    }
  }

  if (!ok) {
    val = 0.0;
  } else if (length >= MAX_STRING_LENGTH) {
    val = 0.0;
  } else
    val = model.fitness[0];

  if (evaluated) {
    Real val_prev;

    val_prev = get_fitness();
    hist_self.fitness_prev[0] = val_prev;
    set_fitness((val + val_prev)/2.0);
    evaluated = 2;
    /*
    if (val != val_prev) {
      printf("Fitnesses are different!!! %.4lf, now-> %.4lf\n",
	     val_prev, val);

      FILE *file1 = fopen("error1.ind", "w");
      write(file1);
      fclose(file1);
    }
    */

  } else {
    evaluated = 1;
  }

  set_fitness(model.fitness);

  return model.fitness[0];
}
#endif
