/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#define NET_VERSION 3

#define MUT_WEIGHT_NUM 10
#define MUT_LINK_NUM 5
#define MUT_PROPERTY_NUM 4
#define MAX_UNITS_RECOMBINE 10
#define MAX_LINKS_RECOMBINE 5

#define MAX_LINKS 15
#define MAX_WEIGHT 1.5
#define MAX_VALUE (PI/2.0)

// bias: 0.05
// weight: 0.05
#define PROPERTY_DELTA 0.05
#define WEIGHT_DELTA 0.05


#define UNIT_MAX_OUTPUT_FUNC 6
#define UNIT_FUNCTIONS 4

#define TYPE_INPUT -2
#define TYPE_OUTPUT -1
#define TYPE_HIDDEN 0


#define FUNCTION_SIGMOID 0
#define FUNCTION_SIGMOID2 1
#define FUNCTION_SUM 2
#define FUNCTION_ASIN 3
#define FUNCTION_SIN 4
#define FUNCTION_SQRT 5
#define FUNCTION_SQR 6
#define FUNCTION_CLOCK 7
#define FUNCTION_MIN 8
#define FUNCTION_MAX 9
#define FUNCTION_SIGN 10
#define FUNCTION_RANDOM 11
#define FUNCTION_PROD 12


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stream.h>
#include <iostream.h>
#include <fstream.h>
#include "ind_net.h"
#include "general.h"

extern Real evaluate_network(int init, Individual *net);
int Net_struct = NET_RECURRENT;
int Net_cont_time = FALSE;
int Net_max_func = UNIT_FUNCTIONS;
int Net_num_iterations = 1;
int Net_switching = FALSE;
int Net_fix_hyperplanes = FALSE;
Real Rec_delta = 1.0;

int Net_groups = 0;
int Net_group_size[20] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			  0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
int Net_group_start[20] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
			  0, 0, 0, 0, 0, 0, 0, 0, 0, 0};

Real Max_output[100];
Real Min_output[100];

Individual :: Individual(void)
{
  num_inputs = 6;
  num_outputs = 6;
  num_hidden = 15;
  win = 0;
  init_hist(&hist_self);
  init_hist(&hist_parent1);
  init_hist(&hist_parent2);
  created = FALSE;
  evaluated = FALSE;
}


Individual :: Individual(Individual *sample)
{
  if (sample != NULL) {
    num_inputs = sample->num_inputs;
    num_outputs = sample->num_outputs;
    num_hidden = sample->num_hidden;
  } else {
    num_inputs = 6;
    num_outputs = 6;
    num_hidden = 15;
  }
  win = 0;
  init_hist(&hist_self);
  init_hist(&hist_parent1);
  init_hist(&hist_parent2);
  created = FALSE;
  evaluated = FALSE;

}


void Individual :: init(void)
{
  int i;

  num_units = num_inputs + num_outputs + num_hidden;
  output_start = num_inputs + num_hidden;
  output_stop = num_units;

  unit_type = (int*)malloc(num_units*sizeof(int));
  function = (int*)malloc(num_units*sizeof(int));
  num_in = (int*)malloc(num_units*sizeof(int));
  active = (int*)malloc(num_units*sizeof(int));
  switching = (int*)malloc(num_units*sizeof(int));
  index_ptr1 = (int*)malloc(num_units*sizeof(int));
  index_ptr2 = (int*)malloc(num_units*sizeof(int));
  scale = (Real*)malloc(num_units*sizeof(Real));
  bias = (Real*)malloc(num_units*sizeof(Real));
  time_const = (Real*)malloc(num_units*sizeof(Real));
  ctr_rate = (Real*)malloc(num_units*sizeof(Real));
  value = (Real*)malloc(num_units*sizeof(Real));
  value2 = (Real*)malloc(num_units*sizeof(Real));
  from_link = (int**)malloc(num_units*sizeof(int*));
  weight = (Real**)malloc(num_units*sizeof(Real*));
  for (i=0; i<num_units; i++) {
    from_link[i] = (int*)malloc(MAX_LINKS*sizeof(int));
    weight[i] = (Real*)malloc(num_units*sizeof(Real));
  }

  created = TRUE;
  init_hist(&hist_self);
  init_hist(&hist_parent1);
  init_hist(&hist_parent2);
}

void Individual :: clear_net(void)
{
  int i, j;

  // clear fitness history?

  for (i=0; i<num_inputs; i++)
    unit_type[i] = TYPE_INPUT;
  for (; i<output_start; i++)
    unit_type[i] = TYPE_HIDDEN;
  for (; i<num_units; i++)
    unit_type[i] = TYPE_OUTPUT;

  for (i=0; i<num_units; i++) {
    function[i] = 0;
    num_in[i] = 0;
    active[i] = FALSE;
    switching[i] = FALSE;
    index_ptr1[i] = 0;
    index_ptr2[i] = 0;
    scale[i] = 0.0;
    bias[i] = 0.0;
    time_const[i] = 0.0;
    ctr_rate[i] = 0.0;
    value[i] = 0.0;
    value2[i] = 0.0;
  }

  for (i=0; i<num_units; i++)
    for (j=0; j<MAX_LINKS; j++)
      from_link[i][j] = 0;

  for (i=0; i<num_units; i++)
    for (j=0; j<num_units; j++)
      weight[i][j] = 0.0;
}


void Individual :: make_random(void)
{
  int i, j, num;

  evaluated = FALSE;
  clear_net();

  /*** Set function type. ***/
  for (i=num_inputs; i<num_units; i++)
    random_function(i);

  // Ensure at least one clock unit.
  // Also, ensures this clock is in same location
  // so as not to get problem of competing conventions.
  if (FUNCTION_CLOCK < Net_max_func)
    function[num_inputs] = FUNCTION_CLOCK;


  /*** Set activation (on/off) ***/
  // Input layer.
  for (i=0; i<num_inputs; i++)
    active[i] = TRUE;

  // Hidden layer.
  /*
  if (Net_cont_time) {
    for (i=num_inputs; i<output_start; i++) {
      if (random_int(2))
	active[i] = TRUE;
      else
	active[i] = FALSE;
    }
  } else {
    for (; i<output_start; i++)
      active[i] = TRUE;
  }
  */
  for (i=num_inputs; i<output_start; i++)
    active[i] = TRUE;


  // Output layer.
  for (; i<num_units; i++)
    active[i] = TRUE;



  /*** Set Properties. ***/
  // Input units have simple properties.
  for (i=0; i<num_inputs; i++) {
    bias[i] = 0.0;
    time_const[i] = 0.0;
    scale[i] = 0.0;
    ctr_rate[i] = 0.0;
    switching[i] = FALSE;
    index_ptr1[i] = 0;
    index_ptr2[i] = 0;
  }

  // Set scale & time_const.
  if (Net_cont_time) {
    for (i=num_inputs; i<num_units; i++)
      scale[i] = 1.0;
    //      scale[i] = reduce(0.8 + random_dbl() * 0.4);

    i = num_inputs;
    if (Net_cont_time == 2) {
      for (; i<num_inputs + num_hidden/3 + (num_hidden%3); i++)
	time_const[i] = 1.0; // First layer is non-continuous time.
    }
    for (; i<num_units; i++) {
      //      time_const[i] = reduce(1.0 - random_dbl() * random_dbl() * random_dbl());
      //      time_const[i] = reduce(1.0 - random_dbl() * random_dbl());
      time_const[i] = reduce(1.0 - random_dbl() * 0.5); // from -0.3
   }
  } else {
    for (i=num_inputs; i<num_units; i++) {
      scale[i] = 1.0;
      time_const[i] = 1.0;
    }
  }
  
  // Set counter rate.
  for (i=num_inputs; i<output_start; i++)
    ctr_rate[i] = reduce(random_dbl() * 0.4);
  for (; i<num_units; i++)
    ctr_rate[i] = 0.0;

  // Randomize switching behavior.
  if (Net_switching) {
    for (i=num_inputs; i<num_units; i++) {
      if (!random_int(4))
	switching[i] = TRUE;
      index_ptr1[i] = random_int(num_units);
      index_ptr2[i] = random_int(num_units-1);
      if (index_ptr1[i] == index_ptr2[i])
	index_ptr2[i] = num_units - 1;
    }
  }


  /*** Randomize connections. ***/
  if (Net_struct == NET_FEED_FORWARD) {
    /*** Input -> Hidden Layer ***/
    // Make sure all inputs are connected to a hidden unit.
    for (i=0; i<num_inputs; i++) {
      num = random_int(3) + 1;
      for (j=0; j<num; j++)
	connect(i, random_int(num_hidden) + num_inputs,
		random_weight());
    }

    // Make sure all hidden units are connected to an input.
    for (i=num_inputs; i<num_inputs + num_hidden; i++) {
      num = random_int(3) + 4;
      for (j=0; j<num; j++)
	connect(random_int(num_inputs), i, random_weight());
    }    

    /*** Hidden -> Output Layer ***/
    // Make sure all hidden units are connected to an output.
    for (i=num_inputs; i<output_start; i++) {
      num = random_int(3) + 1;
      for (j=0; j<num; j++)
	connect(i, random_int(num_outputs) + output_start,
		random_weight());
    }    

    // Make sure all output units are connected from a hidden.
    for (i=output_start; i<num_units; i++) {
      num = random_int(3) + 1;
      for (j=0; j<num; j++)
	connect(random_int(num_hidden) + num_inputs, i, random_weight());
    }

  } else if (Net_struct == NET_NO_INPUTS) {
    /*** Input -> Hidden Layer ***/
    // None.

    // Make sure all hidden units are connected to an input.
    // None.

    /*** Hidden -> Output Layer ***/
    // Make sure all hidden units are connected to an output.
    for (i=num_inputs; i<output_start; i++) {
      num = random_int(3) + 1;
      for (j=0; j<num; j++)
	connect(i, random_int(num_outputs) + output_start,
		random_weight());
    }    

    // Make sure all output units are connected from a hidden.
    for (i=output_start; i<num_units; i++) {
      num = random_int(3) + 1;
      for (j=0; j<num; j++)
	connect(random_int(num_hidden) + num_inputs, i, random_weight());
    }

  } else if (Net_struct == NET_RECURRENT) {
    /*** Input -> Hidden Layer ***/
    // Make sure all inputs are connected to a hidden unit.
    for (i=0; i<num_inputs; i++) {
      num = random_int(2) + 2;
      for (j=0; j<num; j++)
	connect(i, random_int(num_hidden) + num_inputs,
		random_weight());
    }

    // Make sure all hidden units are connected from an input.
    for (i=num_inputs; i<num_inputs + num_hidden; i++) {
      num = random_int(2) + 2;
      for (j=0; j<num; j++)
	connect(random_int(num_inputs), i, random_weight());
    }    

    /*** Hidden -> Output Layer ***/
    // Make sure all hidden units are connected to an output.
    for (i=num_inputs; i<output_start; i++) {
      num = random_int(2) + 1;
      for (j=0; j<num; j++)
	connect(i, random_int(num_outputs) + output_start,
		random_weight());
    }    

    // Make sure all output units are connected from a hidden.
    for (i=output_start; i<num_units; i++) {
      num = random_int(2) + 1;
      for (j=0; j<num; j++)
	connect(random_int(num_hidden) + num_inputs, i, random_weight());
    }

    /*** Output -> Hidden Layer ***/
    // Make sure all hidden units are connected from an output.
    for (i=num_inputs; i<output_start; i++) {
      num = random_int(2) + 2;
      for (j=0; j<num; j++)
	connect(random_int(num_outputs) + output_start, i,
		random_weight());
    }    

    // Make sure all output units are connected to a hidden.
    for (i=output_start; i<num_units; i++) {
      num = random_int(2) + 2;
      for (j=0; j<num; j++)
	connect(i, random_int(num_hidden) + num_inputs, 
		random_weight());
    }

  } else if (Net_struct == NET_FREE) {
    // Free structure.
    if (num_hidden > 0) {
      /*** Input -> Hidden Layer ***/
      // Make sure all inputs are connected to a hidden unit.
      for (i=0; i<num_inputs; i++) {
	num = random_int(2) + 2;
	for (j=0; j<num; j++)
	  connect(i, random_int(num_hidden) + num_inputs,
		  random_weight());
      }

      // Make sure all hidden units are connected from an input.
      for (i=num_inputs; i<num_inputs + num_hidden; i++) {
	num = random_int(3) + 2;
	for (j=0; j<num; j++)
	  connect(random_int(num_inputs), i, random_weight());
      }    

      /*** Input -> Output Layer ***/
      // Make sure all inputs are connected to an output unit.
      for (i=0; i<num_inputs; i++) {
	num = random_int(1);
	for (j=0; j<num; j++)
	  connect(i, random_int(num_outputs) + output_start,
		  random_weight());
      }

      // Make sure all output units are connected from an input.
      for (i=output_start; i<num_units; i++) {
	num = random_int(3) + 2;
	for (j=0; j<num; j++)
	  connect(random_int(num_inputs), i, random_weight());
      }    

      /*** Hidden -> Hidden Layer ***/
      // Make sure all hidden units are connected to a hidden.
      for (i=num_inputs; i<output_start; i++) {
	num = random_int(3) + 1;
	for (j=0; j<num; j++)
	  connect(i, random_int(num_hidden) + num_inputs,
		  random_weight());
      }    
    
      /*** Hidden -> Output Layer ***/
      // Make sure all hidden units are connected to an output.
      for (i=num_inputs; i<output_start; i++) {
	num = random_int(2) + 1;
	for (j=0; j<num; j++)
	  connect(i, random_int(num_outputs) + output_start,
		  random_weight());
      }    

      // Make sure all output units are connected from a hidden.
      for (i=output_start; i<num_units; i++) {
	num = random_int(2) + 2;
	for (j=0; j<num; j++)
	  connect(random_int(num_hidden) + num_inputs, i, random_weight());
      }

      /*** Output -> Output Layer ***/
      // Make sure all output units are connected to an output.
      for (i=output_start; i<num_units; i++) {
	num = random_int(1);
	for (j=0; j<num; j++)
	  connect(i, random_int(num_outputs) + output_start,
		  random_weight());
      }

      /*** Output -> Hidden Layer ***/
      // Make sure all hidden units are connected from an output.
      for (i=num_inputs; i<output_start; i++) {
	num = random_int(1) + 1;
	for (j=0; j<num; j++)
	  connect(random_int(num_outputs) + output_start, i,
		  random_weight());
      }
      
      // Make sure all output units are connected to a hidden.
      for (i=output_start; i<num_units; i++) {
	num = random_int(3) + 1;
      for (j=0; j<num; j++)
	connect(i, random_int(num_hidden) + num_inputs, 
		random_weight());
      }
    } else {  // no hidden layer
      /*** Input -> Output Layer ***/
      // Make sure all inputs are connected to an output unit.
      for (i=0; i<num_inputs; i++) {
	num = random_int(3) + 2;
	for (j=0; j<num; j++)
	  connect(i, random_int(num_outputs) + output_start,
		  random_weight());
      }

      // Make sure all output units are connected from an input.
      for (i=output_start; i<num_units; i++) {
	num = random_int(3) + 3;
	for (j=0; j<num; j++)
	  connect(random_int(num_inputs), i, random_weight());
      }

      /*** Output -> Output Layer ***/
      // Make sure all output units are connected to an output.
      for (i=output_start; i<num_units; i++) {
	num = random_int(3) + 3;
	for (j=0; j<num; j++)
	  connect(i, random_int(num_outputs) + output_start,
		  random_weight());
      }
    }
  } else if (Net_struct == NET_STRUCTURED) {
    // Structured network.
    int layer2 = num_hidden / 3;
    int layer1 = layer2 + (num_hidden % 3);
    int layer2_st = num_inputs + layer1;
    int layer3_st = layer2_st + layer2;

    /*** Input -> Hidden Layer 1 ***/
    if (Net_fix_hyperplanes) {
      // Make sure all inputs are connected to layer 1.
      for (i=0; i<num_inputs; i++) {
	num = random_int(3) + 1;
	for (j=0; j<num; j++)
	  connect(i, random_int(layer1) + num_inputs,
		  random_weight()*random_dbl());
      }

      // ** Make sure all hidden 1 are connect from input layer.
      // First, ensure links from multiple input groups.
      for (i=0; i<layer1; i++) {
	for (j=0; j<Net_groups; j++) {
	  connect(random_int(Net_group_size[j])+Net_group_start[j],
		  i+num_inputs, random_weight());
	}	  

	num = 0;
	//	num = random_int(2) + 1;
	for (j=0; j<num; j++)
	  connect(random_int(num_inputs), i+num_inputs,
		  random_weight()*random_dbl());
      }
    } else {
      // Make sure all inputs are connected to layer 1.
      for (i=0; i<num_inputs; i++) {
	num = random_int(2) + 2;
	for (j=0; j<num; j++)
	  connect(i, random_int(layer1) + num_inputs,
		  random_weight());
      }

      // Make sure all hidden 1 are connect from input layer.
      for (i=0; i<layer1; i++) {
	num = random_int(2) + 2;
	for (j=0; j<num; j++)
	  connect(random_int(num_inputs), i+num_inputs,
		  random_weight());
      }
    }

    /*** Layer 1 -> Layer 2 ***/
    // Make sure all hidden 1 units have an outgoing link.
    for (i=0; i<layer1; i++) {
      num = random_int(3) + 3;
      for (j=0; j<num; j++)
	connect(i+num_inputs,
		random_int(layer2)+layer2_st,
		random_weight());
    }

    // Make sure all hidden 2 units have an ingoing link.
    for (i=0; i<layer2; i++) {
      num = random_int(3) + 3;
      for (j=0; j<num; j++)
	connect(random_int(layer1)+num_inputs,
		i+layer2_st,
		random_weight());
    }


    /*** Layer 2 -> Layer 3 ***/
    // Make sure all hidden 2 units have an output to hidden 3.
    for (i=0; i<layer2; i++) {
      num = random_int(3) + 3;
      for (j=0; j<num; j++)
	connect(i+layer2_st,
		random_int(layer2)+layer3_st,
		random_weight());
    }

    // Make sure all hidden 3 units have an input from hidden 2.
    for (i=0; i<layer2; i++) {
      num = random_int(3) + 3;
      for (j=0; j<num; j++)
	connect(random_int(layer2)+layer2_st,
		i+layer3_st,
		random_weight());
    }

	
    /*** Layer 2 -> Outputs ***/
    // Make sure all hidden 2 units have an input from outputs.
    for (i=0; i<layer2; i++) {
      num = random_int(3) + 3;
      for (j=0; j<num; j++)
	connect(i+layer2_st,
		random_int(num_outputs)+output_start,
		random_weight());
    }
	
    // Make sure all outputs have an input from hidden 2.
    for (i=output_start; i<num_units; i++) {
      num = random_int(3) + 3;
      for (j=0; j<num; j++)
	connect(random_int(layer2)+layer2_st, i,
		random_weight());
    }
	

    /*** Layer 3a -> Layer 1 ***/
    for (i=0; i<layer2/2; i++) {
      num = random_int(3) + 3;
      for (j=0; j<num; j++)
	connect(i+layer3_st,
		random_int(layer1)+num_inputs,
		random_weight());
    }

    /*** Layer 3b -> Layer 2 ***/
    for (; i<layer2; i++) {
      num = random_int(3) + 3;
      for (j=0; j<num; j++)
	connect(i+layer3_st,
		random_int(layer2)+layer2_st,
		random_weight());
    }

  } else if (Net_struct == NET_STRUCTURED2) {
    // Structured network.
    int layer1 = num_hidden / 2 + (num_hidden % 2);
    int layer2 = num_hidden / 2;
    int layer2_st = num_inputs + layer1;

    /*** Input -> Hidden Layer 1 ***/
    // Make sure all inputs are connected to layer 1.
    for (i=0; i<num_inputs; i++) {
      num = random_int(2) + 2;
      for (j=0; j<num; j++)
	connect(i, random_int(layer1) + num_inputs,
		random_weight());
    }

    // Make sure all hidden 1 are connect from input layer.
    for (i=0; i<layer1; i++) {
      num = random_int(2) + 2;
      for (j=0; j<num; j++)
	connect(random_int(num_inputs), i+num_inputs,
		random_weight());
    }


    /*** Layer 1 -> Layer 2 ***/
    // Make sure all hidden 1 units have an outgoing link.
    for (i=0; i<layer1; i++) {
      num = random_int(3) + 3;
      for (j=0; j<num; j++)
	connect(i+num_inputs,
		random_int(layer2)+layer2_st,
		random_weight());
    }

    // Make sure all hidden 2 units have an ingoing link.
    for (i=0; i<layer2; i++) {
      num = random_int(3) + 3;
      for (j=0; j<num; j++)
	connect(random_int(layer1)+num_inputs,
		i+layer2_st,
		random_weight());
    }


    /*** Layer 1 -> Outputs ***/
    // Make sure all hidden 1 units have an output to outputs.
    for (i=0; i<layer1; i++) {
      num = random_int(3) + 3;
      for (j=0; j<num; j++)
	connect(i+num_inputs,
		random_int(num_outputs)+output_start,
		random_weight());
    }
	
    // Make sure all outputs have an input from hidden 1.
    for (i=output_start; i<num_units; i++) {
      num = random_int(3) + 3;
      for (j=0; j<num; j++)
	connect(random_int(layer1)+num_inputs, i,
		random_weight());
    }
	

    /*** Layer 2 -> Layer 1 ***/
    for (i=0; i<layer2; i++) {
      num = random_int(3) + 3;
      for (j=0; j<num; j++)
	connect(i+num_inputs+layer1,
		random_int(layer1)+num_inputs,
		random_weight());
    }
  } else {
    printf("Error: Net structure undefined.\n");
    return;
  }

  // Set bias.
  if (Net_fix_hyperplanes) {
    int f;
    Real max, min;

    for (i=num_inputs; i<num_units; i++) {
      max = 0.0;
      for (j=0; j<num_in[i]; j++) {
	f = from_link[i][j];
	if (weight[i][f] < 0.0)
	  max += absv(weight[i][f]*Min_output[f]);
	else
	  max += absv(weight[i][f]*Max_output[f]);	
      }

      min = 0.0;
      for (j=0; j<num_in[i]; j++) {
	f = from_link[i][j];
	if (weight[i][f] > 0.0)
	  min += absv(weight[i][f]*Min_output[f]);
	else
	  min += absv(weight[i][f]*Max_output[f]);	
      }

      //      cout << "  Max:" << max;
      //      cout << "  Min:" << min;

      if (max > 1.0) max = 1.0;
      if (min > 1.0) min = 1.0;
      bias[i] = (random_dbl() * (min+max)) - min;

      //      cout << "  Bias:" << bias[i] << endl;
    }
  } else {
    for (i=num_inputs; i<num_units; i++)
      bias[i] = reduce(2.0 * random_dbl() - 1.0);
  }
}


void Individual :: init(int inputs, int hidden, int outputs)
{
  num_inputs = inputs;
  num_hidden = hidden;
  num_outputs = outputs;
  output_start = num_inputs + num_hidden;
  output_stop = output_start + num_outputs;
  init();
  clear_net();
  //  make_random();
}

void Individual :: init_hist(history *hist)
{
  int i;

  hist->index = -1;
  hist->operation = OP_INIT;
  hist->how_created = CREATE_INIT;
  hist->num_mutate = 0;
  hist->num_recombine = 0;
  hist->delta = 0.0;
  hist->op1 = 0;
  hist->op2 = 0;
  hist->op3 = 0;
  hist->fitness = 0.0;
  hist->fitness_prev = 0.0;

  for (i=0; i<7; i++)
    hist->op_count[i] = 0;
}


void Individual :: init_hist(void)
{
  int i;

  hist_self.index = -1;
  hist_self.operation = OP_INIT;
  hist_self.how_created = CREATE_INIT;
  hist_self.num_mutate = 0;
  hist_self.num_recombine = 0;
  hist_self.delta = 0.0;
  hist_self.op1 = 0;
  hist_self.op2 = 0;
  hist_self.op3 = 0;
  hist_self.fitness_prev = 0.0;

  for (i=0; i<7; i++)
    hist_self.op_count[i] = 0;
}


int Individual :: create_mutate(void)
{
  if (hist_self.how_created == CREATE_MUTATE)
    return TRUE;

  return FALSE;
}

// Returns TRUE if individual has higher fitness than its first parent.
int Individual :: better(void)
{
  if (create_mutate()) {
    if (hist_self.fitness > hist_parent1.fitness)
      return 1;
    else if (hist_self.fitness == hist_parent1.fitness)
      return 0;
    else
      return -1;
  } else {
    if ((hist_self.fitness > hist_parent1.fitness) &&
	(hist_self.fitness > hist_parent2.fitness))
      return 1;
    else if ((hist_self.fitness == hist_parent1.fitness) ||
	     (hist_self.fitness == hist_parent2.fitness))
      return 0;
    else if ((hist_self.fitness > hist_parent1.fitness) ||
	     (hist_self.fitness > hist_parent2.fitness))
      return 0;
    else
      return -1;
  }
}

Real Individual :: get_fitness(void)
{
  return hist_self.fitness;
}

void Individual :: set_fitness(Real val)
{
  hist_self.fitness = val;
}

void Individual :: add_fitness(Real val)
{
  hist_self.fitness += val;
}

void Individual :: div_fitness(Real val)
{
  hist_self.fitness /= val;
}


void Individual :: set_creation(int type)
{
  hist_self.how_created = type;
}

int Individual :: get_creation(void)
{
  return hist_self.how_created;
}


void Individual :: duplicate_hist(history *hist_from, history *hist_to)
{
  int i;

  hist_to->index = hist_from->index;
  hist_to->operation = hist_from->operation;
  hist_to->how_created = hist_from->how_created;
  hist_to->num_mutate = hist_from->num_mutate;
  hist_to->num_recombine = hist_from->num_recombine;
  hist_to->delta = hist_from->delta;
  hist_to->fitness = hist_from->fitness;

  hist_to->op1 = hist_from->op1;
  hist_to->op2 = hist_from->op2;
  hist_to->op3 = hist_from->op3;
  hist_to->val1 = hist_from->val1;
  for (i=0; i<7; i++)
    hist_to->op_count[i] = hist_from->op_count[i];
}
  
void Individual :: duplicate(Individual *ind)
{
  int i, j;

  if (!created) {
    init(ind->num_inputs, ind->num_hidden, ind->num_outputs);
  }

  evaluated = ind->evaluated;
  duplicate_hist(&(ind->hist_self), &hist_self);
  duplicate_hist(&(ind->hist_parent1), &hist_parent1);
  duplicate_hist(&(ind->hist_parent2), &hist_parent2);

  for (i=0; i<num_units; i++) {
    unit_type[i] = ind->unit_type[i];
    function[i] = ind->function[i];
    num_in[i] = ind->num_in[i];
    active[i] = ind->active[i];
    switching[i] = ind->switching[i];
    index_ptr1[i] = ind->index_ptr1[i];
    index_ptr2[i] = ind->index_ptr2[i];
    scale[i] = ind->scale[i];
    bias[i] = ind->bias[i];
    time_const[i] = ind->time_const[i];
    ctr_rate[i] = ind->ctr_rate[i];
  }

  for (i=0; i<num_units; i++)
    for (j=0; j<MAX_LINKS; j++)
      from_link[i][j] = ind->from_link[i][j];

  for (i=0; i<num_units; i++)
    for (j=0; j<num_units; j++)
      weight[i][j] = ind->weight[i][j];
}


// Mutate a weight connecting to _index_.
void Individual :: mutate_weight(int index, Real alpha)
{
  int i, l;

  if (num_in[index] == 0)
    return;
  
  i = random_int(num_in[index]);
  l = from_link[index][i];

  weight[index][l] += random_norm(0.0, WEIGHT_DELTA*alpha);
  weight[index][l] = reduce(weight[index][l]);
}

void Individual :: mutate_weights(int print_flag, Real alpha)
{
  int i, j, num1, num2, index;
  Real val;

  val = random_norm(0.0, 1.0) * Real(MUT_WEIGHT_NUM);
  if (val < 0.0)
    val = -val;
  num1 = (int)val;
  if (num1 == 0)
    num1 = 1;

  num2 = random_int(MUT_PROPERTY_NUM) + 1;

  hist_self.op1 = num1 + num2;
  //  if (Print[8] && print_flag)
  //    printf("(weights %2d): ", num1 + num2);

  // Mutate bias.
  for (i=0; i<num2; i++) {
    index = random_int(num_units - num_inputs) + num_inputs;
    bias[index] += random_norm(0.0, PROPERTY_DELTA*alpha);
    bias[index] = reduce(bias[index]);
  }    

  // Mutate weights.
  for (i=0; i<num1; i++) {
    index = random_int(num_units - num_inputs) + num_inputs;
    mutate_weight(index, alpha);
  }
}

void Individual :: mutate_weights(Real alpha)
{
  mutate_weights(FALSE, alpha);
}


Real Individual :: random_new_weight(void)
{
  //  return random_weight();
  //  return reduce(MAX_WEIGHT - 2.0*random_dbl()*MAX_WEIGHT);
  //  return reduce(random_norm(0.0, MAX_WEIGHT)/2.0);
  return reduce(random_norm(0.0, MAX_WEIGHT) * random_dbl() / 4.0);
}

Real Individual :: random_weight(void)
{
  //  return reduce((2.0*random_dbl() - 1.0) * random_dbl());
  //  return reduce((2.0*random_dbl()*2.0 - 2.0) * random_dbl());
  return reduce((2.0*random_dbl()*3.0 - 3.0) * random_dbl());
  //  return reduce((2.0*random_dbl()*MAX_WEIGHT - MAX_WEIGHT) * random_dbl());
  //  return reduce(random_norm(0.0, 2.0) * random_dbl() / 2.0) ;
  //  return reduce((MAX_WEIGHT - 2.0*random_dbl()*MAX_WEIGHT)*1.3);
  //  return reduce(random_norm(0.0, MAX_WEIGHT) / 2.0);

  /*
  Real w = reduce(MAX_WEIGHT - random_dbl() * random_dbl() * MAX_WEIGHT);

  if (random_int(2))
    return w;
  else
    return -w;
    */
  //  return reduce((2.0*random_dbl()*MAX_WEIGHT - MAX_WEIGHT) * random_dbl());
  //  return reduce(2.0*random_dbl()*MAX_WEIGHT - MAX_WEIGHT);
}


int Individual :: connect(int f, int t, Real w)
{
  if (hist_self.how_created == CREATE_MUTATE) {
    hist_self.val1 = w;
  }

  if (num_in[t] == MAX_LINKS)
    return FALSE;

  if (weight[t][f] != 0.0)
    return FALSE;

  weight[t][f] = w;
  from_link[t][num_in[t]++] = f;
  return TRUE;
}


void Individual :: random_connect_ff(int index, Real wght)
{
  int l, ll;

  if (num_in[index] == MAX_LINKS)
    return;

  if (unit_type[index] == TYPE_HIDDEN)
    l = random_int(num_inputs);
  else
    l = random_int(num_hidden) + num_inputs;

  ll = l;
  while (weight[index][l] != 0.0) {
    ++l;
    if (l == num_units)
      l = output_start;
    else if (l == output_start)
      l = num_hidden;
    if (l == ll)
      return;
  }

  connect(l, index, wght);
}


// Random connect for a recurrent network.
void Individual :: random_connect_rec(int index, Real wght)
{
  int l, ll;

  if (num_in[index] == MAX_LINKS)
    return;

  if (unit_type[index] == TYPE_HIDDEN) {
    l = random_int(num_inputs + num_outputs);
    if (l >= num_inputs)
      l += num_hidden;
  } else
    l = random_int(num_hidden) + num_inputs;

  ll = l;
  while (weight[index][l] != 0.0) {
    ++l;
    if (l == ll)
      return;
    else if (l == num_units) {
      l = 0;
    } else if (l == output_start) {
      l = num_hidden;
    } else if (l == num_inputs) {
      l = output_start;
    }
  }

  connect(l, index, wght);
}

void Individual :: random_connect_struct(int index, Real wght)
{
  int layer2 = num_hidden / 3;
  int layer1 = layer2 + (num_hidden % 3);
  int layer2_st = num_inputs + layer1;
  int layer3_st = layer2_st + layer2;
  int l;


  if (index < num_inputs + layer1) {
    // Hidden 1: Connect from either input or layer 3a.
    l = random_int(num_inputs + layer2/2);
    if (l >= num_inputs)
      l+= layer1 + layer2;
  } else if (index < num_inputs + layer1 + layer2) {
    // Hidden 2: Connect from hidden 1 or layer 3b.
    l = random_int(layer1 + (layer2 - layer2/2));
    if (l < layer1)
      l += num_inputs;
    else
      l += num_inputs + layer2 + layer2/2;
  } else {
    // Hidden 3: Connect from hidden 2; or
    // Output: also Connect from hidden 2.
    l = random_int(layer2) + num_inputs + layer1;
  }

  //  cout << "Random connect struct: " << l << " -> " << index << "\n";
  connect(l, index, wght);  
}

void Individual :: random_connect_struct2(int index, Real wght)
{
  int layer1 = num_hidden / 2 + (num_hidden % 2);
  int layer2 = num_hidden / 2;
  int layer2_st = num_inputs + layer1;
  int l;


  if (index < num_inputs + layer1) {
    // Hidden 1: Connect from either input or layer 2.
    l = random_int(num_inputs + layer2);
    if (l >= num_inputs)
      l += layer1;
  } else if (index < output_start) {
    // Hidden 2: Connect from hidden 1 or hidden 2.
    l = random_int(num_hidden) + num_inputs;
  } else {
    // Output: connect from hidden 1.
    l = random_int(layer1) + num_inputs;
  }

  //  cout << "Random connect struct: " << l << " -> " << index << "\n";
  connect(l, index, wght);  
}


// Random_connect: connects from a random unit to index with
// weight wght.
void Individual :: random_connect(int index, Real wght)
{
  int l, ll;

  if (Net_struct == NET_FEED_FORWARD) {
    random_connect_ff(index, wght);
    return;
  } else if (Net_struct == NET_RECURRENT) {
    random_connect_rec(index, wght);
    return;
  } else if (Net_struct == NET_STRUCTURED) {
    random_connect_struct(index, wght);
    return;
  } else if (Net_struct == NET_STRUCTURED2) {
    random_connect_struct2(index, wght);
    return;
  } else if (Net_struct == NET_NO_INPUTS) {
    return;
  }


  if (num_in[index] == MAX_LINKS)
    return;

  l = random_int(num_units);
  ll = l;
  while (weight[index][l] != 0.0) {
    if (++l == num_units)
      l = 0;
    if (l == ll)
      return;
  }

  connect(l, index, wght);
}


int Individual :: reduce_link(int index)
{
  int l;

  if (num_in[index] == 0)
    return FALSE;

  l = random_int(num_in[index]);
  hist_self.val1 = weight[index][from_link[index][l]];
  weight[index][from_link[index][l]] *= random_dbl();
  return TRUE;
}

int Individual :: delete_link(int index)
{
  int l;

  if (num_in[index] == 0)
    return FALSE;

  l = random_int(num_in[index]);
  hist_self.val1 = weight[index][from_link[index][l]];
  weight[index][from_link[index][l]] = 0.0;
  
  if (num_in[index]-- < MAX_LINKS) {
    for (; l<num_in[index]; l++)
      from_link[index][l] = from_link[index][l+1];
  }
  from_link[index][l] = 0;
  return TRUE;
}


void Individual :: mutate_links(Real alpha)
{
  int i, ctr1 = 0, ctr2 = 0, num, index, index2;
  Real val;

  hist_self.operation = OP_LINKS;
  hist_self.op1 = 0;
  hist_self.op2 = 0;
  hist_self.val1 = 0.0;
  /*
  val = random_norm(0.0, 0.5) * (Real)MUT_LINK_NUM;
  num = (int)val;
  if (num == 0) {
    if (val < 0.0)
      num = -1;
    else
      num = 1;
  }
  */

  num = random_int(3);
  if (num == 0)
    num = 2;
  else
    num = 1;

  if (random_int(2))
    num = -num;

  if (num > 0) {
    for (i=0; i<num; i++) {
      index = random_int(num_units - num_inputs) + num_inputs;
      val = random_new_weight();
      hist_self.val1 = val;
      random_connect(index, val);
    }
  } else {
    for (i=num; i<0; i++) {
      index = random_int(num_units - num_inputs) + num_inputs;
      //      if (delete_link(index))
      //	ctr1++;
      if (random_int(2)) {
	hist_self.op3 = TRUE;
 	if (delete_link(index))
	  ctr1++;
      } else {
	hist_self.op3 = FALSE;
	if (reduce_link(index))
	  ctr2++;
      }
    }
  }      

  if (Print[8]) {
    if (num >= 0) {
      hist_self.op1 = num;
      //      printf("(lnk    %2d):\n", num);
    } else {
      hist_self.op1 = -ctr1;
      hist_self.op2 = -ctr2;
      //      printf("(lnk    %2d,%2d):\n", -ctr1, -ctr2);
    }
  }
}


void Individual :: random_function(int index)
{
  int min;

  if (Net_max_func == 1) {
    function[index] = 0;
    return;
  }

  if (unit_type[index] == TYPE_OUTPUT) {
    if (Net_max_func < UNIT_MAX_OUTPUT_FUNC)
      min = Net_max_func;
    else
      min = UNIT_MAX_OUTPUT_FUNC;

    if (FUNCTION_SIGMOID2 < min) {
      if (FUNCTION_CLOCK < min) {
	function[index] = random_int(min-2);
	if (function[index] == FUNCTION_CLOCK)
	  function[index] = min - 1;
	else if (function[index] == FUNCTION_SIGMOID2)
	  function[index] = min - 2;
      } else {
	function[index] = random_int(min-1);
	if (function[index] == FUNCTION_SIGMOID2)
	  function[index] = min - 1;
      }
    } else if (FUNCTION_CLOCK < min) {
      function[index] = random_int(min-1);
      if (function[index] == FUNCTION_CLOCK)
	function[index] = min - 1;
    } else
      function[index] = random_int(min);

  } else if (unit_type[index] == TYPE_HIDDEN) {
    if ((Net_max_func == 2) && (FUNCTION_CLOCK == 1)) {
      function[index] = random_int(15);
      if (function[index] > 1)
	function[index] = 0;
    } else {
      if (FUNCTION_CLOCK < Net_max_func) {
	if (!random_int(15))
	  function[index] = FUNCTION_CLOCK;
	else {
	  function[index] = random_int(Net_max_func-1);
	  if (function[index] == FUNCTION_CLOCK) {
	    function[index] = Net_max_func - 1;
	  }
	}
      } else {
	function[index] = random_int(Net_max_func);
      }
    }
  }
}



int Individual :: mutate_properties_all(int index, Real alpha)
{
  int change = FALSE;

  if (Net_cont_time) {
    change = TRUE;

    scale[index] += random_norm(0.0, PROPERTY_DELTA*alpha);
    scale[index] = reduce(scale[index]);

    time_const[index] += random_norm(0.0, PROPERTY_DELTA*alpha);
    if (time_const[index] < 0.0)
      time_const[index] = 0.0;
    else if (time_const[index] > 1.0)
      time_const[index] = 1.0;
    else
      time_const[index] = reduce(time_const[index]);
  }

  if (function[index] == FUNCTION_CLOCK) {
    change = TRUE;

    ctr_rate[index] += random_norm(0.0, PROPERTY_DELTA*alpha);
    if (ctr_rate[index] < 0.0)
      ctr_rate[index] = 0.0;
    else if (ctr_rate[index] > 1.0)
      ctr_rate[index] = 1.0;
    else
      ctr_rate[index] = reduce(ctr_rate[index]);
  }

  if (Net_max_func > 1) {
    if (!random_int(4)) {
      change = TRUE;
      random_function(index);
    }
  }

  return change;
}


// One idea is to mutate ALL properties.
int Individual :: mutate_properties(int index, Real alpha)
{
  float val;
  float scale_p = 0.0;
  float tconst_p = 0.2;
  float clock_p = 0.2;
  float func_p = 0.2;

  if (!Net_cont_time && Net_max_func == 1)
    return FALSE;

  if (!Net_cont_time) {
    scale_p = 0.0;
    tconst_p = 0.0;
  } else if ((Net_cont_time == 3) && (index < num_inputs + num_hidden/3))
    tconst_p = 0.0;

  if (Net_max_func == 1)
    func_p = 0.0;
  if (index >= output_start)
    clock_p = 0.0;

  val = scale_p + tconst_p + clock_p + func_p;
  val *= random_dbl();

  tconst_p += scale_p;
  clock_p += tconst_p;
  func_p += clock_p;

  if (val < scale_p) {
    scale[index] += random_norm(0.0, PROPERTY_DELTA*alpha);
    scale[index] = reduce(scale[index]);
  } else if (val < tconst_p) {
    time_const[index] += random_norm(0.0, PROPERTY_DELTA*alpha);
    if (time_const[index] < 0.0)
      time_const[index] = 0.0;
    else if (time_const[index] > 1.0)
      time_const[index] = 1.0;
    else
      time_const[index] = reduce(time_const[index]);
  } else if (val < clock_p) {
    ctr_rate[index] += random_norm(0.0, PROPERTY_DELTA*alpha);
    if (ctr_rate[index] < 0.0)
      ctr_rate[index] = 0.0;
    else if (ctr_rate[index] > 1.0)
      ctr_rate[index] = 1.0;
    else
      ctr_rate[index] = reduce(ctr_rate[index]);
  } else {
    random_function(index);
  }

  if (!random_int(10) && Net_switching) {
    val = random_dbl();
    if (val < 0.5)
      switching[index] = !switching[index];
    else if (val < 0.75) {
      if (random_int(2)) {
	if (++index_ptr1[index] >= num_units)
	  index_ptr1[index] = 0;
      } else {
	if (--index_ptr1[index] < 0)
	  index_ptr1[index] = num_units - 1;
      }
    } else {
      if (random_int(2)) {
	if (++index_ptr2[index] >= num_units)
	  index_ptr2[index] = 0;
      } else {
	if (--index_ptr2[index] < 0)
	  index_ptr2[index] = num_units - 1;
      }
    }
  }
  return TRUE;
}


void Individual :: mutate_properties(Real alpha)
{
  int i, index, num, ctr = 0;

  hist_self.operation = OP_PROPERTIES;
  num = random_int(MUT_PROPERTY_NUM) + 1;

  for (i=0; i<num; i++) {
    index = num_inputs+random_int(num_hidden + num_outputs);
    if (mutate_properties(index, alpha))
      ctr++;
  }

  hist_self.op1 = num;
  hist_self.op2 = ctr;
}


void Individual :: mutate(Real alpha)
{
  Real choice;

  evaluated = FALSE;
  choice = random_dbl();
  alpha *= random_norm(0.0, 1.0) * random_dbl() / 2.0;

  duplicate_hist(&hist_self, &hist_parent1);
  init_hist(&hist_parent2);

  hist_self.num_mutate++;
  hist_self.how_created = CREATE_MUTATE;
  hist_self.operation = OP_WEIGHTS;
  hist_self.delta = alpha;

  if (!Net_cont_time && !Net_switching) {
    if (Net_max_func == 1)
      choice *= 0.8;

    if (choice < 0.45) {
      hist_self.op_count[0]++;
      mutate_weights(TRUE, alpha);
    } else if (choice < 0.8) {
      hist_self.op_count[1]++;
      mutate_weights(alpha);
      mutate_links(alpha);
    } else {
      hist_self.op_count[2]++;
      mutate_weights(alpha);
      mutate_properties(alpha);
    }
  } else {
    if (choice < 0.4) { // 0.4
      hist_self.op_count[0]++;
      mutate_weights(TRUE, alpha);
    } else {
      mutate_weights(alpha);
      if (choice < 0.7 ) {  // 0.7
	hist_self.op_count[1]++;
	mutate_links(alpha);
      } else {
	hist_self.op_count[2]++;
	mutate_properties(alpha);
      }
    }
  }
}


void Individual :: mutate(void)
{
  mutate(1.0);
}


Real Individual :: recombine(Real val1, Real val2)
{
  if (val1 == 0.0) {
    if (random_int(2))
      return val1;
    else
      return val2;
  } else if (val2 == 0.0) {
    if (random_int(2))
      return val1;
    else
      return val2;
  }

  Real diff;

  diff = Rec_delta * (val1 - val2);
  return reduce(val1 + diff * (random_dbl() * 2.0 - 1.0));
}


void Individual :: configure_links(int num, int index)
{
  int l;

  for (; num > MAX_LINKS; num--) {
    l = random_int(num_units);
    while (weight[index][l] == 0.0) {
      if (++l == num_units)
	l = 0;
    }
    weight[index][l] = 0.0;
  }

  num_in[index] = 0;
  for (l=0; l<num_units; l++) {
    if (weight[index][l] != 0.0)
      from_link[index][num_in[index]++] = l;
  }
}


int Individual :: recombine_weights(int print_flag,
				    Individual *parent2)
{
  int i, j, num = 0;

  // Recombine bias:  this can be considered an link from
  // a special unit that always has value 1.
  for (i=num_inputs; i<num_units; i++) {
    if (function[i] != FUNCTION_CLOCK)
      bias[i] = recombine(bias[i], parent2->bias[i]);
  }

  // Recombine Weight Matrix
  for (i=num_inputs; i<num_units; i++)
    for (j=0; j<num_units; j++) {
      if ((weight[i][j] != 0.0) &&
	  (parent2->weight[i][j] != 0.0) &&
	  (weight[i][j] != parent2->weight[i][j])) {
	if (random_int(2)) {
	  num++;
	  weight[i][j] = recombine(weight[i][j],
				   parent2->weight[i][j]);
	}
      }
    }


  hist_self.op1 = num;

  if (num == 0)
    return FALSE;
  return TRUE;
}


int Individual :: recombine_weights(Individual *parent2)
{
  return recombine_weights(FALSE, parent2);
}


int Individual :: recombine_links(int terse, Individual *parent2)
{
  int i, j, k, ctr = 0, num, index, links;
  int new_total, new_ctr;
  int swap_total, swap_ctr;

  // Recombine Weight Matrix

  hist_self.operation = OP_LINKS;

  new_total = 0;
  swap_total = 0;
  num = random_int(MAX_LINKS_RECOMBINE) + 1;
  for (i=0; i<num; i++) {
    new_ctr = 0;
    swap_ctr = 0;

    for (k=0; k<4; k++) {
      if (new_ctr != 0)
	break;

      links = 0;
      index = random_int(num_units - num_inputs) + num_inputs;
      for (j=0; j<num_units; j++) {
	if (weight[index][j] == parent2->weight[index][j])
	  continue;

	if ((weight[index][j] != 0.0) ||
	    (parent2->weight[index][j] != 0.0)) {
	  if (random_int(2)) {
	    if (weight[index][j] * parent2->weight[index][j] == 0.0)
	      new_ctr++;
	    else if (k > 0)
	      continue;
	    else
	      swap_ctr++;
	    ctr++;
	    weight[index][j] = parent2->weight[index][j];
	  }
	}
      }
    }

    // Test that there are not too many links.
    for (j=0; j<num_units; j++)
      if (weight[index][j] != 0)
	links++;
    configure_links(links, index);

    for (j=0; j<num_units; j++) {
      if (weight[index][j] != parent2->weight[index][j])
	break;
    }
    if (j == num_units)
      mutate_weight(index, 0.1);

    new_total += new_ctr;
    swap_total += swap_ctr;

    if (new_total >= MAX_LINKS_RECOMBINE)
      break;
  }      

  hist_self.op1 = num;
  hist_self.op2 = swap_total;
  hist_self.op3 = new_total;

  if (ctr == 0)
    return FALSE;
  return TRUE;
}
  

int Individual :: recombine_propertiesA(Individual *parent2)
{
  int rec_flag, tries;
  int i, j, ctr = 0, num, index;
  float val;

  num = random_int(MAX_UNITS_RECOMBINE) + 1;

  for (i=0; i<num; i++) {
    rec_flag = FALSE;

    for (j=0; j<2; j++) {
      index = random_int(num_units - num_inputs) + num_inputs;
      tries = 0;
      while ((tries++ < 5) && (!rec_flag)) {
	val = random_dbl();
	if (val <= 0.3) {
	  if (scale[index] != parent2->scale[index]) {
	    rec_flag = TRUE;
	    scale[index] = recombine(scale[index], parent2->scale[index]);
	  }
	} else if (val <= 0.6) {
	  if (time_const[index] != parent2->time_const[index]) {
	    rec_flag = TRUE;
	    time_const[index] = recombine(time_const[index], parent2->time_const[index]);
	  }
	} else if ((val <= 0.8) || (Net_max_func == 1)) {
	  if (active[index] != parent2->active[index]) {
	    rec_flag = TRUE;
	    active[index] = parent2->active[index];
	  }
	} else {
	  if (function[index] != parent2->function[index]) {
	    rec_flag = TRUE;
	    function[index] = parent2->function[index];
	  }
	}
      }
    }

    if (rec_flag)
      ctr++;
  }

  if (Print[8])
    printf("(prop A %2d):  ", ctr);

  if (ctr == 0)
    return FALSE;
  return TRUE;
}


int Individual :: recombine_propertiesB(int terse, Individual *parent2)
{
  int rec_flag;
  int i, j, ctr = 0, num, index;
  float val;

  hist_self.operation = OP_PROPERTIES;
  num = random_int(MAX_UNITS_RECOMBINE) + 2;

  for (i=0; i<num; i++) {
    rec_flag = FALSE;

    for (j=0; j<5; j++) {
      index = random_int(num_units - num_inputs) + num_inputs;

      if (scale[index] != parent2->scale[index]) {
	rec_flag = TRUE;
	scale[index] = recombine(scale[index], parent2->scale[index]);
      }

      if (time_const[index] != parent2->time_const[index]) {
	rec_flag = TRUE;
	time_const[index] = recombine(time_const[index], parent2->time_const[index]);
	if (time_const[index] < 0.0)
	  time_const[index] = 0.0;
	else if (time_const[index] > 1.0)
	  time_const[index] = 1.0;
      }

      if (function[index] == FUNCTION_CLOCK) {
	if (ctr_rate[index] != parent2->ctr_rate[index]) {
	  rec_flag = TRUE;
	  ctr_rate[index] = recombine(ctr_rate[index], parent2->ctr_rate[index]);
	  if (ctr_rate[index] < 0.0)
	    ctr_rate[index] = 0.0;
	  else if (ctr_rate[index] > 1.0)
	    ctr_rate[index] = 1.0;
	}
      }

      if (unit_type[index] == TYPE_HIDDEN) {
	if (active[index] != parent2->active[index]) {
	  if (random_int(2)) {
	    rec_flag = TRUE;
	    active[index] = parent2->active[index];
	  }
	}
      }

      if (function[index] != parent2->function[index]) {
	if (random_int(2)) {
	  rec_flag = TRUE;
	  function[index] = parent2->function[index];
	}
      }

      if (rec_flag) {
	ctr++;
	break;
      }
    }
  }

  hist_self.op1 = ctr;

  if (ctr == 0)
    return FALSE;
  return TRUE;
}


int Individual :: recombine_propertiesC(Individual *parent2)
{
  int rec_flag;
  int i, j, ctr = 0, num, index;
  float val;

  num = random_int(MAX_UNITS_RECOMBINE) + 1;

  for (i=0; i<num; i++) {
    rec_flag = FALSE;

    for (j=0; j<5; j++) {
      index = random_int(num_units - num_inputs) + num_inputs;

      if (scale[index] != parent2->scale[index]) {
	rec_flag = TRUE;
	scale[index] = parent2->scale[index];
      }

      if (time_const[index] != parent2->time_const[index]) {
	rec_flag = TRUE;
	time_const[index] = parent2->time_const[index];
      }
      
      if (unit_type[index] == TYPE_HIDDEN) {
	if (active[index] != parent2->active[index]) {
	  rec_flag = TRUE;
	  active[index] = parent2->active[index];
	}
      }

      if (function[index] != parent2->function[index]) {
	rec_flag = TRUE;
	function[index] = parent2->function[index];
      }

      if (rec_flag == TRUE)
	break;
    }

    if (rec_flag)
      ctr++;
  }

  hist_self.op1 = ctr;

  if (ctr == 0)
    return FALSE;
  return TRUE;
}


int Individual :: recombine_properties(Individual *parent2)
{
  int val = random_int(2);

  return recombine_propertiesB(FALSE, parent2);

  if (val == 0)
    return recombine_propertiesA(parent2);
  else if (val == 1)
    return recombine_propertiesC(parent2);

  return recombine_propertiesB(FALSE, parent2);
}


void Individual :: recombine(Individual *parent2)
{
  Real val = random_dbl();

  // Assume that the current individual is already a copy of parent1.
  evaluated = FALSE;
  duplicate_hist(&hist_self, &hist_parent1);
  duplicate_hist(&(parent2->hist_self), &hist_parent2);
  hist_self.num_recombine++;
  hist_self.how_created = CREATE_RECOMBINE;
  hist_self.operation = OP_WEIGHTS;

  if (!Net_cont_time) {
    if (Net_max_func == 1)
      val *= 0.8;

    if (val < 0.45) {
      hist_self.op_count[3]++;
      recombine_weights(TRUE, parent2);
    } else if (val < 0.8) {
      hist_self.op_count[4]++;
      recombine_weights(parent2);
      recombine_links(FALSE, parent2);
    } else {
      hist_self.op_count[5]++;
      recombine_weights(parent2);
      recombine_propertiesB(FALSE, parent2);
    }
  } else {
    if (val < 0.4) { // 0.4
      hist_self.op_count[3]++;
      recombine_weights(TRUE, parent2);
    } else if (val < 0.7) {  // 0.7
      hist_self.op_count[4]++;
      recombine_weights(parent2);
      recombine_links(FALSE, parent2);
    } else {
      hist_self.op_count[5]++;
      recombine_weights(parent2);
      recombine_propertiesB(FALSE, parent2);
    }
  }
}



Real Individual::sum_inputs(int index)
{
  int i, f;
  float sum = bias[index];

  for (i=0; i<num_in[index]; i++) {
    f = from_link[index][i];
    sum += value[f] * weight[index][f];
  }

  if (Print[10])
    printf(" %10.6f ", sum * scale[index]);

  return sum * scale[index];
}


void Individual :: reset(void)
{
  int i;
  for (i=0; i<num_units; i++)
    value[i] = 0.0;
}


void Individual :: next_iteration(void)
{
  int index, i, j, f;
  Real new_val, val;

  for (index=num_inputs; index<num_units; index++) {
    if ((!active[index]) && (unit_type[index] == TYPE_HIDDEN))
      continue;

    if (Print[10])
      printf("Value %2d:", index);

    switch (function[index])
      {
      case FUNCTION_SIGMOID:
	new_val = tanh(sum_inputs(index));
	break;
      
      case FUNCTION_SIGMOID2:
	new_val = 1.0 / (1.0 + exp(- sum_inputs(index)));
	break;
      
      case FUNCTION_SIN:
	new_val = sin(sum_inputs(index));
	break;
      
      case FUNCTION_ASIN:
	new_val = sum_inputs(index);
	if (new_val > 1.0)
	  new_val = 1.0;
	else if (new_val < -1.0)
	  new_val = -1.0;
	new_val = asin(new_val);
	break;

      case FUNCTION_SQRT:
	new_val = sum_inputs(index);
	if (new_val < 0.0)
	  new_val = - sqrt(-new_val);
	else
	  new_val = sqrt(new_val);
	break;

      case FUNCTION_SQR:
	new_val = sum_inputs(index);
	new_val = sqr(new_val);
	break;

      case FUNCTION_MIN:
	new_val = bias[index]/scale[index];
	for (i=0; i<num_in[index]; i++) {
	  f = from_link[index][i];
	  val = value[f] * weight[index][f];
	  if (val < new_val)
	    new_val = val;
	}
	new_val *= scale[index];
	break;

      case FUNCTION_MAX:
	new_val = bias[index]/scale[index];
	for (i=0; i<num_in[index]; i++) {
	  f = from_link[index][i];
	  val = value[f] * weight[index][f];
	  if (val > new_val)
	    new_val = val;
	}
	new_val *= scale[index];
	break;

      case FUNCTION_SIGN:
	new_val = sum_inputs(index);
	if (new_val < 0.0)
	  new_val = -1.0;
	else if (new_val > 0.0)
	  new_val = 1.0;
	new_val *= scale[index];
	break;

      case FUNCTION_SUM:
	new_val = sum_inputs(index);
	break;

      case FUNCTION_CLOCK:
	new_val = value[index] + ctr_rate[index];
	if (new_val > MAX_VALUE)
	  new_val = 0.0;
	else if (new_val < -MAX_VALUE)
	  new_val = 0.0;
	break;

      case FUNCTION_RANDOM:
	new_val = bias[index] + 2.0 * random_dbl() - 1.0;
	new_val *= scale[index];
	break;
      }

    if (Print[10])
      printf("%10.6f", new_val);

    if (function[index] != FUNCTION_CLOCK) {
      if (!switching[index]) {
	new_val = time_const[index] * new_val
	  + (1.0 - time_const[index]) * value[index];
      } else {
	new_val = time_const[index] * new_val
	  + (1.0 - time_const[index]) * value2[index];
      }
    }

    if (new_val > MAX_VALUE)
      new_val = MAX_VALUE;
    else if (new_val < -MAX_VALUE)
      new_val = -MAX_VALUE;
    if (!switching[index]) {
      value[index] = reduce2(new_val);
    } else {
      value2[index] = new_val;
      if (value2[index] >= 0.0)
	value[index] = value[index_ptr1[index]];
      else
	value[index] = value[index_ptr2[index]];
    }


    if (Print[10])
      printf(" -> %10.6f\n", value[index]);
  }

  if (Print[13]) {
    int stop = num_inputs + num_hidden/3 + (num_hidden%3);
    for (i=num_inputs; i<stop; i++)
      printf("%5.2f ", value[i]);
    printf("\n");
  }
}
      

void Individual :: update(void)
{
  for (int i=0; i<Net_num_iterations; i++)
    next_iteration();
}


Real Individual :: reevaluate(void)
{
  Real val_prev, val;

  if (num_units <= 0) {
    cout << "Error: " << num_units << " units.\n" << endl;
    set_fitness(0.0);
    return 0.0;
  }

  val_prev = get_fitness();
  val = evaluate();
  hist_self.fitness_prev = val_prev;
  set_fitness((val + val_prev)/2.0);
  evaluated = 2;
  return val;
}


Real Individual :: evaluate(int init)
{
  int i;
  Real val;

  if (num_units <= 0) {
    cout << "Error: " << num_units << " units.\n" << endl;
    set_fitness(0.0);
    return 0.0;
  }

  for (i=0; i<num_units; i++)
    value[i] = 0.0;
  val = evaluate_network(init, this);

  if (evaluated) {
    Real val_prev;

    val_prev = get_fitness();
    hist_self.fitness_prev = val_prev;
    set_fitness((val + val_prev)/2.0);
    evaluated = 2;
    return val;
  }    

  evaluated = 1;
  set_fitness(val);
  return val;
}


Real Individual :: evaluate(void)
{
  evaluate(FALSE);
}


void Individual ::  set_input(Real *val)
{
  int i;
  for (i=0; i<num_inputs; i++)
    value[i] = val[i];
}

void Individual :: get_output(Real *val)
{
  int i, j = 0;
  for (i=output_start; i<output_stop; i++)
    val[j++] = value[i];
}


void Individual :: kill(void)
{
  cout << "free - exiting\n" << endl;
  exit(0);
}


void Individual :: print_fitness(void)
{
  printf("%6.1f", hist_self.fitness);
}



void Individual :: print_history2(history *hist)
{
  if (hist->how_created == CREATE_INIT)
    cout << "[-:";
  else if (hist->how_created == CREATE_MUTATE)
    cout << "[m:";
  else if (hist->how_created == CREATE_RECOMBINE)
    cout << "[r:";
  else 
    cout << "[?:";

  if (hist->operation == OP_WEIGHTS)
    cout << "w:";
  else if (hist->operation == OP_LINKS)
    cout << "l:";
  else if (hist->operation == OP_PROPERTIES)
    cout << "p:";
  else if (hist->operation == OP_INIT)
    cout << "-:";
  else
    cout << "?:";
  printf("%3d:%3d]", hist->num_mutate, hist->num_recombine);
}



void Individual :: fprint_history2(history *hist, FILE *log)
{
  if (hist->how_created == CREATE_INIT)
    fprintf(log, "[-:");
  else if (hist->how_created == CREATE_MUTATE)
    fprintf(log, "[m:");
  else if (hist->how_created == CREATE_RECOMBINE)
    fprintf(log, "[r:");
  else 
    fprintf(log, "[?:");

  if (hist->operation == OP_WEIGHTS)
    fprintf(log, "w:");
  else if (hist->operation == OP_LINKS)
    fprintf(log, "l:");
  else if (hist->operation == OP_PROPERTIES)
    fprintf(log, "p:");
  else if (hist->operation == OP_INIT)
    fprintf(log, "-:");
  else
    fprintf(log, "?:");

  fprintf(log, "%3d:%3d]", hist->num_mutate, hist->num_recombine);
}


void Individual :: print_history(history *hist)
{
  printf("%6.1f ", hist->fitness);
  print_history2(hist);
}


void Individual :: print_history_self(void)
{
  print_history(&hist_self);
}

void Individual :: print_history_self2(void)
{
  print_history2(&hist_self);
}

void Individual :: fprint_history_self2(FILE *log)
{
  fprint_history2(&hist_self, log);
}


void Individual :: print_history(void)
{
  printf("[%3d:%3d:%3d:%3d:%3d:%3d]", hist_self.op_count[0],
	 hist_self.op_count[1], hist_self.op_count[2], 
	 hist_self.op_count[3], hist_self.op_count[4], 
	 hist_self.op_count[5]);
}

void Individual :: fprint_history(FILE *log)
{
  fprintf(log, "[%3d:%3d:%3d:%3d:%3d:%3d]", hist_self.op_count[0],
	  hist_self.op_count[1], hist_self.op_count[2], 
	  hist_self.op_count[3], hist_self.op_count[4], 
	  hist_self.op_count[5]);
}

void Individual :: print_results(void)
{
  if (hist_self.how_created == CREATE_MUTATE) {
    if (hist_self.fitness > hist_parent1.fitness)
      cout << "  Higher!";
    else if (hist_self.fitness == hist_parent1.fitness)
      cout << "  Same.";
  } else if (hist_self.how_created == CREATE_RECOMBINE) {
    if ((hist_self.fitness > hist_parent1.fitness) &&
	(hist_self.fitness > hist_parent2.fitness))
      cout << "  Highest!";
    else if ((hist_self.fitness == hist_parent1.fitness) ||
	     (hist_self.fitness == hist_parent2.fitness))
      cout << "  Same.";
    else if ((hist_self.fitness > hist_parent1.fitness) ||
	     (hist_self.fitness > hist_parent2.fitness))
      cout << "  Middle.";
  }
}


void Individual :: print_history_full(void)
{
  Real diff, val;


  /*  
  printf("[%3d:%3d:%3d:%3d:%3d:%3d]", hist_parent1.op_count[0],
	 hist_parent1.op_count[1], hist_parent1.op_count[2], 
	 hist_parent1.op_count[3], hist_parent1.op_count[4], 
	 hist_parent1.op_count[5]);
  printf("[%3d:%3d:%3d:%3d:%3d:%3d]", hist_parent2.op_count[0],
	 hist_parent2.op_count[1], hist_parent2.op_count[2], 
	 hist_parent2.op_count[3], hist_parent2.op_count[4], 
	 hist_parent2.op_count[5]);
  cout << "\n";
  */

  if (evaluated == 2) {
    printf("Evl                                      ");
    val = get_fitness();
    diff = val - hist_self.fitness_prev;
    set_fitness(hist_self.fitness_prev);
    print_history(&hist_self);
    set_fitness(val);
    printf(" -> ");
    printf("%6.1f", val + diff);
    printf(" diff: %6.1f ", 2.0*diff);
    print_history();
    return;
  } else if (hist_self.how_created == CREATE_INIT) {
    print_history(&hist_self);
    return;
  } else if (hist_self.how_created == CREATE_MUTATE) {
    cout << "Mut ";
  } else if (hist_self.how_created == CREATE_RECOMBINE) {
    cout << "Rec ";
  }

  if (hist_self.operation == OP_INIT)
    return;

  if (hist_self.operation == OP_WEIGHTS) {
    printf("(weights %3d) ", hist_self.op1);
  } else if (hist_self.operation == OP_LINKS) {
    if (hist_self.how_created == CREATE_MUTATE) {
      if (hist_self.op1 > 0) {
	printf("(lnk %2d:%5.2lf)", hist_self.op1, hist_self.val1);
      } else {
	if (hist_self.op3) {
	  // deleted
	  printf("(lk %1d:%1d:%5.2lf)", -hist_self.op1, -hist_self.op2, hist_self.val1);
	  //	  printf("(lnk %2d:%5.2lf)", hist_self.op1+hist_self.op2, hist_self.val1);
	} else {
	  printf("(lk'%1d:%1d:%5.2lf)", -hist_self.op1, -hist_self.op2, hist_self.val1);
	  //	  printf("(lnk'%2d:%5.2lf)", hist_self.op1+hist_self.op2, hist_self.val1);
	}
      }
    } else {
      printf("(lnk %d:%2d:%2d) ", hist_self.op1,
	     hist_self.op2, hist_self.op3);
    }
  } else if (hist_self.operation == OP_PROPERTIES) {
    if (hist_self.how_created == CREATE_MUTATE)
      printf("(prop %2d:%2d)  ", hist_self.op1, hist_self.op2);
    else
      printf("(prop    %2d)  ", hist_self.op1);
  }

  if (hist_self.how_created == CREATE_MUTATE) {
    printf("              %6.3f   ", hist_self.delta);
    print_history(&hist_parent1);
  } else {
    print_history(&hist_parent1);
    cout << " + ";
    print_history(&hist_parent2);
  }

  cout << " -> ";
  print_history(&hist_self);
  print_history();
  print_results();
}


void Individual :: info(void)
{
  print_history();
}


int Individual :: read_old(FILE *file)
{
  char func;
  int i, j, ii, num, version;
  Real val;

  num = fscanf(file, "%d %d %d %lf %d\n", &num_inputs, &num_hidden,
	       &num_outputs, &hist_self.fitness, &version);

  printf("%d %d %d %.7lf %d\n", num_inputs, num_hidden, num_outputs,
	 hist_self.fitness, version);

  if (num != 5) {
      cout << "Individual: error in reading structure\n" << endl;
      return FALSE;
  }

  init();
  clear_net();

  for (i=0; i<num_units; i++) {
    num = fscanf(file, "%d %d %c %d %d %d %d %d %lf %lf %lf %lf\n",
		 &ii, &(unit_type[i]), &func, &(num_in[i]),
		 &(active[i]), &(switching[i]), 
		 &(index_ptr1[i]), &(index_ptr2[i]),
		 &(scale[i]), &(bias[i]), 
		 &(time_const[i]), &(ctr_rate[i]));
    if (num != 12) {
      cout << "Individual: error parsing for unit " << i << "." << endl;
      return FALSE;
    }

    if (func == 'g')
      function[i] = FUNCTION_SIGMOID;
    else if (func == 'G')
      function[i] = FUNCTION_SIGMOID2;
    else if (func == '+')
      function[i] = FUNCTION_SUM;
    else if (func == 's')
      function[i] = FUNCTION_SIN;
    else if (func == 'a')
      function[i] = FUNCTION_ASIN;
    else if (func == 'q')
      function[i] = FUNCTION_SQRT;
    else if (func == '2')
      function[i] = FUNCTION_SQR;
    else if (func == 'r')
      function[i] = FUNCTION_RANDOM;
    else if (func == '*')
      function[i] = FUNCTION_PROD;
    else if (func == '>')
      function[i] = FUNCTION_MAX;
    else if (func == '<')
      function[i] = FUNCTION_MIN;
    else if (func == '|')
      function[i] = FUNCTION_SIGN;
    else if (func == 'c')
      function[i] = FUNCTION_CLOCK;
    else if (func == '!')
      function[i] = FUNCTION_CLOCK;

    if (i != ii) {
      cout << "Individual: error reading unit " << i << "." << endl;
      return FALSE;
    }
  }

  for (i=0; i<num_units; i++) {
    for (j=0; j<num_in[i]; j++) {
      num = fscanf(file, "%d %lf %d", &from_link[i][j],
		   &val, &ii);
      if (num != 3) {
	cout << "Individual: error parsing for unit:" << i;
	cout << "  link: " << j << "." << endl;
	return FALSE;
      }

      if (i != ii) {
	cout << "Individual: error reading unit:" << i;
	cout << "  link: " << j << "." << endl;
	cout << "Incorrect _to_ field." << endl;
	return FALSE;
      }

      weight[i][from_link[i][j]] = val;
    }
  }

  return TRUE;
}


int Individual :: read_version2(FILE *file)
{
  char func;
  int i, j, ii, num;
  Real val;

  // Read version 2.
  Net_num_iterations = 1;

  // Read number of units and fitness.
  num = fscanf(file, "%d %d %d %lf\n", &num_inputs, &num_hidden,
	       &num_outputs, &hist_self.fitness);
  printf("%d %d %d %.7lf\n", num_inputs, num_hidden, num_outputs,
	 hist_self.fitness);

  if (num != 4) {
      cout << "Individual: error units & fitness\n" << endl;
      return FALSE;
  }

  // Read net structure/type.
  num = fscanf(file, "%d %d %d %d\n", &Net_struct,
	       &Net_cont_time, &Net_max_func, &Net_switching);
  printf("%d %d %d %d\n", Net_struct,
	 Net_cont_time, Net_max_func, Net_switching);
  if (num != 4) {
      cout << "Individual: error in reading structure\n" << endl;
      return FALSE;
  }

  // Initialize (allocate memory).
  init();
  clear_net();

  // Read units.
  for (i=0; i<num_units; i++) {
    num = fscanf(file, "%d %d %c %d %d %d %d %d %lf %lf %lf %lf\n",
		 &ii, &(unit_type[i]), &func, &(num_in[i]),
		 &(active[i]), &(switching[i]), 
		 &(index_ptr1[i]), &(index_ptr2[i]),
		 &(scale[i]), &(bias[i]), 
		 &(time_const[i]), &(ctr_rate[i]));
    if (num != 12) {
      cout << "Individual: error parsing for unit " << i << "." << endl;
      return FALSE;
    }

    if (func == 'g')
      function[i] = FUNCTION_SIGMOID;
    else if (func == 'G')
      function[i] = FUNCTION_SIGMOID2;
    else if (func == '+')
      function[i] = FUNCTION_SUM;
    else if (func == 's')
      function[i] = FUNCTION_SIN;
    else if (func == 'a')
      function[i] = FUNCTION_ASIN;
    else if (func == 'q')
      function[i] = FUNCTION_SQRT;
    else if (func == '2')
      function[i] = FUNCTION_SQR;
    else if (func == 'r')
      function[i] = FUNCTION_RANDOM;
    else if (func == '*')
      function[i] = FUNCTION_PROD;
    else if (func == '>')
      function[i] = FUNCTION_MAX;
    else if (func == '<')
      function[i] = FUNCTION_MIN;
    else if (func == '|')
      function[i] = FUNCTION_SIGN;
    else if (func == 'c')
      function[i] = FUNCTION_CLOCK;
    else if (func == '!')
      function[i] = FUNCTION_CLOCK;

    if (i != ii) {
      cout << "Individual: error reading unit " << i << "." << endl;
      return FALSE;
    }
  }

  // Read links.
  for (i=0; i<num_units; i++) {
    for (j=0; j<num_in[i]; j++) {
      num = fscanf(file, "%d %lf %d", &from_link[i][j],
		   &val, &ii);
      if (num != 3) {
	cout << "Individual: error parsing for unit:" << i;
	cout << "  link: " << j << "." << endl;
	return FALSE;
      }

      if (i != ii) {
	cout << "Individual: error reading unit:" << i;
	cout << "  link: " << j << "." << endl;
	cout << "Incorrect _to_ field." << endl;
	return FALSE;
      }

      weight[i][from_link[i][j]] = val;
    }
  }

  return TRUE;
}


int Individual :: read(FILE *file)
{
  char func;
  int i, j, ii, num, version;
  Real val;

  // Read version.
  num = fscanf(file, "Version: %d\n", &version);
  if (num != 1) {
    cout << "Individual: error in reading version\n" << endl;
    return FALSE;
  } else if (version == 2) {
    cout << "Individual: Using version #" << version << endl;
    return read_version2(file);
  } else if (version != NET_VERSION) {
    cout << "Individual: version #%d\n" << version << endl;
    return FALSE;
  }
  cout << "Reading version: " << version << "\n" << endl;

  // Read number of units and fitness.
  num = fscanf(file, "%d %d %d %lf\n", &num_inputs, &num_hidden,
	       &num_outputs, &hist_self.fitness);
  printf("%d %d %d %.7lf\n", num_inputs, num_hidden, num_outputs,
	 hist_self.fitness);

  if (num != 4) {
      cout << "Individual: error units & fitness\n" << endl;
      return FALSE;
  }

  // Read net structure/type.
  num = fscanf(file, "%d %d %d %d %d %d\n", &Net_struct,
	       &Net_cont_time, &Net_max_func, &Net_switching,
	       &Net_num_iterations, &ii);
  printf("%d %d %d %d %d\n", Net_struct, Net_cont_time,
	 Net_max_func, Net_switching, Net_num_iterations);
  if (num != 6) {
      cout << "Individual: error in reading structure\n" << endl;
      return FALSE;
  }

  // Initialize (allocate memory).
  init();
  clear_net();

  // Read units.
  for (i=0; i<num_units; i++) {
    num = fscanf(file, "%d %d %c %d %d %d %d %d %lf %lf %lf %lf\n",
		 &ii, &(unit_type[i]), &func, &(num_in[i]),
		 &(active[i]), &(switching[i]), 
		 &(index_ptr1[i]), &(index_ptr2[i]),
		 &(scale[i]), &(bias[i]), 
		 &(time_const[i]), &(ctr_rate[i]));
    if (num != 12) {
      cout << "Individual: error parsing for unit " << i << "." << endl;
      return FALSE;
    }

    if (func == 'g')
      function[i] = FUNCTION_SIGMOID;
    else if (func == 'G')
      function[i] = FUNCTION_SIGMOID2;
    else if (func == '+')
      function[i] = FUNCTION_SUM;
    else if (func == 's')
      function[i] = FUNCTION_SIN;
    else if (func == 'a')
      function[i] = FUNCTION_ASIN;
    else if (func == 'q')
      function[i] = FUNCTION_SQRT;
    else if (func == '2')
      function[i] = FUNCTION_SQR;
    else if (func == 'r')
      function[i] = FUNCTION_RANDOM;
    else if (func == '*')
      function[i] = FUNCTION_PROD;
    else if (func == '>')
      function[i] = FUNCTION_MAX;
    else if (func == '<')
      function[i] = FUNCTION_MIN;
    else if (func == '|')
      function[i] = FUNCTION_SIGN;
    else if (func == 'c')
      function[i] = FUNCTION_CLOCK;
    else if (func == '!')
      function[i] = FUNCTION_CLOCK;

    if (i != ii) {
      cout << "Individual: error reading unit " << i << "." << endl;
      return FALSE;
    }
  }

  // Read links.
  for (i=0; i<num_units; i++) {
    for (j=0; j<num_in[i]; j++) {
      num = fscanf(file, "%d %lf %d", &from_link[i][j],
		   &val, &ii);
      if (num != 3) {
	cout << "Individual: error parsing for unit:" << i;
	cout << "  link: " << j << "." << endl;
	return FALSE;
      }

      if (i != ii) {
	cout << "Individual: error reading unit:" << i;
	cout << "  link: " << j << "." << endl;
	cout << "Incorrect _to_ field." << endl;
	return FALSE;
      }

      weight[i][from_link[i][j]] = val;
    }
  }

  return TRUE;
}


void Individual :: write(FILE *file)
{
  char func;
  int i, j;


  fprintf(file, "Version: %d\n", NET_VERSION);
  fprintf(file, "%d %d %d %.7lf\n", num_inputs, num_hidden,
	  num_outputs, hist_self.fitness);
  fprintf(file, "%d %d %d %d %d 0\n", Net_struct,
	  Net_cont_time, Net_max_func, Net_switching, 
	  Net_num_iterations);

  for (i=0; i<num_units; i++) {
    if (function[i] == FUNCTION_SIGMOID)
      func = 'g';
    else if (function[i] == FUNCTION_SIGMOID2)
      func = 'G';
    else if (function[i] == FUNCTION_SUM)
      func = '+';
    else if (function[i] == FUNCTION_SIN)
      func = 's';
    else if (function[i] == FUNCTION_ASIN)
      func = 'a';
    else if (function[i] == FUNCTION_SQR)
      func = '2';
    else if (function[i] == FUNCTION_SQRT)
      func = 'q';
    else if (function[i] == FUNCTION_RANDOM)
      func = 'r';
    else if (function[i] == FUNCTION_PROD)
      func = '*';
    else if (function[i] == FUNCTION_MAX)
      func = '>';
    else if (function[i] == FUNCTION_MIN)
      func = '<';
    else if (function[i] == FUNCTION_SIGN)
      func = '|';
    else if (function[i] == FUNCTION_CLOCK)
      func = '!';

    fprintf(file, "%2d %2d %c %2d %1d %1d %2d %2d %10.7lf %10.7lf %10.7lf %10.7lf\n",
	    i, unit_type[i], func, num_in[i], active[i],
	    switching[i],  index_ptr1[i], index_ptr2[i], 
	    scale[i], bias[i], 
	    time_const[i], ctr_rate[i]);
  }

  for (i=0; i<num_units; i++) {
    for (j=0; j<num_in[i]; j++) {
      fprintf(file, "%2d %10.7lf %2d\n", from_link[i][j],
	      weight[i][from_link[i][j]], i);
    }
  }
}


void Individual :: read(ifstream *infile)
{
  cout << "Ind_net: read(stream) - not implemented." << endl;
}

void Individual :: write(ofstream *outfile)
{
  cout << "Ind_net: write(stream) - not implemented." << endl;
}




