/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/


#ifndef IND_NET_HEADER_FILE
#define IND_NET_HEADER_FILE

#include <iostream.h>
#include <fstream.h>
#include "general.h"


#define CREATE_INIT -1
//define CREATE_REEVALUATED -1
#define CREATE_MUTATE 0
#define CREATE_RECOMBINE 1

#define OP_INIT -1
#define OP_WEIGHTS 1
#define OP_LINKS 2
#define OP_PROPERTIES 3


#define NET_FEED_FORWARD 0
#define NET_RECURRENT 1
#define NET_FREE 2
#define NET_STRUCTURED 3
#define NET_STRUCTURED2 4
#define NET_NO_INPUTS 5

extern int Net_groups;
extern int Net_group_size[20];
extern int Net_group_start[20];

extern int Net_struct;
extern int Net_cont_time;
extern int Net_max_func;
extern int Net_num_iterations;
extern int Net_switching;
extern int Net_fix_hyperplanes;
extern Real Rec_delta;

extern Real Max_output[100];
extern Real Min_output[100];

class Network;
class Graphic;

typedef struct History
{
  int index;
  int how_created;
  int operation;
  int op1;
  int op2;
  int op3;
  int num_mutate;
  int num_recombine;
  int op_count[7];
  Real delta;
  Real val1;
  Real fitness;
  Real fitness_prev;
} history;

class Individual
{
public:
  //  Real fitness, fitness_p1, fitness_p2;

  // Data structures defining a network.
  int created;
  int evaluated;

  //  int operation;
  history hist_self;
  history hist_parent1;
  history hist_parent2;
  //  int how_created;
  int output_start, output_stop;
  int num_inputs, num_outputs, num_hidden, num_units;
  //  int num_mutate, num_recombine;
  //  int op_count[7];
  int *unit_type;
  int *function;
  int *num_in;
  int *active;
  int *switching;
  int *index_ptr1;
  int *index_ptr2;
  Real *scale;
  Real *bias;
  Real *time_const;
  Real *ctr_rate;
  Real *value;
  Real *value2;
  int **from_link; // 'from' index for weight matrix.
  Real **weight; // weight[to][from]
  Graphic *win;


  /*************************/

  Individual(void); // perhaps input data struct
  Individual(Individual *sample);
  void clear_net(void);
  void make_random(void);
  void init(void);
  void init(int inputs, int hidden, int outputs);
  int create_mutate(void);
  int better(void);
  Real get_fitness(void);
  void set_fitness(Real val);
  void add_fitness(Real val);
  void div_fitness(Real val);
  void set_creation(int type);
  int get_creation(void);
  void kill(void);
  void init_hist(history *hist);
  void init_hist(void);
  void duplicate_hist(history *hist_from, history *hist_to);
  void duplicate(Individual *ind);


  int connect(int f, int t, Real w);
  void mutate_weight(int index, Real rate);
  void mutate_weights(int print_flag, Real rate);
  void mutate_weights(Real rate);
  Real random_new_weight(void);
  Real random_weight(void);
  void random_connect_ff(int index, Real wght);
  void random_connect_rec(int index, Real wght);
  void random_connect_struct(int index, Real wght);
  void random_connect_struct2(int index, Real wght);
  void random_connect(int index, Real wght);
  int reduce_link(int index);
  int delete_link(int index);
  void mutate_links(Real alpha);
  void random_function(int index);
  int mutate_properties_all(int index, Real alpha);
  int mutate_properties(int index, Real alpha);
  void mutate_properties(Real alpha);
  void mutate(Real alpha);
  void mutate(void);  // perhaps input data struct

  void configure_links(int num, int index);
  Real recombine(Real val1, Real val2);
  int recombine_weights(int print_flag, Individual *parent2);
  int recombine_weights(Individual *parent2);
  int recombine_links(int terse, Individual *parent2);
  int recombine_propertiesA(Individual *parent2);
  int recombine_propertiesB(int terse, Individual *parent2);
  int recombine_propertiesC(Individual *parent2);
  int recombine_properties(Individual *parent2);
  void recombine(Individual *parent2);

  void set_input(Real *inputs);
  void get_output(Real *outputs);
  Real sum_inputs(int index);
  void reset(void);
  void next_iteration(void);
  void update(void);
  Real evaluate(int init);
  Real evaluate(void);
  Real reevaluate(void);

  void print_fitness(void);
  void print_history(history *hist);
  void print_history2(history *hist);
  void fprint_history2(history *hist, FILE *log);
  void print_history_self(void);
  void print_history_self2(void);
  void fprint_history_self2(FILE *log);
  void print_history(void);
  void fprint_history(FILE *log);
  void print_history_full(void);
  void print_results(void);
  void info(void);

  void read(ifstream *infile);
  void write(ofstream *outfile);

  int read(FILE *file);
  int read_version2(FILE *file);
  int read_old(FILE *file);
  void write(FILE *file);
};

#endif
