/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#ifndef IND_TREE_HEADER_FILE
#define IND_TREE_HEADER_FILE

#include "general.h"
#include <iostream.h>
#include <fstream.h>


#define CREATE_INIT -1
//define CREATE_REEVALUATED -1
#define CREATE_MUTATE 0
#define CREATE_RECOMBINE 1

class Tree;

typedef struct History
{
  int index;
  int how_created;
  int operation;
  int op1;
  int op2;
  int op3;
  int num_mutate;
  int num_recombine;
  int op_count[7];
  Real delta;
  Real val1;
  Real fitness;
  Real fitness_prev;
} history;

class Individual
{
public:
  int created;
  int evaluated;

  int num_inputs;
  Real *input;
  int num_outputs;
  Real *output;

  Tree *tree;

  history hist_self;
  history hist_parent1;
  history hist_parent2;


  /*************************/

  Individual(void); // perhaps input data struct
  Individual(Individual *sample);
  void clear(void);
  void make_random(void);
  void init(void);
  int create_mutate(void);
  int better(void);
  Real get_fitness(void);
  void set_fitness(Real val);
  void add_fitness(Real val);
  void div_fitness(Real val);
  void set_creation(int type);
  int get_creation(void);
  void init_hist(history *hist);
  void init_hist(void);
  void duplicate_hist(history *hist_from, history *hist_to);
  void duplicate(Individual *ind);

  int mutate(Real alpha);
  int mutate(void);  // perhaps input data struct
  int recombine(Individual *parent2);

  int set_input(Real *val);
  int get_output(Real *val);
  Real calculate(void);
  void reset(void);
  Real evaluate(int init);
  Real evaluate(void);

  void print_fitness(void);
  void print_history(history *hist);
  void print_history2(history *hist);
  void print_history_self(void);
  void print_history_self2(void);
  void print_history(void);
  void print_history_full(void);
  void print_results(void);
  void info(void);

  int read_history(history *hist, FILE *file);
  int read(FILE *file);
  void write_history(history *hist, FILE *file);
  void write(FILE *file);
};

#endif
