/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stream.h>
#include <iostream.h>
#include <fstream.h>
#include "general.h"
#include "ind_vec.h"

#define IND_VERSION 1

extern Real evaluate_vec(int init, int size, Real *val);
Real Rec_delta = 1.0;


Individual :: Individual(void)
{
  init_hist(&hist_self);
  init_hist(&hist_parent1);
  init_hist(&hist_parent2);
  created = FALSE;
  evaluated = FALSE;
  size = 1;
}

Individual :: Individual(Individual *sample)
{
  init_hist(&hist_self);
  init_hist(&hist_parent1);
  init_hist(&hist_parent2);
  created = FALSE;
  evaluated = FALSE;
  if (sample != NULL) {
    size = sample->size;
  } else {
    size = 1;
  }
}


void Individual :: init(void)
{
  if (created)
    free(gene);

  if (size > 0)
    gene = (Real *)malloc(size*sizeof(Real));

  init_hist(&hist_self);
  init_hist(&hist_parent1);
  init_hist(&hist_parent2);
  created = TRUE;
}


void Individual :: init(int length)
{
  size = length;
  init();
}


void Individual :: init_hist(history *hist)
{
  int i;

  hist->index = -1;
  hist->how_created = CREATE_INIT;
  hist->num_mutate = 0;
  hist->num_recombine = 0;
  hist->delta = 0.0;
  hist->op1 = 0;
  hist->op2 = 0;
  hist->op3 = 0;
  hist->fitness = 0.0;
  hist->fitness_prev = 0.0;

  for (i=0; i<7; i++)
    hist->op_count[i] = 0;
}

void Individual :: init_hist(void)
{
  int i;

  hist_self.index = -1;
  hist_self.how_created = CREATE_INIT;
  hist_self.num_mutate = 0;
  hist_self.num_recombine = 0;
  hist_self.delta = 0.0;
  hist_self.op1 = 0;
  hist_self.op2 = 0;
  hist_self.op3 = 0;
  hist_self.fitness_prev = 0.0;

  for (i=0; i<7; i++)
    hist_self.op_count[i] = 0;
}


int Individual :: create_mutate(void)
{
  if (hist_self.how_created == CREATE_MUTATE)
    return TRUE;

  return FALSE;
}

// Returns TRUE if individual has higher fitness than its first parent.
int Individual :: better(void)
{
  if (create_mutate()) {
    if (hist_self.fitness > hist_parent1.fitness)
      return 1;
    else if (hist_self.fitness == hist_parent1.fitness)
      return 0;
    else
      return -1;
  } else {
    if ((hist_self.fitness > hist_parent1.fitness) &&
	(hist_self.fitness > hist_parent2.fitness))
      return 1;
    else if ((hist_self.fitness == hist_parent1.fitness) ||
	     (hist_self.fitness == hist_parent2.fitness))
      return 0;
    else if ((hist_self.fitness > hist_parent1.fitness) ||
	     (hist_self.fitness > hist_parent2.fitness))
      return 0;
    else
      return -1;
  }
}

Real Individual :: get_fitness(void)
{
  return hist_self.fitness;
}

void Individual :: set_fitness(Real val)
{
  hist_self.fitness = val;
}

void Individual :: add_fitness(Real val)
{
  hist_self.fitness += val;
}

void Individual :: div_fitness(Real val)
{
  hist_self.fitness /= val;
}


void Individual :: set_creation(int type)
{
  hist_self.how_created = type;
}

int Individual :: get_creation(void)
{
  return hist_self.how_created;
}


void Individual :: duplicate_hist(history *hist_from, history *hist_to)
{
  int i;

  hist_to->index = hist_from->index;
  hist_to->operation = hist_from->operation;
  hist_to->how_created = hist_from->how_created;
  hist_to->num_mutate = hist_from->num_mutate;
  hist_to->num_recombine = hist_from->num_recombine;
  hist_to->delta = hist_from->delta;
  hist_to->fitness = hist_from->fitness;

  hist_to->op1 = hist_from->op1;
  hist_to->op2 = hist_from->op2;
  hist_to->op3 = hist_from->op3;
  hist_to->val1 = hist_from->val1;
  for (i=0; i<7; i++)
    hist_to->op_count[i] = hist_from->op_count[i];
}



void Individual :: print_fitness(void)
{
  printf("%6.1f", hist_self.fitness);
}



void Individual :: print_history2(history *hist)
{
  if (hist->how_created == CREATE_INIT)
    cout << "[-:";
  else if (hist->how_created == CREATE_MUTATE)
    cout << "[m:";
  else if (hist->how_created == CREATE_RECOMBINE)
    cout << "[r:";
  else 
    cout << "[?:";

  printf("%3d:%3d]", hist->num_mutate, hist->num_recombine);
}


void Individual :: fprint_history2(history *hist, FILE *log)
{
  if (hist->how_created == CREATE_INIT)
    fprintf(log, "[-:");
  else if (hist->how_created == CREATE_MUTATE)
    fprintf(log, "[m:");
  else if (hist->how_created == CREATE_RECOMBINE)
    fprintf(log, "[r:");
  else 
    fprintf(log, "[?:");

  fprintf(log, "%3d:%3d]", hist->num_mutate, hist->num_recombine);
}

void Individual :: print_history(history *hist)
{
  printf("%6.1f ", hist->fitness);
  print_history2(hist);
}

void Individual :: print_history_self(void)
{
  print_history(&hist_self);
}

void Individual :: print_history_self2(void)
{
  print_history2(&hist_self);
}

void Individual :: fprint_history_self2(FILE *log)
{
  fprint_history2(&hist_self, log);
}

void Individual :: print_history(void)
{
  printf("[%3d:%3d:%3d:%3d:%3d:%3d]", hist_self.op_count[0],
	 hist_self.op_count[1], hist_self.op_count[2], 
	 hist_self.op_count[3], hist_self.op_count[4], 
	 hist_self.op_count[5]);
}

void Individual :: fprint_history(FILE *log)
{
  fprintf(log, "[%3d:%3d:%3d:%3d:%3d:%3d]", hist_self.op_count[0],
	  hist_self.op_count[1], hist_self.op_count[2], 
	  hist_self.op_count[3], hist_self.op_count[4], 
	  hist_self.op_count[5]);
}

void Individual :: print_results(void)
{
  if (hist_self.how_created == CREATE_MUTATE) {
    if (hist_self.fitness > hist_parent1.fitness)
      cout << "  Higher!";
    else if (hist_self.fitness == hist_parent1.fitness)
      cout << "  Same.";
  } else if (hist_self.how_created == CREATE_RECOMBINE) {
    if ((hist_self.fitness > hist_parent1.fitness) &&
	(hist_self.fitness > hist_parent2.fitness))
      cout << "  Highest!";
    else if ((hist_self.fitness == hist_parent1.fitness) ||
	     (hist_self.fitness == hist_parent2.fitness))
      cout << "  Same.";
    else if ((hist_self.fitness > hist_parent1.fitness) ||
	     (hist_self.fitness > hist_parent2.fitness))
      cout << "  Middle.";
  }
}


void Individual :: print_history_full(void)
{
  Real diff, val;

  if (evaluated == 2) {
    printf("Evl                        ");
    val = get_fitness();
    diff = val - hist_self.fitness_prev;
    set_fitness(hist_self.fitness_prev);
    print_history(&hist_self);
    set_fitness(val);
    printf(" -> ");
    printf("%6.1f", val + diff);
    printf(" diff: %6.1f ", 2.0*diff);
    print_history();
    return;
  } else if (hist_self.how_created == CREATE_INIT) {
    print_history(&hist_self);
    return;
  } else if (hist_self.how_created == CREATE_MUTATE) {
    cout << "Mut ";
  } else if (hist_self.how_created == CREATE_RECOMBINE) {
    cout << "Rec ";
  }

  //  if (hist_self.operation == OP_INIT)
  //    return;

  if (hist_self.how_created == CREATE_MUTATE) {
    printf("              %6.3f   ", hist_self.delta);
    print_history(&hist_parent1);
  } else {
    print_history(&hist_parent1);
    cout << " + ";
    print_history(&hist_parent2);
  }

  cout << " -> ";
  print_history(&hist_self);
  print_history();
  print_results();
}


void Individual :: info(void)
{
  print_history();
}



void Individual :: make_random(void)
{
  int i;

  if (!created) {
    cout << "Individual not created." << endl;
    init();
  }

  evaluated = FALSE;
  /*  gene[0] = 0.0;
  gene[1] = -60.0;
  gene[2] = 0.0;
  gene[3] = 0.0;
  gene[4] = -50.0;
  gene[5] = 0.0; */
  for (i=0; i<size; i++)
    gene[i] = 200.0 * random_dbl() - 100.0;
  /*
  gene[i++] = 0.0;
  gene[i++] = 50.0;
  gene[i++] = 0.0;
  */
}


void Individual :: duplicate(Individual *ind)
{
  int i;

  if (!created) {
    size = ind->size;
    init();
  }

  evaluated = ind->evaluated;
  duplicate_hist(&(ind->hist_self), &hist_self);
  duplicate_hist(&(ind->hist_parent1), &hist_parent1);
  duplicate_hist(&(ind->hist_parent2), &hist_parent2);

  for (i=0; i<size; i++)
    gene[i] = ind->gene[i];
}



int Individual :: mutate(Real alpha)
{
  int i, j;

  evaluated = FALSE;

  duplicate_hist(&hist_self, &hist_parent1);
  init_hist(&hist_parent2);
  hist_self.num_mutate++;
  hist_self.how_created = CREATE_MUTATE;

  for (i=0; i<6; i++) {
    j = random_int(size);
    gene[j] += random_norm(0.0, 0.7 * alpha);
  }

  return TRUE;
}


int Individual :: mutate(void)
{
  return mutate(1.0);
}


Real Individual :: recombine(Real val1, Real val2)
{
  if (val1 == 0.0) {
    if (random_int(2))
      return val1;
    else
      return val2;
  } else if (val2 == 0.0) {
    if (random_int(2))
      return val1;
    else
      return val2;
  }

  Real diff;

  diff = Rec_delta * (val1 - val2);
  return reduce(val1 + diff * (random_dbl() * 2.0 - 1.0));
}


int Individual :: recombine(Individual *parent2)
{
  evaluated = FALSE;
  parent2->evaluated = FALSE;

  // History for this individual.
  duplicate_hist(&hist_self, &hist_parent1);
  duplicate_hist(&(parent2->hist_self), &hist_parent2);
  hist_self.num_recombine++;
  hist_self.how_created = CREATE_RECOMBINE;

  // History for parent2.
  parent2->duplicate_hist(&(parent2->hist_self), &(parent2->hist_parent1));
  duplicate_hist(&hist_self, &(parent2->hist_parent2));
  parent2->hist_self.num_recombine++;
  parent2->hist_self.how_created = CREATE_RECOMBINE;

  for (int i=0; i<size; i++) {
    gene[i] = recombine(gene[i], parent2->gene[i]);
  }

  return TRUE;
}


int Individual :: read_history(history *hist, FILE *file)
{
  int num;

  num = fscanf(file, "%lf\n", &(hist->fitness));
  if (num != 1) {
    cout << "Read history:: Error reading fitness" << endl;
    exit -1;
  }

  num = fscanf(file, "%d %d %d %d %d\n", &(hist->index),
	       &(hist->how_created), &(hist->operation),
	       &(hist->num_mutate), &(hist->num_recombine));
  if (num != 5) {
    cout << "Read history:: Error reading." << endl;
    exit -1;
  }    
}


int Individual :: read(FILE *file)
{
  int num, version;

  cout << "Ind vec - read" << endl;
  
  num = fscanf(file, "Version: %d\n", &version);
  if (num != 1) {
    cout << "Individual: error in reading version\n" << endl;
    return FALSE;
  } else if (version != IND_VERSION) {
    cout << "Individual: Using version #%d\n" << version << endl;
    return FALSE;
  }    


  if (!read_history(&hist_self, file)) {
    cout << "Individ: error in history.\n" << endl;
    return FALSE;
  }
  
  num = fscanf(file, "%d\n", &size);
  if (num != 1) {
    cout << "Individual: error in reading size\n" << endl;
    return FALSE;
  }
  init();

  for (int i=0; i<size; i++) {
    num = fscanf(file, "%lf\n", &(gene[i]));
    if (num != 1) {
      cout << "Individual: error in reading value: " << i << "\n" << endl;
      return FALSE;
    }
  }

  return TRUE;
}


void Individual :: write_history(history *hist, FILE *file)
{
  fprintf(file, "%.7lf\n", hist->fitness);
  fprintf(file, "%d %d %d %d %d\n", hist->index,
	  hist->how_created, hist->operation,
	  hist->num_mutate, hist->num_recombine);
}


void Individual :: write(FILE *file)
{
  int i;

  fprintf(file, "Version: %d\n", IND_VERSION);
  write_history(&hist_self, file);

  fprintf(file, "%d\n", size);
  if (size > 0) {
    for (i=0; i<size; i++)
      fprintf(file, "%.7lf\n", gene[i]);
  } else {
    cout << "Ind::write()  Error - size is 0." << endl;
    exit -1;
  }
}




int Individual :: set_input(Real *val)
{
  return TRUE;
}

int Individual :: get_output(Real *val)
{

  return TRUE;
}




Real Individual :: evaluate(int init)
{
  Real val;

  val = evaluate_vec(init, size, gene);
  if (evaluated) {
    Real val_prev;

    val_prev = get_fitness();
    hist_self.fitness_prev = val_prev;
    set_fitness((val + val_prev)/2.0);
    evaluated = 2;
    return val;
  }    

  evaluated = 1;
  set_fitness(val);
  return val;
}

Real Individual :: evaluate(void) {
  return evaluate(FALSE);
}
