/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stream.h>
#include <iostream.h>
#include <fstream.h>
#include "general.h"
#include "network2.h"
#include "individ.h"



Individ :: Individ(void)
{
  created = FALSE;
}

void Individ :: init(void)
{
  created = TRUE;
  how_created = CREATE_INIT;
  operation = OP_INIT;
}

void Individ :: clear(void)
{
}


void Individ :: make_random(void)
{
  int i, j, num;

  clear();

  player[i]->make_random();
}


void Individ :: init(int inputs, int hidden, int outputs)
{
  int i;

  num_inputs = inputs;
  num_hidden = hidden;
  num_outputs = outputs;

  init();
  for (i=0; i<num_players; i++) {
    player[i]->num_inputs = num_inputs;
    player[i]->num_hidden = num_hidden;
    player[i]->num_outputs = num_outputs;
  }
  clear_networks();
  make_random();
}


void Individ :: duplicate(Individ *ind)
{
  int i, j;

  if (!created) {
    init(ind->num_inputs, ind->num_hidden, ind->num_outputs);
  }

  fitness = ind->fitness;
  operation = ind->operation;
  num_mutate = ind->num_mutate;
  num_recombine = ind->num_recombine;
  how_created = ind->how_created;
  for (i=0; i<7; i++)
    op_count[i] = ind->op_count[i];

  for (i=0; i<num_players; i++)
    player[i]->duplicate(ind->player[i]);
}

void Individ :: mutate(Real alpha)
{
  int i;
  Real choice;

  for (i=0; i<num_players; i++)
    player[i]->mutate(alpha);

  if (Print[8])
    printf("%6.3f", absv(alpha));

}


void Individ :: mutate(void)
{
  mutate(1.0);
}


void Individ :: recombine(Individ *parent2)
{
  int i;

  for (i=0; i<num_players; i++)
    player[i]->recombine(parent2->player[i]);
}



void Individ :: kill(void)
{
  cout << "free - exiting\n" << endl;
  exit(0);
}



void Individ :: print(void)
{
  printf("%6.1f [", fitness);
  if (how_created == CREATE_INIT)
    cout << "-:";
  else if (how_created == CREATE_MUTATE)
    cout << "m:";
  else if (how_created == CREATE_RECOMBINE)
    cout << "r:";
  else 
    cout << "?:";

  if (operation == OP_WEIGHTS)
    cout << "w:";
  else if (operation == OP_LINKS)
    cout << "l:";
  else if (operation == OP_PROPERTIES)
    cout << "p:";
  else if (operation == OP_INIT)
    cout << "-:";
  else
    cout << "?:";
  printf("%3d:%3d]", num_mutate, num_recombine);
}

void Individ :: print_history(void)
{
  printf("[%3d:%3d:%3d:%3d:%3d:%3d]", op_count[0], op_count[1],
	 op_count[2], op_count[3], op_count[4], op_count[5]);
}

void Individ :: print_results(void)
{
  if (how_created == CREATE_MUTATE) {
    if (fitness > fitness_p1)
      cout << "  Higher!";
    else if (fitness == fitness_p1)
      cout << "  Same.";
  } else if (how_created == CREATE_RECOMBINE) {
    if ((fitness > fitness_p1) &&
	(fitness > fitness_p2))
      cout << "  Highest!";
    else if ((fitness == fitness_p1) ||
	     (fitness == fitness_p2))
      cout << "  Same.";
    else if ((fitness > fitness_p1) ||
	     (fitness > fitness_p2))
      cout << "  Middle.";
  }
}


void Individ :: info(void)
{
  print();
}


int Individ :: read(FILE *file)
{
  char func;
  int i, j, ii, num;
  Real val;

  num = fscanf(file, "%d %lf\n", &num_players, &fitness);
  if (num != 2) {
      cout << "Individ: error in num players.\n" << endl;
      return FALSE;
  }

  num = fscanf(file, "%d %d %d %d\n", &operation, &how_created,
	       &num_mutate, &num_recombine);
  if (num != 4) {
      cout << "Individ: error in creation history.\n" << endl;
      return FALSE;
  }

  num = fscanf(file, "%d %d %d %d %d %d\n", &(op_count[0]),
	       &(op_count[1]), &(op_count[2]), &(op_count[3]),
	       &(op_count[4]), &(op_count[5]));
  if (num != 6) {
      cout << "Individ: error in creation history statistics.\n" << endl;
      return FALSE;
  }

  init();

  for (i=0; i<num_players; i++)
    player[i]->read(file);

  return TRUE;
}


void Individ :: write(FILE *file)
{
  int i;

  fprintf(file, "%d %.7lf\n", num_players, fitness);
  fprintf(file, "%d %d %d %d\n", operation, how_created,
	  num_mutate, num_recombine);
  fprintf(file, "%d %d %d %d %d %d\n", op_count[0], op_count[1],
	  op_count[2], op_count[3], op_count[4], op_count[5]);

  for (i=0; i<num_players; i++)
    player[i]->write(file);
}


void Individ :: read(ifstream *infile)
{
  cout << "Ind_net: read(stream) - not implemented." << endl;
}

void Individ :: write(ofstream *outfile)
{
  cout << "Ind_net: write(stream) - not implemented." << endl;
}




