/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stream.h>
#include <fstream.h>
#include "general.h"
#include "graphic.h"
#include "lsystem.h"
#include "build_bricks.h"
#include "bars2d.h"
#include "bars3d.h"
#include "programs.h"
#include "pop.h"
#include "ind_lsys.h"
#include "algebra.h"
#include "prod_body.h"

Graphic *win1 = 0;

int Num_terminals = 0;
char Terminal[50];
Real Terminal_prob[50];

int Fitness_func = 1;
int Terminal_set = 5;
int Build_type = 1;
int Model_type = 3;

int Pen_down = TRUE;

int Num_models = 0;
BarStruct **BarModel;
BarStruct3D **BarModel3D;
int Model_id;

int Max_joints = MAX_JOINTS;
int Number_trials = 3;



int Heading[3][3];

int TUp_mat[3][3] = {{0, 0, -1},
		     {0, 1, 0},
		     {1, 0, 0}};
int TDown_mat[3][3] = {{0, 0, 1},
		       {0, 1, 0},
		       {-1, 0, 0}};
int TRight_mat[3][3] = {{0, 1, 0},
			{-1, 0, 0},
			{0, 0, 1}};
int TLeft_mat[3][3] = {{0, -1, 0},
		       {1, 0, 0},
		       {0, 0, 1}};
int TClockwise_mat[3][3] = {{1, 0, 0},
			    {0, 0, 1},
			    {0, -1, 0}};
int TCntrClockwise_mat[3][3] = {{1, 0, 0},
				{0, 0, -1},
				{0, 1, 0}};


void init_heading(void)
{
  Heading[0][0] = 1;
  Heading[0][1] = 0;
  Heading[0][2] = 0;

  Heading[1][0] = 0;
  Heading[1][1] = 0;
  Heading[1][2] = 1;

  Heading[2][0] = 0;
  Heading[2][1] = -1;
  Heading[2][2] = 0;
}

void turn_left(void)
{
  int i, j;
  int heading[3][3];

  mmm3_mult(heading, Heading, TLeft_mat);
  for (i=0; i<3; i++)
    for (j=0; j<3; j++)
      Heading[i][j] = heading[i][j];
}

void turn_right(void)
{
  int i, j;
  int heading[3][3];

  mmm3_mult(heading, Heading, TRight_mat);
  for (i=0; i<3; i++)
    for (j=0; j<3; j++)
      Heading[i][j] = heading[i][j];
}

void turn_up(void)
{
  int i, j;
  int heading[3][3];

  mmm3_mult(heading, Heading, TUp_mat);
  for (i=0; i<3; i++)
    for (j=0; j<3; j++)
      Heading[i][j] = heading[i][j];
}

void turn_down(void)
{
  int i, j;
  int heading[3][3];

  mmm3_mult(heading, Heading, TDown_mat);
  for (i=0; i<3; i++)
    for (j=0; j<3; j++)
      Heading[i][j] = heading[i][j];
}

void turn_clockwise(void)
{
  int i, j;
  int heading[3][3];

  mmm3_mult(heading, Heading, TClockwise_mat);
  for (i=0; i<3; i++)
    for (j=0; j<3; j++)
      Heading[i][j] = heading[i][j];
}

void turn_cntr_clockwise(void)
{
  int i, j;
  int heading[3][3];

  mmm3_mult(heading, Heading, TCntrClockwise_mat);
  for (i=0; i<3; i++)
    for (j=0; j<3; j++)
      Heading[i][j] = heading[i][j];
}


/**********************************************/


int display_structure(int model_id)
{
  if (Model_type == 1)
    print_structure();

  // if (win1) {
  if (Config.do_graphics) {
    if (Model_type == 2)
      BarModel[model_id]->draw_bars(0);

    else if (Model_type == 3)
      BarModel3D[model_id]->draw_bars(0);

    else {
      draw_structure(0);
    }
    return TRUE;
  }
  return FALSE;
}


void view_structure(int arg)
{
  if (win1) {
    if (Model_type == 1) {
      display_structure(0);
    } else {
      display_structure(Model_id);
    }
  }
}


int valid_structure(void)
{
  if (Model_type == 1) {
    return test_balance();
  }
  return TRUE;
}


int build_structure(ModelStruct *model)
{
  int ret_val = TRUE;
  int length = model->lstring_length;
  Real *args = model->build_args;
  LString *command = model->lstring;

  //  printf("build structure: m:%d  b:%d\n", Model_type, Build_type);

  model->error = FALSE;
  if (Model_type == 1) {
    if (Build_type == 1) {
      ret_val = build_structure_param(length, command, args);

    } else if (Build_type == 2) {
      ret_val = build_structure_param_rel2(length, command, args);

    } else if (Build_type == 3) {
      ret_val = build_structure_param_rel(length, command, args);

    } else if (Build_type == 4) {
      ret_val = build_structure_param_obstruct(length, command, args);
    }

  } else if (Model_type == 2) {
    if (model->model_id == -1) {
      model->model_id = Num_models;
      Num_models++;
    }

    if (Build_type == 1) {
      ret_val = BarModel[model->model_id]->build_structure_param(length, command, args);
    } else if (Build_type == 2) {
      ret_val = BarModel[model->model_id]->build_structure60(length, command, args);
    }

  } else if (Model_type == 3) {
    if (model->model_id == -1) {
      model->model_id = Num_models;
      Num_models++;
    }

    ret_val = BarModel3D[model->model_id]->build_structure(model);
  } else if (Model_type == 4) {
      ret_val = build_program(length, command, args);
  }

  if (!Option[13]) {
    if (!valid_structure()) {
      if (Print[10])
	printf("Fail because not valid structure.\n");
      return FALSE;
    }
  }

  model->new_string = FALSE;
  if (!ret_val) {
    // build failed.
    if (Print[10])
      printf("Fail because build_structure() failed.\n");
    return FALSE;
  }

  return TRUE;
}


int calc_fitness(ModelStruct *model)
{
  if (Model_type == 1) {
    fitness_struct(model);
    /*
    if (Fitness_func == 1) {
      fitness_table4(model->fv_length, model->fitness);

    } else if (Fitness_func == 2) {
      fitness_staircase1(model->fv_length, model->fitness);

    } else if (Fitness_func == 3) {
      fitness_staircase2(model->fv_length, model->fitness);

    } else if (Fitness_func == 4) {
      fitness_staircase3(model->fv_length, model->fitness);

    } else if (Fitness_func == 5) {
      model->fitness[0] = fitness_tiles();
    }
    */

  } else if (Model_type == 2) {
    BarModel[model->model_id]->fitness_struct(model);

  } else if (Model_type == 3) {
    BarModel3D[model->model_id]->fitness_struct(model);

  } else if (Model_type == 4) {
    fitness_program(model);

  } else {
    printf("calc_fitness() - error: unknown fitness function: %d.", Fitness_func);
    return FALSE;
  }

  return TRUE;
}



int evaluate_structure(ModelStruct *model)
{
  int valid;

  model->error = FALSE;
  for (int i=0; i<model->fv_length; i++) {
    model->fitness[i] = 0.0;
  }

  /*
  if (!random_int(3)) {
    model->fitness[0] = random_dbl() + 1.0;
  } else {
    model->fitness[0] = 0.0;
  }    
  return TRUE;
  */


  if (Option[10]) {
    if (Print[10])
      printf("Evaluating...  %d\n", model->model_id);

    if (Model_type == 1) {
      valid = fitness_brick_blind_watchmaker(model->fv_length, model->fitness);
      if (!valid)
	return FALSE;

    } else if (Model_type == 2) {
      valid = BarModel[model->model_id]->fitness_blind_watchmaker(model->fv_length, model->fitness);
      if (!valid)
	return FALSE;

    } else if (Model_type == 3) {
      valid = BarModel3D[model->model_id]->fitness_blind_watchmaker(model->fv_length, model->fitness);
      if (!valid)
	return FALSE;
    }

    if (Print[10])
      printf("Succeeded !!!\n");

    draw_structure(0);
    model->fitness[0] = 1.0;
    return TRUE;
  }

  valid = calc_fitness(model);

  Model_id = model->model_id;

#ifndef OPTIMIZE
  if (Print[10])
    printf("Evaluating...  %d\n", Model_id);
#endif

  if (Config.do_graphics && Print[16]) {
    win1->RunViewer(-1, view_structure);
  } else if (Config.do_graphics && valid) {
    win1->RunViewer(1, view_structure);
    if (Config.save_graphics) {
      char name[20];
      sprintf(&name[0], "lsys.raw");
      win1->WriteWindow(name);
    }
  }

#ifndef OPTIMIZE
  if (model->fitness[0] >= 1.0) {
    if (Print[10])
      printf("Succeeded !!!\n");
  }
#endif
  return TRUE;
}


Real evaluate_structure(void)
{
  Real fitness;
  ModelStruct model;

  model.fv_length = Num_fit_dims;
  model.fitness = &fitness;

  evaluate_structure(&model);
  return fitness;
}


////////////////////////////////////////////////////////


void configure_graphics(Configuration config)
{
  if (config.do_graphics) {
    //    win1 = new Graphic(400, 400, "Automated Modular Design");
    //    win1->Initialize(15.0, 40.0, 0.0, 400.0, -250.0, -300.0);

    if (Option[10]) {
      win1 = new Graphic(1000, 1000, "Encoding for Automated Design");
      win1->Initialize(0.0, 0.0, 0.0, -43.0, -48.0, -100.0);
    } else {
      if (Option[16]) {
	win1 = new Graphic(320, 240, "Encoding for Automated Design");
      } else if (!Option[14]) {
	win1 = new Graphic(400, 400, "Encoding for Automated Design");
      } else if (Option[14] == 1) {
	win1 = new Graphic(320, 240, "Encoding for Automated Design");
      } else if (Option[14] == 2) {
	win1 = new Graphic(1080, 720, "Encoding for Automated Design");
	//	win1 = new Graphic(1200, 800, "Encoding for Automated Design");
      } else if (Option[14] == 3) {
	//	win1 = new Graphic(1080, 720, "Encoding for Automated Design");
	win1 = new Graphic(900, 900, "Encoding for Automated Design");
      } else {
	win1 = new Graphic(320, 240, "Encoding for Automated Design");
      }
	//	win1 = new Graphic(700, 700, "Encoding for Automated Design");
      //      win1 = new Graphic(320, 320, "Encoding for Automated Design");
      win1->Initialize(45.0, 40.0, 29.0, 110.0, -110.0, -90.0);
      //      win1->Initialize(0.0, 0.0, 0.0, 0.0, 0.0, -200);
      //      win1->Initialize(0.0, 0.0, 0.0, -10.0, -200);
    }

    if ((Option[8] == 2) || (Option[8] == 4) || (Option[8] == 6) ||
	(Option[8] == 10)) {
      win1->SetClearColor(1.0, 1.0, 1.0, 1.0);
    }
    win1->StartDrawing();
    random_dbl();
    win1->DoneDrawing();

    if (config.start_frame >= 0) {
      printf("Start frame set to: %d\n", config.start_frame);
      win1->frame_num = config.start_frame;
    }
  }
}


Real play_game(int init, Real *fitnessA, Real *fitnessB,
	       Individual *individ1, Individual *individ2)
{
  return 0.0;
}


void set_terminal_set(int val)
{
  int index;

  Terminal_set = val;
  if (Terminal_set == 1) {
    Num_terminals = 6;
    Terminal[0] = TERM_DOWN;
    Terminal[1] = TERM_UP;
    Terminal[2] = TERM_LEFT;
    Terminal[3] = TERM_RIGHT;
    Terminal[4] = TERM_FORWARD;
    Terminal[5] = TERM_BACKWARD;
    Terminal_prob[0] = 1.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 1.0;
    Terminal_prob[3] = 1.0;
    Terminal_prob[4] = 1.0;
    Terminal_prob[5] = 1.0;

  } else if (Terminal_set == 2) {
    Num_terminals = 6;
    Terminal[0] = TERM_DOWN;
    Terminal[1] = TERM_UP;
    Terminal[2] = TERM_LEFT;
    Terminal[3] = TERM_RIGHT;
    Terminal[4] = TERM_PENUP;
    Terminal[5] = TERM_PENDOWN;
    Terminal_prob[0] = 1.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 1.0;
    Terminal_prob[3] = 1.0;
    Terminal_prob[4] = 1.0;
    Terminal_prob[5] = 1.0;

  } else if (Terminal_set == 3) {
    Num_terminals = 8;
    Terminal[0] = TERM_DOWN;
    Terminal[1] = TERM_UP;
    Terminal[2] = TERM_LEFT;
    Terminal[3] = TERM_RIGHT;
    Terminal[4] = TERM_FORWARD;
    Terminal[5] = TERM_BACKWARD;
    Terminal[6] = TERM_PENUP;
    Terminal[7] = TERM_PENDOWN;
    Terminal_prob[0] = 1.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 1.0;
    Terminal_prob[3] = 1.0;
    Terminal_prob[4] = 1.0;
    Terminal_prob[5] = 1.0;
    Terminal_prob[6] = 1.0;
    Terminal_prob[7] = 1.0;

  } else if (Terminal_set == 4) {
    Num_terminals = 12;
    Terminal[0] = TERM_DOWN;
    Terminal[1] = TERM_UP;
    Terminal[2] = TERM_LEFT;
    Terminal[3] = TERM_RIGHT;
    Terminal[4] = TERM_FORWARD;
    Terminal[5] = TERM_BACKWARD;

    Terminal[6] = TERM_TURN_DOWN;
    Terminal[7] = TERM_TURN_UP;
    Terminal[8] = TERM_TURN_LEFT;
    Terminal[9] = TERM_TURN_RIGHT;
    Terminal[10] = TERM_TURN_CLOCKWISE;
    Terminal[11] = TERM_TURN_CNTR_CLOCKWISE;

    Terminal_prob[0] = 0.8;
    Terminal_prob[1] = 0.8;
    Terminal_prob[2] = 0.8;
    Terminal_prob[3] = 0.8;
    Terminal_prob[4] = 0.8;
    Terminal_prob[5] = 0.8;
    Terminal_prob[6] = 1.0;
    Terminal_prob[7] = 1.0;
    Terminal_prob[8] = 1.0;
    Terminal_prob[9] = 1.0;
    Terminal_prob[10] = 1.0;
    Terminal_prob[11] = 1.0;

  } else if (Terminal_set == 5) {
    Num_terminals = 8;
    Terminal[0] = TERM_FORWARD;
    Terminal[1] = TERM_BACKWARD;

    Terminal[2] = TERM_TURN_DOWN;
    Terminal[3] = TERM_TURN_UP;
    Terminal[4] = TERM_TURN_LEFT;
    Terminal[5] = TERM_TURN_RIGHT;
    Terminal[6] = TERM_TURN_CLOCKWISE;
    Terminal[7] = TERM_TURN_CNTR_CLOCKWISE;

    Terminal_prob[0] = 1.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 1.0;
    Terminal_prob[3] = 1.0;
    Terminal_prob[4] = 1.0;
    Terminal_prob[5] = 1.0;
    Terminal_prob[6] = 1.0;
    Terminal_prob[7] = 1.0;

  } else if (Terminal_set == 6) {
    Num_terminals = 8;
    Terminal[0] = TERM_DOWN;
    Terminal[1] = TERM_UP;
    Terminal[2] = TERM_LEFT;
    Terminal[3] = TERM_RIGHT;
    Terminal[4] = TERM_FORWARD;
    Terminal[5] = TERM_BACKWARD;
    Terminal[6] = TERM_TURN_CLOCKWISE;
    Terminal[7] = TERM_TURN_CNTR_CLOCKWISE;

    Terminal_prob[0] = 1.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 1.0;
    Terminal_prob[3] = 1.0;
    Terminal_prob[4] = 1.0;
    Terminal_prob[5] = 1.0;
    Terminal_prob[6] = 1.0;
    Terminal_prob[7] = 1.0;

  } else if (Terminal_set == 7) {
    Num_terminals = 5;
    Terminal[0] = TERM_FORWARD;
    Terminal[1] = TERM_JOINT_CLOCKWISE;
    Terminal[2] = TERM_JOINT_COUNTER;
    Terminal[3] = TERM_TURN_CLOCKWISE;
    Terminal[4] = TERM_TURN_CNTR_CLOCKWISE;

    Terminal_prob[0] = 1.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 1.0;
    Terminal_prob[3] = 1.0;
    Terminal_prob[4] = 1.0;

  } else if (Terminal_set == 8) {
    Num_terminals = 8;
    Terminal[0] = TERM_FORWARD;
    Terminal[1] = TERM_BACKWARD;
    Terminal[2] = TERM_JOINT_CLOCKWISE;
    Terminal[3] = TERM_JOINT_COUNTER;
    Terminal[4] = TERM_TURN_LEFT;
    Terminal[5] = TERM_TURN_RIGHT;
    Terminal[6] = TERM_TURN_DOWN;
    Terminal[7] = TERM_TURN_UP;

    Terminal_prob[0] = 1.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 1.0;
    Terminal_prob[3] = 1.0;
    Terminal_prob[4] = 1.0;
    Terminal_prob[5] = 1.0;
    Terminal_prob[6] = 1.0;
    Terminal_prob[7] = 1.0;

  } else if (Terminal_set == 9) {
    Num_terminals = 12;
    Terminal[0] = TERM_FORWARD;
    Terminal[1] = TERM_BACKWARD;
    Terminal[2] = TERM_JOINT_REVOLUTE;
    Terminal[3] = TERM_JOINT_TWIST;

    Terminal[4] = TERM_TURN_DOWN;
    Terminal[5] = TERM_TURN_UP;
    Terminal[6] = TERM_TURN_LEFT;
    Terminal[7] = TERM_TURN_RIGHT;
    Terminal[8] = TERM_TURN_CLOCKWISE;
    Terminal[9] = TERM_TURN_CNTR_CLOCKWISE;

    Terminal[10] = TERM_OFFSET_INCREASE;
    Terminal[11] = TERM_OFFSET_DECREASE;

    Terminal_prob[0] = 2.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 0.25;
    Terminal_prob[3] = 0.25;
    Terminal_prob[4] = 1.0;
    Terminal_prob[5] = 1.0;
    Terminal_prob[6] = 1.0;
    Terminal_prob[7] = 1.0;
    Terminal_prob[8] = 1.5;
    Terminal_prob[9] = 1.5;
    Terminal_prob[10] = 2.0;
    Terminal_prob[11] = 2.0;

  } else if (Terminal_set == 10) {
    Num_terminals = 6;
    Terminal[0] = TERM_FORWARD;
    Terminal[1] = TERM_BACKWARD;
    Terminal[2] = TERM_JOINT_CLOCKWISE;
    Terminal[3] = TERM_JOINT_COUNTER;
    Terminal[4] = TERM_TURN_CLOCKWISE;
    Terminal[5] = TERM_TURN_CNTR_CLOCKWISE;

    Terminal_prob[0] = 1.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 0.25;
    Terminal_prob[3] = 0.25;
    Terminal_prob[4] = 2.0;
    Terminal_prob[5] = 2.0;

  } else if (Terminal_set == 11) {
    Num_terminals = 8;
    Terminal[0] = TERM_FORWARD;
    Terminal[1] = TERM_BACKWARD;
    Terminal[2] = TERM_JOINT_CLOCKWISE;
    Terminal[3] = TERM_JOINT_COUNTER;
    Terminal[4] = TERM_TURN_CLOCKWISE;
    Terminal[5] = TERM_TURN_CNTR_CLOCKWISE;
    Terminal[6] = TERM_OFFSET_INCREASE;
    Terminal[7] = TERM_OFFSET_DECREASE;

    Terminal_prob[0] = 1.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 0.25;
    Terminal_prob[3] = 0.25;
    Terminal_prob[4] = 2.0;
    Terminal_prob[5] = 2.0;
    Terminal_prob[6] = 1.0;
    Terminal_prob[7] = 1.0;

  } else if (Terminal_set == 12) {
    Num_terminals = 10;
    Terminal[0] = TERM_FORWARD;
    Terminal[1] = TERM_BACKWARD;
    Terminal[2] = TERM_JOINT_CLOCKWISE;
    Terminal[3] = TERM_JOINT_COUNTER;
    Terminal[4] = TERM_JOINT_CLOCKWISE2;
    Terminal[5] = TERM_JOINT_COUNTER2;
    Terminal[6] = TERM_TURN_CLOCKWISE;
    Terminal[7] = TERM_TURN_CNTR_CLOCKWISE;
    Terminal[8] = TERM_OFFSET_INCREASE;
    Terminal[9] = TERM_OFFSET_DECREASE;

    Terminal_prob[0] = 1.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 0.25;
    Terminal_prob[3] = 0.25;
    Terminal_prob[4] = 0.25;
    Terminal_prob[5] = 0.25;
    Terminal_prob[6] = 2.0;
    Terminal_prob[7] = 2.0;
    Terminal_prob[8] = 1.0;
    Terminal_prob[9] = 1.0;

  } else if (Terminal_set == 13) {
    Num_terminals = 14;
    Terminal[0] = TERM_FORWARD;
    Terminal[1] = TERM_BACKWARD;
    Terminal[2] = TERM_JOINT_REVOLUTE;
    Terminal[3] = TERM_JOINT_REVOLUTE2;
    Terminal[4] = TERM_JOINT_TWIST;
    Terminal[5] = TERM_JOINT_TWIST180;

    Terminal[6] = TERM_TURN_DOWN;
    Terminal[7] = TERM_TURN_UP;
    Terminal[8] = TERM_TURN_LEFT;
    Terminal[9] = TERM_TURN_RIGHT;
    Terminal[10] = TERM_TURN_CLOCKWISE;
    Terminal[11] = TERM_TURN_CNTR_CLOCKWISE;

    Terminal[12] = TERM_OFFSET_INCREASE;
    Terminal[13] = TERM_OFFSET_DECREASE;

    Terminal_prob[0] = 2.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 0.25;
    Terminal_prob[3] = 0.25;
    Terminal_prob[4] = 0.25;
    Terminal_prob[5] = 0.25;
    Terminal_prob[6] = 1.0;
    Terminal_prob[7] = 1.0;
    Terminal_prob[8] = 1.0;
    Terminal_prob[9] = 1.0;
    Terminal_prob[10] = 1.5;
    Terminal_prob[11] = 1.5;
    Terminal_prob[12] = 2.0;
    Terminal_prob[13] = 2.0;

  } else if (Terminal_set == 14) {
    // Same as #13, different probabilities.
    Num_terminals = 14;
    Terminal[0] = TERM_FORWARD;
    Terminal[1] = TERM_BACKWARD;
    Terminal[2] = TERM_JOINT_REVOLUTE;
    Terminal[3] = TERM_JOINT_REVOLUTE2;
    Terminal[4] = TERM_JOINT_TWIST;
    Terminal[5] = TERM_JOINT_TWIST180;

    Terminal[6] = TERM_TURN_DOWN;
    Terminal[7] = TERM_TURN_UP;
    Terminal[8] = TERM_TURN_LEFT;
    Terminal[9] = TERM_TURN_RIGHT;
    Terminal[10] = TERM_TURN_CLOCKWISE;
    Terminal[11] = TERM_TURN_CNTR_CLOCKWISE;

    Terminal[12] = TERM_OFFSET_INCREASE;
    Terminal[13] = TERM_OFFSET_DECREASE;

    Terminal_prob[0] = 1.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 0.25;
    Terminal_prob[3] = 0.25;
    Terminal_prob[4] = 0.25;
    Terminal_prob[5] = 0.25;

    Terminal_prob[6] = 2.0;
    Terminal_prob[7] = 2.0;
    Terminal_prob[8] = 2.0;
    Terminal_prob[9] = 2.0;
    Terminal_prob[10] = 2.5;
    Terminal_prob[11] = 2.5;
    Terminal_prob[12] = 2.0;
    Terminal_prob[13] = 2.0;

  } else if (Terminal_set == 15) {
    // Same as #13, different probabilities.
    // Plus color sensors.
    Num_terminals = 15;
    Terminal[0] = TERM_FORWARD;
    Terminal[1] = TERM_BACKWARD;
    Terminal[2] = TERM_JOINT_REVOLUTE;
    Terminal[3] = TERM_JOINT_REVOLUTE2;
    Terminal[4] = TERM_JOINT_TWIST;
    Terminal[5] = TERM_JOINT_TWIST180;
    Terminal[6] = TERM_COLOR_SENSOR;

    Terminal[7] = TERM_TURN_DOWN;
    Terminal[8] = TERM_TURN_UP;
    Terminal[9] = TERM_TURN_LEFT;
    Terminal[10] = TERM_TURN_RIGHT;
    Terminal[11] = TERM_TURN_CLOCKWISE;
    Terminal[12] = TERM_TURN_CNTR_CLOCKWISE;

    Terminal[13] = TERM_OFFSET_INCREASE;
    Terminal[14] = TERM_OFFSET_DECREASE;

    Terminal_prob[0] = 1.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 0.25;
    Terminal_prob[3] = 0.25;
    Terminal_prob[4] = 0.25;
    Terminal_prob[5] = 0.25;
    Terminal_prob[6] = 0.25;

    Terminal_prob[7] = 2.0;
    Terminal_prob[8] = 2.0;
    Terminal_prob[9] = 2.0;
    Terminal_prob[10] = 2.0;
    Terminal_prob[11] = 2.5;
    Terminal_prob[12] = 2.5;
    Terminal_prob[13] = 2.0;
    Terminal_prob[14] = 2.0;

  } else if (Terminal_set == 16) {
    // Same as #14, no twist.
    Num_terminals = 12;
    Terminal[0] = TERM_FORWARD;
    Terminal[1] = TERM_BACKWARD;
    Terminal[2] = TERM_JOINT_REVOLUTE;
    Terminal[3] = TERM_JOINT_REVOLUTE2;

    Terminal[4] = TERM_TURN_DOWN;
    Terminal[5] = TERM_TURN_UP;
    Terminal[6] = TERM_TURN_LEFT;
    Terminal[7] = TERM_TURN_RIGHT;
    Terminal[8] = TERM_TURN_CLOCKWISE;
    Terminal[9] = TERM_TURN_CNTR_CLOCKWISE;

    Terminal[10] = TERM_OFFSET_INCREASE;
    Terminal[11] = TERM_OFFSET_DECREASE;

    Terminal_prob[0] = 1.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 0.25;
    Terminal_prob[3] = 0.25;

    Terminal_prob[4] = 2.0;
    Terminal_prob[5] = 2.0;
    Terminal_prob[6] = 2.0;
    Terminal_prob[7] = 2.0;
    Terminal_prob[8] = 2.5;
    Terminal_prob[9] = 2.5;
    Terminal_prob[10] = 2.0;
    Terminal_prob[11] = 2.0;

  } else if (Terminal_set == 17) {
    // Building bodies
    index = 0;

    Terminal[index] = TERM_FORWARD;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_BACKWARD;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_JOINT_REVOLUTE;
    Terminal_prob[index++] = 0.5;
    Terminal[index] = TERM_JOINT_REVOLUTE2;
    Terminal_prob[index++] = 0.5;
    Terminal[index] = TERM_JOINT_TWIST;
    Terminal_prob[index++] = 0.5;
    Terminal[index] = TERM_JOINT_TWIST180;
    Terminal_prob[index++] = 0.5;

    Terminal[index] = TERM_TURN_DOWN;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_UP;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_LEFT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_RIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CLOCKWISE;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CNTR_CLOCKWISE;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_OFFSET_INCREASE;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_OFFSET_DECREASE;
    Terminal_prob[index++] = 1.0;

    Num_terminals = index;

  } else if (Terminal_set == 18) {
    // Same as #16, no revolute2.
    // Building bodies
    index = 0;

    Terminal[index] = TERM_FORWARD;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_BACKWARD;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_JOINT_REVOLUTE;
    Terminal_prob[index++] = 0.5;

    Terminal[index] = TERM_TURN_DOWN;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_UP;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_LEFT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_RIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CLOCKWISE;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CNTR_CLOCKWISE;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_OFFSET_INCREASE;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_OFFSET_DECREASE;
    Terminal_prob[index++] = 1.0;

    Num_terminals = index;

  } else if (Terminal_set == 20) {
    // Building networks and bodies.
    Num_terminals = 25;

    Terminal[0] = TERM_FORWARD;
    Terminal[1] = TERM_BACKWARD;
    Terminal[2] = TERM_JOINT_REVOLUTE;
    Terminal[3] = TERM_JOINT_REVOLUTE2;
    Terminal[4] = TERM_JOINT_TWIST;
    Terminal[5] = TERM_JOINT_TWIST180;

    Terminal[6] = TERM_TURN_DOWN;
    Terminal[7] = TERM_TURN_UP;
    Terminal[8] = TERM_TURN_LEFT;
    Terminal[9] = TERM_TURN_RIGHT;
    Terminal[10] = TERM_TURN_CLOCKWISE;
    Terminal[11] = TERM_TURN_CNTR_CLOCKWISE;

    Terminal[12] = TERM_SPLIT_LINK;
    Terminal[13] = TERM_DUPLICATE_LINK;
    Terminal[14] = TERM_DELETE_LINK;
    Terminal[15] = TERM_NEXT_LINK;
    Terminal[16] = TERM_PARENT_NEURON;
    Terminal[17] = TERM_MERGE_NEURONS;
    Terminal[18] = TERM_INCREASE_WEIGHT;
    Terminal[19] = TERM_DECREASE_WEIGHT;
    Terminal[20] = TERM_REVERSE_LINK;
    Terminal[21] = TERM_SPLIT_LINK_NEG;
    Terminal[22] = TERM_DUPLICATE_LINK_NEG;
    Terminal[23] = TERM_SET_OUTPUT_FUNCTION;
    Terminal[24] = TERM_MAKE_LOOP;

    Terminal_prob[0] = 1.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 0.25;
    Terminal_prob[3] = 0.25;
    Terminal_prob[4] = 0.25;
    Terminal_prob[5] = 0.25;

    Terminal_prob[6] = 2.0;
    Terminal_prob[7] = 2.0;
    Terminal_prob[8] = 2.0;
    Terminal_prob[9] = 2.0;
    Terminal_prob[10] = 2.5;
    Terminal_prob[11] = 2.5;

    Terminal_prob[12] = 1.0;
    Terminal_prob[13] = 1.0;
    Terminal_prob[14] = 1.0;
    Terminal_prob[15] = 5.0;
    Terminal_prob[16] = 5.0;
    Terminal_prob[17] = 3.0;
    Terminal_prob[18] = 1.0;
    Terminal_prob[19] = 1.0;
    Terminal_prob[20] = 1.0;
    Terminal_prob[21] = 1.0;
    Terminal_prob[22] = 1.0;
    Terminal_prob[23] = 1.0;
    Terminal_prob[24] = 5.0;

  } else if (Terminal_set == 21) {
    // Building networks and bodies.
    Num_terminals = 23;

    Terminal[0] = TERM_FORWARD;
    Terminal[1] = TERM_BACKWARD;
    Terminal[2] = TERM_JOINT_REVOLUTE;
    Terminal[3] = TERM_JOINT_REVOLUTE2;

    Terminal[4] = TERM_TURN_DOWN;
    Terminal[5] = TERM_TURN_UP;
    Terminal[6] = TERM_TURN_LEFT;
    Terminal[7] = TERM_TURN_RIGHT;
    Terminal[8] = TERM_TURN_CLOCKWISE;
    Terminal[9] = TERM_TURN_CNTR_CLOCKWISE;

    Terminal[10] = TERM_SPLIT_LINK;
    Terminal[11] = TERM_DUPLICATE_LINK;
    Terminal[12] = TERM_DELETE_LINK;
    Terminal[13] = TERM_NEXT_LINK;
    Terminal[14] = TERM_PARENT_NEURON;
    Terminal[15] = TERM_MERGE_NEURONS;
    Terminal[16] = TERM_INCREASE_WEIGHT;
    Terminal[17] = TERM_DECREASE_WEIGHT;
    Terminal[18] = TERM_REVERSE_LINK;
    Terminal[19] = TERM_SPLIT_LINK_NEG;
    Terminal[20] = TERM_DUPLICATE_LINK_NEG;
    Terminal[21] = TERM_SET_OUTPUT_FUNCTION;
    Terminal[22] = TERM_MAKE_LOOP;

    Terminal_prob[0] = 1.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 0.25;
    Terminal_prob[3] = 0.25;

    Terminal_prob[4] = 2.0;
    Terminal_prob[5] = 2.0;
    Terminal_prob[6] = 2.0;
    Terminal_prob[7] = 2.0;
    Terminal_prob[8] = 2.5;
    Terminal_prob[9] = 2.5;

    Terminal_prob[10] = 1.0;
    Terminal_prob[11] = 1.0;
    Terminal_prob[12] = 1.0;
    Terminal_prob[13] = 2.0;
    Terminal_prob[14] = 2.0;
    Terminal_prob[15] = 1.0;
    Terminal_prob[16] = 1.0;
    Terminal_prob[17] = 1.0;
    Terminal_prob[18] = 1.0;
    Terminal_prob[19] = 1.0;
    Terminal_prob[20] = 1.0;
    Terminal_prob[21] = 1.0;
    Terminal_prob[22] = 2.0;

  } else if (Terminal_set == 22) {
    // Building networks and bodies.
    index = 0;

    Terminal[index] = TERM_FORWARD;
    Terminal_prob[index++] = 2.0;
    Terminal[index] = TERM_BACKWARD;
    Terminal_prob[index++] = 2.0;
    Terminal[index] = TERM_JOINT_REVOLUTE;
    Terminal_prob[index++] = 0.5;
    Terminal[index] = TERM_JOINT_REVOLUTE2;
    Terminal_prob[index++] = 0.5;

    Terminal[index] = TERM_TURN_DOWN;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_UP;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_LEFT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_RIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CLOCKWISE;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CNTR_CLOCKWISE;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_SPLIT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DUPLICATE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_MAKE_LOOP;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_MERGE_NEURONS;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_NEXT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_PARENT_NEURON;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_REVERSE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_INCREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DECREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_SET_OUTPUT_FUNCTION;
    Terminal_prob[index++] = 1.0;

    Num_terminals = index;

  } else if (Terminal_set == 23) {
    // Building networks and bodies.
    index = 0;

    Terminal[index] = TERM_FORWARD;
    Terminal_prob[index++] = 2.0;
    Terminal[index] = TERM_BACKWARD;
    Terminal_prob[index++] = 2.0;
    Terminal[index] = TERM_JOINT_REVOLUTE;
    Terminal_prob[index++] = 0.7;
    Terminal[index] = TERM_JOINT_REVOLUTE2;
    Terminal_prob[index++] = 0.7;
    Terminal[index] = TERM_JOINT_TWIST;
    Terminal_prob[index++] = 0.5;
    Terminal[index] = TERM_JOINT_TWIST180;
    Terminal_prob[index++] = 0.5;

    Terminal[index] = TERM_TURN_DOWN;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_UP;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_LEFT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_RIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CLOCKWISE;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CNTR_CLOCKWISE;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_SPLIT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DUPLICATE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_MAKE_LOOP;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_MERGE_NEURONS;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_NEXT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_PARENT_NEURON;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_REVERSE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_INCREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DECREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_SET_OUTPUT_FUNCTION;
    Terminal_prob[index++] = 1.0;

    Num_terminals = index;

  } else if (Terminal_set == 25) {
    // Building networks and bodies w/ sensors.
    index = 0;

    Terminal[index] = TERM_FORWARD;
    Terminal_prob[index++] = 2.0;
    Terminal[index] = TERM_BACKWARD;
    Terminal_prob[index++] = 2.0;
    Terminal[index] = TERM_JOINT_REVOLUTE;
    Terminal_prob[index++] = 0.5;
    Terminal[index] = TERM_JOINT_REVOLUTE2;
    Terminal_prob[index++] = 0.5;
    Terminal[index] = TERM_COLOR_SENSOR;
    Terminal_prob[index++] = 0.5;

    Terminal[index] = TERM_TURN_DOWN;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_UP;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_LEFT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_RIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CLOCKWISE;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CNTR_CLOCKWISE;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_SPLIT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DUPLICATE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_MAKE_LOOP;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_MERGE_NEURONS;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_NEXT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_PARENT_NEURON;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_REVERSE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_INCREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DECREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_SET_OUTPUT_FUNCTION;
    Terminal_prob[index++] = 1.0;

    Num_terminals = index;

  } else if (Terminal_set == 26) {
    // Building networks and bodies w/ sensors.
    index = 0;

    Terminal[index] = TERM_FORWARD;
    Terminal_prob[index++] = 2.0;
    Terminal[index] = TERM_BACKWARD;
    Terminal_prob[index++] = 2.0;
    Terminal[index] = TERM_JOINT_REVOLUTE;
    Terminal_prob[index++] = 0.7;
    Terminal[index] = TERM_JOINT_REVOLUTE2;
    Terminal_prob[index++] = 0.7;
    Terminal[index] = TERM_JOINT_TWIST;
    Terminal_prob[index++] = 0.5;
    Terminal[index] = TERM_JOINT_TWIST180;
    Terminal_prob[index++] = 0.5;
    Terminal[index] = TERM_COLOR_SENSOR;
    Terminal_prob[index++] = 0.5;

    Terminal[index] = TERM_TURN_DOWN;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_UP;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_LEFT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_RIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CLOCKWISE;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CNTR_CLOCKWISE;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_SPLIT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DUPLICATE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_MAKE_LOOP;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_MERGE_NEURONS;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_NEXT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_PARENT_NEURON;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_REVERSE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_INCREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DECREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_SET_OUTPUT_FUNCTION;
    Terminal_prob[index++] = 1.0;

    Num_terminals = index;

  } else if (Terminal_set == 28) {
    // Building networks and bodies for remote control.
    index = 0;

    Terminal[index] = TERM_FORWARD;
    Terminal_prob[index++] = 2.0;
    Terminal[index] = TERM_BACKWARD;
    Terminal_prob[index++] = 2.0;
    Terminal[index] = TERM_JOINT_REVOLUTE;
    Terminal_prob[index++] = 0.7;
    Terminal[index] = TERM_JOINT_REVOLUTE2;
    Terminal_prob[index++] = 0.7;
    Terminal[index] = TERM_JOINT_TWIST;
    Terminal_prob[index++] = 0.5;
    Terminal[index] = TERM_JOINT_TWIST180;
    Terminal_prob[index++] = 0.5;
    Terminal[index] = TERM_CONNECT_TO_INPUT;
    Terminal_prob[index++] = 4.0;

    Terminal[index] = TERM_TURN_DOWN;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_UP;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_LEFT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_RIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CLOCKWISE;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CNTR_CLOCKWISE;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_SPLIT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DUPLICATE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_MAKE_LOOP;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_MERGE_NEURONS;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_NEXT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_PARENT_NEURON;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_REVERSE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_INCREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DECREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_SET_OUTPUT_FUNCTION;
    Terminal_prob[index++] = 1.0;

    Num_terminals = index;

  } else if (Terminal_set == 30) {
    // Building networks.
    Num_terminals = 13;
    Terminal[0] = TERM_ADD_INPUT;
    Terminal[1] = TERM_ADD_OUTPUT;
    Terminal[2] = TERM_SPLIT_LINK;
    Terminal[3] = TERM_DUPLICATE_LINK;
    Terminal[4] = TERM_DELETE_LINK;
    Terminal[5] = TERM_NEXT_LINK;

    Terminal[6] = TERM_PARENT_NEURON;
    Terminal[7] = TERM_MERGE_NEURONS;
    Terminal[8] = TERM_INCREASE_WEIGHT;
    Terminal[9] = TERM_DECREASE_WEIGHT;
    Terminal[10] = TERM_REVERSE_LINK;
    Terminal[11] = TERM_SET_OUTPUT_FUNCTION;
    Terminal[12] = TERM_MAKE_LOOP;

    Terminal_prob[0] = 1.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 1.0;
    Terminal_prob[3] = 1.0;
    Terminal_prob[4] = 1.0;
    Terminal_prob[5] = 3.0;
    Terminal_prob[6] = 3.0;
    Terminal_prob[7] = 3.0;
    Terminal_prob[8] = 1.0;
    Terminal_prob[9] = 1.0;
    Terminal_prob[10] = 1.0;
    Terminal_prob[11] = 1.0;
    Terminal_prob[12] = 1.0;

  } else if (Terminal_set == 31) {
    // Building networks.
    Num_terminals = 15;
    Terminal[0] = TERM_ADD_INPUT;
    Terminal[1] = TERM_ADD_OUTPUT;
    Terminal[2] = TERM_SPLIT_LINK;
    Terminal[3] = TERM_DUPLICATE_LINK;
    Terminal[4] = TERM_DELETE_LINK;
    Terminal[5] = TERM_NEXT_LINK;

    Terminal[6] = TERM_PARENT_NEURON;
    Terminal[7] = TERM_MERGE_NEURONS;
    Terminal[8] = TERM_INCREASE_WEIGHT;
    Terminal[9] = TERM_DECREASE_WEIGHT;
    Terminal[10] = TERM_REVERSE_LINK;
    Terminal[11] = TERM_SPLIT_LINK_NEG;
    Terminal[12] = TERM_DUPLICATE_LINK_NEG;
    Terminal[13] = TERM_SET_OUTPUT_FUNCTION;
    Terminal[14] = TERM_MAKE_LOOP;

    Terminal_prob[0] = 1.0;
    Terminal_prob[1] = 1.0;
    Terminal_prob[2] = 1.0;
    Terminal_prob[3] = 1.0;
    Terminal_prob[4] = 1.0;
    Terminal_prob[5] = 1.0;
    Terminal_prob[6] = 1.0;
    Terminal_prob[7] = 1.0;
    Terminal_prob[8] = 1.0;
    Terminal_prob[9] = 1.0;
    Terminal_prob[10] = 1.0;
    Terminal_prob[11] = 1.0;
    Terminal_prob[12] = 1.0;
    Terminal_prob[13] = 1.0;
    Terminal_prob[14] = 1.0;

  } else if (Terminal_set == 32) {
    // Building networks.
    index = 0;

    Terminal[index] = TERM_ADD_INPUT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_ADD_OUTPUT;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_SPLIT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DUPLICATE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_MAKE_LOOP;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_MERGE_NEURONS;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_NEXT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_PARENT_NEURON;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_REVERSE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_INCREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DECREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_SET_OUTPUT_FUNCTION;
    Terminal_prob[index++] = 1.0;

    Num_terminals = index;

  } else if (Terminal_set == 33) {
    // Building networks.
    index = 0;

    Terminal[index] = TERM_ADD_INPUT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_ADD_OUTPUT;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_SPLIT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DUPLICATE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_MAKE_LOOP;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_MERGE_NEURONS;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_NEXT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_PARENT_NEURON;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_REVERSE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_INCREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DECREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_NEGATE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_SET_OUTPUT_FUNCTION;
    Terminal_prob[index++] = 1.0;

    Num_terminals = index;

  } else if (Terminal_set == 40) {
    // Building networks.
    index = 0;

    Terminal[index] = TERM_ADD_INPUT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_ADD_OUTPUT;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_SPLIT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DUPLICATE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_MAKE_LOOP;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_MERGE_NEURONS;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_NEXT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_PARENT_NEURON;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_REVERSE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_INCREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DECREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;

    Num_terminals = index;

  } else if (Terminal_set == 41) {
    // Building bodies & networks - only simple sigmoid function.
    index = 0;

    Terminal[index] = TERM_FORWARD;
    Terminal_prob[index++] = 4.0;
    Terminal[index] = TERM_BACKWARD;
    Terminal_prob[index++] = 4.0;
    Terminal[index] = TERM_JOINT_REVOLUTE;
    Terminal_prob[index++] = 2.0;
    Terminal[index] = TERM_JOINT_REVOLUTE2;
    Terminal_prob[index++] = 2.0;

    Terminal[index] = TERM_TURN_DOWN;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_UP;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_LEFT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_RIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CLOCKWISE;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CNTR_CLOCKWISE;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_SPLIT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DUPLICATE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_MAKE_LOOP;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_MERGE_NEURONS;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_NEXT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_PARENT_NEURON;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_REVERSE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_INCREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DECREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;

    Num_terminals = index;    

  } else if (Terminal_set == 42) {
    // Building bodies & networks - only simple sigmoid function.
    index = 0;

    Terminal[index] = TERM_FORWARD;
    Terminal_prob[index++] = 2.0;
    Terminal[index] = TERM_BACKWARD;
    Terminal_prob[index++] = 2.0;
    Terminal[index] = TERM_JOINT_REVOLUTE;
    Terminal_prob[index++] = 0.7;
    Terminal[index] = TERM_JOINT_REVOLUTE2;
    Terminal_prob[index++] = 0.7;
    Terminal[index] = TERM_JOINT_TWIST;
    Terminal_prob[index++] = 0.5;
    Terminal[index] = TERM_JOINT_TWIST180;
    Terminal_prob[index++] = 0.5;

    Terminal[index] = TERM_TURN_DOWN;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_UP;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_LEFT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_RIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CLOCKWISE;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_TURN_CNTR_CLOCKWISE;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_SPLIT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DUPLICATE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_MAKE_LOOP;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_MERGE_NEURONS;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_NEXT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_PARENT_NEURON;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_REVERSE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_INCREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DECREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;

    Num_terminals = index;

  } else if (Terminal_set == 50) {
    // Building networks for parity problem.
    index = 0;

    Terminal[index] = TERM_ADD_INPUT;
    Terminal_prob[index++] = 0.25;
    Terminal[index] = TERM_ADD_OUTPUT;
    Terminal_prob[index++] = 0.1;

    Terminal[index] = TERM_SPLIT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DUPLICATE_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_MAKE_LOOP;
    Terminal_prob[index++] = 1.0;

    Terminal[index] = TERM_MERGE_NEURONS;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_NEXT_LINK;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_PARENT_NEURON;
    Terminal_prob[index++] = 3.0;
    Terminal[index] = TERM_REVERSE_LINK;
    Terminal_prob[index++] = 2.0;
    /*
    Terminal[index] = TERM_SET_OUTPUT_FUNCTION;
    Terminal_prob[index++] = 1.0;
    */
    Terminal[index] = TERM_INCREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_DECREASE_WEIGHT;
    Terminal_prob[index++] = 1.0;

    Num_terminals = index;

  } else if (Terminal_set == 60) {
    // Building programs for sorting.
    index = 0;

    Terminal[index] = TERM_LOAD_ACC;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_STORE_ACC;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_COMPARE_ACC;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_PUSH_ACC;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_POP_ACC;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_CLEAR_FLAG;
    Terminal_prob[index++] = 1.0;

    Num_terminals = index;

  } else if (Terminal_set == 61) {
    // Building programs for sorting.
    index = 0;

    Terminal[index] = TERM_COMPARE_DATA;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_SWAP_DATA;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_CLEAR_FLAG;
    Terminal_prob[index++] = 1.0;
    // add HALT instruction? (end of program)
    // add BREAK instruction? (pop up two levels)

    Num_terminals = index;

  } else if (Terminal_set == 62) {
    // Building programs for sorting.
    index = 0;

    Terminal[index] = TERM_COMPARE_DATA;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_SWAP_DATA;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_CLEAR_FLAG;
    Terminal_prob[index++] = 1.0;
    Terminal[index] = TERM_HALT;
    Terminal_prob[index++] = 0.05;
    // add BREAK instruction? (pop up two levels)

    Num_terminals = index;

  } else if (Terminal_set == 65) {
    // Building programs for sorting.
    index = 0;

    Terminal[index] = TERM_SWAP_DATA;
    Terminal_prob[index++] = 1.0;
    Prob_branch = 0.0; // No branches/IFs because no compares.
    Num_terminals = index;
  }


  int i;
  Real sum = Terminal_prob[0];
  for (i=1; i<Num_terminals; i++) {
    sum += Terminal_prob[i];
    Terminal_prob[i] += Terminal_prob[i-1];
  }
  for (i=0; i<Num_terminals; i++)
    Terminal_prob[i] /= sum;
}


void configure_pop(Pop *pop)
{
  Max_bodies = 2;
  Num_arguments = 2;
  Num_operations = NUM_RECOMB_OPS + NUM_MUTATE_OPS;
  Max_operator = Num_operations + 3;
  Num_productions = NUM_PRODUCTIONS;

  Num_fit_dims = 4;
  //  Num_fit_dims = 8;


  // Set constants
  //  Terminal_set = 5;
  Fitness_func = 1;
  Terminal_set = 5; // or 4
  Build_type = 1;
  Model_type = 1;

  if (Option[1]) {
    Fitness_func = Option[1];
    printf("Using fitness function #%d.\n", Fitness_func);
  }

  if (Option[2]) {
    Terminal_set = Option[2];
  }

  if (Option[3]) {
    Build_type = Option[3];
    printf("Build type set to: %d.\n", Build_type);
  }

  if (Option[4]) {
    Model_type = Option[4];
    printf("Model type set to: %d.\n", Model_type);
    if (Model_type == 1) {
      if (Terminal_set > 6)
	Terminal_set = 5;

    } else if (Model_type == 2) {
      if (Terminal_set < 7)
	Terminal_set = 10;

    } else if (Model_type == 3) {
      if (Build_type == 1) {
	// Doing creatures.
	if (Terminal_set < 7)
	  Terminal_set = 9;
	if (!Option[3]) {
	  Build_type = 2;
	  printf("Build type set to: %d.\n", Build_type);
	}

      } else if (Build_type == 5) {
	// Doing networks.
	if (Terminal_set < 20)
	  Terminal_set = 20;
	else if ((Terminal_set < 40) && (Terminal_set >= 30)) {
	  Fitness_func = 10;
	  printf("Using fitness function #10.\n");
	}

      } else if (Build_type == 10) {
	// Doing trees.
	Terminal_set = 5;
      }
    }

  } else if (Model_type == 4) {
    Pvalue_type = 1;
  }

  if (Model_type == 1) {
    Num_fit_dims = 8;
  } else if ((Model_type == 3) && (Build_type == 5) && (Terminal_set == 50)) {
    Num_fit_dims = 6;
  }

  printf("Using terminal set #%d.\n", Terminal_set);
  set_terminal_set(Terminal_set);
  if (Terminal_set < 20) {
    Prob_branch_edge = 0.0;
  } else if (Terminal_set >= 50) {
    Prob_branch_edge = 0.0;
  }


  if (Option[5]) {
    if (Option[5] < 0)
      Num_arguments = 0;
    else
      Num_arguments = Option[5];
    printf("Using %d argument(s).\n", Num_arguments);
  }

  if (Option[6]) {
    Max_bodies = Option[6];
    printf("Maximum number of production bodies now: %d\n", Max_bodies);
  }

  if (Option[7]) {
    Num_productions = Option[7];
    printf("Number of productions set to: %d.\n", Num_productions);
  }

  if (Option[15]) {
    Max_joints = Option[15];
    printf("Maximum number of joints set to: %d.\n", Max_joints);
  }

  if ((Model_type != 3) && (Build_type != 5)) {
    Prob_branch_edge = 0.0;
  }

  if ((Num_productions == 1) && (Num_arguments == 0)) {
    Prob_mult_seq = 0.0;
    Max_prod_length = Max_string_length - 100;
    Prod_length = Max_string_length;
    printf("Doing direct encoding.\n");
  }

  Rec_delta = 1.0;
  Individual *sample = new Individual();
  sample->iterations = 8;
  sample->init();
  pop->set_sample_individ(sample);

  pop->structure = STRUCT_CANONICAL;
  //  pop->structure = STRUCT_MULTI_OBJECTIVE;
  //  pop->structure = STRUCT_DETERMINISTIC_CROWDING;
  //  pop->structure = STRUCT_ELITIST_STEADY_STATE;
  pop->num_individs = 200; // 200
  pop->nghbr_size = 25;
  //  pop->nghbr_type = NGHBR_DEME;
  pop->nghbr_type = NGHBR_NONE;
  pop->recomb_rate = 0.50;
  pop->adaptive_mut = FALSE;
  pop->elitism = 2;
  if (pop->structure == STRUCT_MULTI_OBJECTIVE) {
    pop->elitism = 0;
    Num_fit_dims = 8;
  }
  pop->pressure = 0.03;
  pop->set_tournament_size(4);
  pop->set_replace_size(2);
  pop->max_age = 0;
  pop->max_generations = 100;

  if (Option[10]) {
    pop->structure = STRUCT_BLIND_WATCHMAKER;
    pop->num_individs = 16;

    BarModel = (BarStruct**)malloc(32*sizeof(BarStruct*));
    for (int i=0; i<32; i++)
      BarModel[i] = new BarStruct();

    BarModel3D = (BarStruct3D**)malloc(32*sizeof(BarStruct3D*));
    for (int i=0; i<32; i++)
      BarModel3D[i] = new BarStruct3D();

  } else {
    BarModel = (BarStruct**)malloc(sizeof(BarStruct*));
    BarModel[0] = new BarStruct();

    BarModel3D = (BarStruct3D**)malloc(sizeof(BarStruct3D*));
    BarModel3D[0] = new BarStruct3D();
  }

  if (Config.num_trials > 1) {
    printf("Pop size: %d; Gen: %d, Pressure: %.4f, Elitism: %d.\n",
	   pop->num_individs, pop->max_generations,
	   pop->pressure, pop->elitism);
    printf("fit:%d  term-set:%d  build:%d  model:%d  #args:%d  #bodys:%d  #prod:%d\n",
	   Fitness_func, Terminal_set, Build_type, Model_type,
	   Num_arguments, Max_bodies, Num_productions);
    printf("prod-length: %d, max-length: %d, string-len: %d.\n",
	   Prod_length, Max_prod_length, Max_string_length);
  }
}

void test_iteration1(Individual *ind)
{
  int j;

  cout << "test_iteration1" << endl;

  if (win1 == 0) {
    win1 = new Graphic(400, 400, "Encoding for Automated Design");
    win1->Initialize(0.0, 0.0, 0.0, 0.0, -250);

    win1->StartDrawing();
    win1->DoneDrawing();
  }

  //  build_structure(ind);

  for (j=0; j<100000000; j++)
    ;
}


void Run_BlindWatchmaker(Deme *deme)
{
  if (win1 == 0)
    return;
  
  win1->BlindWatchMaker(deme);
}
