/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#ifndef BUILDER_HEADER_FILE
#define BUILDER_HEADER_FILE

#include "general.h"

class Graphic;
class Individual;
class Pop;
class Deme;

// Build commands
#define NUM_BUILD_COMMANDS 6
#define UP_RIGHT '\''
#define UP 'u'
#define UP_LEFT '`'
#define DOWN_RIGHT 'r'
#define DOWN 'd'
#define DOWN_LEFT 'l'
#define PUSH '['
#define POP ']'

#define TERM_PUSH_EDGE '('
#define TERM_POP_EDGE ')'

#define TERM_PUSH '['
#define TERM_POP ']'
#define TERM_UP 'u'
#define TERM_DOWN 'd'
#define TERM_LEFT 'l'
#define TERM_RIGHT 'r'
#define TERM_FORWARD 'f'
#define TERM_BACKWARD 'b' // was conflict with "duplicate negative" for neural nets (but not oscillators)

#define TERM_JOINT 'j'
#define TERM_JOINT_COUNTER 'L'
#define TERM_JOINT_COUNTER2 'l'
#define TERM_JOINT_CLOCKWISE '|'
#define TERM_JOINT_CLOCKWISE2 'i'
#define TERM_JOINT_FORWARD 'F'
#define TERM_JOINT_BACKWARD 'B'
#define TERM_JOINT_REVOLUTE 'z'
#define TERM_JOINT_REVOLUTE2 'Z'
#define TERM_JOINT_TWIST 'x'
#define TERM_JOINT_TWIST180 'X'

#define TERM_COLOR_SENSOR 's'

#define TERM_TURN_UP '^'
#define TERM_TURN_DOWN 'v'
#define TERM_TURN_LEFT '<'
#define TERM_TURN_RIGHT '>'
#define TERM_TURN_CLOCKWISE '/'
#define TERM_TURN_CNTR_CLOCKWISE '\\'

#define TERM_OFFSET_INCREASE '+'
#define TERM_OFFSET_DECREASE '-'

#define TERM_PENUP '!'
#define TERM_PENDOWN '#'


// Network stuff.
// Unused symbols: HKTUVWXYZ
#define TERM_CONNECT_TO_INPUT 'Q'
#define TERM_ADD_INPUT 'I'
#define TERM_ADD_OUTPUT 'O'
#define TERM_SPLIT_LINK 'S'
#define TERM_DUPLICATE_LINK 'D'
#define TERM_ADJUST_WEIGHT 'W'
#define TERM_DELETE_LINK 'C'
#define TERM_NEXT_LINK 'N'
#define TERM_PARENT_NEURON 'B'
//define TERM_DELETE_JOIN 'J'
#define TERM_MERGE_NEURONS 'M'
#define TERM_REVERSE_LINK 'R'
#define TERM_INCREASE_WEIGHT 'A'
#define TERM_DECREASE_WEIGHT 'E'  // 'L' ; 'E'
#define TERM_SET_OUTPUT_FUNCTION 'F'
#define TERM_MAKE_LOOP 'L'
#define TERM_NEGATE_WEIGHT 'G'

#define TERM_SPLIT_LINK_NEG 'a'
#define TERM_DUPLICATE_LINK_NEG 'b' // previously 'b' other is 'c'
 
// Program commands.
#define TERM_LOAD_ACC 'l'
#define TERM_STORE_ACC 's'
#define TERM_PUSH_ACC 'v'
#define TERM_POP_ACC '^'
#define TERM_COMPARE_ACC '?'
#define TERM_CLEAR_FLAG 'c'
#define TERM_COMPARE_DATA 't'
#define TERM_SWAP_DATA 'w'
#define TERM_HALT 'h'
#define TERM_BREAK 'b'
// define TERM_IF_FLAG - done with brackets.




// Types of bars/joints.
#define BAR_TYPE_FIXED 0
#define BAR_TYPE_JOINT 1
#define BAR_TYPE_COUNTER 2
#define BAR_TYPE_CLOCKWISE 3
#define BAR_TYPE_COUNTER2 4
#define BAR_TYPE_CLOCKWISE2 5
#define BAR_TYPE_SENSOR 6

#define MAX_JOINTS 50
#define DRAW_TOLERANCE 20.0



extern int Max_joints;
extern int Number_trials;

extern int Heading[3][3];
extern int TUp_mat[3][3];
extern int TDown_mat[3][3];
extern int TRight_mat[3][3];
extern int TLeft_mat[3][3];
extern int TClockwise_mat[3][3];
extern int TCntrClockwise_mat[3][3];

void init_heading(void);
void turn_left(void);
void turn_right(void);
void turn_up(void);
void turn_down(void);
void turn_clockwise(void);
void turn_cntr_clockwise(void);



typedef struct Terminal_Structure
{
  char symbol;
  int arity;  // number of arguments.
  int arg_type[3]; // type of each arg.
  Real min[3];  // minimum for each arg.
  Real max[3];  // maximum for each arg.
} Terminal_Struct;


typedef struct tLString
{
  char symbol;
  Real arg;
} LString;


typedef struct Model_Struct
{
  int model_id;
  int new_string;
  int lstring_length;
  int lstring_waste;
  LString *lstring;
  Real *build_args;
  int fv_length;
  Real *fitness;
  int error;
  Real offset[3];
} ModelStruct;


extern int Fitness_func;
extern int Terminal_set, Build_type;
extern int Model_type;

extern int Pen_down;

extern int Num_terminals;
extern char Terminal[50];
extern Real Terminal_prob[50];

extern Graphic *win1;

int display_structure(int model_id);
void view_structure(int arg);

int build_structure(ModelStruct *model);
Real evaluate_structure(void);
int evaluate_structure(ModelStruct *model);

void configure_graphics(Configuration config);
Real play_game(int init, Real *fitnessA, Real *fitnessB,
	       Individual *individ1, Individual *individ2);
void set_terminal_set(int val);
void configure_pop(Pop *pop);
void test_iteration1(Individual *ind);

void Run_BlindWatchmaker(Deme *deme);

#endif
