/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <stream.h>
#include <fstream.h>
#include <math.h>
#include <string.h>
#include "general.h"
#include "pop.h"
#include "deme.h"
#include "individual.h"


void configure_pop(Pop *pop);
void configure_graphics(Configuration config);

extern void test_iteration1(Individual *individ);


extern Real play_game(int init, Real *fitnessA, Real *fitnessB,
		      Individual *individ1, Individual *individ2);

char *filename_stem;

int Start_trial = 0;

void run_test1(char *individ_name)
{
  int i, j, dist, dist1, dist2;
  Individual *ind1, *ind2;
  FILE *file;

  file = fopen(individ_name, "r");
  ind1 = new Individual();
  ind1->read(file);
  fclose(file);
  ind1->evaluate();

  //  test_iteration1(ind1);

  ind2 = new Individual();

  for (i=0; i<100; i++) {
    ind2->duplicate(ind1);
    j = random_int(25)+1;
    for (; j>0; j--) {
      ind2->mutate();
    }
    ind2->evaluate();

    dist1 = ind1->phenotype_distance(ind2);
    dist2 = ind1->phenotype_distance2(ind2);

    if (dist1 > dist2)
      dist = dist1;
    else
      dist = dist2;

    printf("%d %.2f\n",
	   dist,
	   ind2->get_fitness());

    /*
    printf("%d %d %d %.2f %.2f   len:%d:%d\n",
	   dist, dist1, dist2,
	   ind1->get_fitness(),
	   ind2->get_fitness(),
	   ind1->length, ind2->length);
    */
  }

  file = fopen("orig.ind", "w");
  ind1->write(file);
  fclose(file);

  file = fopen("mutant.ind", "w");
  ind2->write(file);
  fclose(file);
}  

void run_test2(char *individ_name)
{
  int i, dist, dist1, dist2;
  int num_prod, max_bodies;
  Individual *ind1, *ind2;
  FILE *file;

  file = fopen(individ_name, "r");
  ind1 = new Individual();
  ind1->read(file);
  fclose(file);
  ind1->evaluate();

  //  test_iteration1(ind1);

  ind2 = new Individual();
  
  ind2->init_pheno(TRUE);

  num_prod = Num_productions;
  max_bodies = Max_bodies;

  for (i=0; i<1; i++) {
    if (!ind2->duplicate_pheno(ind1))
      break;
    ind2->mutate_pheno();

    Num_productions = 1;
    Max_bodies = 1;
    ind2->evaluate();
    Num_productions = num_prod;
    Max_bodies = max_bodies;


    dist1 = ind1->phenotype_distance(ind2);
    dist2 = ind1->phenotype_distance2(ind2);

    if (dist1 > dist2)
      dist = dist1;
    else
      dist = dist2;

    printf("%d %d %d %.2f %.2f   len:%d:%d\n",
	   dist, dist1, dist2,
	   ind1->get_fitness(),
	   ind2->get_fitness(),
	   ind1->length, ind2->length);
  }

  /*
  file = fopen("orig.ind", "w");
  ind1->write(file);
  fclose(file);
  */

  file = fopen("mutant.ind", "w");
  ind2->write_pheno(file);
  fclose(file);
}  


void play_individ(int f_index, char *file_name)
{
  Individual *ind1, *ind2;
  Pop *pop;
  FILE *file;

  //  infile = new ifstream(argv[1], ios::in);
  ind1 = new Individual();
  file = fopen(file_name, "r");


  if (f_index >= 0) {
    pop = new Pop();
    pop->read(file);
    ind1 = pop->deme[0]->get_individ(f_index);

  } else {
    if (ind1->read(file))
      fclose(file);
    else {
      return;
    }
  }

  printf("Fitness: %.2lf\n", ind1->get_fitness());
  /*
  file = fopen("read.ind", "w");
  ind2 = new Individual();
  ind2->duplicate(ind1);
  ind2->write(file);
  fclose(file);
  */

  Print[16] = TRUE;

#ifdef VECTOR
  test_iteration1(ind1);
#elif CELLULAR
  test_iteration1(ind1);
#elif MATRIX
  test_iteration1(ind1);
#else
  ind1->evaluate();
  cout << "Fitness: " << ind1->get_fitness();
#endif
  file = fopen("read.ind", "w");
  ind2 = new Individual();
  ind2->duplicate(ind1);
  ind2->write(file);
  fclose(file);
}  


void play_individs(char *individ_name1, char *individ_name2)
{
  Real valA, valB;
  Individual *ind1, *ind2;
  FILE *file;

  cout << individ_name1 << endl;

  cout << individ_name2 << endl;

  file = fopen(individ_name1, "r");
  ind1 = new Individual();
  ind1->read(file);
  fclose(file);

  file = fopen(individ_name2, "r");
  ind2 = new Individual();
  ind2->read(file);
  fclose(file);

  play_game(FALSE, &valA, &valB, ind1, ind2);

  cout << "fitness A: " << valA;
  cout << "  fitness B: " << valB;
  cout << endl;

  ind1->evaluate(FALSE);
  ind2->evaluate(FALSE);
}  



void evolve(char *pop_name, Pop *pop, char *seed)
{
  int i;
  char best_name[50];
  FILE *file;

  if (pop_name == NULL) {
    pop->create();
  } else {
    file = fopen(pop_name, "r");
    pop->create();
    pop->read(file);
  }

  for (i=Start_trial; i<Config.num_trials; i++) {
    print_time();
    printf("\n");

    if (pop_name != NULL) {
      ;
    } else if (seed == NULL) {
      pop->initialize();
    } else {
      Individual *ind_seed = new Individual();
      file = fopen(seed, "r");

      ind_seed->read(file);
      pop->initialize(ind_seed);
      //      ind_seed->kill();
    }

    pop->evolve();
    if (Config.num_trials == 1) {
      pop->print_stats();

    } else {
      printf("\nIteration: %d.\n", i);
      pop->print_last_stats();

      if (i<10) {
	sprintf(best_name, "%s.00%d.ind", filename_stem, i);
      } else if (i<100) {
	sprintf(best_name, "%s.0%d.ind", filename_stem, i);
      } else {
	sprintf(best_name, "%s.%d.ind", filename_stem, i);
      }
      pop->write_best_individ(best_name);
    }
  }

  if (Config.num_trials > 1)
    pop->print_final_stats();
}


void coevolve(char *pop_name, Pop *pop, char *seed)
{
  FILE *file;

  if (pop_name == NULL) {
    pop->create();

    if (seed == NULL)
      pop->initialize();
    else {
      Individual *ind_seed = new Individual();
      file = fopen(seed, "r");

      ind_seed->read(file);
      //      pop->set_network_struct(ind_seed->num_inputs,
      //			      ind_seed->num_hidden,
      //			      ind_seed->num_outputs);
      pop->initialize(ind_seed);
    }
  } else {
    //    infile = new ifstream(pop_name, ios::in);
    //    pop->read(infile);
    //    infile->close();
  }

  pop->coevolve(1000);
  pop->print_stats();
}


void init_vars(void)
{
  int i;

  for (i=0; i<50; i++)
    Print[i] = FALSE;
  for (i=0; i<50; i++)
    Option[i] = 0;

  for (i=0; i<100; i++) {
    Operation_hist[i] = 0;
    Op_better_hist[i] = 0;
    Operation[i] = 0;
    Op_better[i] = 0;
  }
}


int main(int argc, char** argv)
{
  bool replay = FALSE;
  bool test_flag = FALSE;
  bool test_flag2 = FALSE;
  int i, j, index, f_index = -1;
  char *log_file = 0;
  char *filename1 = 0, *filename2 = 0;
  char *pop_name1 = NULL;
  int num_trials = 1;
  int num_generations = -1;
  Real recomb_rate = -1.0;
  Real selective_pressure = -1.0;
  Pop *pop;

  /*
  cout.width(6);
  cout.fill(' ');
  */

  init_random(-1);
  init_vars();
  
  Config.start_frame = -1;
  Config.frame_rate = 1;
  Config.do_graphics = TRUE;
  Config.draw_history = FALSE;
  Config.save_graphics = FALSE;
  Config.filename_stem[0] = 0;
  strcpy(Config.filename_stem, argv[0]);
  filename_stem = argv[0];
  sprintf(Config.filename_error, "%s_last.ind", argv[0]);

  for (i=1; i<argc; i++) {
    if (argv[i][0] == '-') {
      switch (argv[i][1]) {
      case 'p':
	if (++i >= argc) {
	  cout << "Missing argument for -p." << endl;
	  exit(-1);
	}

	index = atoi(argv[i]);
	if ((index < 0) || (index > 49)) {
	  cout << "Missing argument for -p." << endl;
	  exit(-1);
	}
	Print[index] = FALSE;
	break;

      case 'P':
	if (++i >= argc) {
	  cout << "Missing argument for -P." << endl;
	  exit(-1);
	}

	if (argv[i][0] == 'A') {
	  for (j=0; j<10; j++)
	    Print[j] = TRUE;
	} else if (argv[i][0] == 'a') {
	  for (j=0; j<10; j++)
	    Print[j] = FALSE;
	} else {
	  index = atoi(argv[i]);
	  if ((index < 0) || (index > 50)) {
	    cout << "Missing argument for -P." << endl;
	    exit(-1);
	  }
	  Print[index] = TRUE;
	}
	break;

      case 'f':
	if (++i < argc) {
	  strcpy(Config.filename_stem, argv[i]);
	} else {
	  cout << "Missing argument for -f." << endl;
	  exit(-1);
	}
	break;

      case 'F':
	if (++i < argc)
	  Config.frame_rate = atoi(argv[i]);
	else {
	  cout << "Missing argument for -F." << endl;
	  exit(-1);
	}
	break;

      case 'g':
	Config.do_graphics = FALSE;
	break;

      case 'G':
	Config.do_graphics = TRUE;
	break;

      case 'i':
	if (++i < argc) {
	  filename1 = argv[i];
	} else {
	  cout << "Missing argument for -i." << endl;
	  exit(-1);
	}
	break;

      case 'l':
      case 'L':
	if (++i < argc) {
	  log_file = argv[i];
	} else {
	  cout << "Missing argument for -l/L." << endl;
	  exit(-1);
	}
	break;

      case 'N':
	if (++i >= argc) {
	  cout << "Missing arguments for -N." << endl;
	  exit(-1);
	}

	num_generations = atoi(argv[i]);
	if ((num_generations < 0) || (num_generations > 100000)) {
	  printf("Invalid number of generations: %d.\n", num_generations);
	  exit(-1);
	}
	break;

      case 'o':
      case 'O':
	if (++i+1 >= argc) {
	  cout << "Missing arguments for -o." << endl;
	  exit(-1);
	}

	index = atoi(argv[i]);
	if ((index < 0) || (index > 49)) {
	  printf("Invalid index %d for -o.\n", index);
	  exit(-1);
	}
	Option[index] = atoi(argv[++i]);
	break;

      case 'r':
	if (++i < argc) {
	  replay = TRUE;
	  filename1 = argv[i];
	} else {
	  cout << "Missing argument for -r." << endl;
	  exit(-1);
	}

	if (++i < argc) {
	  if (argv[i][0] != '-') {
	    filename2 = argv[i];
	  } else
	    i--;
	}
	break;

      case 'R':
	if (++i < argc) {
	  replay = TRUE;
	  filename1 = argv[i];
	} else {
	  cout << "Missing filename for -R." << endl;
	  exit(-1);
	}

	if (++i < argc) {
	  if (argv[i][0] != '-') {
	    f_index = atoi(argv[i]);
	  } else {
	    cout << "Missing index for -R." << endl;
	    exit(-1);
	  }
	}
	break;

      case 's':
	if (++i < argc) {
	  init_random(atoi(argv[i]));
	} else {
	  cout << "Missing argument for -s." << endl;
	  exit(-1);
	}
	break;

      case 'S':
	if (++i < argc) {
	  pop_name1 = argv[i];
	} else {
	  cout << "Missing argument for -S." << endl;
	  exit(-1);
	}
	break;

      case 't':
	if (++i >= argc) {
	  cout << "Missing argument for -t." << endl;
	  exit(-1);
	}

	Start_trial = atoi(argv[i]);
	printf("Start trial set to: %d.\n", Start_trial);
	break;

      case 'T':
	if (++i >= argc) {
	  cout << "Missing argument for -T." << endl;
	  exit(-1);
	}

	num_trials = atoi(argv[i]);
	printf("Number of trials set to: %d.\n", num_trials);
	break;

      case 'U':
	if (++i >= argc) {
	  cout << "Missing argument for -U." << endl;
	  exit(-1);
	}
	selective_pressure = atof(argv[i]);
	break;

      case 'w':
      case 'W':
	Config.save_graphics = TRUE;
	break;

      case 'x':
	if (++i >= argc) {
	  cout << "Missing argument for -x." << endl;
	  exit(-1);
	}
	recomb_rate = atof(argv[i]);
	break;

      case 'z':
	test_flag = TRUE;
	break;

      case 'Z':
	test_flag2 = TRUE;
	break;

      case '#':
	if (++i < argc)
	  Config.start_frame = atoi(argv[i]);
	else {
	  cout << "Missing argument for -#." << endl;
	  exit(-1);
	}
	break;


      }
    } else {
      printf("Error in command line >%c<.\n", argv[i][1]);
      exit(-1);
    }
  }


  //  init_sim();
  Config.num_trials = num_trials;
  configure_graphics(Config);

  pop = new Pop(argv[0]);
  //  pop->structure = STRUCT_STEADY_STATE;
  pop->structure = STRUCT_CANONICAL;
  pop->num_demes = 1;
  pop->num_individs = 50;
  if (log_file != 0)
    pop->set_logfile(log_file);
  configure_pop(pop);
  if (num_generations > 0) {
    pop->max_generations = num_generations;
    printf("Using: %d generations.\n", pop->max_generations);
  }
  if (recomb_rate >= 0.0) {
    printf("Recomb rate = %.4f.\n", recomb_rate);
    pop->recomb_rate = recomb_rate;
  }
  if (selective_pressure >= 0.0) {
    printf("Selective pressure = %.4f.\n", selective_pressure);
    pop->pressure = selective_pressure;
  }

  if (test_flag2) {
    Config.draw_history = TRUE;
  }

  if (replay) {
    if (pop->do_coevolve)
       play_individs(filename1, filename2);
    else {
      if (test_flag) {
	//	run_test1(filename1);
	run_test2(filename1);
      } else
	play_individ(f_index, filename1);
    }

  } else {
    if (pop->do_coevolve) {
      coevolve(NULL, pop, filename1);
    } else {
      evolve(pop_name1, pop, filename1);
    }
  }
  
  cout << endl;
  return 0;
}

