/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby
This file contains code for evolution with a multi-objective
fitness function and was written by Edwin de Jong
(dejong@cs.uu.nl) and is copyright (C) Edwin de Jong 2002.

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/
#include <stdlib.h>
#include <iostream.h>
#include "multi_obj.h"

bool betterinatleastone(int i, int j, int **fitnesses, int num_dims){
  bool better=false;
  for (int obj=0; obj<num_dims; obj++)
    if (fitnesses[i][obj]>fitnesses[j][obj])
      better = true;
  return better;
}

bool equalorbetterinall(int i, int j, int **fitnesses, int num_dims){
 bool bettereq=true;
  for (int obj=0; obj<num_dims; obj++)
    if (fitnesses[i][obj]<fitnesses[j][obj])
      bettereq = false;
  return bettereq;
 }

void compute_domination_ranks(int pop_size, double **nrdom, int **fitnesses, int num_dims){
  bool betterinone;

  //computes nrdom (-1 for indivs to be deleted)
  
  for (int index=0; index < pop_size; index++){
    nrdom[index][0] = 0;
    for(int j = 0; j < pop_size; j++){
      if (j != index){
	betterinone = betterinatleastone(j, index, fitnesses, num_dims);
	if ( equalorbetterinall(j, index, fitnesses, num_dims) ){
	  if (!betterinone){
	    if (j < index)
	      nrdom[index][0]=-1;//delete
	  }		    
	  else if (nrdom[index][0]>=0)//wasn't to be deleted
	    nrdom[index][0]++;
	}
      }
    }
  }
}

int compare_nrdom(const void *pa, const void *pb){
  double *a=*((double**)pa);
  double *b=*((double**)pb);
  if (a[0]<b[0])
    return -1;
  else if (a[0]==b[0])
    return 0;
  else 
    return 1;
}


int sort_multi_objective(int pop_size, int num_dims, int **fitnesses, float *exp_childs, int max_pop_size)
{

  //popsz: #indivs in pop+gen (<=2*max_pop_size)
  //fitnesses: [0] =ftn (higher=better) [1]=sz (lower=better)
  //exp_childs is set to expected #offspring 
  //(zero for delete), sums up to maxpopsize(e.g. 100) or less

  double **nrdom;
  nrdom = new double*[pop_size];
  for (int i=0;i<pop_size;i++){
    nrdom[i]=new double[2];
    nrdom[i][1]=i; //[0]=nrdom, [1]=original index
  }

  compute_domination_ranks(pop_size, nrdom, fitnesses, num_dims);

  /*for (int i=0;i<pop_size;i++){
    cout<<"ranks: "<<nrdom[i][0]<<" "<<nrdom[i][1]<<endl;
    }*/

  int n=0;
  for (int i=0;i<pop_size;i++)
    if (nrdom[i][0]>=0)
      n++;
  if (n<=max_pop_size){//all individuals may stay
    for (int i=0;i<pop_size;i++)
      if (nrdom[i][0]>=0)
	exp_childs[i]=1;
      else
	exp_childs[i]=0;

    return 0;
  }


  double *a=new double[2];
  qsort(nrdom,pop_size,sizeof(a),compare_nrdom);
  delete a;

  int index=0;
  while(nrdom[index][0]==-1)
    index++;//find first that will not be deleted
  int tobedeleted=index;
  int from,to;
  double avg,xfrom,xto;
  double expc=(double)n*(1.0+1.0/(double)n)/2.0;//current expectation of #selected
  double factor=(double)max_pop_size/expc;//scale to desired popsz
  while(index<pop_size){
    from=index;
    to=from;
    while((to<pop_size-1)&&(nrdom[to+1][0]==nrdom[to][0]))
      to++;
    xfrom=1.0-((double)(from-tobedeleted))/((double)(pop_size-tobedeleted));
    xto=1.0-((double)(to-tobedeleted))/((double)(pop_size-tobedeleted));
    avg=factor*(double)(xfrom+xto)/2.0;
    for(int i=from;i<=to;i++)
      nrdom[i][0]=avg;
    index=to+1;
  }

  int oldindex;
  for (int i=0;i<pop_size;i++){
    oldindex=(int)(nrdom[i][1]);
    if (nrdom[i][0]>=0)
      exp_childs[oldindex]=nrdom[i][0];
    else
      exp_childs[oldindex]=0;
  }

  for (int i=0;i<pop_size;i++)
    delete nrdom[i];
  delete nrdom;

  return 0;
}



/*

int main(int argc, char *argv[]){
  int **fitnesses;
  fitnesses = new int*[13];
  for (int i=0;i<13;i++)
    fitnesses[i]=new int[2];

  int pop_size=13;
  int num_dims=2;

  fitnesses[0][0]=0;
  fitnesses[0][1]=1;
  fitnesses[1][0]=0;
  fitnesses[1][1]=1;
  fitnesses[2][0]=2;
  fitnesses[2][1]=2;
  fitnesses[3][0]=3;
  fitnesses[3][1]=3;
  fitnesses[4][0]=0;
  fitnesses[4][1]=1;
  fitnesses[5][0]=5;
  fitnesses[5][1]=5;
  fitnesses[6][0]=6;
  fitnesses[6][1]=6;
  fitnesses[7][0]=7;
  fitnesses[7][1]=7;
  fitnesses[8][0]=7;
  fitnesses[8][1]=8;
  fitnesses[9][0]=9;
  fitnesses[9][1]=9;
  fitnesses[10][0]=10;
  fitnesses[10][1]=10;
  fitnesses[11][0]=11;
  fitnesses[11][1]=11;
  fitnesses[12][0]=12;
  fitnesses[12][1]=12;

  float exp_childs[pop_size];

  sort_multi_objective(pop_size, fitnesses, (float*)&exp_childs, 10, num_dims);

  for(int i=0;i<pop_size;i++)
    cout<<i<<" ftn "<<fitnesses[i][0]<<" "<<fitnesses[i][1]<<" exp "<<exp_childs[i]<<endl;
      
}
*/
