/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#ifndef NETWORK_HEADER_FILE
#define NETWORK_HEADER_FILE

#include <fstream.h>
#include "neuron.h"

#define MAX_NEURONS 200
#define MAX_NETWORK_INPUTS 50
#define MAX_NETWORK_OUTPUTS 50
#define NEURON_CHG 2
#define MAX_NEURON_CHG 2
#define PROPERTY_CHG 2


#define NEURON_MAX_OUTPUT_FUNC 4
#define NEURON_FUNCTIONS 9

#define NEURON_SIGMOID 0
#define NEURON_SUM 1
#define NEURON_SAW_WAVE 2
#define NEURON_SIN 3
#define NEURON_ASIN 4
#define NEURON_MIN 5
#define NEURON_MAX 6
#define NEURON_SIGN 7
#define NEURON_PROD 8
#define NEURON_SIGMOID2 9



#define MAX_WEIGHT_NUM 10

class Neuron;

typedef struct NeuronStruct
{
  int Type;
  int Function;
  double Value;

  int Num_inputs;
  int Input[MAX_INPUTS];
} NeuronInfo;


typedef struct LinkIndexStruct
{
  int from_index;
  int from_child;
  int from_child_link_index;
  int to_index;
} LinkIndex;


void copy_linkindex(LinkIndex *to_li, LinkIndex *from_li);

class Network
{
public:
  static int init_neurons;

public:
  int Made;
  int Num_inputs;
  int Num_outputs; // last N neurons are output;
  int Num_neurons;
  Neuron *Input_neuron[MAX_NETWORK_INPUTS];
  Neuron *Output_neuron[MAX_NETWORK_OUTPUTS];
  Neuron *Neuron_list[MAX_NEURONS];


  int Neuron_input_list[MAX_NEURONS]; // Indexes of input neurons.
  int Neuron_output_list[MAX_NEURONS]; // Indexes of output neurons.

  int Neuron_function[MAX_NEURONS];
  int Neuron_type[MAX_NEURONS];
  int Neuron_num_inputs[MAX_NEURONS];
  int Neuron_input_index[MAX_NEURONS][MAX_INPUTS];
  Real Neuron_input_weight[MAX_NEURONS][MAX_INPUTS];
  Real Neuron_memory[MAX_NEURONS];
  Real Neuron_value[MAX_NEURONS];
  Real Neuron_value_prev[MAX_NEURONS];

  Real Neuron_max[MAX_NEURONS];
  Real Neuron_min[MAX_NEURONS];

  

  //  NeuronInfo Neuron_data[MAX_NEURONS];

  void empty(void);
  
  /******************************/

  Network(void);
  int init(void);
  int make_random(void);
  void reset(void);
  void clear(void);
  void duplicate(Network *net);
  void print_stats(void);
  void print_state(void);
  void print_structure(void);
  void print_neuron_state(int index);
  void print_structure2(void);
  void print_dot_file(FILE *file);
  int output_changed(void);
  void get_output(double *val);
  void set_input(double * val);
  void set_first_value(double val);
  void next_iteration(void);

  int valid_network(void);
  int print_link_index(LinkIndex *current_link);
  int valid_link_index(LinkIndex *current_link);
  int update_link(LinkIndex *current_link);
  int find_existing_link(LinkIndex *current_link);

  int first_neuron(Real weight, LinkIndex *current_link);
  int create_inputs(int num);
  int add_input(Real weight, LinkIndex *link_index);
  int connect_to_input(Real weight, int input, LinkIndex *link_index);
  int add_output(Real weight, LinkIndex *link_index);
  int cut_link(LinkIndex *current_link);
  int duplicate_link(LinkIndex *current_link);
  int duplicate_link(Real weight, LinkIndex *current_link);
  int adjust_weight(Real value, LinkIndex *current_link);
  int negate_weight(Real weight, LinkIndex *current_link);
  int set_input_function(int function, Real value, LinkIndex *current_link);
  int set_output_function(int function, LinkIndex *current_link);
  int split_link2(Real weight, LinkIndex *current_link);
  int split_link(Real weight, LinkIndex *current_link);
  int make_loop(Real weight, LinkIndex *current_link);
  int next_link(int num, LinkIndex *current_link);
  int parent_link(int num, LinkIndex *current_link);
  int merge_neurons(LinkIndex *current_link);
  int reverse_link(LinkIndex *current_link);
  int delete_output(int index);

  void rebuild_network(void);
  void finish_build(void);
  int num_used_inputs(void);
  Real average_num_inputs(void);
  Real average_output_range(void);
  int num_unique_output_sources(void);
  int num_active_outputs(void);
  int num_links(void);

  void read(ifstream *infile);
  void write(ofstream *outfile);

  int read(FILE *file);
  void write(FILE *file);
};

#endif
