/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/


#ifndef NEURON_HEADER_FILE
#define NEURON_HEADER_FILE

#include <fstream.h>

#define NEURON_INPUT -2
#define NEURON_OUTPUT -1
#define NEURON_HIDDEN 0
#define NEURON_DELETED -3
#define NEURON_UNUSED -4

#define MAX_INPUTS 20 // previously 10
#define MAX_OUTPUTS 20

#define WAVE_STEP 0.01
#define MAX_WEIGHT 2.0
#define MAX_VALUE 1.0  // PI

class Network;
class Neuron;

extern int neurons_avail;
extern Neuron *first_free_neuron;

void release_neuron(Neuron *n);
Neuron *next_free_neuron(void);


typedef struct LinkStruct
{
  Real weight;
  Neuron *input;  
} Link;


class Neuron
{
private:

public:
  int Type;
  int Function;
  int Id;
  int Same_as; // Points to location in Input/Output order this neuron should take because of merges.
  int Merged_to; // If merged, this points to Id of new neuron.
  Neuron *prev_neuron;
  Neuron *next_neuron;
  Network *network;

  int Num_inputs;
  Link Input[MAX_INPUTS];

  double Scale;
  double Offset;
  double Time_const;
  double Value;
  double Memory;
  double Speed;
  
  /******************************/

  Neuron(void);
  Neuron(Network *n);
  void init(void);
  void reset(void);
  void print_state(void);
  void print_structure(void);
  void print_stats(void);
  void print_stats(char *string);
  int valid_neuron(void);
  int is_input(Neuron *output);

  int add_input(Real weight, Neuron *in_neuron);
  int duplicate_input(int index);
  void copy_link(int index_to, int index_from);
  void cut_input(int index);
  void remove_input(int index);
  int dead_link(int index);
  void remove_dead_links(void);
  int has_live_inputs(void);
  int set_function(int function);
  
  void remove(void);

  double sum_inputs(void);
  void update(void);
  int duplicate(Neuron *original);
  int reverse_link(int index);
  void delete_inputs_from(Neuron *source);
  void replace_input_neuron(Neuron *original, Neuron *replacement);


  void read(ifstream *infile, Neuron **neurons);
  void write(ofstream *outfile);

  int read(FILE *file);
  void write(FILE *file);
};

#endif
