/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/


#ifndef POP_HEADER_FILE
#define POP_HEADER_FILE

// structure types
#define STRUCT_CANONICAL 1
#define STRUCT_NEIGHBOR 2
#define STRUCT_STEADY_STATE 3
#define STRUCT_STEADY_STATE2 4
#define STRUCT_HILL_CLIMB 5
#define STRUCT_TEST_ONE 6
#define STRUCT_DETERMINISTIC_CROWDING 7
#define STRUCT_BLIND_WATCHMAKER 8
#define STRUCT_ELITIST_STEADY_STATE 9
#define STRUCT_MULTI_OBJECTIVE 10

// Neighborhood types.
#define NGHBR_NONE 0
#define NGHBR_DEME 1
#define NGHBR_NEAR 2


#define MAX_AGE 5

class Individual;
class Deme;
class Stats;


class Pop
{
public:
  int created;
  int do_coevolve;
  int structure;  // keep in deme
  int num_demes; 
  int num_individs; // keep in deme
  int adaptive_mut;
  int elitism;
  int max_age;
  int switching;
  int tournament_size;
  int replace_size;
  int nghbr_size;
  int nghbr_type;
  int num_opponents;
  int generation;
  int max_generations;
  int hof_size;
  double pressure;
  double recomb_rate;
  double max_fitness;
  double avg_fitness;
  double best_ever;
  char pop_name[50];
  char log_file[50];
  FILE *log;
  Individual *sample_individ;
  Deme **deme;
  Stats *stats_avg, *stats_best, *stats_best_vec, *stats_avg_vec;
  Stats *stats_trials_avg, *stats_trials_best;

  /******************************/

  Pop(void);
  Pop(char *n);
  void init(char *n);
  void set_name(char *n);
  void set_logfile(char *n);
  void set_sample_individ(Individual *sample);
  void set_tournament_size(int size);
  void set_replace_size(int size);
  void uncreate(void);
  void create(void);
  void initialize(void);
  void initialize(Individual *ind_seed);
  void evaluate_init(void);
  void evaluate_steady_state(void);
  void evaluate_pop_init(void);
  void coevaluate_init(void);
  void coevaluate_hof(void);
  void coevaluate(void);
  void post_evaluate(void);
  void make_next_generation(void);
  void update_final_stats(void);
  void update_stats(void);
  void evolve(int iterations);
  void evolve(void);
  void store_hof(void);
  void coevolve(int iterations);
  void mutate_all(void);
  void info(void);
  void print_all(void);
  void print_all2(void);
  void print_gen_results(void);
  void print_last_stats(void);
  void print_stats(void);
  void print_final_stats(void);
  void write(ofstream *outfile);
  void read(ifstream *infile);
  void write_best_individ(char *filename);
  void write_backup(void);

  void write_best(FILE *file);
  void write_best_new(FILE *file);
  int read_individ(FILE *file);
  int read(FILE *file);
  void write(FILE *file);
};

#endif
