/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stream.h>
#include <iostream.h>
#include <fstream.h>
#include "general.h"
#include "lsystem.h"
#include "ind_lsys.h"
#include "prod_body.h"

#define PROB_TERMINAL 1.5 // 2.5 0.7
#define PROB_PROD 0.25

#define PROB_SEQUENCE 1.5 // 1.0
#define PROB_BRANCH 1.0  // 0.7
#define PROB_MULT_SEQ 0.2
//define PROB_MULT_SEQ 0.2
#define PROB_BRANCH_EDGE 1.0

#define PROB_PVAL_VALUE 0.5
#define PROB_PVAL_CONSTANT 0.0
#define PROB_PVAL_VARIABLE 0.25
#define PROB_PVAL_FUNCTION 0.0

#define PROB_PVAL_DEFAULT 0.7

#define PROB_MUT_CHAR 1.0
#define PROB_MUT_ADD 0.25
#define PROB_MUT_DEL 0.25
#define PROB_MUT_REORDER 0.25
#define PROB_MUT_COND 0.25
#define PROB_MUT_REPL 0.25
#define PROB_MUT_ENCAPSULATE 0.0

#define PROB_MUT_CHAR_PARAM 1.0
#define PROB_MUT_CHAR_SYMBOL 0.3
#define PROB_MUT_CHAR_TYPE 0.01 // 0.05
Real Prob_mut_char_param = PROB_MUT_CHAR_PARAM;
Real Prob_mut_char_symbol = PROB_MUT_CHAR_SYMBOL;
Real Prob_mut_char_type = PROB_MUT_CHAR_TYPE;

//define MAX_SEQ_LENGTH 3
#define MIN_SEQ_LENGTH 2
#define MAX_SEQ_LENGTH 5


int Max_prod_length = MAX_PROD_LENGTH;
int Prod_length = PROD_LENGTH;
int Max_string_length = MAX_STRING_LENGTH;
int Num_pbc_ops = NUM_PBC_OPS;


/*
PVALUE_TYPES:
0: original
1: greater range and include 0s.
*/
#define PVALUE_TYPE 1
#define PVALUE_MAX_CREATE 10
int Pvalue_type = 0;
int Pvalue_max_create = PVALUE_MAX_CREATE;



int is_term_char(char c);
int Num_args = 1;

Real Prob_prod = PROB_PROD;
Real Prob_mult_seq = PROB_MULT_SEQ;
Real Prob_branch_edge = PROB_BRANCH_EDGE;
Real Prob_branch = PROB_BRANCH;

void set_probabilities(void)
{
  //  if (Num_productions == 0)
  //    Prob_mut_char_type = 0.0;
}


int is_term_char(char c)
{
  int i;
  for (i=0; i<Num_terminals; i++)
    if (c == Terminal[i])
      return TRUE;
  return FALSE;
}


int is_op(char c)
{
  if (c == OP_ADD)
    return TRUE;
  else if (c == OP_SUB)
    return TRUE;
  else if (c == OP_EQUAL)
    return TRUE;
  else if (c == COMP_GREATER)
    return TRUE;
  else if (c == COMP_LESS)
    return TRUE;
  else if (c == OP_MULT)
    return TRUE;
  else if (c == OP_DIV)
    return TRUE;
  else if (c == OP_POW)
    return TRUE;
  return FALSE;
}



Prod_Body :: Prod_Body(void)
{
  length = 0;
  used = 0;
  set_probabilities();
  initialize();
}

void Prod_Body :: initialize_pvalue(Prod_Value *pvalue)
{
  pvalue->type = PBC_TYPE_NONE;
  pvalue->symbol = 0;
  pvalue->value = 0.0;
}

void Prod_Body :: initialize_param(Prod_Param *param)
{
  param->op = PBC_OP_NONE;
  initialize_pvalue(&(param->arg1));
  initialize_pvalue(&(param->arg2));
}

void Prod_Body :: initialize_pchar(int index)
{
  int i;

  string[index].type = PBC_TYPE_NONE;
  string[index].symbol = 0;
  string[index].match = 0;

  string[index].param = (Prod_Param*)malloc(Num_args*sizeof(Prod_Param));
  if (string[index].param == 0) {
    printf("Prod_Body :: initialize_pchar() malloc failed.\n");
    return;
  }

  for (i=0; i<Num_args; i++)
    initialize_param(&(string[index].param[i]));
}


void Prod_Body :: initialize(void)
{
  int i;

  length = 0;
  Num_args = Num_arguments;
  if (Num_args == 0)
    Num_args = 1;

  string = (Prod_Char*)malloc((Prod_length+10) * sizeof(Prod_Char));
  for (i=0; i<Prod_length; i++)
    initialize_pchar(i);
}



bool Prod_Body :: is_block(int i)
{
  if (string[i].type == PBC_TYPE_BLOCK)
    return TRUE;
  return FALSE;
}

bool Prod_Body :: is_block_start(int i)
{
  if (string[i].type == PBC_TYPE_BLOCK) {
    if ((string[i].symbol == SYMBOL_PUSH) ||
	(string[i].symbol == SYMBOL_BLK_START) ||
	(string[i].symbol == SYMBOL_PUSH_EDGE))
      return TRUE;
  }
  return FALSE;
}

bool Prod_Body :: is_mult_start(int i)
{
  if (string[i].type == PBC_TYPE_BLOCK) {
    if (string[i].symbol == SYMBOL_BLK_START)
      return TRUE;
  }
  return FALSE;
}

bool Prod_Body :: is_mult_end(int i)
{
  if (string[i].type == PBC_TYPE_BLOCK) {
    if (string[i].symbol == SYMBOL_BLK_END)
      return TRUE;
  }
  return FALSE;
}

bool Prod_Body :: is_block_end(int i)
{
  if (string[i].type == PBC_TYPE_BLOCK) {
    if ((string[i].symbol == SYMBOL_POP) ||
	(string[i].symbol == SYMBOL_BLK_END) ||
	(string[i].symbol == SYMBOL_POP_EDGE))
      return TRUE;
  }
  return FALSE;
}

bool Prod_Body :: is_production(int i)
{
  if (string[i].type == PBC_TYPE_PRODUCTION)
    return TRUE;
  return FALSE;
}

bool Prod_Body :: is_terminal(int i)
{
  if (string[i].type == PBC_TYPE_TERMINAL)
    return TRUE;
  return FALSE;
}

int Prod_Body :: num_productions(void)
{
  int i;
  int num = 0;

  for (i=0; i<length; i++) {
    if (is_production(i))
      num++;
  }
  return num;
}


bool Prod_Body :: valid_cond(void)
{
  if ((cond.arg2.type != PBC_TYPE_VALUE) ||
      (cond.arg2.value > 1000.0)) {
    // Print the condition statement.
    printf("Invalid cond: ");
    if (cond.comparator == PBC_COMP_TRUE)
      printf("   (T)   ");
    else {
      printf("   (");
      fprint_pvalue(stdout, &cond.arg1);
      
      if (cond.comparator == COMP_GREATER)
	printf(">");
      else if (cond.comparator == COMP_LESS)
	printf("<");
      
      fprint_pvalue(stdout, &cond.arg2);
      printf(")");
    }
    printf("\n");
    while (1);
    return FALSE;
  }

  return TRUE;
}

bool Prod_Body :: valid(void)
{
  return match_parens();
}


void copy_pvalue(Prod_Value *pv_to, Prod_Value *pv_from)
{
  pv_to->type = pv_from->type;
  pv_to->symbol = pv_from->symbol;
  pv_to->value = pv_from->value;
}


void copy_param(Prod_Param *p_from, Prod_Param *p_to)
{
  p_to->op = p_from->op;
  copy_pvalue(&p_to->arg1, &p_from->arg1);
  copy_pvalue(&p_to->arg2, &p_from->arg2);
}


void copy_pchar(Prod_Char *pc_from, 
		 Prod_Char *pc_to)
{
  pc_to->type = pc_from->type;
  pc_to->symbol = pc_from->symbol;
  pc_to->match = pc_from->match;

  for (int i=0; i<Num_args; i++)
    copy_param(&pc_from->param[i], &pc_to->param[i]);
}


void Prod_Body :: copy_symbol(int s_from, int s_to)
{
  copy_pchar(&string[s_from], &string[s_to]);
}

void Prod_Body :: move_symbol(int s_from, int s_to)
{
  copy_pchar(&string[s_from], &string[s_to]);

  // If a block, update match.
  if (is_block(s_from))
    string[string[s_from].match].match = s_to;
}


int Prod_Body :: insert_space(int s_from, int len)
{
  if (length + len >= Prod_length)
    return FALSE;

  int i;
  for (i=length-1; i>=s_from; i--) {
    move_symbol(i, i+len);
  }
  length += len;
  return TRUE;
}


// returns -1 if symbol is useful, or body is empty.
// otherwise returns the index of symbol that is wasted.
int Prod_Body :: symbol_test(int index)
{
  if (index >= length)
    return -1;

  if (is_block(index)) {
    int index2 = string[index].match;
    if (absv(index-index2) == 1)
      return index;
  }

  return -1;  
}


// Simple delete - removes the specified symbol.
int Prod_Body :: delete_single_symbol(int index)
{
  length--;
  for (int i=index; i<length; i++)
    move_symbol(i+1, i);
  return TRUE;
}

int Prod_Body :: delete_block(int index, int del_inside)
{
  int i, b_end, len;

#ifndef OPTIMIZE
  if (Print[9]) {
    printf("delete block:%d (%d)\n", index, del_inside);
    fprint(stdout);
  }
#endif

  b_end = string[index].match;
  if (b_end < index) {
    // swap order so index < b_end.
    i = index;
    index = b_end;
    b_end = i;
  }
    
  if (del_inside) {
    // delete inside of block.
    len = b_end - index + 1;
    for (i=b_end+1; i<length; i++)
      move_symbol(i, i-len);
    length -= len;

  } else {
    // delete just the brackets.
    delete_single_symbol(b_end);
    delete_single_symbol(index);
  }

#ifndef OPTIMIZE
  if (Print[9]) {
    printf("deleted from: %d to %d (%d).\n", index,
	   b_end, del_inside);
    fprint(stdout);
  }
#endif

  return TRUE;
}


int Prod_Body :: delete_symbol(int index, int del_inside)
{
  if (index >= length)
    return FALSE;

#ifndef OPTIMIZE2
  parent->op_used(OP_MUT_DEL);
#endif

#ifndef OPTIMIZE
  if (Print[9]) {
    printf("delete symbol:%d (%d)\n", index, del_inside);
    fprint(stdout);
  }
#endif

  if (!is_block(index))
    delete_single_symbol(index);
  else
    delete_block(index, del_inside);

  int i = symbol_test(index);

#ifndef OPTIMIZE
  if (Print[9]) {
    printf("Usefulness(%d): %d\n", index, i);
    fprint(stdout);
  }
#endif
    

  if (i > -1)
    delete_symbol(i, del_inside);

#ifndef OPTIMIZE
  if (Print[9]) {
    fprint(stdout);
  }
#endif
    
  return TRUE;
}

void Prod_Body :: delete_sequence(int l1, int l2)
{
  int i, len;
  
  len = l2 - l1 + 1;
  if (len > 0) {
    for (i=l2+1; i<length; i++)
      move_symbol(i, i-len);
    length -= len;
    
    i = symbol_test(l1);
    if (i > -1)
      delete_symbol(i, FALSE);
  } 
}


void Prod_Body :: delete_string(void)
{
  length = 0;
}

void repair_redundancy(void)
{
#ifdef OLD
  int i;
  int repeat;

  do {
    repeat = FALSE;

    // Test for empty blocks.
    match_paren(body);
    for (i=0; i<body->length; i++) {
      if (is_block(i, body)) {
	if (body->param[i] == i+1) {
	  // empty block.
	  delete_symbol(i, FALSE, body);
	  match_paren(body);
	  i--;
	}
      }
    }
    
    // Test for redundant (embedded) multipliers.
    for (i=1; i<body->length; i++) {
      if (is_mult(i-1, body) &&
	  is_mult(i, body)) {
	// Remove first multiplier.
	delete_symbol(i-1, FALSE, body);
	i--;
	repeat = TRUE;
      }
    }

    // Test for un-used multiplier blocks.
    for (i=0; i<body->length; i++) {
      if (string[i].symbol == SYMBOL_BLK_START) {
	if (i==0) {
	  delete_symbol(i, FALSE, body);
	  i--;
	  repeat = TRUE;
	} else {
	  if (!is_mult(i-1, body)) {
	    delete_symbol(i, FALSE, body);
	    i--;
	    repeat = TRUE;
	  }
	}
      }
    }
    
    // Test for redundant branch blocks.
    for (i=1; i<body->length; i++) {
      if ((body->symbol[i-1] == SYMBOL_PUSH) &&
	  (string[i].symbol == SYMBOL_POP)) {
	// Remove first branch.
	delete_symbol(i-1, FALSE, body);
	i--;
	repeat = TRUE;
      }
    }

    // Test for redundant branch edge blocks.
    for (i=1; i<body->length; i++) {
      if ((body->symbol[i-1] == SYMBOL_PUSH_EDGE) &&
	  (string[i].symbol == SYMBOL_POP_EDGE)) {
	// Remove first branch.
	delete_symbol(i-1, FALSE, body);
	i--;
	repeat = TRUE;
      }
    }
  } while (repeat);
#endif
}


int Prod_Body :: cond_repair_pvalue(char v, Prod_Value *pvalue)
{
  int i, val;

  if (pvalue->type == PBC_TYPE_CONSTANT) {
    // try to replace with another constant.
    int num = 0;
    int cnstnt[Num_constants];

    for (i=0; i<Num_constants; i++) {
      if ((constant[i].value < max[v]) &&
	  (constant[i].value > min[v])) {
	cnstnt[num++] = i;
      }

      if (num > 0) {
	// Pick a valid constant at random.
	pvalue->symbol = char(cnstnt[random_int(num)]);
	return TRUE;
      }
      // else set to a value.
    }
  }

  pvalue->type = PBC_TYPE_VALUE;

  if (max[v]-min[v]<1.0) {
    pvalue->value = min[v];
    return FALSE;
  }
    
  val = int(min[v]) + random_int(int(max[v]-min[v]));
  pvalue->value = Real(val);

  return TRUE;
}


// Increases threshold so less tests succeed.
int Prod_Body :: make_cond_smaller_old1(void)
{
  char v2;
  char v = cond.arg1.symbol;
  int i, val;

  if (is_cond_true()) {
    // Condition is always true -> put to variable.
    cond.arg1.type = PBC_TYPE_VARIABLE;
    cond.comparator = COMP_GREATER;
    cond.arg2.type = PBC_TYPE_VALUE;
    cond.arg2.value = 0.0;

    if (Num_arguments == 1) {
      cond.arg1.symbol = 0;
      cond_repair_pvalue(0, &cond.arg2);
      return TRUE;

    } else {
      v = random_int(Num_arguments);
      for (i=0; i<Num_arguments; i++) {
	v2 = (v+i)%Num_arguments;
	if (max[v2] > min[v2]) {
	  cond.arg1.symbol = v2;
	  cond_repair_pvalue(v2, &cond.arg2);
	  return TRUE;
	}
      }
    }

    return FALSE;
  }

  if ((max[v]-min[v]<1.0) ||
      (max[v] - cond.arg2.value<1.0)) {
    // No room to make smaller.
    if (Num_arguments > 1) {
      // try a different variable.
      for (i=1; i<Num_arguments; i++) {
	v2 = (v+i)%Num_arguments;
	if (max[v2] > min[v2]) {
	  cond.arg1.symbol = v2;
	  cond_repair_pvalue(v2, &cond.arg2);
	  return TRUE;
	}
      }
    }
    return FALSE; // Could not make smaller.
  }

  val = int(cond.arg2.value);
  val += random_int(int(max[v]) - val);
  cond.arg2.value = Real(val);

  return TRUE;
}

// Increases threshold so less tests succeed.
int Prod_Body :: make_cond_smaller(void)
{
  char v2;
  char v = cond.arg1.symbol;
  int i, val;
  int maxv, minv;

  if (is_cond_true()) {
    // Condition is always true -> don't touch.
    return FALSE;
  }

  // Make sure second arg of condition is a value.
  if (cond.arg2.type != PBC_TYPE_VALUE) {
    cond.arg2.type = PBC_TYPE_VALUE;
  }

  maxv = int(max[v]);
  minv = int(min[v]);

  val = maxv - int(cond.arg2.value);
  if ((maxv == minv) || (val <= 0) || (!random_int(3))) {
    // Try a different argument.

    // Look at different arguments sequentially.
    for (i=1; i<=Num_arguments; i++) {
      v2 = (v+i)%Num_arguments;
      maxv = int(max[v2]);
      minv = int(min[v2]);

      if (maxv > minv) {
	// Use this argument.
	cond.arg1.symbol = v2;
	make_valid_cond_arg2();
	return TRUE;
      }
    }
    cond.arg2.value = min[v];
    return FALSE; // Could not make smaller.
  }

  if (int(cond.arg2.value) < minv) {
    val = maxv - minv;
    if (val == 0) {
      cond.arg2.value = Real(minv);
    } else {
      cond.arg2.value = Real(minv + random_int(val));
    }

  } else {
    val = maxv - int(cond.arg2.value);
    cond.arg2.value += Real(random_int(val)+1.0);
  }

  return TRUE;
}


// Decreases threshold so more tests succeed.
int Prod_Body :: make_cond_larger(void)
{
  char v2;
  char v = cond.arg1.symbol;
  int i, val;
  int maxv, minv;

  if (is_cond_true()) {
    // Always true -> can't get any bigger.
    return FALSE;
  }


  maxv = int(max[v]);
  minv = int(min[v]);

  val = int(cond.arg2.value) - minv;
  if ((maxv == minv) || (val < 0) || (!random_int(3))) {
      // try a different variable.
    if (Num_arguments > 1) { 
      for (i=1; i<=Num_arguments; i++) {
	v2 = (v+i)%Num_arguments;
	maxv = int(max[v2]);
	minv = int(min[v2]-1);

	if (maxv >= minv) {
	  cond.arg1.symbol = v2;
	  make_valid_cond_arg2();
	  //	  cond_repair_pvalue(v2, &cond.arg2);
	  return TRUE;
	}
      }
    }
    cond.arg2.value = min[v]-1.0;
    return FALSE; // Could not make larger.
  }


  if (int(cond.arg2.value) > maxv) {
    val = maxv - minv + 1;
  } else {
    val = int(cond.arg2.value) - minv + 1;
  }
  cond.arg2.value = Real(minv + random_int(val) - 1.0);
    
  return TRUE;
}



// Not used.
int Prod_Body :: repair(void)
{
  return cond_repair_pvalue(cond.arg1.symbol,
			    &cond.arg2);
}


char random_terminal(void)
{
  Real val = random_dbl();
  int i = 0;
  while (Terminal_prob[i] < val)
    i++;
  //  return Terminal[random_int(Num_terminals)];
  return Terminal[i];
}


char random_func(void)
{
  return random_int(Num_pbc_ops)+1;
}


char random_production(void)
{
  return ((char)random_int(Num_productions));
}


void Prod_Body :: make_random_production(int i)
{
  int p;
  string[i].type = PBC_TYPE_PRODUCTION;
  string[i].symbol = random_production();

  for (p=0; p<Num_arguments; p++)
    random_full_param(i, p);
}



// No functions.
void random_pvalue(Prod_Value *pvalue, int deflt_flag)
{
  Real choice;

  if (deflt_flag) {
    choice = random_dbl();
    if (choice < PROB_PVAL_DEFAULT) {
      pvalue->type = PBC_TYPE_VALUE;
      pvalue->value = 1.0;
      return;
    }
  }

  if (Num_arguments == 0) {
    choice = random_dbl() * (PROB_PVAL_VALUE +
			     PROB_PVAL_CONSTANT +
			     PROB_PVAL_FUNCTION);
    if (choice > PROB_PVAL_VALUE + PROB_PVAL_CONSTANT)
      choice += PROB_PVAL_VARIABLE;
  } else {
    choice = random_dbl() * (PROB_PVAL_VALUE +
			     PROB_PVAL_CONSTANT +
			     PROB_PVAL_VARIABLE +
			     PROB_PVAL_FUNCTION);
  }

  if (choice < PROB_PVAL_VALUE) {
    // Should this allow 0 values??
    if (Pvalue_type == 0) {
      pvalue->type = PBC_TYPE_VALUE;
      pvalue->value = Real(random_int(5) + 1);
    } else {
      pvalue->type = PBC_TYPE_VALUE;
      pvalue->value = Real(random_int(Pvalue_max_create));
    }

    // INTEGER*


  } else if (choice < PROB_PVAL_VALUE + PROB_PVAL_CONSTANT) {
    pvalue->type = PBC_TYPE_CONSTANT;
    pvalue->symbol = random_int(Num_constants);
    
  } else if (choice < PROB_PVAL_VALUE + PROB_PVAL_CONSTANT
	     + PROB_PVAL_VARIABLE) {
    pvalue->type = PBC_TYPE_VARIABLE;
    pvalue->symbol = random_int(Num_arguments);

  } else {
    // Function: not implemented.
    printf("Prod_Body :: random_pvalue() - function not implemented.\n");
  }
}


void Prod_Body :: random_single_param(int i, int p, int deflt_flag)
{
  string[i].param[p].op = PBC_OP_NONE;
  random_pvalue(&string[i].param[p].arg1, deflt_flag);
}

void Prod_Body :: random_full_param(int i, int p)
{
  //  string[i].param[p].op = random_int(Num_pbc_ops) + 1;
  string[i].param[p].op = random_func();

  // Forcing first arg to be a variable.
  //  random_pvalue(&string[i].param[p].arg1);
  string[i].param[p].arg1.type = PBC_TYPE_VARIABLE;
  string[i].param[p].arg1.symbol = random_int(Num_arguments);

  do {
    if (string[i].param[p].op != PBC_OP_NONE)
      random_pvalue(&string[i].param[p].arg2, FALSE);
  } while ((string[i].param[p].arg1.type ==
	    string[i].param[p].arg2.type) &&
	   (string[i].param[p].arg1.symbol ==
	    string[i].param[p].arg2.symbol));
}


void Prod_Body :: make_random_terminal(int i)
{
  string[i].type = PBC_TYPE_TERMINAL;
  string[i].symbol = random_terminal();

  random_single_param(i, 0, TRUE);
}

void Prod_Body :: make_random_sequence(int start, int length)
{
  int i;
  Real choice;

  if (Num_productions <= 1)
    Prob_prod = 0.0;
  else
    Prob_prod = PROB_PROD;


  for (i=start; i<start+length; i++) {
    choice = random_dbl() * (PROB_TERMINAL + Prob_prod);

    if (choice < PROB_TERMINAL) {
      // symbol is a terminal.
      make_random_terminal(i);

    } else {
      // symbol is a production.
      make_random_production(i);
    }
  }
}


// Length does not include the [ and ] characters.
void Prod_Body :: make_random_branch(int start, int length)
{
  int end = start + length + 1;

  // Put branch start symbol '['
  string[start].type = PBC_TYPE_BLOCK;
  string[start].symbol = SYMBOL_PUSH;
  string[start].match = end;
  
  // Insert random sequence.
  make_random_sequence(start+1, length);

  // Put branch end symbol ']'
  string[end].type = PBC_TYPE_BLOCK;
  string[end].symbol = SYMBOL_POP;
  string[end].match = start;
}


// Length does not include the < and > characters.
void Prod_Body :: make_random_branch_edge(int start, int length)
{
  int end = start + length + 1;

  // Put branch start symbol '<'
  string[start].type = PBC_TYPE_BLOCK;
  string[start].symbol = SYMBOL_PUSH_EDGE;
  string[start].match = end;
  
  // Insert random sequence.
  make_random_sequence(start+1, length);

  // Put branch end symbol '>'
  string[end].type = PBC_TYPE_BLOCK;
  string[end].symbol = SYMBOL_POP_EDGE;
  string[end].match = start;
}


void Prod_Body :: make_random_mult(int start, int length)
{
  int end = start + length + 1;

  // Put branch start symbol '['
  string[start].type = PBC_TYPE_BLOCK;
  string[start].symbol = SYMBOL_BLK_START;
  string[start].match = end;
  
  // Insert random sequence.
  make_random_sequence(start+1, length);

  // Put branch end symbol ']'
  string[end].type = PBC_TYPE_BLOCK;
  string[end].symbol = SYMBOL_BLK_END;
  string[end].match = start;

  random_single_param(end, 0, FALSE);
  if (string[end].param[0].arg1.type == PBC_TYPE_VALUE) {
    // Put random values in rangoe 2-3, not 1-5.
    string[end].param[0].arg1.value = Real(random_int(2)+2);
  }
}


void Prod_Body :: duplicate_pchar(int to_loc, int from_loc, Prod_Body *from_body)
{
  copy_pchar(&from_body->string[from_loc], &string[to_loc]);
}

void Prod_Body :: duplicate_string(Prod_Body *body)
{
  length = body->length;
  for (int i=0; i<length; i++) {
    copy_pchar(&body->string[i], &string[i]);
  }
}

void Prod_Body :: duplicate_cond(Prod_Body *body)
{
  cond.comparator = body->cond.comparator;
  copy_pvalue(&cond.arg1, &body->cond.arg1);
  copy_pvalue(&cond.arg2, &body->cond.arg2);
}


// This body becomds a copy of input.
void Prod_Body :: duplicate(Prod_Body *body)
{
  int i;

  used = body->used;
  tested = body->tested;

  duplicate_cond(body);

  for (i=0; i<Num_arguments; i++) {
    max[i] = body->max[i];
    min[i] = body->min[i];
  }

  duplicate_string(body);
  if (!valid()) {
    printf("prod_body :: duplicate() - invalid!\n");
  }
}

void print_string_detailed(LString *string);
void print_string(LString *string);

// Makes a duplicate of an Lstring.
int Prod_Body :: duplicate(int len, LString *lstring)
{
  int i;

  for (i=0; i<len; i++) {
    if (lstring[i].symbol == TERM_PUSH) {
      string[i].type = PBC_TYPE_BLOCK;
      string[i].symbol = SYMBOL_PUSH;

    } else if (lstring[i].symbol == TERM_POP) {
      string[i].type = PBC_TYPE_BLOCK;
      string[i].symbol = SYMBOL_POP;

    } else if (lstring[i].symbol == TAPE_END) {
      break;

    } else {
      string[i].type = PBC_TYPE_TERMINAL;
      string[i].symbol = lstring[i].symbol;
      string[i].param[0].op = PBC_OP_NONE;
      string[i].param[0].arg1.type = PBC_TYPE_VALUE;
      string[i].param[0].arg1.value = lstring[i].arg;
    }

    //    fprint_char(stdout, i);
  }

  length = i;

  if (!valid()) {
    printf("prod_body::duplicate_pheno() - invalid body: %d:%d\n", len, length);
    fprint(stdout);
    printf("\nlstring:\n");
    print_string_detailed(lstring);
    print_string(lstring);
    printf("\n");
    return FALSE;
  }

  return TRUE;
}



Real calc_prod_value(Real *arg, Constant *constant, Prod_Value *pval)
{
  Real value;

  if (pval->type == PBC_TYPE_NONE)
    return 0.0;
  else if (pval->type == PBC_TYPE_VALUE)
    value = pval->value;
  else if (pval->type == PBC_TYPE_VARIABLE)
    value = arg[pval->symbol];
  else if (pval->type == PBC_TYPE_CONSTANT) {
    value = constant[pval->symbol].value;
    constant[pval->symbol].used++;

  } else {
    printf("calc prod value:: error, invalid arg type: %d.\n", pval->type);
    while (1) ;
  }

  return value;
}


#define MAX_ARG_VAL 10000000000.0
#define MIN_ARG_VAL -10000000000.0
Real calc_prod_param(Real *arg, Constant *constant,
		     Prod_Param *param)
{
  Real val1, val2, result;

  if (param->op == PBC_OP_NONE)
    return calc_prod_value(arg, constant, &(param->arg1));

  if (param->op == PBC_OP_EQUAL)
    return calc_prod_value(arg, constant, &(param->arg2));

  val1 = calc_prod_value(arg, constant, &(param->arg1));
  val2 = calc_prod_value(arg, constant, &(param->arg2));

  if (param->op == PBC_OP_ADD) {
    result = val1 + val2;    
  } else if (param->op == PBC_OP_SUB) {
    result =  val1 - val2;
  } else if (param->op == PBC_OP_MULT) {
    result = val1 * val2;
  } else if (param->op == PBC_OP_DIV) {
    if (absv(val2) <= 1.0) { // Don't allow division to increase size of value.
      result = val1; // treat as if val2 = 1.
    } else {
      result = val1 / val2;
    }
  } else if (param->op == PBC_OP_POW) {
    result = pow(val1, val2);
  } else {
    // shouldn't get here.
    result = 0.0;
  }

  if (result > MAX_ARG_VAL) {
    result = MAX_ARG_VAL;
  } else if (result < MIN_ARG_VAL) {
    result = MIN_ARG_VAL;
  }

  return result;
}


// Set body condition to always true.
void Prod_Body :: set_cond_true(void)
{
  cond.comparator = PBC_COMP_TRUE;
}

// Set body condition to: variable > 0.
void Prod_Body :: set_cond_greater_zero(void)
{
  if (Num_arguments == 0) {
    set_cond_true();
    return;
  }

  cond.arg1.type = PBC_TYPE_VARIABLE;
  cond.arg1.symbol = random_int(Num_arguments);

  cond.arg2.type = PBC_TYPE_VALUE;
  cond.arg2.value = 0.0;
}


int Prod_Body :: is_cond_true(void)
{
  return (cond.comparator == PBC_COMP_TRUE);
}

int Prod_Body :: test_cond(Real *arg)
{
  int i;
  Real val1, val2;

  tested++;
  if (tested == 1) {
    for (i=0; i<Num_arguments; i++) {
      min[i] = arg[i];
      max[i] = arg[i];
    }
  } else {
    for (i=0; i<Num_arguments; i++) {
      if (arg[i] < min[i])
	min[i] = arg[i];
      else if (arg[i] > max[i])
	max[i] = arg[i];
    }
  }

  if (cond.comparator == PBC_COMP_TRUE)
    return TRUE;

  val1 = calc_prod_value(arg, constant, &(cond.arg1));
  val2 = calc_prod_value(arg, constant, &(cond.arg2));

  if (cond.comparator == COMP_GREATER)
    return (val1 > val2);  
  else if (cond.comparator == COMP_LESS)
    return (val1 < val2);

  return FALSE;
}




int Prod_Body :: write_body(Real *arg, Tape *tape_dest)
{
  int i, j, t = 0;

  used++;
  if (used == 1) {
    for (i=0; i<Num_arguments; i++) {
      min[i] = arg[i];
      max[i] = arg[i];
    }
  } else {
    for (i=0; i<Num_arguments; i++) {
      if (arg[i] < min[i])
	min[i] = arg[i];
      else if (arg[i] > max[i])
	max[i] = arg[i];
    }
  }

  i=0;
  while (i < length) {
#ifndef OPTIMIZE
    if (Print[3]) {
      fprint_char(stdout, i);
    }
#endif

    tape_dest[t].type = string[i].type;
    tape_dest[t].symbol = string[i].symbol;

    if (is_production(i)) {
      for (j=0; j<Num_arguments; j++)
	tape_dest[t].arg[j] = calc_prod_param(arg, constant,
					      &(string[i].param[j]));
    } else {
      // if multiplier block start, get arg from branch end character.
      if (is_block(i) && (string[i].symbol == SYMBOL_BLK_START)) {
	tape_dest[t].arg[0] = calc_prod_param(arg, constant,
					      &(string[string[i].match].param[0]));
	if (tape_dest[t].arg[0] == 0.0) {
	  // Empty multiplier.
	  if (string[i].match < i) {
	    printf("Prod_body ::  Error with empty multiplier %d(%c) -> %d.\n", i,
		   string[i].symbol, string[i].match);
	    fprint(stdout);
	  }
	  i = string[i].match + 1;
	  continue;
	}

      } else
	tape_dest[t].arg[0] = calc_prod_param(arg, constant,
					      &(string[i].param[0]));
    }

    t++;
    i++;
  }

  if (Print[3]) {
    printf("(i=%d;l=%d)", i, length);
  }
    
  return t;
}




void Prod_Body :: mutate_pvalue(int index, Prod_Value *pvalue)
{
  char prev_type;
  char prev_symbol;
  Real prev_value;

  if ((!random_int(4)) ||
      ((pvalue->type == PBC_TYPE_VARIABLE) && (Num_arguments == 1)) ||
      ((pvalue->type == PBC_TYPE_CONSTANT) && (Num_constants == 1)) ||
      ((pvalue->type == PBC_TYPE_FUNCTION) && (Num_functions == 1))) {
    // Make completely random.

    prev_symbol = pvalue->symbol;
    prev_type = pvalue->type;
    prev_value = pvalue->value;

    while (1) {
      if ((string[index].type != PBC_TYPE_TERMINAL) ||
	  (index == -1))
	random_pvalue(pvalue, FALSE);
      else
	random_pvalue(pvalue, TRUE);

      if ((prev_type != pvalue->type) ||
	  (prev_symbol != pvalue->symbol) ||
	  (prev_value != pvalue->value))
	break;
    }
    return;
  }

  prev_symbol = pvalue->symbol;
  if (pvalue->type == PBC_TYPE_VALUE) {
    // INTEGER*

    if (random_int(2))
      pvalue->value += 1.0;
    else {
      pvalue->value -= 1.0;
      //      if (pvalue->value <= 0.0)
      //	pvalue->value = 0.0;
    }
    prev_symbol++; // make different.
    
  } else if (pvalue->type == PBC_TYPE_VARIABLE) {
    pvalue->symbol = random_int(Num_arguments);

  } else if (pvalue->type == PBC_TYPE_CONSTANT) {
    pvalue->symbol = random_int(Num_constants);

  } else if (pvalue->type == PBC_TYPE_FUNCTION) {
    pvalue->symbol = random_int(Num_functions);
  }

  prev_type = pvalue->type;
  while (1) {
    random_pvalue(pvalue, FALSE);
    if ((prev_type != pvalue->type) ||
	(prev_symbol != pvalue->symbol))
      break;
  }
}


void Prod_Body :: mutate_param_single(int index)
{
  mutate_pvalue(index, &string[index].param[0].arg1);
}


// Only for productions.
// Note: if no arguments -> cannot mutate anything.
int Prod_Body :: mutate_param_full(int index)
{
  int p;

  if (Num_arguments == 0)
    return FALSE;

  p = random_int(Num_arguments);

  if (!random_int(4)) {
    // Mutate op 25% of the time.
    int prev_op = string[index].param[p].op;
    //    string[index].param[p].op = random_int(Num_pbc_ops)+1;
    string[index].param[p].op = random_func();
    if (string[index].param[p].op != prev_op)
      return TRUE;
    // If op is unchanged -> mutate an argument.
  }

  // Mutate an argument.
  if (random_int(2)) {
    mutate_pvalue(index, &string[index].param[p].arg1);
  } else {
    mutate_pvalue(index, &string[index].param[p].arg2);
  }

  return TRUE;
}


int Prod_Body :: mutate_param(int index)
{
#ifndef OPTIMIZE
  if (Print[9]) {
    printf("Mutate param: %d\n", index);
  }
#endif

  if (is_production(index)) {
    parent->duplicate_used_bodies(string[index].symbol);
    return mutate_param_full(index);
  } else {
    mutate_param_single(index);
  }

  return TRUE;
}

void Prod_Body :: reset_single_param(int index)
{
  string[index].param[0].arg1.type = PBC_TYPE_VALUE;
  string[index].param[0].arg1.value = 1.0;
}


int Prod_Body :: mutate_symbol(int index)
{
  char prev_symbol;

#ifndef OPTIMIZE
  if (Print[9]) {
    printf("Mutate symbol: %d\n", index);
  }
#endif

  prev_symbol = string[index].symbol;
  if (string[index].type == PBC_TYPE_TERMINAL) {
    string[index].symbol = random_terminal();
    if (string[index].symbol == prev_symbol) {
      // If same symbol as original - recreate and reset.
      string[index].symbol = random_terminal();
      reset_single_param(index);
    }

  } else if (string[index].type == PBC_TYPE_PRODUCTION) {
    string[index].symbol = random_production();
    if (string[index].symbol == prev_symbol) {
      // If same symbol as original - make whole new production.
      make_random_production(index);
    }

  } else if (string[index].type == PBC_TYPE_BLOCK) {
    // Set index to be end of block location.
    if (index < string[index].match)
      index = string[index].match;

    if (string[index].symbol == SYMBOL_POP) {
      // Switch to multiplier block.
      string[index].symbol = SYMBOL_BLK_END;
      string[string[index].match].symbol = SYMBOL_BLK_START;
      random_single_param(index, 0, FALSE);

    } else {
      // Switch to branch block.
      string[index].symbol = SYMBOL_POP;
      string[string[index].match].symbol = SYMBOL_PUSH;
      reset_single_param(index);
    }
  }

  return TRUE;
}


int Prod_Body :: mutate_type(int index)
{
  // Can't mutate block types, only terminals and productions.
  if ((string[index].type == PBC_TYPE_BLOCK) ||
      (Num_productions <= 1))
    return FALSE;

#ifndef OPTIMIZE
  if (Print[9]) {
    printf("Mutate type: %d\n", index);
  }
#endif

  if (string[index].type == PBC_TYPE_TERMINAL) {
    make_random_production(index);
  } else { 
    make_random_terminal(index);
  }

  return TRUE;
}


int Prod_Body :: mutate_character_orig(void)
{
  int i;
  int choice, result, index;

  // Select a mutatable character.
  for (i=0; i<5; i++) {
    index = random_int(length);
    if (is_block(index) && 
	((Prob_branch == 0.0) || (Prob_mult_seq == 0.0)))
      continue;
    break;
  }

  // Verify it is mutatable.
  if (is_block(index) && 
      ((Prob_branch == 0.0) || (Prob_mult_seq == 0.0))) {
    // Specifically, if only 1 block type allowed, cannot
    // mutate a block.
    for (; index<length; index++) {
      if (!is_block(index))
	break;
    }
    if (index == length)
      return FALSE;
  }


#ifndef OPTIMIZE2
  parent->op_used(OP_MUT_CHAR);
#endif

#ifndef OPTIMIZE
  if (Print[9]) {
    fprint(stdout);
    printf("Mutate character: %d\n", index);
  }
#endif

  // Mutate: type, symbol, parameter.
  // Note: twice as likely to mutate a parameter.
  if (is_block(index)) {
    if (index > string[index].match)
      index = string[index].match;

    if (string[index].symbol == SYMBOL_BLK_END)
      choice = random_int(3) + 1;
    else
      choice = 1;

  } else if (Num_productions == 0) {
    choice = random_int(3) + 1;

  } else if (is_terminal(index)) {
    choice = random_int(2);

  } else {
    choice = random_int(4);
  }

  if (choice == 0) {
    result = mutate_type(index);

  } else if (choice == 1) {
    result = mutate_symbol(index);

  } else {
    result = mutate_param(index);
  }

#ifndef OPTIMIZE
  if (Print[9]) {
    fprint(stdout);
  }
#endif

  return result;
}


int Prod_Body :: mutate_character(void)
{
  int i;
  int choice, result, index;

  if (length == 0)
    return FALSE;  

  // Test for bodies containing only block symbols.
  for (i=0; i<length; i++) {
    if (!is_block(i))
      break;
  }
  if (i == length) {
    // If a body containing only block symbols is found, empty it.
    length = 0;
    return FALSE;
  }

  // Select a mutatable character.
  while (1) {
    index = random_int(length);
    if (is_block(index)) {
      if (!is_mult_end(index))
	continue;
    }
    break;
  }

#ifndef OPTIMIZE2
  parent->op_used(OP_MUT_CHAR);
#endif

#ifndef OPTIMIZE
  if (Print[9]) {
    fprint(stdout);
    printf("Mutate character: %d\n", index);
  }
#endif

  if (Num_productions <= 1)
    Prob_mut_char_type = 0.0;
  else
    Prob_mut_char_type = PROB_MUT_CHAR_TYPE;
    

  // Mutate: 
  //  choice = 0  -> parameter.
  //  choice = 1  -> symbol.
  //  choice = 2  -> type.
  if (is_mult_end(index)) {
    // If multiplier symbol, can only mutate the multiplier amount.
    choice = 0;
  } else {
    Real sum = Prob_mut_char_param + Prob_mut_char_symbol + Prob_mut_char_type;
    sum *= random_dbl();

    if (sum < Prob_mut_char_param)
      choice = 0;
    else if (sum < Prob_mut_char_param + Prob_mut_char_symbol)
      choice = 1;
    else 
      choice = 2;
  }

  if (choice == 0) {
    result = mutate_param(index);

  } else if (choice == 1) {
    result = mutate_symbol(index);

  } else {
    // Actually -> make random.
    //    result = mutate_type(index);
    make_random_sequence(index, 1);
    result = TRUE;
  }

#ifndef OPTIMIZE
  if (Print[9]) {
    fprint(stdout);
  }
#endif

  return result;
}


int Prod_Body :: insert_prod_rule(int i)
{
  int p, r, ins1;

  r = parent->random_unused_rule();
  if (r == -1) {
    return FALSE;
  }

  if (!parent->randomize_rule(r)) {
    return FALSE;
  }


  ins1 = TRUE;
  if (random_int(2)) {
    if (insert_space(i, 3)) {
      ins1 = FALSE;

      string[i].type = PBC_TYPE_BLOCK;
      string[i].symbol = SYMBOL_PUSH;
      string[i].match = i+2;

      // Put branch end symbol ']'
      string[i+2].type = PBC_TYPE_BLOCK;
      string[i+2].symbol = SYMBOL_POP;
      string[i+2].match = i;

      i++;
    }
  }

  if (ins1) {
    if (!insert_space(i, 1)) {
#ifndef OPTIMIZE
      if (Print[9])
	printf("prod_body :: insert_prod_rule() - error - could not insert enough space.\n");
#endif
      return FALSE;
    }
    
  }


  string[i].type = PBC_TYPE_PRODUCTION;
  string[i].symbol = r;
  for (p=0; p<Num_arguments; p++) {
    random_full_param(i, p);
  }

  return TRUE;
}

int Prod_Body :: add_symbol(int i)
{
  int type;
  int seq_len;
  int tot_len;
  Real choice;

#ifndef OPTIMIZE2
  parent->op_used(OP_MUT_ADD);
#endif

#ifndef OPTIMIZE
  if (Print[9]) {
    cout << "Add symbol: " << i << endl;
    fprint(stdout);
  }
#endif

  if (!random_int(3)) {
    // Add a single symbol.
    if (insert_prod_rule(i)) {
      return TRUE;
    }

    // If insert_prod_rule fails, insert a single character.
    type = 0;
    seq_len = 1;
    tot_len = 1;

  } else {
    choice = random_dbl() * (Prob_mult_seq + Prob_branch + PROB_SEQUENCE + Prob_branch_edge);
    if (choice < PROB_SEQUENCE) {
      type = 0;
    } else if (choice < PROB_SEQUENCE + Prob_branch) {
      type = 1;
    } else if (choice < PROB_SEQUENCE + Prob_branch + Prob_mult_seq) {
      type = 2;
    } else {
      type = 3;
    }

    seq_len = random_int(3)+1;
    if (type == 0)
      tot_len = seq_len; // simple sequence
    else
      tot_len = seq_len + 2; // a block sequence
  }

  if (!insert_space(i, tot_len)) {
#ifndef OPTIMIZE
    if (Print[9])
      printf("error - could not insert enough space.\n");
#endif
    return FALSE;
  }

#ifndef OPTIMIZE
  if (Print[9]) {
    printf("from: %d  to: %d\n", i, i+tot_len);
  }
#endif

  if (type == 0) {
    make_random_sequence(i, seq_len);
  } else if (type == 1) {
    make_random_branch(i, seq_len);
  } else if (type == 2) {
    make_random_mult(i, seq_len);
  } else {
    make_random_branch_edge(i, seq_len);
  }

#ifndef OPTIMIZE
  if (Print[9])
    fprint(stdout);
#endif

  return TRUE;
}


void Prod_Body :: make_random_condition(void)
{
  //  if ((!random_int(5)) || Num_arguments == 0) {
  if (Num_arguments == 0) {
    set_cond_true();
    return;
  }

  cond.comparator = COMP_GREATER;

  // Ensure that at least one of arg1 and arg2 use a variable.
  //  if (random_int(2)) {
  if (TRUE) {
    // Put variable in first arg.
    cond.arg1.type = PBC_TYPE_VARIABLE;
    cond.arg1.symbol = random_int(Num_arguments);

    do {
      random_pvalue(&cond.arg2, FALSE);
    } while ((cond.arg1.type == cond.arg2.type) &&
	     (cond.arg2.symbol == cond.arg2.symbol));
  } else {
    // Put variable in first arg.
    cond.arg2.type = PBC_TYPE_VARIABLE;
    cond.arg2.symbol = random_int(Num_arguments);

    random_pvalue(&cond.arg1, FALSE);
  }
}


void Prod_Body :: make_random(void)
{
  int i;
  int len, size;
  Real choice;


  make_random_condition();


  //  size = random_int(10);
  if (Num_productions > 1) {
    //    size = random_int(4) + random_int(4) + 1;
    //    size = random_int(6) + random_int(6) + random_int(6) + 1;
    size = random_int(6) + random_int(7) + random_int(7) + 4;
    //    size = random_int(7) + random_int(7) + random_int(7) + 4;
  } else {
    //    size = random_int(100) + 4;
    size = random_int(50) + random_int(50) + 2;
  }

  if (size > Max_prod_length)
    size = 1;
  else if (size == 0)
    size = 1;

  
  length = size;
  for (i=0; i<size; ) {
    if (size-i <= 2) {
      choice = 0.0;
      len = size-i;
    } else {      
      choice = random_dbl() * (Prob_mult_seq + Prob_branch + PROB_SEQUENCE + Prob_branch_edge);
      if (MAX_SEQ_LENGTH <= MIN_SEQ_LENGTH) {
	len = MIN_SEQ_LENGTH;
      } else {
	len = random_int(MAX_SEQ_LENGTH-MIN_SEQ_LENGTH+1) + MIN_SEQ_LENGTH;
      }
      if (len+i > size)
	len = size - i;
    }

    if (choice < PROB_SEQUENCE) {
      make_random_sequence(i, len);
      i += len;

    } else if (choice < Prob_branch + PROB_SEQUENCE) {
      // Branch sequence.
      if (len+i > size-2)
	len = size - i - 2;
      make_random_branch(i, len);
      i += len + 2;

    } else if (choice < Prob_branch + PROB_SEQUENCE + Prob_mult_seq) {
      // Multiplier sequence.
      if (len+i > size-2)
	len = size - i - 2;
      make_random_mult(i, len);
      i += len + 2;

    } else {
      // Branch edge sequence.
      if (len+i > size-2)
	len = size - i - 2;
      make_random_branch_edge(i, len);
      i += len + 2;
    }
  }


#ifndef OPTIMIZE
  if (!valid()) {
    printf("Prod_Body :: make random() - invalid body.\n");
    fprint(stdout);
  }
#endif
}


int Prod_Body :: reorder(void)
{
  int i, tries;

  if ((length <= 1) || (length == Prod_length))
    return FALSE;

  if (length == 2) {
    i = 0;
    if (is_block(0) && (is_block(1)))
      return FALSE;

  } else {
    tries = 0;
    while (TRUE) {
      if (tries++ > 10)
	return FALSE;
      i = random_int(length-1);
      if (is_block(i) && (is_block(i+1)))
	continue;
      break;
    }
  }

#ifndef OPTIMIZE2
  parent->op_used(OP_MUT_REORDER);
#endif

#ifndef OPTIMIZE
  if (Print[9]) {
    cout << "Re-order: " << i << endl;
    fprint(stdout);
  }
#endif

  move_symbol(i, Prod_length-1);
  move_symbol(i+1, i);
  move_symbol(Prod_length-1, i+1);

  if ((string[i+1].match == i+2) &&
      is_block_start(i+1))
    delete_block(i+2, TRUE);

  if ((string[i].match == i+1) &&
      is_block_start(i))
    delete_block(i+1, TRUE);

#ifndef OPTIMIZE
  if (Print[9]) {
    fprint(stdout);
  }
#endif

  return TRUE;
}


int Prod_Body :: replicate_old1(void)
{
  int i;

  // replicate.
  if (length > Max_prod_length)
    return FALSE;
  else if (length == 0)
    return FALSE;

  int loc1, loc2;
  for (i=0; i<5; i++) {
    random_subblock(&loc1, &loc2);
    // Ensure the block is non-empty.
    if (loc1 <= loc2)
      break;
  }

#ifndef OPTIMIZE2
  parent->op_used(OP_MUT_REPLICATE);
#endif

#ifndef OPTIMIZE
  if (Print[9]) {
    fprint(stdout);
    printf("Replicate: loc1:%d  loc2:%d\n", loc1, loc2);    
  }
#endif

  if (loc2 < loc1)
    // Got an empty block.
    return FALSE;

  if ((!is_mult_start(loc1)) || (string[loc1].match != loc2)) {
    // If already a multiplier block, just generate a new multiplier.
    loc2++;
    insert_space(loc2, 1);
    insert_space(loc1, 1);
    loc2++;
      
    string[loc1].type = PBC_TYPE_BLOCK;
    string[loc1].symbol = SYMBOL_BLK_START;

    string[loc2].type = PBC_TYPE_BLOCK;
    string[loc2].symbol = SYMBOL_BLK_END;

    random_single_param(loc2, 0, FALSE);

  } else {
    char prev_type = string[loc2].param[0].arg1.type;
    char prev_symbol = string[loc2].param[0].arg1.symbol;
    Real prev_value = string[loc2].param[0].arg1.value;

    while (1) {
      // Ensure new parameter is different.
      random_single_param(loc2, 0, FALSE);
      if (prev_type != string[loc2].param[0].arg1.type)
	break;
      else if (prev_type == PBC_TYPE_VALUE) {
	if (prev_value != string[loc2].param[0].arg1.value)
	  break;
      } else if (prev_symbol != string[loc2].param[0].arg1.symbol) {
	break;
      }
    }
  }

#ifndef OPTIMIZE
  if (Print[9]) {
    fprint(stdout);
  }
#endif

  return TRUE;
}


// Copy a block of code.
int Prod_Body :: replicate(void)
{
  int i;
  int loc1, loc2, len=0;

  // replicate.
  if (length > Max_prod_length)
    return FALSE;
  else if (length == 0)
    return FALSE;

  // Find a block that isn't empty and not too long.
  for (i=0; i<5; i++) {
    random_subblock(&loc1, &loc2);

    len = loc2 - loc1 + 1;
    if (len + length >= Prod_length)
      // Too long.
      continue;

    if (loc1 <= loc2)
      // Ok.
      break;
  }

  if (loc2 < loc1)
    // Got an empty block.
    return FALSE;
  else if (len + length >= Prod_length)
    // Too long.
    return FALSE;

#ifndef OPTIMIZE2
  parent->op_used(OP_MUT_REPLICATE);
#endif

#ifndef OPTIMIZE
  if (Print[9]) {
    fprint(stdout);
    printf("Replicate: loc1:%d  loc2:%d\n", loc1, loc2);    
  }
#endif


  // Duplicate the sequence.
  loc2++;
  if (!insert_space(loc2, len)) {
    printf("Prod_Body :: replicate() - error, could not insert space.\n");
    return FALSE;
  }

  for (i=loc1; i<loc2; i++)
    copy_symbol(i, i+len);
  if (!match_parens()) {
    printf("Prod_Body :: replicate() - error, parens do not match.\n");
  }

#ifndef OPTIMIZE
  if (Print[9]) {
    fprint(stdout);
  }
#endif

  return TRUE;
}


// Adjusts arg2 to split the variable min-max range.
// In doing this, it makes arg2 a value.
int Prod_Body :: repair_arg2(void)
{
  int v = cond.arg1.symbol;
  
  if (used == 0)
    return FALSE;
  else if (max[v] <= min[v])
    return FALSE;

  cond.arg2.type = PBC_TYPE_VALUE;
  if (max[v]-min[v] < 1.0) {
    cond.arg2.value = min[v];
  } else {
    cond.arg2.value = Real(random_int(int(max[v])-int(min[v]))) +
      min[v];
  }
  return TRUE;
}


int Prod_Body :: usable_cond(void)
{
  int v1 = cond.arg1.symbol;
  Real val2;

  if (cond.arg2.type == PBC_TYPE_VALUE) {
    val2 = cond.arg2.value;
    if ((val2 < max[v1]) && (val2 >= min[v1]))
      return TRUE;

  } else if (cond.arg2.type == PBC_TYPE_VARIABLE) {
    int v2 = cond.arg2.symbol;
    if (v1 == v2)
      return FALSE;
    
    if (((max[v2] < max[v1]) && (max[v2] > min[v1])) ||
	((min[v2] < max[v1]) && (min[v2] > min[v1])))
      return TRUE;

  } else if (cond.arg2.type == PBC_TYPE_CONSTANT) {
    val2 = constant[cond.arg2.symbol].value;
    if ((val2 < max[v1]) && (val2 > min[v1]))
      return TRUE;
  }

  return FALSE;
}


// Returns true if the condition was always true.
int Prod_Body :: always_true(void)
{
  return (tested == used);
}


int Prod_Body :: mutate_cond_old1(void)
{
  if (Num_arguments == 0)
    return FALSE;

#ifndef OPTIMIZE2
  parent->op_used(OP_MUT_COND);
#endif

#ifndef OPTIMIZE
  if (Print[9]) {
    printf("Mutate condition\n");
    fprint(stdout);
  }
#endif

  if (!used || (always_true() && !is_cond_true())) {
    int b = parent->repair_rule(phead);
    if (b < 0)
      return FALSE;
    return TRUE;
    //    repair();
  }


  if ((!random_int(4)) && (Num_arguments > 1)) {
    // mutate first arg.
    cond.arg1.symbol = random_int(Num_arguments);
  }


  int prev_type = cond.arg2.type;
  int prev_symbol = cond.arg2.symbol;

  mutate_pvalue(-1, &cond.arg2);
  if (!usable_cond()) {
    if (!repair_arg2()) {
      // Should do major fix.
      // Restore condition and try again.
      cond.arg2.type = prev_type;
      cond.arg2.symbol = prev_symbol;
      mutate_pvalue(-1, &cond.arg2);
    }
  }
      

#ifndef OPTIMIZE
  if (Print[9]) {
    fprint(stdout);
  }
#endif

  return TRUE;
}


// Sets condition arg 2 to a random value
// within the decision range for the variable.
int Prod_Body :: make_valid_cond_arg2(void)
{
  int v = cond.arg1.symbol;
  int maxv = int(max[v]);
  int minv = int(min[v]);

  cond.arg2.type = PBC_TYPE_VALUE;
  if (maxv-minv <= 0) {
    cond.arg2.value = Real(minv-1.0);
    return FALSE;
  }

  if (maxv-minv < 1.0)
    cond.arg2.value = Real(minv);
  else {
    cond.arg2.value = Real(random_int(maxv-minv) + minv);
  }
  return TRUE;
}


int Prod_Body :: mutate_cond(void)
{
  int v;

  if (Num_arguments == 0)
    return FALSE;

  if (!tested) {
    return FALSE;
  }

  if (is_cond_true())
    return FALSE;


#ifndef OPTIMIZE2
  parent->op_used(OP_MUT_COND);
#endif

#ifndef OPTIMIZE
  if (Print[9]) {
    printf("Mutate condition\n");
    fprint(stdout);
  }
#endif

  if ((!random_int(4)) && (Num_arguments > 1)) {
    // Mutate first arg.
    int prev_var = cond.arg1.symbol;

    if (Num_arguments == 2) {
      if (cond.arg1.symbol == 0) {
	cond.arg1.symbol = 1;
      } else {
	cond.arg1.symbol = 0;
      }

      if (make_valid_cond_arg2()) {
	return TRUE;
      }
      cond.arg1.symbol = prev_var;

    } else {
      v = random_int(Num_arguments-1);
      if (v == cond.arg1.symbol) {
	cond.arg1.symbol = Num_arguments-1;
      } else {
	cond.arg1.symbol = v;
      }

      if (make_valid_cond_arg2()) {
	return TRUE;
      }
      cond.arg1.symbol = prev_var;
    }
  }

  // Mutate second arg.
  if (!usable_cond()) {
    return make_valid_cond_arg2();
  } else if (cond.arg2.type == PBC_TYPE_CONSTANT) {
    return make_valid_cond_arg2();
  }

  // Type of second arg is value.
  v = cond.arg2.symbol;
  cond.arg2.type = PBC_TYPE_VALUE;

  if (cond.arg2.value > min[v]) {
    make_cond_smaller();
    //    cond.arg2.value += 1.0;

  } else if (max[v] <= cond.arg2.value) {
    make_cond_larger();
    //    cond.arg2.value -= 1.0;

  } else {
    if (random_int(2)) {
      make_cond_smaller();
      //      cond.arg2.value += 1.0;
    } else {
      make_cond_larger();
      //      cond.arg2.value -= 1.0;      
    }
  }

#ifndef OPTIMIZE
  if (Print[9]) {
    fprint(stdout);
  }
#endif

  return TRUE;
}





int Prod_Body :: mutate_old1(void)
{
  int result=-2;
  int i, choice=-1;
  Real val;

  Real op_prob[6];

  if (!used) {
    printf("Mutate: not used!  <%d:%d>\n", phead->used, used);
    parent->repair_rule(phead);
  }

  // choice:
  // 0 - mutate character - op 6
  // 1 - add symbol    - op 7
  // 2 - delete symbol - op 8
  // 3 - reorder       - op 9
  // 4 - condition     - op 10
  // 5 - replicate     - op 11
  // 6 - split: if Pa uses Pb, copy Pb to Pc, mutate Pc and change reference in Pa.

  // Set probabilities for each mutation type.
  op_prob[0] = PROB_MUT_CHAR;
  op_prob[1] = PROB_MUT_ADD;
  op_prob[2] = PROB_MUT_DEL;
  op_prob[3] = PROB_MUT_REORDER;
  op_prob[4] = PROB_MUT_COND;
  op_prob[5] = PROB_MUT_REPL;
  op_prob[6] = PROB_MUT_ENCAPSULATE;


  // Adjust probabilities to 0 if case cannot be performed.
  if (length == 0) {
    op_prob[0] = 0.0;
    op_prob[2] = 0.0;
    op_prob[3] = 0.0;
    op_prob[5] = 0.0;
    op_prob[6] = 0.0;

  } else if (length == 1) {
    op_prob[3] = 0.0;

  } else if (length > Max_prod_length) {
    op_prob[1] = 0.0;
    if (length == Prod_length)
      op_prob[3] = 0.0;
    op_prob[5] = 0.0;
  }
  if (used == 0) {
    op_prob[4] = 0.0;
  }


  // Randomly select a mutation type.
  for (i=1; i<7; i++)
    op_prob[i] += op_prob[i-1];
  val = random_dbl() * op_prob[6];

  for (i=0; i<7; i++) {
    if (val < op_prob[i]) {
      choice = i;
      break;
    }
  }


  if (choice == 0) {
    // Pick character to mutate.
    if (length == 0)
      return FALSE;
    result = mutate_character();
      
  } else if (choice == 1) {
    // Add symbol    
    i = random_int(length+1);
    result = add_symbol(i);

  } else if (choice == 2) {
    // Delete symbol
    i = random_int(length);
    result = delete_symbol(i, random_int(2));
    
  } else if (choice == 3) {
    // reorder.
    result = reorder();

  } else if (choice == 4) {
    result = mutate_cond();

  } else if (choice == 5) {
    // replicate.
    result = replicate();

  } else {
    printf("Prod_Body :: Mutate - error invalid choice.\n");
  }


#ifndef OPTIMIZE
  if (!valid()) {
    printf("prod_body::mutate() - error, invalid body.\n");
    printf("Mutation type: %d, i:%d\n\n", choice, i);
    exit(-1);
    return FALSE;
  }

  if (Print[9]) {
    if (!result)
      printf("Prod_Body :: Mutate - failed.  No mutation.\n");
    cout << endl;
  }
#endif
  return TRUE;
}


int Prod_Body :: mutate(void)
{
  int result=-2;
  int i, choice=-1;
  Real val;

  Real op_prob[7];

  // choice:
  // 0 - mutate character - op 6
  // 1 - add symbol    - op 7
  // 2 - delete symbol - op 8
  // 3 - reorder       - op 9
  // 4 - condition     - op 10
  // 5 - replicate     - op 11
  // 6 - split: if Pa uses Pb, copy Pb to Pc, mutate Pc and change reference in Pa.

  // Set probabilities for each mutation type.
  op_prob[0] = PROB_MUT_CHAR;
  op_prob[1] = PROB_MUT_ADD;
  op_prob[2] = PROB_MUT_DEL;
  op_prob[3] = PROB_MUT_REORDER;
  op_prob[4] = PROB_MUT_COND;
  op_prob[5] = PROB_MUT_REPL;
  op_prob[6] = PROB_MUT_ENCAPSULATE;


  if (!tested) {
    //    printf("Mutate: not tested!  <%d:%d>\n", phead->used, used);
    //    return FALSE;
    op_prob[4] = 0.0;
  }


  // Adjust probabilities to 0 if case cannot be performed.
  if (length == 0) {
    op_prob[0] = 0.0;
    op_prob[2] = 0.0;
    op_prob[3] = 0.0;
    op_prob[5] = 0.0;
    op_prob[6] = 0.0;

  } else if (length == 1) {
    op_prob[3] = 0.0;

  } else if (length > Max_prod_length) {
    op_prob[1] = 0.0;
    if (length == Prod_length)
      op_prob[3] = 0.0;
    op_prob[5] = 0.0;
  }
  if ((used == 0) || (Max_bodies < 2)) {
    op_prob[4] = 0.0;
  }
  //  op_prob[5] = 0.0; // only for testing


  // Randomly select a mutation type.
  for (i=1; i<7; i++)
    op_prob[i] += op_prob[i-1];
  val = random_dbl() * op_prob[6];

  for (i=0; i<7; i++) {
    if (val < op_prob[i]) {
      choice = i;
      break;
    }
  }


  if (choice == 0) {
    // Pick character to mutate.
    if (length == 0)
      return FALSE;
    result = mutate_character();
      
  } else if (choice == 1) {
    // Add symbol    
    i = random_int(length+1);
    result = add_symbol(i);

  } else if (choice == 2) {
    // Delete symbol
    i = random_int(length);
    result = delete_symbol(i, random_int(2));
    
  } else if (choice == 3) {
    // reorder.
    result = reorder();

  } else if (choice == 4) {
    result = mutate_cond();

  } else if (choice == 5) {
    // replicate.
    result = replicate();

  } else {
    printf("Prod_Body :: Mutate - error invalid choice.\n");
  }

#ifndef OPTIMIZE
  if (!valid()) {
    printf("prod_body :: mutate() - error, invalid body.\n");
    printf("Mutation type: %d, i:%d\n\n", choice, i);
    exit(-1);
    return FALSE;
  }

  if (Print[9]) {
    if (!result)
      printf("Prod_Body :: Mutate - failed.  No mutation.\n");
    cout << endl;
  }
#endif
  return TRUE;
}



int Prod_Body :: mutate_pheno(void)
{
  int result=-2;
  int i, choice=-1;
  Real val;

  Real op_prob[7];

  if (!tested) {
    printf("Mutate: not tested!  <%d:%d>\n", phead->used, used);
    return FALSE;
  }

  // choice:
  // 0 - mutate character - op 6
  // 1 - add symbol    - op 7
  // 2 - delete symbol - op 8
  // 3 - reorder       - op 9
  // 4 - condition     - op 10
  // 5 - replicate     - op 11
  // 6 - split: if Pa uses Pb, copy Pb to Pc, mutate Pc and change reference in Pa.

  // Set probabilities for each mutation type.
  op_prob[0] = PROB_MUT_CHAR;
  op_prob[1] = PROB_MUT_ADD;
  op_prob[2] = PROB_MUT_DEL;
  op_prob[3] = PROB_MUT_REORDER;
  op_prob[4] = PROB_MUT_COND;
  op_prob[5] = PROB_MUT_REPL;
  op_prob[6] = PROB_MUT_ENCAPSULATE;


  // Adjust probabilities to 0 if case cannot be performed.
  if (length == 0) {
    op_prob[0] = 0.0;
    op_prob[2] = 0.0;
    op_prob[3] = 0.0;
    op_prob[5] = 0.0;
    op_prob[6] = 0.0;

  } else if (length == 1) {
    op_prob[3] = 0.0;

  } else if (length > Max_prod_length) {
    op_prob[1] = 0.0;
    if (length == Prod_length)
      op_prob[3] = 0.0;
    op_prob[5] = 0.0;
  }
  op_prob[4] = 0.0;
  op_prob[6] = 0.0;
  //  op_prob[5] = 0.0; // only for testing


  // Randomly select a mutation type.
  for (i=1; i<7; i++)
    op_prob[i] += op_prob[i-1];
  val = random_dbl() * op_prob[6];

  for (i=0; i<7; i++) {
    if (val < op_prob[i]) {
      choice = i;
      break;
    }
  }


  if (choice == 0) {
    // Pick character to mutate.
    if (length == 0)
      return FALSE;
    result = mutate_character();
      
  } else if (choice == 1) {
    // Add symbol    
    i = random_int(length+1);
    result = add_symbol(i);

  } else if (choice == 2) {
    // Delete symbol
    i = random_int(length);
    result = delete_symbol(i, random_int(2));
    
  } else if (choice == 3) {
    // reorder.
    result = reorder();

  } else if (choice == 4) {
    result = mutate_cond();

  } else if (choice == 5) {
    // replicate.
    result = replicate();

  } else {
    printf("Prod_Body :: Mutate_pheno() - error invalid choice.\n");
  }

#ifndef OPTIMIZE
  if (!valid()) {
    printf("prod_body :: mutate_pheno() - error, invalid body.\n");
    printf("Mutation type: %d, i:%d\n\n", choice, i);
    exit(-1);
    return FALSE;
  }

  if (Print[9]) {
    if (!result)
      printf("Prod_Body :: mutate_pheno() - failed.  No mutation.\n");
    cout << endl;
  }
#endif
  return TRUE;
}



int Prod_Body :: block_beginning(int i)
{
  if (i >= length - 1)
    return 0;

  for (; i>0; i--) {
    if (is_block_start(i))
      return i;

    if (is_block_end(i))
      i = string[i].match;
  }
  return 0;
}
      
int Prod_Body :: block_ending(int i)
{
  if (i < 0)
    return length-1;

  if (is_block_start(i)) 
    return string[i].match;

  for (; i<length; i++) {    
    if (is_block_end(i))
      return i;

    if (is_block_start(i))
      i = string[i].match;
  }
  return length - 1;
}
      


// Select a random sequence of terminals and productions.
void Prod_Body :: pick_block(int *loc1, int *loc2)
{
  int i, j;

  // Test if block is empty.
  if (length == 0) {
    *loc1 = 0;
    *loc2 = -1;
    return;
  }

  // Go to beginning of subblock.
  i = block_beginning(*loc1);
  j = block_ending(i);

#ifndef OPTIMIZE
  if (Print[9]) {
    printf(" pick_block: %d-> (%d - %d)\n", *loc1, i, j);
  }
#endif

  *loc1 = i;
  *loc2 = j;
}


// Randomly select a block in body for recombining.
// A block is a {} or [] block or the entire production
// if it contains no such block.
void Prod_Body :: random_block(int *l1, int *l2)
{
  int loc1, loc2;

  // Test if block is empty.
  if (length == 0) {
    *l1 = 0;
    *l2 = -1;
    return;
  }

  // Get location in body.
  loc1 = random_int(length);

  if (is_block(loc1)) {
    // Re-order so loc1 < loc2.
    if (loc1 < string[loc1].match)
      loc2 = string[loc1].match;
    else {
      loc2 = loc1;
      loc1 = string[loc2].match;
    }
	
  } else {
    pick_block(&loc1, &loc2);
  }

  *l1 = loc1;
  *l2 = loc2;
}


int Prod_Body :: block_length(int l1, int l2)
{
  int len = 0;
#ifndef OPTIMIZE
  int tmp = l1;
#endif

  for (; l1<=l2; l1++) {
    len++;
    if (is_block_start(l1))
      l1 = string[l1].match;
  }

#ifndef OPTIMIZE
  if (Print[9])
    printf("block length: %d -> %d = %d\n", tmp, l2, len);
#endif

  return len;  
}


// Return the location in body of the len-th symbol after start.
// This will not end on a block start, unless len=0 and
// a block start is the first symbol.
int Prod_Body :: get_loc(int start, int len)
{
  int l1 = start;
#ifndef OPTIMIZE
  int l = len;
#endif

  while (len > 0) {
    if (l1 >= length) {
      break;
    }

    len--;
    if (is_block_start(l1))
      l1 = string[l1].match;
    l1++;
  }

  if (l1 >= length)
    l1 = length-1;

#ifndef OPTIMIZE
  if (Print[9])
    printf("get_loc: start:%d, len:%d -> %d\n",
	   start, l, l1);
#endif
  return l1;
}


void Prod_Body :: random_subblock(int *l1, int *l2)
{
  int len, start;
  int loc1, loc2;

  random_block(l1, l2);
#ifndef OPTIMIZE
  if (Print[9])
    printf("random_subblock: %d %d  (len:%d)\n", *l1, *l2, length);
#endif

  if (*l2 < *l1) {
    // body is empty.
    return;
  }

  loc1 = *l1;
  loc2 = *l2;

  // If loc1&loc2 are beginning and end of a block,
  // move them to first and last characters of the block.
  if (is_block(loc1)) {
    if (string[loc1].match == loc2) {
      loc1++;
      loc2--;
    }
  }

  len = block_length(loc1, loc2);
  if (len <= 0)
    return;

  start = random_int(len+1);
  // If start == len, pick entire block.
  // eg. don't do following block of code.
  if (start != len) {
    len = random_int(len - start + 1);

#ifndef OPTIMIZE
    if (Print[9])
      printf("start: %d   len: %d\n", start, len);
#endif

    loc1 = get_loc(loc1, start);

    if (len == 0)
      loc2 = loc1 - 1;
    else {
      if (len == 1)
	loc2 = loc1;
      else
	loc2 = get_loc(loc1, len - 1);

      // Make sure loc2 is at end, not start of block.
      if (is_block_start(loc2))
	loc2 = string[loc2].match;
    }

    *l1 = loc1;
    *l2 = loc2;
  }
}



int Prod_Body :: match_parens(char end_type, char end_symbol, int i)
{
  int index = i-1;

  /*
  if (Print[7]) {
    printf("Match parens starting at: %d\n", int(i));
    fprint(stdout);
  }
  */

  while (1) {
    if ((end_type == string[i].type) &&
	(end_symbol == string[i].symbol)) {
      if (is_block_end(i)) {
	string[i].match = index;
	string[index].match = i;
      }
      return i;

    } else if (i >= length) {
      if ((end_type == PBC_TYPE_NONE) &&
	  (end_symbol == TAPE_END))
	return i;

      printf("prod_body :: Match paren() - error, premature termination:@ %d(%d:%d)\n",
	       i, string[i].type, string[i].symbol);
      fprint(stdout);
      return -1;

    } else if (is_block_end(i)) {
      printf("prod_body :: Match paren() - error, unexpected block end:@ %d(%d:%d)\n",
	       i, string[i].type, string[i].symbol);
      fprint(stdout);
      return -1;
    }

    if (is_block_start(i)) {
      if (string[i].symbol == SYMBOL_PUSH) {
	if (!match_parens(PBC_TYPE_BLOCK, SYMBOL_POP, i+1))
	  return -1;
      } else if (string[i].symbol == SYMBOL_PUSH_EDGE) {
	if (!match_parens(PBC_TYPE_BLOCK, SYMBOL_POP_EDGE, i+1))
	  return -1;
      } else {
	if (!match_parens(PBC_TYPE_BLOCK, SYMBOL_BLK_END, i+1))
	  return -1;
      }
      /*
      if (string[i].match < 0) {
	printf("i: %d,  match: %d.\n", i, string[i].match);
	fprint(stdout);
	while (1);
      }
      */

      i = string[i].match;
    }

    i++;
  }
}


int Prod_Body :: match_parens(void)
{
  //  fprint(stdout);
  int len = match_parens(PBC_TYPE_NONE, TAPE_END, 0);
  return (len == length);
}


// Writes a block from body overtop of this object.
int Prod_Body :: recombine_blocks(Prod_Body *body)
{
  int i, diff;
  int locA1, locA2, locB1, locB2;

#ifndef OPTIMIZE2
  parent->op_used(OP_REC_PROD);
#endif
  for (i=0; i<5; i++) {
    // Repeat up to 5 times to ensure that length
    // of body2 does not become too large.
      
    // Select subblock in body.
    // Ensuring both blocks are not empty.
    while (1) {
      // Select subblock in other body.
      body->random_subblock(&locB1, &locB2);

      // Select subblock in this object.
      random_subblock(&locA1, &locA2);

      if (locB2 >= locB1)
	break;
      if (locA2 >= locA1)
	break;
    }

    diff = (locB2-locB1) - (locA2-locA1);
    if (length + diff < Max_prod_length)
      break;
  }

  // Check that new body length is not too long.
  if (length + diff >= Max_prod_length) {
#ifndef OPTIMIZE
    if (Print[9])
      printf("Recomb_block():: new length too big (%d+%d).  No recomb\n",
	     length, diff);
#endif
    return FALSE;
  }

  // Adjust to fit the incoming block.
  if (diff > 0) {
    // Must expand the string.
    insert_space(locA1, diff);

  } else if (diff < 0) {
    // Must shrink the string.
    for (i=locA2+1; i<length; i++)
      move_symbol(i, i+diff);

    length += diff;
  }

  // Copy block from body to this object.
  diff = locB2 - locB1;
  for (i=0; i<=diff; i++)
    copy_pchar(&(body->string[locB1+i]),
	       &string[locA1+i]);

  match_parens();
  repair_redundancy();

#ifndef OPTIMIZE
  if (Print[9]) {
    if (locA2 < locA1)
      printf("%d->%d  inserted at:  %d\n", locB1, locB2,
	     locA1);
    else
      printf("%d->%d  over:  %d->%d\n", locB1, locB2,
	     locA1, locA2);
    fprint(stdout);
    cout << endl;
  }

  if (!valid()) {
    printf("prod_body :: recombine_blocks() - Error - invalid body!\n");
    if (locA2 < locA1) {
      printf("   %d->%d  inserted at:  %d\n", locB1, locB2,
	     locA1);
    } else {
      printf("   %d->%d  over:  %d->%d\n", locB1, locB2,
	     locA1, locA2);
    }
    printf("  length: %d, diff: %d;  Max_prod_length: %d\n",
	   length, diff, Max_prod_length);
    return -1;
  }
#endif

  return TRUE;
}



// Recombines a block from body overtop of a block in this object.
int Prod_Body :: recombine(Prod_Body *body)
{
  int result;

#ifndef OPTIMIZE
  if (Print[9])
    printf("Recombine\n");
#endif


  if ((body->length == 0) &&
      (length == 0)) {
#ifndef OPTIMIZE
    if (Print[9])
      printf("No recombination :: both lengths equal 0.\n");
#endif
    return FALSE;
  }

  // Maybe add chance for body to completely overwrite this object.
  result = recombine_blocks(body);
  if (result == -1) {
    printf("prod_body :: recombine() - error with recombine_blocks\n");
    return -1;
  }
  return result;
}


void fprint_pvalue(FILE *file, Prod_Value *pvalue)
{
  if (pvalue->type == PBC_TYPE_VALUE)
    fprintf(file, "%.1lf", pvalue->value);
  else if (pvalue->type == PBC_TYPE_CONSTANT)
    fprintf(file, "k%d", int(pvalue->symbol));
  else if (pvalue->type == PBC_TYPE_VARIABLE)
    fprintf(file, "n%d", int(pvalue->symbol));
  else if (pvalue->type == PBC_TYPE_FUNCTION)
    fprintf(file, "f%d", int(pvalue->symbol));
}


void fprint_param(FILE *file, Prod_Param *param)
{
  fprint_pvalue(file, &param->arg1);

  if (param->op == PBC_OP_ADD)
    fprintf(file, "%c", OP_ADD);
  else if (param->op == PBC_OP_SUB)
    fprintf(file, "%c", OP_SUB);
  else if (param->op == PBC_OP_MULT)
    fprintf(file, "%c", OP_MULT);
  else if (param->op == PBC_OP_DIV)
    fprintf(file, "%c", OP_DIV);
  else if (param->op == PBC_OP_POW)
    fprintf(file, "%c", OP_POW);
  else if (param->op == PBC_OP_EQUAL)
    fprintf(file, "%c", OP_EQUAL);

  if (param->op != PBC_OP_NONE)
    fprint_pvalue(file, &param->arg2);
}


int Prod_Body :: fprint_char(FILE *file, int i)
{
  int j;

  if (string[i].type == PBC_TYPE_NONE) {
    printf("prod_body :: fprint() - Error at %d, unspecified type.\n", i);
    printf("%c:%d:", string[i].symbol,
	   int(string[i].symbol));
    fprint_pvalue(stdout, &string[i].param[0].arg1);
    printf("\n");
    return FALSE;
    
  } else if (string[i].type == PBC_TYPE_TERMINAL) {
    fprintf(file, "%c", string[i].symbol);
    fprintf(file, "(");
    fprint_pvalue(file, &string[i].param[0].arg1);
    fprintf(file, ") ");
    
  } else if (string[i].type == PBC_TYPE_PRODUCTION) {
    fprintf(file, "P%d", int(string[i].symbol));
    if (Num_arguments > 0)
      fprintf(file, "(");
    for (j=0; j<Num_arguments; j++) {	
      fprint_param(file, &string[i].param[j]);
      if (j < Num_arguments-1) {
	fprintf(file, ",");
      }
    }
    if (Num_arguments > 0)
      fprintf(file, ") ");
    else
      fprintf(file, " ");
      
  } else if (string[i].type == PBC_TYPE_BLOCK) {
    fprintf(file, "%c", string[i].symbol);
    if (string[i].symbol == SYMBOL_BLK_END) {
      fprintf(file, "(");
      fprint_pvalue(file, &string[i].param[0].arg1);
      fprintf(file, ") ");
    } else if ((string[i].symbol == SYMBOL_POP) ||
	       (string[i].symbol == SYMBOL_POP_EDGE)) {
      fprintf(file, " ");
    }
  }

  return TRUE;
}


int Prod_Body :: fprint(FILE *file)
{
  int i;

  // Print the condition statement.
  if (cond.comparator == PBC_COMP_TRUE)
    fprintf(file, "   (T)   ");
  else {
    fprintf(file, "   (");
    fprint_pvalue(file, &cond.arg1);

    if (cond.comparator == COMP_GREATER)
      fprintf(file, ">");
    else if (cond.comparator == COMP_LESS)
      fprintf(file, "<");

    fprint_pvalue(file, &cond.arg2);
    fprintf(file, ")");
  }
  fprintf(file, " :- ");
  

  // Print the body.
  for (i=0; i<length; i++) {
    if (!fprint_char(file, i)) {
      return FALSE;
    }
  }

  //  fprintf(file, ". <%d>\n", used);
  fprintf(file, ". <%d>", used);

  for (i=0; i<Num_arguments; i++)
    fprintf(file, " (%.1f:%.1f)", min[i], max[i]);
  fprintf(file, "\n");

  return TRUE;
}


int read_integer(char *line)
{
  char prev;
  int i = 0;
  int val;

  while ((line[i] != ' ') && (line[i] != ',')
	 && (line[i] != ')') && (line[i] != '>'))
    i++;
  prev = line[i];
  line[i] = 0;

  val = atoi(line);
  line[i] = prev;

  return val;
}


int read_pvalue(char *line, Prod_Value *pvalue)
{
  if (line[0] == 'k') {
    pvalue->type = PBC_TYPE_CONSTANT;
    pvalue->symbol = read_integer(&line[1]);

  } else if (line[0] == 'n') { 
    pvalue->type = PBC_TYPE_VARIABLE;
    pvalue->symbol = read_integer(&line[1]);

  } else if (line[0] == 'f') { 
    pvalue->type = PBC_TYPE_FUNCTION;
    pvalue->symbol = read_integer(&line[1]);

  } else {
    if ((line[0] != '-') &&
	((line[0] < '0') || (line[0] > '9'))) {
      // not a number.
      printf("Prod_Body :: read_pvalue() - error reading value at: %s.\n", line);
      return FALSE;
    }

    int i = 1;
    while (!is_op(line[i]) && (line[i] != ',')
	   && (line[i] != ')'))
      i++;
    
    char prev = line[i];
    line[i] = 0;
    pvalue->type = PBC_TYPE_VALUE;
    pvalue->value = atof(line);
    line[i] = prev;
  }

  return TRUE;
}


int read_param(char *line, Prod_Param *param)
{
  char op;
  int i = 0;

  // Read arg 1.
  if (!read_pvalue(line, &param->arg1)) {
    printf("read_param() - error reading arg 1.\n");
    printf("!%s", line);
    return FALSE;
  }

  // Advance along line.
  while (!is_op(line[i]) && (line[i] != ')'))
    i++;

  // Read operator.
  op = line[i];
  if (op == OP_ADD)
    param->op = PBC_OP_ADD;
  else if (op == OP_SUB)
    param->op = PBC_OP_SUB;
  else if (op == OP_MULT)
    param->op = PBC_OP_MULT;
  else if (op == OP_DIV)
    param->op = PBC_OP_DIV;
  else if (op == OP_POW)
    param->op = PBC_OP_POW;
  else if (op == OP_EQUAL)
    param->op = PBC_OP_EQUAL;
  else if (op == ')') {
    param->op = PBC_OP_NONE;
    return TRUE;
      
  } else {
    printf("read_param() - error reading arg operator: %c:%d.\n", op, int(op));
    return FALSE;
  }
  i++;

  // Read arg 2.
  if (!read_pvalue(&line[i], &param->arg2)) {
    printf("read_param() - error reading arg 2.\n");
    printf("!%s", &line[i]);
    return FALSE;
  }

  return TRUE;
}


#define MAX_STRING_LEN 75000

int Prod_Body :: read(FILE *file)
{
  char *line; // [MAX_STRING_LEN];
  int i, j, len;

  line = (char*)malloc(MAX_STRING_LEN*sizeof(char));
  if (line == 0) {
    printf("prod_body :: read() - error malloc'ing line.\n");
    return FALSE;
  }

  if (!fgets(line, MAX_STRING_LEN, file)) {
    cout << "Prod_Body() :: read() - error, empty line." << endl;
    return FALSE;
  }


  // Read Condition first.
  i=0;
  while (line[i] != '(')
    i++;
  i++;
  if (line[i] == 'T') {
    // Condition is always true.
    set_cond_true();
    while (line[i] != ')')
      i++;

  } else {    
    // Read truth condition.
    if (!read_pvalue(&(line[i]), &cond.arg1)) {
      printf("Prod_Body() :: read() - error reading arg 1 at: %d.\n", i);
      return FALSE;
    }
    while (!is_op(line[i]))
      i++;

    if (line[i] == '>') {
      cond.comparator = COMP_GREATER;
    } else if (line[i] == '<') {
      cond.comparator = COMP_LESS;
    } else {
      printf("Prod_Body() :: read() - error, unexpected comparator op: %c.\n", line[i]);
      return FALSE;
    }
    i++;

    if (!read_pvalue(&(line[i]), &cond.arg2)) {
      printf("Prod_Body() :: read() - error reading arg 2 at: %d.\n", i);
      return FALSE;
    }
    if (line[i] == '-')
      i++;
    while (!is_op(line[i]))
      i++;
  }


  // Read production.
  while (line[i] != '-')
    i++;
  i += 2;

  len=-1;
  while(line[i] != '.') {
    len++;
    if (line[i] == 'P') {
      string[len].type = PBC_TYPE_PRODUCTION;
      string[len].symbol = read_integer(&line[i+1]);

      if (Num_arguments > 0) {
	while (line[i] != '(')
	  i++;
	i++;

	for (j=0; j<Num_arguments; j++) {
	  if (!read_param(&line[i], &string[len].param[j])) {
	    printf("Prod_Body() :: read() - sym #%d, error reading param at: %d.\n", len, i);
	    printf("!%s", &line[i]);
	    return FALSE;
	  }
	  while ((line[i] != ')') && (line[i] != ','))
	    i++;
	  i++;
	}
      } else {
	// No arguments.
	while (line[i] != ' ')
	  i++;
      }

    } else if ((line[i] == '[') ||
	       (line[i] == ']') ||
	       (line[i] == SYMBOL_PUSH_EDGE) ||
	       (line[i] == SYMBOL_POP_EDGE) ||
	       (line[i] == '{')) {
      string[len].type = PBC_TYPE_BLOCK;
      string[len].symbol = line[i];
      if ((line[i] == ']') || (line[i] == SYMBOL_POP_EDGE))
	i++; // Advance to the empty space.

    } else if (line[i] == '}') {
      string[len].type = PBC_TYPE_BLOCK;
      string[len].symbol = line[i];
      i += 2;
      if (!read_pvalue(&line[i], &string[len].param[0].arg1)) {
	printf("Prod_Body() :: read() - error reading '}' param at: %d.\n", i);
	printf("!%s", &line[i]);
	return FALSE;
      }
      while (line[i] != ' ')
	i++;

    } else if (is_term_char(line[i])) {
      string[len].type = PBC_TYPE_TERMINAL;
      string[len].symbol = line[i];
      i += 2;
      if (!read_pvalue(&line[i], &string[len].param[0].arg1)) {
	printf("Prod_Body() :: read() - error reading terminal '%c' param at: %d.\n",
	       string[len].symbol, i);
	return FALSE;
      }
      while (line[i] != ' ')
	i++;
    }

    // line[i] should equal ' ', advance to next symbol.
    i++;
  }
  length = len + 1;

  i++;
  used = read_integer(&line[i]);
  free(line);

  return match_parens();
}


int Prod_Body :: read_old(FILE *file)
{
#ifdef OLD
  char symbol, line[MAX_STRING_LEN];
  int i, j, len;
  Real mult;



  if (!fgets(line, MAX_STRING_LEN, file)) {
    cout << "Read pbody() error, empty line." << endl;
    return FALSE;
  }

  i=0;
  while (((i==' ') || (i=='-') || (i==':')))
    i++;

  len=0;
  mult = 1.0;
  while((line[i] != 0) && (line[i] != '\n') &&
	(line[i] != '.')) {
    symbol = line[i];
    if ((symbol <= '9') &&
	(symbol >= '0')) {
      // integer multiplier
      mult = Real(symbol-'0');

    } else if ((symbol <= 'Z') &&
	(symbol >= 'A')) {
      // Production
      string[len].type = PBC_TYPE_PRODUCTION;
      string[len].symbol = symbol - 'A';
      string[len].param[0].arg1.type = PBC_TYPE_VALUE;
      string[len].param[0].arg1.value = mult;

      body->type[len] = STYPE_PRODUCTION;
      body->param[len] = line[i+3];

      for (j=i+4; line[j]!=')'; j++) {
	if (j >= MAX_STRING_LEN) {
	  printf("Read pbody() - error reading production %c [%s]\n",
		 symbol, line);
	  return FALSE;
	}
      }

      line[j] = 0;
      body->arg[len] = atof(&(line[i+4]));
      i = j;

    } else if (is_block(symbol)) {
      // Block symbol.
      body->symbol[len] = symbol;
      body->type[len] = STYPE_BLOCK;
      body->param[len] = 0;
      body->arg[len] = 0.0;

    } else if (is_term(symbol))  {
      // Terminal.
      body->symbol[len] = symbol;
      body->type[len] = STYPE_TERM;
      body->param[len] = 0;
      body->arg[len] = 0.0;

    } else if (symbol == 'n') {
      // Argument multiplier.
      body->symbol[len] = symbol;
      body->type[len] = STYPE_MULTIPLIER;
      body->param[len] = 0;
      body->arg[len] = 0.0;

    } else {
      len--;
    }

    i++;
    len++;
  }

  while (line[i] == ' ')
    i++;

  if ((line[i] != 0) && (line[i] != '\n')) {
    sscanf(&(line[i]), "(%d)\n",
	   &body->used);
  }
  
  body->length = len;
  valid_pbody(body);
  match_paren(body);
#endif
  return TRUE;
}
