/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <values.h>
#include <stream.h>
#include <fstream.h>
#include "general.h"
#include "graphic.h"
#include "lsystem.h"
#include "programs.h"



int Prog_length;
LString *Program;


#define PROG_STACK_SIZE 100

Real Test_data[29][8] = {
  {8.0, 7.0, 6.0, 5.0, 4.0, 3.0, 2.0, 1.0},
  {7.0, 8.0, 6.0, 5.0, 4.0, 3.0, 2.0, 1.0},
  {6.0, 7.0, 8.0, 5.0, 4.0, 3.0, 2.0, 1.0},
  {5.0, 7.0, 6.0, 8.0, 4.0, 3.0, 2.0, 1.0},
  {4.0, 7.0, 6.0, 5.0, 8.0, 3.0, 2.0, 1.0},
  {3.0, 7.0, 6.0, 5.0, 4.0, 8.0, 2.0, 1.0},
  {2.0, 7.0, 6.0, 5.0, 4.0, 3.0, 8.0, 1.0},
  {1.0, 7.0, 6.0, 5.0, 4.0, 3.0, 2.0, 8.0},
  {8.0, 6.0, 7.0, 5.0, 4.0, 3.0, 2.0, 1.0},
  {8.0, 5.0, 6.0, 7.0, 4.0, 3.0, 2.0, 1.0},
  {8.0, 4.0, 6.0, 5.0, 7.0, 3.0, 2.0, 1.0},
  {8.0, 3.0, 6.0, 5.0, 4.0, 7.0, 2.0, 1.0},
  {8.0, 2.0, 6.0, 5.0, 4.0, 3.0, 7.0, 1.0},
  {8.0, 1.0, 6.0, 5.0, 4.0, 3.0, 2.0, 7.0},
  {8.0, 7.0, 5.0, 6.0, 4.0, 3.0, 2.0, 1.0},
  {8.0, 7.0, 4.0, 5.0, 6.0, 3.0, 2.0, 1.0},
  {8.0, 7.0, 3.0, 5.0, 4.0, 6.0, 2.0, 1.0},
  {8.0, 7.0, 2.0, 5.0, 4.0, 3.0, 6.0, 1.0},
  {8.0, 7.0, 1.0, 5.0, 4.0, 3.0, 2.0, 6.0},
  {8.0, 7.0, 6.0, 4.0, 5.0, 3.0, 2.0, 1.0},
  {8.0, 7.0, 6.0, 3.0, 4.0, 5.0, 2.0, 1.0},
  {8.0, 7.0, 6.0, 2.0, 4.0, 3.0, 5.0, 1.0},
  {8.0, 7.0, 6.0, 1.0, 4.0, 3.0, 2.0, 5.0},
  {8.0, 7.0, 6.0, 5.0, 3.0, 4.0, 2.0, 1.0},
  {8.0, 7.0, 6.0, 5.0, 2.0, 3.0, 4.0, 1.0},
  {8.0, 7.0, 6.0, 5.0, 1.0, 3.0, 2.0, 4.0},
  {8.0, 7.0, 6.0, 5.0, 4.0, 2.0, 3.0, 1.0},
  {8.0, 7.0, 6.0, 5.0, 4.0, 1.0, 2.0, 3.0},
  {8.0, 7.0, 6.0, 5.0, 4.0, 3.0, 1.0, 2.0}
};



void print_data(int size, Real *data)
{
  for (int i=0; i<size; i++) {
    printf("%d  ", (int)data[i]);
  }
}


// Returns TRUE if program runs successfully, or
// FALSE if the program crashes.
int run_program(int size, Real *data)
{
  int i;
  int flag = FALSE;
  int execute_code = TRUE;
  int index;
  Real acc = 0.0;
  Real tmp;

  int stack_ptr = 0;
  sProg_Stack prog_stack[PROG_STACK_SIZE];

  int acc_ptr = 0;
  Real acc_stack[PROG_STACK_SIZE];


  for (i=0; i<Prog_length; i++) {
    if (execute_code) {
      // Execute code.
      if (Program[i].symbol == TERM_PUSH) { // This is "IF-FLAG".
	if (stack_ptr == PROG_STACK_SIZE) {
	  printf("Program stack exceeded.\n");
	  return FALSE;
	}
	prog_stack[stack_ptr].execute_code = TRUE;
	stack_ptr++;
	if (!flag) {
	  execute_code = FALSE;
	}

      } else if (Program[i].symbol == TERM_POP) {
	if (stack_ptr == 0) {
	  printf("Progam stack over-popped.\n");
	  return FALSE;
	}
	stack_ptr--;

      } else if (Program[i].symbol == TERM_LOAD_ACC) {
	index = int(Program[i].arg) % size;
	index = absv(index);
	acc = data[index];

      } else if (Program[i].symbol == TERM_STORE_ACC) {
	index = int(Program[i].arg) % size;
	index = absv(index);
	data[index] = acc;

      } else if (Program[i].symbol == TERM_PUSH_ACC) {
	if (acc_ptr == PROG_STACK_SIZE) {
	  if (Print[9]) {
	    printf("run_program :: accumulation stack exceeded: %d.\n",
		   acc_ptr);
	  }
	  //	  return FALSE;
	} else {
	  acc_stack[acc_ptr] = acc;
	  acc_ptr++;
	}

      } else if (Program[i].symbol == TERM_POP_ACC) {
	if (acc_ptr == 0) {
	  if (Print[9]) {
	    printf("run_program :: accumulation stack over-popped.\n");
	  }
	  //	  return FALSE;
	} else {
	  acc_ptr--;
	  acc = acc_stack[acc_ptr];
	}

      } else if (Program[i].symbol == TERM_COMPARE_ACC) {
	index = int(Program[i].arg) % size;
	index = absv(index);
	if (data[index] < acc) {
	  flag = TRUE;
	} else {
	  flag = FALSE;
	}

      } else if (Program[i].symbol == TERM_COMPARE_DATA) {
	index = int(Program[i].arg) % (size-1);
	index = absv(index);
	if (data[index] < data[index+1]) {
	  flag = TRUE;
	} else {
	  flag = FALSE;
	}

      } else if (Program[i].symbol == TERM_SWAP_DATA) {
	index = int(Program[i].arg) % (size-1);
	index = absv(index);
	tmp = data[index];
	data[index] = data[index+1];
	data[index+1] = tmp;

      } else if (Program[i].symbol == TERM_CLEAR_FLAG) {
	flag = FALSE;

      } else if (Program[i].symbol == TERM_BREAK) {
	return TRUE;

      } else if (Program[i].symbol == TERM_HALT) {
	return TRUE;

      } else if (Program[i].symbol == 0) {
	if (Print[9])
	  printf("run_program() :: execute: unexpected early termination of command sequence @:%d(!%d)\n",
		 i, Prog_length);
	return FALSE;

      } else {
	printf("run_program() :: unknown command: (%c:%d)\n", 
	       Program[i].symbol, int(Program[i].symbol));
	return FALSE;
      }     

      /*
      printf("%c : ", Program[i].symbol);
      print_data(size, data);
      */
    } else {

      // Don't execute code - wait until end of block.
      if (Program[i].symbol == TERM_PUSH) { // This is "IF-FLAG".
	if (stack_ptr == PROG_STACK_SIZE) {
	  printf("Program stack exceeded.\n");
	  return FALSE;
	}
	prog_stack[stack_ptr].execute_code = FALSE;
	stack_ptr++;

      } else if (Program[i].symbol == TERM_POP) {
	if (stack_ptr == 0) {
	  printf("Progam stack over-popped.\n");
	  return FALSE;
	}
	stack_ptr--;
	execute_code = prog_stack[stack_ptr].execute_code;

      } else if (Program[i].symbol == 0) {
	if (Print[0])
	  printf("run_program() :: not execute : unexpected early termination of command sequence @:%d(!%d)\n",
		 i, Prog_length);
	return FALSE;

      }
    }
  }


  return TRUE;
}


int data_same(int size, Real *data, Real *data2)
{
  int i;

  for (i=0; i<size; i++) {
    if (data[i] != data2[i]) {
      return FALSE;
    }
  }
  return TRUE;
}


// Returns number of in-order pairs.
int is_sorted(int size, Real *data)
{
  int i, score = 0;

  for (i=0; i<size-1; i++) {
    if (data[i] == 0.0) {
      continue;
    }
    if (data[i] <= data[i+1]) {
      score++;
    }
  }
  return score;
}


int sort_score(int size, Real *data)
{
  int score = size*size;

  for (int i=1; i<=size; i++) {
    score -= absv(i - (int)data[i-1]);
    //    printf("absv: %d,  score: %d.\n", absv(i - (int)data[i-1]), score);
  }
  return score;
}


// A bubble sort to count number of numbers still
// in array.
int num_not_missing(int size, Real *data)
{
  int i, j;
  int score = 0;

  for (i=1; i<=size; i++) {
    for (j=0; j<size; j++) {
      if ((int)data[j] == i) {
	break;
      }
    }
    if (j != size) {
      score++;
    }
  }
  return score;
}



int sort_data(int size, Real *data)
{
  int result;
  if (Print[7]) {
    print_data(size, data);
  }
  result = run_program(size, data);
  if (result) {
    if (Print[7]) {
      printf(" ->   ");
      print_data(size, data);
      printf("\n");
    }
    return sort_score(size, data);
  }
  printf("\n");
  return -1;
}



// Sort.
int fitness_program3(ModelStruct *model)
{
  Real *fitness_vec = model->fitness;

  static Real max_fitness = 0.0;

  int i, j, size;
  Real score, tmp, fitness = 0.0;
  Real data[100];
  Real data2[100];
  
  size = 8;
  for (i=0; i<size; i++) {
    data[i] = (Real)(size-i);
  }


  for (i=0; i<5; i++) {
    for (j=0; j<size; j++) {
      data2[j] = data[j];
    }

    score = sort_data(size, data2);
    if (score == -1) {
      fitness = 0.0;
      break;
    }

    fitness += (Real)score;

    tmp = data[i];
    data[i] = data[i+1];
    data[i+1] = tmp;
  }

  fitness_vec[0] = fitness;

  if (Print[6]) {
    printf("Program #2: size:%d, fitness: %.0f.\n",
	   size, fitness);
  }

  if (fitness > max_fitness) {
    max_fitness = fitness;
    printf("New max: %2d: ", (int)fitness);
    for (i=0; i<size; i++) {
      printf("%d  ", (int)data[i]);
    }
    printf("\n");
  }

  return TRUE;
}


// Sort.
int fitness_sort(ModelStruct *model)
{
  Real *fitness_vec = model->fitness;

  static Real max_fitness = 0.0;

  int i, j, size;
  Real score, fitness = 0.0;
  Real data[100];

  size = 8;
  
  for (i=0; i<29; i++) {
    for (j=0; j<size; j++) {
      data[j] = Test_data[i][j];
    }

    score = sort_data(size, data);
    if (score == -1) {
      fitness = 0.0;
      break;
    }

    fitness += (Real)score;
  }

  fitness_vec[0] = fitness;

  if (Print[6]) {
    printf("Program #2: size:%d, fitness: %.0f.\n",
	   size, fitness);
  }

  if (fitness > max_fitness) {
    max_fitness = fitness;
    printf("New max: %2d: ", (int)fitness);
    for (i=0; i<size; i++) {
      printf("%d  ", (int)data[i]);
    }
    printf("\n");
  }

  return TRUE;
}



int boolean_sort_score(int length, Real *data)
{
  int i;
  int num_zeros = 0;
  int score = length * length;

  for (i=0; i<length; i++) {
    if (data[i] == 0.0) {
      num_zeros++;
    }
  }

  for (i=0; i<num_zeros; i++) {
    if (data[i] == 1.0) {
      score -= num_zeros - i;
    }
  }

  num_zeros--;
  for (; i<length; i++) {
    if (data[i] == 0.0) {
      score -= i - num_zeros;
    }
  }

  return score;
}


#define MAX_DATA_LENGTH 100

int fitness_boolean_sort(ModelStruct *model)
{
  Real *fitness_vec = model->fitness;

  int i, length, score;
  Real fitness = 0.0;
  Real data[MAX_DATA_LENGTH];
  Real data2[MAX_DATA_LENGTH];
  Real fitness2 = 0.0;

  length = (int)model->build_args[0];

  for (i=0; i<length; i++) {
    data[i] = 0;
  }

  while (1) {
    for (i=0; i<length; i++) {
      data2[i] = data[i];
    }
    if (!run_program(length, data2)) {
      // program crashed -> end loop.
      fitness = 0.0;
      if (Print[7]) {
	printf("programs :: fitness_boolean_sort() - program crashed.\n");
      }
      break;
    }
    score = boolean_sort_score(length, data2);
    fitness += (Real)score;

    if (Print[7]) {
      int score2 = 0;
      for (i=0; i<length; i++) {
	if (data2[i] == 1.0) {
	  data2[i] = 0.0;
	} else {
	  data2[i] = 1.0;
	}
      }
      score2 = boolean_sort_score(length, data2);
      fitness2 += (Real)score2;
      print_data(length, data);
      printf(" ->   ");
      print_data(length, data2);
      printf(" : %3d %3d (%3.0f:%3.0f)\n",
	     score, score2, fitness, fitness2);
    }

    
    // "Add 1".
    for (i=0; i<length; i++) {
      if (data[i] == 0.0) {
	data[i] = 1.0;
	break;
      }
      data[i] = 0.0;
    }

    if (i == length) {
      // No more inputs to test -> done!
      break;
    }
  }

  if (length == 3) {
    fitness -= 54.0; // Max is -> 18
  } else if (length == 4) {
    fitness -= 188.0; // Max is -> 68
  } else if (length == 5) {
    fitness -= 590.0; // Max is -> 210
  } else if (length == 6) {
    fitness -= 1692.0; // Max is -> 612
  } else if (length == 7) {
    fitness -= 4620.0; // Max is -> 1652
  } else if (length == 8) {
    fitness -= 12100.0; // Max is -> 4284
  }
  fitness_vec[0] = fitness;

  if (Print[6]) {
    printf("Program #3: length:%d, fitness: %.0f, cmnds: %d.\n",
	   length, fitness, Prog_length);
  }

  /*
  if (fitness > max_fitness) {
    max_fitness = fitness;
    printf("New max: %2d: ", (int)fitness);
    for (i=0; i<length; i++) {
      printf("%d  ", (int)data[i]);
    }
    printf("\n");
  }
  */

  return TRUE;
}


// Reverse a list.
int fitness_reverse(ModelStruct *model)
{
  int fv_len = model->fv_length;
  Real *fitness_vec = model->fitness;

  static Real max_fitness = 0.0;

  int i, result, size;
  int sorted = 0, not_missing = 0, sscore = 0;
  Real fitness = 0.0;
  Real data[100];
  Real data_bk[100];
  
  size = 8;
  for (i=0; i<size; i++) {
    data[i] = (Real)(size-i);
  }

  for (i=0; i<size; i++) {
    data_bk[i] = data[i];
  }

  result = run_program(size, data);
  if (result) {
    if (data_same(size, data, data_bk)) {
      fitness = 0.0;
      
    } else {
      sorted = is_sorted(size, data);
      not_missing = num_not_missing(size, data);
      sscore = sort_score(size, data);
      //      fitness = (Real)(sorted+not_missing);
      fitness = (Real)sscore;
      /*      
      for (i=0; i<size; i++) {
	if ((int)data[i] == i+1) {
	  fitness += 1.0;
	}
      }
      */
    }
  }

  fitness_vec[0] = fitness;
  if (fv_len > 7) { 
    fitness_vec[1] = (Real)sscore;
    fitness_vec[2] = (Real)sorted;
    fitness_vec[3] = (Real)not_missing;
    fitness_vec[4] = 0.0;
    fitness_vec[5] = 0.0;

  } else if (fv_len > 1) {
    fitness_vec[1] = 0.0;
  }

  if (Print[6]) {
    printf("Program #1: size:%d, sort:%d, !miss:%d,  sscore:%d,  fitness: %.0f.\n",
	   size, sorted, not_missing, sscore, fitness);
  }

  /*
  if ((int)fitness > 0) {
    printf("%2d: ", (int)fitness);
    for (i=0; i<size; i++) {
      printf("%d  ", (int)data[i]);
    }
    printf("\n");
  }
  */

  if (fitness > max_fitness) {
    max_fitness = fitness;
    printf("New max: %2d: ", (int)fitness);
    for (i=0; i<size; i++) {
      printf("%d  ", (int)data[i]);
    }
    printf("\n");
  }

  return TRUE;
}


int fitness_program(ModelStruct *model)
{
  if (Fitness_func == 1) {
    return fitness_reverse(model);
  } else if (Fitness_func == 2) {
    return fitness_sort(model);
  } else if (Fitness_func >= 3) {
    return fitness_boolean_sort(model);
  }

  return FALSE;
}

int build_program(int length, LString *command,
			  Real *args)
{
  Prog_length = length;
  Program = command;
  /*
  printf("Program length: %d, cmnd: %c:%d.\n", length,
	 command[0].symbol, command[0].symbol);
  */
  return TRUE;
}
  

