/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/


#define stat_mptr() multi_ptr
#define stat_nm() nmulti
#define stat_mlen() multi_len
#define stat_m(i) stat_mptr()[i]
#define stat_mavg_ptr(i) stat_m(i).average
#define stat_mstd_ptr(i) stat_m(i).std_dev
#define stat_mavg(i, j) stat_mavg_ptr(i)[j]
#define stat_mstd(i, j) stat_mstd_ptr(i)[j]
#define stat_mtrials() mtrials
#define stat_mstart(i) stat_mmctr(i) = 0
#define stat_mnext() stat_mctr()++
#define stat_mctr() m_ctr
#define stat_mmctr(i) stat_m(i).ctr

#define stat_sptr() single_ptr
#define stat_ns() nsingle
#define stat_s(i) stat_sptr()[i]
#define stat_savg(i) stat_s(i).average
#define stat_sstd(i) stat_s(i).std_dev
#define stat_strials() strials
#define stat_sctr() sctr

typedef struct tStat_multi
{
   int ctr;
   double *average;
   double *std_dev;
} Stat_multi;

typedef struct tStat_single
{
   double average;
   double std_dev;
} Stat_single;


class Stats
{
private:
  int Built;

public:

  int num_multi;
  int multi_len;
  int num_single;
  int m_ctr;
  int s_ctr;
  int m_trials;
  int s_trials;
  struct tStat_multi *multi_ptr;
  struct tStat_single *single_ptr;

  Stats(void);
  void free(void);
  void init(void);
  int built(void);
  int unbuild(void);
  int build(void);
  void multi_start(void);
  void single_start(void);
  void multi_add(double val);
  void multi_add(double *val);
  void single_add(double val);
  void multi_again(void);
  void multi_next(void);
  void send_stats(Stats *stats);
  void print(void);
  void print_last(void);
  void print_formatted(void);
  void print_short_formatted(void);
};
