/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <stream.h>
#include <fstream.h>
#include "ind_tree.h"
#include "tree.h"
#include "general.h"

#define TREE_VERSION 1

#define TREE_NUM_NONTERMINAL 9
#define TREE_IFLT 0
#define TREE_ADD 1
#define TREE_SUBTRACT 2
#define TREE_MULTIPLY 3
#define TREE_DIVIDE 4
#define TREE_SIN 5
#define TREE_COS 6
#define TREE_REMAINDER 7
#define TREE_ASIN 8

#define TREE_NUMBER 9
#define TREE_TERMINAL1 10
#define TREE_VARIABLE 11



#define TREE_MUTATE_SIZE 0.1

#define TREE_MAX_RESULTS 6


int num_trees = 0;
int tree_num;
Tree *tree_next = NULL;


Tree *tree_new(void)
{
  if (tree_num > 0) {
    Tree *v_tmp;

    v_tmp = tree_next;
    tree_next = v_tmp->child[0];
    tree_num--;
    v_tmp->init();
    return v_tmp;
  }

  num_trees++;
  return new Tree();
}


Tree :: Tree(void)
{
  init();
}


void Tree :: init(void)
{
  parent = NULL;
  for (int i=0; i<TREE_MAX_CHILDREN; i++)
    child[i] = NULL;
  num_childs = 0;

  type = VC_NONE;
  value = 0.0;
  size1 = 0;
  size2 = 0;
}



void Tree :: free_tree(void)
{
  for (int i=0; i<num_childs; i++) {
    if (child[i] == NULL) {
      cout << "Error: free tree, freeing child #" << i << endl;
      exit;
    }
    child[i]->free_tree();
  }
  free(this);
}


void Tree :: delete_tree(void)
{
  for (int i=0; i<num_childs; i++) {
    if (child[i] == NULL) {
      cout << "Error: free tree, freeing child #" << i << endl;
      exit;
    }
    child[i]->delete_tree();
  }

  child[0] = tree_next;
  tree_next = this;
  tree_num++;
}


Tree *Tree :: make_random_subtree_size(int size)
{
  Tree *child;

  if (size == 0) {
    return NULL;
  }

}


int Tree :: verify(void)
{
  int i, s1, s2, childs, ok = TRUE;

  if (return_type == 0) {
    s1 = 1;
    s2 = 0;

#ifdef OLD
    if (parent == NULL) {
      cout << "Verify:: error root of child has return type 0" << endl;
      ok = FALSE;
    }
#endif
  } else {
    s1 = 0;
    s2 = 1;
  }
  
  if (type == TREE_IFLT)
    childs = 4;
  else if ((type == TREE_ADD) || (type == TREE_SUBTRACT) ||
	   (type == TREE_MULTIPLY) || (type == TREE_DIVIDE) ||
	   (type == TREE_REMAINDER)) {
    childs = 2;
  } else if ((type == TREE_NUMBER) || (type == TREE_TERMINAL1)
	     || (type == TREE_VARIABLE)) {
    childs = 0;
  } else {
    childs = 1;
  }
  if (childs != num_childs) {
    cout << "Verify:: error num_childs:"<<num_childs<<" != " << childs << endl;
    ok = FALSE;
  }


  for (i=0; i<num_childs; i++) {
    s1 += child[i]->size1;
    s2 += child[i]->size2;
    if (child[i]->child_id != i) {
      cout << "Verify:: error child_id:"<<child[i]->child_id<<" != " << i << endl;
      ok = FALSE;
    }
  }

  if (s1 != size1) {
    cout << "Verify:: error size1:"<<size1<<" != " << s1 << endl;
    ok = FALSE;
  }

  if (s2 != size2) {
    cout << "Verify:: error size2:"<<size2<<" != " << s2 << endl;
    ok = FALSE;
  }

  for (i=0; i<num_childs; i++) {
    if (!child[i]->verify()) {
      ok = FALSE;
      cout << "Verify :: child failed." << endl;
    }
  }
  return ok;
}    



void Tree :: make_random_depth(int depth, int rtype)
{
  int i;

  //  cout << "make random depth: " << depth << "  rtype: " << rtype << endl;

  return_type = rtype;

  if (depth <= 1) {
    for (i=0; i<num_childs; i++) {
      child[i]->delete_tree();
    }
    num_childs = 0;

    if (rtype == 0) {
      if (random_int(2)) {
	// normal number
	type = TREE_NUMBER;
	value = 2.0 * random_dbl() - 1.0;
      } else {
	// variable
	type = TREE_VARIABLE;
	value = random_int(source->num_inputs);
      }
      size1 = 1;
      size2 = 0;
    } else {
      type = TREE_TERMINAL1;
      value = (Real)random_int(TREE_NUM_TERMINAL1);
      size1 = 0;
      size2 = 1;
    }

    return;
  }

  int prev_childs = num_childs;

  if (return_type == 0)
    type = random_int(TREE_NUM_NONTERMINAL-1)+1;
  else
    type = TREE_IFLT;

  if (type == TREE_IFLT)
    num_childs = 4;
  else if ((type == TREE_ADD) || (type == TREE_SUBTRACT) ||
	   (type == TREE_MULTIPLY) || (type == TREE_DIVIDE) ||
	   (type == TREE_REMAINDER)) {
    num_childs = 2;
  } else {
    num_childs = 1;
  }

  if (prev_childs < num_childs) {
    for (i=prev_childs; i<num_childs; i++) {
      child[i] = tree_new();
      child[i]->child_id = i;
      child[i]->parent = this;
      child[i]->source = source;
    }
  } else if (num_childs < prev_childs) {
    for (i=num_childs; i<prev_childs; i++) {
      child[i]->delete_tree();
      child[i] = NULL;
    }
  }
    
  depth--;
  for (i=0; i<num_childs; i++) {
    if (i < 2) // child is type 0
      child[i]->make_random_depth(depth - random_int(2), 0);
    else
      child[i]->make_random_depth(depth - random_int(2), return_type);
  }

  if (return_type == 0) {
    size1 = 1;
    size2 = 0;
  } else {
    size1 = 0;
    size2 = 1;
  }

  for (i=0; i<num_childs; i++) {
    size1 += child[i]->size1;
    size2 += child[i]->size2;
  }    
}


// Copies "top" part of vertex over this tree,
// but not child info.
void Tree :: duplicate(Tree *vertex, Tree *parnt)
{
  int i;

  type = vertex->type;
  return_type = vertex->return_type;
  value = vertex->value;
  size1 = vertex->size1;
  size2 = vertex->size2;
  parent = parnt;

  child_id = vertex->child_id;

  int prev_childs = num_childs;
  num_childs = vertex->num_childs;

  if (prev_childs < num_childs) {
    for (i=prev_childs; i<num_childs; i++) {
      child[i] = tree_new();
      child[i]->parent = this;
      child[i]->source = source;
    }
  } else if (num_childs < prev_childs) {
    for (i=num_childs; i<prev_childs; i++) {
      child[i]->delete_tree();
      child[i] = NULL;
    }
  }
    
  //  cout << "Vertex::Duplicate()  num childs:" << num_childs;
  //  cout << "  size:" << size1 << endl;

  for (i=0; i<num_childs; i++) {
    child[i]->source = source;
    child[i]->duplicate(vertex->child[i], this);
    if (child[i]->child_id != i) {
      cout << "Duplicate:: error child_id:"<<child[i]->child_id<<" != " << i;
      cout << "  (original):" << vertex->child[i]->child_id << endl;
      exit -1;
    }
  }
}


// Recalculates subtree size from this vertex up to the root.
void Tree :: resize(void)
{
  if (return_type == 0) {
    size1 = 1;
    size2 = 0;
  } else {
    size1 = 0;
    size2 = 1;
  }

  for (int i=0; i<num_childs; i++) {
    size1 += child[i]->size1;
    size2 += child[i]->size2;
  }

  if (parent != NULL)
    parent->resize();
}

void Tree :: re_source(void)
{
  int i;
  for (i=0; i<num_childs; i++) {
    child[i]->source = source;
    child[i]->re_source();
  }
}


Tree *Tree :: get_tree1(int num)
{
  //  cout << "get_tree1: num:" << num << "  size:" << size1 << endl;

  if (num > size1) {
    cout << "subtree_loc1:: error, num too large: " << num << endl;
    return NULL;
  } else if (num == 0) {
    cout << "subtree_loc1:: error, num = 0." << endl;
    return NULL;
  }

  if (return_type == 0) {
    if (num == 1)
      return this;
    num--;
  }

  for (int i=0; i<num_childs; i++) {
    //    cout << "get_tree1: num:" << num << "  child size:" << child[i]->size1 << endl;
    if (num <= child[i]->size1)
      return child[i]->get_tree1(num);
    num -= child[i]->size1;
  }

  cout << "get_tree1():: error, no more children." << endl;
  return NULL;
}


Tree *Tree :: get_tree2(int num)
{
  if (num > size2) {
    cout << "subtree_loc2:: error, num too large: " << num << endl;
    return NULL;
  } else if (num == 0) {
    cout << "subtree_loc2:: error, num = 0." << endl;
    return NULL;
  }

  if (return_type == 1) {
    if (num == 1)
      return this;
    num--;
  }

  for (int i=0; i<num_childs; i++) {
    if (num <= child[i]->size2)
      return child[i]->get_tree2(num);
    num -= child[i]->size2;
  }

  return NULL;
}


int Tree :: mutate(void)
{
  if (type == TREE_NUMBER) {
    if (!random_int(5)) {
      type = TREE_VARIABLE;
      value = random_int(source->num_inputs);
    } else {
      value += 2.0 * random_dbl() * TREE_MUTATE_SIZE - TREE_MUTATE_SIZE;
    }
  } else if (type == TREE_VARIABLE) {
    value = random_int(source->num_inputs + 1) - 1;
    if (value < 0) {
      type = TREE_NUMBER;
      value = 2.0 * random_dbl() - 1.0;
    }
  } else if (type == TREE_TERMINAL1) {
    value = (Real)random_int(TREE_NUM_TERMINAL1);
  } else if (type == TREE_IFLT) {
    ;
  } else if ((type == TREE_ADD) || (type == TREE_SUBTRACT) ||
	     (type == TREE_MULTIPLY) || (type == TREE_DIVIDE) ||
	     (type == TREE_REMAINDER)) {
    type = random_int(5);
    if (type == 0)
      type = TREE_ADD;
    else if (type == 1)
      type = TREE_SUBTRACT;
    else if (type == 2)
      type = TREE_MULTIPLY;
    else if (type == 3)
      type = TREE_DIVIDE;
    else if (type == 4)
      type = TREE_REMAINDER;
  } else if ((type == TREE_SIN) || (type == TREE_COS) ||
	     (type == TREE_ASIN)) {
    type = random_int(3);
    if (type == 0)
      type = TREE_SIN;
    else if (type == 1)
      type = TREE_COS;
    else if (type == 2)
      type = TREE_ASIN;
  } else {
    cout << "Tree::Mutate() - invalide type" << endl;
  }
}


Real remainder(Real val1, Real val2)
{
  // can do this with divide and int()!!!
  if (val1 < 0.0)
    return -remainder(-val1, val2);

  int val3 = int(val1 / val2);
  val1 -= Real(val3) * val2;
  return val1;

  /*
  if (val1 >= 0.0) {
    if (val2 > val1)
      return val1;

    val3 = 10.0;
    while (val1 > val3 * val2)
      val3 *= 10.0;
    val1 -= (val3 / 10.0) * val2;

    while (val1 > val2)
      val1 -= val2;
    return val1;
  }

  return -remainder(-val1, val2);
  */
}
  


Real Tree :: calculate(void)
{
  Real value1, value2;

  if (type == TREE_NUMBER) {
    return value;
  } else if (type == TREE_TERMINAL1) {
    return value;
  } else if (type == TREE_VARIABLE) {
    return source->input[int(value)];
  } else if (type == TREE_SIN) {
    return sin(child[0]->calculate());
  } else if (type == TREE_COS) {
    return cos(child[0]->calculate());
  } else if (type == TREE_ASIN) {
    value1 = child[0]->calculate();
    if (value1 > 1.0)
      value1 = 1.0;
    else if (value1 < -1.0)
      value1 = -1.0;
    return asin(value1);
  } else if (type == TREE_ADD) {
    return child[0]->calculate() + child[1]->calculate();
  } else if (type == TREE_SUBTRACT) {
    return child[0]->calculate() - child[1]->calculate();
  } else if (type == TREE_MULTIPLY) {
    return child[0]->calculate() * child[1]->calculate();
  } else if (type == TREE_DIVIDE) {
    value1 = child[0]->calculate();
    value2 = child[1]->calculate();
    if (value2 == 0.0)
      return 1000000.0;
    return value1 / value2;
  } else if (type == TREE_REMAINDER) {
    value1 = child[0]->calculate();
    value2 = absv(child[1]->calculate());
    if (value2 == 0.0)
      return 0.0;

    return remainder(value1, value2);

    if (value1 < 0.0) {
      while (-value1 > value2)
	value1 += value2;
    } else {
      while (value1 > value2)
	value1 -= value2;
    }
  } else if (type == TREE_IFLT) {
    value1 = child[0]->calculate();
    value2 = child[1]->calculate();
    if (value1 < value2)
      return child[2]->calculate();
    else
      return child[3]->calculate();
  } else {
    cout << "Tree::calculate() - invalid type:" << type << endl;
  }

  return 0.0;
}


void Tree :: read(FILE *file)
{
  char func;
  int num;

  num = fscanf(file, "%c %lf %d %d %d %d %d\n", &func,
	       &value, &return_type, &size1, &size2,
	       &child_id, &num_childs);
  if (num != 7) {
    cout << "Tree::read()  Error reading data. " << num << endl;
    printf("%c %.7lf %d %d %d %d %d\n", func, value,
	  return_type, size1, size2, child_id, num_childs);
    exit -1;
  }

  if (func == '?')
    type = TREE_IFLT;
  else if (func == '+')
    type = TREE_ADD;
  else if (func == '-')
    type = TREE_SUBTRACT;
  else if (func == '*')
    type = TREE_MULTIPLY;
  else if (func == '/')
    type = TREE_DIVIDE;
  else if (func == 's')
    type = TREE_SIN;
  else if (func == 'c')
    type = TREE_COS;
  else if (func == '%')
    type = TREE_REMAINDER;
  else if (func == 'a')
    type = TREE_ASIN;
  else if (func == '#')
    type = TREE_NUMBER;
  else if (func == 'T')
    type = TREE_TERMINAL1;
  else if (func == 'V')
    type = TREE_VARIABLE;


  for (int i=0; i<num_childs; i++) {
    child[i] = tree_new();
    child[i]->parent = this;
    child[i]->read(file);
  }
}

void Tree :: write(FILE *file)
{
  char func;

  if (type == TREE_IFLT)
    func = '?';
  else if (type == TREE_ADD)
    func = '+';
  else if (type == TREE_SUBTRACT)
    func = '-';
  else if (type == TREE_MULTIPLY)
    func = '*';
  else if (type == TREE_DIVIDE)
    func = '/';
  else if (type == TREE_SIN)
    func = 's';
  else if (type == TREE_COS)
    func = 'c';
  else if (type == TREE_REMAINDER)
    func = '%';
  else if (type == TREE_ASIN)
    func = 'a';
  else if (type == TREE_NUMBER)
    func = '#';
  else if (type == TREE_TERMINAL1)
    func = 'T';
  else if (type == TREE_VARIABLE)
    func = 'V';
  else {
    cout << "Tree::write() - invalid type" << endl;
  }

  fprintf(file, "%c %10.7lf %d %3d %3d %2d %d\n", func, value,
	  return_type, size1, size2, child_id, num_childs);

  for (int i = 0; i<num_childs; i++)
    child[i]->write(file);
}
