/***************************************************
GENRE: An evolutionary design system.
Copyright (C) 1999-2003, Gregory S. Hornby.
hornby@email.arc.nasa.gov
http://ic.arc.nasa.gov/people/hornby

This library is free software; you can redistribute
it and/or modify it under the terms of the GNU Lesser
General Public License as published by the Free
Software Foundation; either version 2.1 of the License,
or (at your option) any later version.

This library is distributed in the hope that it will
be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

The text of the GNU Lesser General Public License is
included with this library in the file LICENSE.TXT.
If this file is not included with this library write
to the Free Software Foundation, Inc., 59 Temple Place,
Suite 330, Boston, MA  02111-1307  USA
****************************************************/

#ifndef TREE_HEADER_FILE
#define TREE_HEADER_FILE

#include "general.h"
#include <fstream.h>

#define TREE_LEFT 0
#define TREE_RIGHT 1

#define TREE_NEW 0

#define TREE_SPLIT 1 // mother and daughter not connected
#define TREE_REVERSE 2 // reverse a link
#define TREE_CUT_IN 3   // cut a link
#define TREE_CUT_OUT 4
#define TREE_CHAIN 5 // mother and daughter connected
#define TREE_RECURSE 6  // recursion

#define TREE_NUM_TERMINAL1 6

#define TREE_MAX_CHILDREN 4

#define VC_NUM_VALUES 1
#define VC_NONE 0

class Tree;
class Individual;

extern int tree_num, num_trees;

Tree *tree_new(void);

class Tree
{
public:
  int type, return_type;
  Real value;
  int size1, size2;
  int num_childs;
  int child_id;  // Which child of the parent this is.
  Tree *parent;
  Tree *child[TREE_MAX_CHILDREN];
  Individual *source;

  Tree(void);
  void init(void);
  void randomize_tree(void);
  void duplicate(Tree *vertex, Tree *parnt);
  void free_tree(void);
  void delete_tree(void);
  Tree *make_random_subtree_size(int size);
  void make_random_depth(int depth, int rtype);
  int mutate(void);

  int verify(void);
  void resize(void);
  void re_source(void);

  Tree *get_tree1(int num);
  Tree *get_tree2(int num);
  Tree *swap_tree(Tree *tree);
 
  Real calculate(void);

  void read(FILE *file);
  void write(FILE *file);
};

#endif
